/* This library implements an AVL tree, i.e a height-balanced binary search tree in which the heights of a nodes two sub-trees
   are not allowed to differ by more than one.

   The package can process strings only. The read and write functions take twice the time of their UltraSets equivalents.

   The underlying implementation has been taken from Martin Broadhurst's exemplary website
   http://www.martinbroadhurst.com/data-structures.html */

#define avl_c
#define LUA_LIB

#include <stdlib.h>

#include "agena.h"
#include "agnxlib.h"
#include "agenalib.h"
#include "agnhlps.h"

#define checkavl(L, n)      (Avl *)luaL_checkudata(L, n, "avl")

#if !(defined(LUA_DOS) || defined(__OS2__) || defined(LUA_ANSI))
#define AGENA_AVLLIBNAME "avl"
LUALIB_API int (luaopen_avl) (lua_State *L);
#endif

/* `header` *************************************************************************************************/

struct avltreenode {
  struct avltreenode *left;
  struct avltreenode *right;
  struct avltreenode *parent;
  unsigned int leftheight;
  unsigned int rightheight;
  int data;
};

typedef struct avltreenode avltreenode;

typedef int  (*avltree_cmpfn)(lua_State *L, int idx1, int idx2);
typedef void (*avltree_forfn)(int);

struct avltree {
  avltreenode *root;
  size_t count;
  avltree_cmpfn compare;
};

typedef struct avltree avltree;

/* `private` ********************************************************************************************* */

static int avl_compare (lua_State *L, int idx1, int idx2) {
  if (lua_type(L, idx1) != lua_type(L, idx2)) return 2;
  if (lua_equal(L, idx1, idx2)) return 0;
  if (lua_lessthan(L, idx1, idx2)) return -1;
  return 1;
}


avltree *avltree_create (avltree_cmpfn compare) {
  avltree *tree = malloc(sizeof(avltree));
  if (tree != NULL) {
    tree->root = NULL;
    tree->compare = compare;
    tree->count = 0;
  }
  return tree;
}

static void avltreenode_delete (lua_State *L, avltreenode *node) {
  luaL_unref(L, LUA_REGISTRYINDEX, node->data);
  xfree(node);
}

static void avltree_empty_recursive (lua_State *L, avltreenode *root) {
  if (root->left)
    avltree_empty_recursive(L, root->left);
  if (root->right)
    avltree_empty_recursive(L, root->right);
  avltreenode_delete(L, root);
}

void avltree_empty (lua_State *L, avltree *tree) {
  if (tree->root) {
    avltree_empty_recursive(L, tree->root);
    tree->root = NULL;
    tree->count = 0;
  }
}

void avltree_delete (lua_State *L, avltree *tree) {
  if (tree) {
    avltree_empty(L, tree);
    xfree(tree);
  }
}

static void avltree_for_each_recursive (const avltreenode *root, avltree_forfn fun) {
  if (root->left != NULL)
    avltree_for_each_recursive(root->left, fun);
  fun(root->data);
  if (root->right != NULL)
    avltree_for_each_recursive(root->right, fun);
}

void avltree_for_each (const avltree *tree, avltree_forfn fun) {
  if (tree->root)
    avltree_for_each_recursive(tree->root, fun);
}

struct avlsearchresult {
  avltreenode *node;
  avltreenode *parent;
};

typedef struct avlsearchresult avlsearchresult;

static int avltree_search (lua_State *L, const avltree *tree, avlsearchresult *result, int idx, int *height) {
  int found;
  found = 0; *height = 0;
  result->node = tree->root;
  while (!found && result->node != NULL) {
    luaL_checkstack(L, 2, "too many calls");
    lua_rawgeti(L, LUA_REGISTRYINDEX, result->node->data);
    lua_pushvalue(L, idx);
    int rv = tree->compare(L, -2, -1);
    agn_poptoptwo(L);
    (*height)++;
    if (rv == 0)
      found = 1;
    else {
      result->parent = result->node;
      if (rv > 0)
        result->node = result->node->left;
      else if (rv < 0)
        result->node = result->node->right;
    }
  }
  return found;
}

static avltreenode *avltreenode_create (lua_State *L, int idx) {
  avltreenode *node = malloc(sizeof(avltreenode));
  if (node) {
    node->left = NULL;
    node->right = NULL;
    node->parent = NULL;
    node->leftheight = 0;
    node->rightheight = 0;
    lua_pushvalue(L, idx);  /* value to be put into registry table */
    node->data = luaL_ref(L, LUA_REGISTRYINDEX);  /* store unique reference to value and pop value */
  }
  return node;
}

static int avltreenode_get_max_height (const avltreenode *node) {
  int height;
  if (node->leftheight > node->rightheight)
    height = node->leftheight;
  else
    height = node->rightheight;
  return height;
}

static void avltreenode_fix_height (avltreenode *node) {
  node->leftheight = 0;
  node->rightheight = 0;
  if (node->left)
    node->leftheight = avltreenode_get_max_height(node->left) + 1;
  if (node->right)
    node->rightheight = avltreenode_get_max_height(node->right) + 1;
}

static void avltree_rotate_left (avltree *tree, avltreenode *node) {
  avltreenode *right = node->right;
  if (node == tree->root)
    tree->root = right;
  else if (node == node->parent->left)
    node->parent->left = right;
  else
    node->parent->right = right;
  right->parent = node->parent;
  if (right->left) {
    node->right = right->left;
    node->right->parent = node;
  } else
    node->right = NULL;
  right->left = node;
  node->parent = right;
  avltreenode_fix_height(node);
  avltreenode_fix_height(right);
}

static void avltree_rotate_right (avltree *tree, avltreenode *node) {
  avltreenode *left = node->left;
  if (node == tree->root)
    tree->root = left;
  else if (node == node->parent->left)
    node->parent->left = left;
  else
    node->parent->right = left;
  left->parent = node->parent;
  if (left->right) {
    node->left = left->right;
    node->left->parent = node;
  } else
    node->left = NULL;
  left->right = node;
  node->parent = left;
  avltreenode_fix_height(node);
  avltreenode_fix_height(left);
}

static int avltreenode_get_balance_factor (const avltreenode *node) {
  return node->leftheight - node->rightheight;
}

static void avltree_rebalance (avltree *tree, avltreenode *node) {
  avltreenode *current = node;
  while (current != NULL) {
    avltreenode *parent = current->parent;
    int balance;
    avltreenode_fix_height(current);
    balance = avltreenode_get_balance_factor(current);
    if (balance == -2) {
      /* right heavy */
      const int rightbalance = avltreenode_get_balance_factor(current->right);
      if (rightbalance < 0)
        avltree_rotate_left(tree, current);
      else {
        avltree_rotate_right(tree, current->right);
        avltree_rotate_left(tree, current);
      }
    } else if (balance == 2) {
      /* left heavy */
      const int leftbalance = avltreenode_get_balance_factor(current->left);
      if (leftbalance > 0)
        avltree_rotate_right(tree, current);
      else {
        avltree_rotate_left(tree, current->left);
        avltree_rotate_right(tree, current);
      }
    }
    current = parent;
  }
}

static void avltree_add (lua_State *L, avltree *tree, int idx) {
  int height;
  avlsearchresult result;
  result.node = NULL;
  result.parent = NULL;
  (void)height;
  if (!avltree_search(L, tree, &result, idx, &height)) {
    avltreenode *node = avltreenode_create(L, idx);
    if (result.node == tree->root)
      tree->root = node;
    else {
      lua_pushvalue(L, idx);
      lua_rawgeti(L, LUA_REGISTRYINDEX, result.parent->data);
      int rv = tree->compare(L, -2, -1);
      agn_poptoptwo(L);
      if (rv < 0)
        result.parent->left = node;
      else
        result.parent->right = node;
      node->parent = result.parent;
      avltree_rebalance(tree, node);
    }
    tree->count++;
  }
}

static int avltree_find (lua_State *L, const avltree *tree, int idx, int *height) {
  avlsearchresult result;
  result.node = NULL;
  result.parent = NULL;
  if (avltree_search(L, tree, &result, idx, height))
    return result.node->data;
  return LUA_NOREF;
}


static avltreenode *avltreenode_find_min (avltreenode *node) {
  avltreenode *current = node;
  while (current->left) {
    current = current->left;
  }
  return current;
}

static void avltree_remove_node (lua_State *L, avltree *tree, avltreenode *node) {
  if (node->left && node->right) {  /* node with 2 children */
    avltreenode *successor = avltreenode_find_min(node->right);
    node->data = successor->data;
    avltree_remove_node(L, tree, successor);
  } else {
    avltreenode *parent = node->parent;
    if (node->left) {  /* node with only left child */
      if (node->parent) {
        if (node == node->parent->left) {
          node->parent->left = node->left;
          node->parent->left->parent = node->parent;
        } else {
          node->parent->right = node->left;
          node->parent->right->parent = node->parent;
        }
      } else {
        tree->root = node->left;
        tree->root->parent = NULL;
      }
    } else if (node->right) {  /* node with only right child */
      if (node->parent) {
        if (node == node->parent->left) {
          node->parent->left = node->right;
          node->parent->left->parent = node->parent;
        } else {
          node->parent->right = node->right;
          node->parent->right->parent = node->parent;
        }
      } else {
        tree->root = node->right;
        tree->root->parent = NULL;
      }
    } else {  /* node with no children */
      if (node->parent) {
        if (node == node->parent->left) {
          node->parent->left = NULL;
        } else {
          node->parent->right = NULL;
        }
      } else {
        tree->root = NULL;
      }
    }
    avltreenode_delete(L, node);
    avltree_rebalance(tree, parent);
    tree->count--;
  }
}

static int avltree_remove (lua_State *L, avltree *tree, int idx) {
  int height, rc;
  avlsearchresult result;
  result.node = NULL;
  result.parent = NULL;
  (void)height;
  rc = avltree_search(L, tree, &result, idx, &height);
  if (rc) {
    avltree_remove_node(L, tree, result.node);
  }
  return rc;
}

static size_t avltree_get_count (const avltree *tree) {
  return tree->count;
}


/* Agena Library Functions ******************************************************************************* */

typedef struct {
  avltree *tree;
  int registry;  /* registry, for attribute information, stored to index 0; 2.15.1 */
} Avl;


static int avl_new (lua_State *L) {
  Avl *a;
  a = (Avl *)lua_newuserdata(L, sizeof(Avl));
  if (!a)
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "avl.new");
  lua_setmetatabletoobject(L, -1, "avl", 0);
  agn_setutypestring(L, -1, "avl");
  lua_createtable(L, 0, 0);
  a->registry = luaL_ref(L, LUA_REGISTRYINDEX);
  a->tree = avltree_create((avltree_cmpfn)avl_compare);
  return 1;
}


static int avl_include (lua_State *L) {
  Avl *a = checkavl(L, 1);
  avltree_add(L, a->tree, 2);
  return 0;
}


static int avl_remove (lua_State *L) {
  Avl *a = checkavl(L, 1);
  lua_pushboolean(L, avltree_remove(L, a->tree, 2));
  return 1;
}


static int avl_find (lua_State *L) {
  int height, nrets, idx;
  Avl *a = checkavl(L, 1);
  nrets = 1;
  idx = avltree_find(L, a->tree, 2, &height);
  if (idx == LUA_NOREF) {
    lua_pushnil(L);
  } else {
    lua_rawgeti(L, LUA_REGISTRYINDEX, idx);
    lua_pushinteger(L, height);
    nrets++;
  }
  return nrets;
}

/* Metamethods ******************************************************************************************* */

static int mt_size (lua_State *L) {
  Avl *a = checkavl(L, 1);
  lua_pushnumber(L, avltree_get_count(a->tree));
  return 1;
}


static int mt_in (lua_State *L) {
  int height;
  Avl *a = checkavl(L, 2);
  lua_pushboolean(L, avltree_find(L, a->tree, 1, &height) != LUA_NOREF);
  return 1;
}


static int mt_notin (lua_State *L) {
  int height;
  Avl *a = checkavl(L, 2);
  lua_pushboolean(L, avltree_find(L, a->tree, 1, &height) == LUA_NOREF);
  return 1;
}


static int mt_empty (lua_State *L) {
  Avl *a = checkavl(L, 1);
  lua_pushboolean(L, avltree_get_count(a->tree) == 0);
  return 1;
}


static int mt_filled (lua_State *L) {
  Avl *a = checkavl(L, 1);
  lua_pushboolean(L, avltree_get_count(a->tree) != 0);
  return 1;
}


static int mt_gc (lua_State *L) {
  Avl *a;
  (void)L;
  lua_lock(L);
  a = checkavl(L, 1);
  avltree_delete(L, a->tree);
  luaL_unref(L, LUA_REGISTRYINDEX, a->registry);  /* delete registry table */
  lua_setmetatabletoobject(L, 1, NULL, 1);
  lua_unlock(L);
  return 0;
}


static int mt_tostring (lua_State *L) {  /* at the console, the avl tree is formatted as follows: */
  if (luaL_isudata(L, 1, "avl"))  /* 2.15.1 change */
    lua_pushfstring(L, "avl(%p)", lua_topointer(L, 1));
  else {
    void *p = lua_touserdata(L, 1);
    lua_pushfstring(L, (p != NULL) ? "userdata(%p)" : "unknown(%p)", lua_topointer(L, 1));
  }
  return 1;
}


static const struct luaL_Reg avl_avllib [] = {  /* metamethods for unrolled singly-linked lists `n' */
//  {"__index",      ulist_getitem},  /* n[p], with p the index, counting from 1 */
//  {"__writeindex", ulist_setitem},  /* n[p] := value, with p the index, counting from 1 */
  {"__gc",         mt_gc},        /* please do not forget garbage collection */
  {"__in",         mt_in},        /* `in` operator for ulists */
  {"__notin",      mt_notin},     /* `notin` operator for ulists */
  {"__size",       mt_size},      /* retrieve the number of entries in `n' */
  {"__empty",      mt_empty},     /* metamethod for `empty` operator */
  {"__filled",     mt_filled},    /* metamethod for `filled` operator */
  {"__tostring",   mt_tostring},  /* for output at the console, e.g. print(n) */
  {NULL, NULL}
};

static const luaL_Reg avllib[] = {
  {"find",         avl_find},
  {"new",          avl_new},
  {"include",      avl_include},
  {"remove",      avl_remove},
  {NULL, NULL}
};


/*
** Open avl library
*/
LUALIB_API int luaopen_avl (lua_State *L) {
  /* metamethods for AVL trees */
  luaL_newmetatable(L, "avl");
  luaL_register(L, NULL, avl_avllib);
  /* register library */
  luaL_register(L, AGENA_AVLLIBNAME, avllib);
  return 1;
}