/*
** $Id: calc.c, initiated June 20, 2008 $
** Calculus library
** See Copyright Notice in agena.h
*/

#include <stdlib.h>
#include <math.h>
#include <string.h>

#define calc_c
#define LUA_LIB

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"
#include "agncmpt.h"
#include "agnhlps.h"

#include "lstate.h"  /* for agn_get(dbl)epsilon */
#include "cephes.h"
#include "calc.h"
#include "interp.h"
#include "rlhmath.h"


#if !(defined(LUA_DOS) || defined(__OS2__) || defined(LUA_ANSI))
#define AGENA_CALCLIBNAME "calc"
LUALIB_API int (luaopen_calc) (lua_State *L);
#endif


static lua_Number aux_fnunivar (lua_State *L, lua_Number x, int issueerror, int *error) {  /* 2.21.4 */
  luaL_checkstack(L, 2, "not enough stack space");  /* 2.31.7 fix */
  lua_pushvalue(L, 1);
  lua_pushnumber(L, x);
  return agn_ncall(L, 1, error, issueerror);
}

#define checkdblarray(L, n) (LongDoubleArray *)luaL_checkudata(L, n, "calc")

static int aux_newdblarray (lua_State *L, size_t n, const char *procname) {  /* creates numeric C array userdata object, rewritten 2.9.0 */
  LongDoubleArray *a;
  if (n < 1)  /* 2.35.1 fix */
    luaL_error(L, "Error in " LUA_QS ": structure is empty.", procname);
  a = (LongDoubleArray *)lua_newuserdata(L, sizeof(LongDoubleArray));  /* push userdata on stack */
  a->v = calloc(n, sizeof(long double));
  if (!a->v)  /* 2.35.1 fix */
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", procname);
  a->size = n;
  /* attach proper metatable with __gc method to userdata; name of mt name must be different from the one of
     the numarray package to avoid crashes during gc. */
  luaL_getmetatable(L, "calc");
  lua_setmetatable(L, -2);
  return 1;  /* leave userdata on the top of the stack */
}

/* 2.8.2, there allegedly is a bug in Lua when more than 256 arguments are pushed into the userdata stack; aux_newdblarray
   pushes userdata a. */
#define aux_pushud(L,a,nops,procname) { \
  if (nops > 256) \
    luaL_error(L, "Error in " LUA_QS ": received too many coefficients.", procname); \
  aux_newdblarray(L, nops, procname); \
  a = (LongDoubleArray *)lua_touserdata(L, -1); \
}

static int arraygc (lua_State *L) {  /* __gc method */
  LongDoubleArray *a = checkdblarray(L, 1);  /* 2.9.0 */
  xfree(a->v);
  lua_setmetatabletoobject(L, 1, NULL, 0);
  return 0;
}

static const struct luaL_Reg numarray_arraylib [] = {
  {"__gc", arraygc},
  {NULL, NULL}
};


static int calc_fsum (lua_State *L) {
  /* computes the sum of f(a), ..., f(b); 0.10.0 - April 20, 2008;
     12 % faster than an Agena implementation;
     extended 0.30.0, December 30, 2009: the procedure now uses a modified Kahan summation algorithm
     to avoid roundoff errors; the modified Kahan algorithm has been developed by Kazufumi Ozawa
     and has been published in his paper `Analysis and Improvement of Kahan's Summation Algorithm`,
     available at http://ci.nii.ac.jp/naid/110002673355/en
     changed 2.34.0: changed to Kahan-Babuska summation to avoid roundoff errors, see:
     `A generalized Kahan-Babuka Summation-Algorithm` by Andreas Klein, April 21, 2005 */
  long double a, b;  /* changed to longdoubles 2.34.10 */
  int nargs;
  volatile long double s, cs, ccs, t, c, cc, x;
  s = cs = ccs = 0;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  nargs = lua_gettop(L);  /* 2.1.4 */
  /* luaL_checkstack is called by agnL_fncall */
  a = agn_checknumber(L, 2);  /* start value */
  b = agn_checknumber(L, 3);  /* stop value */
  while (a <= b) {
    x = agnL_fncall(L, 1, a, 4, nargs);
    t = s + x;
    c = (fabsl(s) >= fabsl(x)) ? (s - t) + x : (x - t) + s;
    s = t;
    t = cs + c;
    cc = (fabsl(cs) >= fabsl(c)) ? (cs - t) + c : (c - t) + cs;
    cs = t;
    ccs += cc;
    a++;
  }
  lua_pushnumber(L, s + cs + ccs);  /* no rounding in between */
  return 1;
}


static int calc_fprod (lua_State *L) {
  /* computes the product of f(a), ..., f(b); 0.29.0 - Nov 23, 2009 */
  long double a, b, r;
  int error;
  r = 1;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  while (a <= b) {
    r *= aux_fnunivar(L, a, 1, &error);
    a++;
  }
  lua_pushnumber(L, r);
  return 1;
}


/* DE-Quadrature, Numerical Automatic Integrator for Improper Integral
      method    : Double Exponential (DE) Transformation
      dimension : one
      table     : use
   functions
      intde  : integrator of f(x) over (a,b).

   written by Takuya Ooura in C and available at http://www.kurims.kyoto-u.ac.jp/~ooura/intde.html.
   See intde2.c file.
   For an explanation of the algorithm used, see: http://en.wikipedia.org/wiki/Tanh-sinh_quadrature

   Copyright(C) 1996 Takuya OOURA.
   You may use, copy, modify this code for any purpose and without fee.
   You may distribute this ORIGINAL package. */

#define NSLOTS 8000

void intdeini (int lenaw, long double tiny, long double eps, long double *aw) {
  long double efs = 0.1, hoff = 8.5;  /* adjustable parameters */
  int noff, nk, k, j;
  long double tinyln, epsln, h0, ehp, ehm, h, t, ep, em, xw, wg;
  tinyln = -tools_logl(tiny);
  epsln = 1 - tools_logl(efs*eps);
  h0 = hoff/epsln;
  ehp = tools_expl(h0);
  ehm = 1/ehp;
  aw[2] = eps;
  aw[3] = tools_expl(-ehm*epsln);
  aw[4] = tools_sqrtl(efs*eps);
  noff = 5;
  aw[noff] = 0.5;
  aw[noff + 1] = h0;
  aw[noff + 2] = M_PIO2ld*h0*0.5;
  h = 2;
  nk = 0;
  k = noff + 3;
  do {
    t = h*0.5;
    do {
      em = tools_expl(h0*t);
      ep = M_PIO2ld*em;
      em = M_PIO2ld/em;
      j = k;
      do {
        if (j >= NSLOTS - 2) break;  /* 2.34.10 invalid array access patch */
        xw = 1/(1 + tools_expl(ep - em));
        wg = xw*(1 - xw)*h0;
        aw[j] = xw;
        aw[j + 1] = wg*4;
        aw[j + 2] = wg*(ep + em);
        ep *= ehp;
        em *= ehm;
        j += 3;
      } while (ep < tinyln && j <= lenaw - 3);
      t += h;
      k += nk;
    } while (t < 1);
    h *= 0.5;
    if (nk == 0) {
      if (j > lenaw - 6) j -= 3;
      nk = j - noff;
      k += nk;
      aw[1] = nk;
    }
  } while (2*k - noff - 3 <= lenaw);
  aw[0] = k - 3;
}

static int calc_intde (lua_State *L) {
  int noff, lenawm, nk, k, j, jtmp, jm, m, klim;
  long double r, a, b, eps, err, epsh, ba, ir, xa, fa, fb, errt, errh, errd, h, iback, irback;
  /* "If there are fewer initializers in a brace-enclosed list than there are elements or members of an aggregate,
      or fewer characters in a string literal used to initialize an array of known size than there are elements
      in the array, the remainder of the aggregate shall be initialized implicitly the same as objects that have
      static storage duration." https://stackoverflow.com/questions/5636070/zero-an-array-in-c-code
      2.34.10 invalid array access patch */
  long double aw[NSLOTS] = {0};
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agnL_checknumber(L, 2);
  b = agnL_checknumber(L, 3);
  eps = agnL_optnonnegative(L, 4, 1.0e-15);  /* 2.13.0 */
  intdeini(NSLOTS, 1.0e-307, eps, aw);
  noff = 5;
  lenawm = (int)(aw[0] + 0.5);
  nk = (int)(aw[1] + 0.5);
  epsh = aw[4];
  ba = b - a;
  r = agnL_fnunicall(L, 1, (a + b)*aw[noff]);  /* 2.25.0 tweak */
  ir = r*aw[noff + 1];
  r *= aw[noff + 2];
  err = fabsl(r);
  k = nk + noff;
  j = noff;
  do {
    j += 3;
    if (j >= NSLOTS - 2) break;  /* 2.34.10 fix against invalid array access when run in Valgrind */
    xa = ba*aw[j];
    fa = agnL_fnunicall(L, 1, a + xa);  /* 2.25.0 tweak */
    fb = agnL_fnunicall(L, 1, b - xa);  /* 2.25.0 tweak */
    ir += (fa + fb)*aw[j + 1];
    fa *= aw[j + 2];
    fb *= aw[j + 2];
    r += fa + fb;
    err += fabsl(fa) + fabsl(fb);
  } while (aw[j] > epsh && j < k);
  errt = err*aw[3];
  errh = err*epsh;
  errd = 1 + 2*errh;
  jtmp = j;
  while (fabsl(fa) > errt && j < k) {  /* 2.34.10 fix against invalid array access when run in Valgrind */
    j += 3;
    if (j >= NSLOTS - 2) break;  /* 2.34.10 fix against invalid array access when run in Valgrind */
    fa = agnL_fnunicall(L, 1, a + ba*aw[j]);  /* 2.25.0 tweak */
    ir += fa*aw[j + 1];
    fa *= aw[j + 2];
    r += fa;
  }
  jm = j;
  j = jtmp;
  while (fabsl(fb) > errt && j < k) {  /* 2.34.10 fix against invalid array access when run in Valgrind */
    j += 3;
    if (j >= NSLOTS - 2) break;  /* 2.34.10 fix against invalid array access when run in Valgrind */
    fb = agnL_fnunicall(L, 1, b - ba*aw[j]);  /* 2.25.0 tweak */
    ir += fb*aw[j + 1];
    fb *= aw[j + 2];
    r += fb;
  }
  if (j < jm) jm = j;
  jm -= noff + 3;
  h = 1;
  m = 1;
  klim = k + nk;
  while (errd > errh && klim <= lenawm) {
    iback = r;
    irback = ir;
    do {
      jtmp = k + jm;
      for (j = k + 3; j < NSLOTS - 2 && j <= jtmp; j += 3) {  /* 2.34.10 invalid array access patch */
        xa = ba*aw[j];
        fa = agnL_fnunicall(L, 1, a + xa);  /* 2.25.0 tweak */
        fb = agnL_fnunicall(L, 1, b - xa);  /* 2.25.0 tweak */
        ir += (fa + fb)*aw[j + 1];
        r += (fa + fb)*aw[j + 2];
      }
      k += nk;
      j = jtmp;
      do {
        j += 3;
        if (j >= NSLOTS - 2) break;  /* 2.34.10 invalid array access patch */
        fa = agnL_fnunicall(L, 1, a + ba*aw[j]);  /* 2.25.0 tweak */
        ir += fa*aw[j + 1];
        fa *= aw[j + 2];
        r += fa;
      } while (fabsl(fa) > errt && j < k);
      j = jtmp;
      do {
        j += 3;
        if (j >= NSLOTS - 2) break;  /* 2.34.10 invalid array access patch */
        fb = agnL_fnunicall(L, 1, b - ba*aw[j]);  /* 2.25.0 tweak */
        ir += fb*aw[j + 1];
        fb *= aw[j + 2];
        r += fb;
      } while (fabsl(fb) > errt && j < k);
    } while (k < klim);
    errd = h*(fabsl(r - 2*iback) + fabsl(ir - 2*irback));
    h *= 0.5;
    m *= 2;
    klim = 2*klim - noff;
  }
  r *= h*ba;
  if (errd > errh) {
    err = -errd*(m*fabsl(ba));
  } else {
    err = err*aw[2]*(m*fabsl(ba));
  }
  if (err < 0)
    lua_pushfail(L);
  else
    lua_pushnumber(L, r);
  lua_pushnumber(L, err);
  return 2;
}


void intdeiini (int lenaw, long double tiny, long double eps, long double *aw) {
  /* ---- adjustable parameter ---- */
  long double efs = 0.1, hoff = 11.0;
  /* ------------------------------ */
  int noff, nk, k, j;
  long double tinyln, epsln, h0, ehp, ehm, h, t, ep, em, xp, xm,
    wp, wm;
  tinyln = -tools_logl(tiny);
  epsln = 1 - tools_logl(efs*eps);
  h0 = hoff/epsln;
  ehp = tools_expl(h0);
  ehm = 1/ehp;
  aw[2] = eps;
  aw[3] = tools_expl(-ehm*epsln);
  aw[4] = tools_sqrtl(efs*eps);
  noff = 5;
  aw[noff] = 1;
  aw[noff + 1] = 4*h0;
  aw[noff + 2] = 2*M_PIO4ld*h0;
  h = 2;
  nk = 0;
  k = noff + 6;
  do {
    t = h*0.5;
    do {
      em = tools_expl(h0*t);
      ep = M_PIO4ld*em;
      em = M_PIO4ld/em;
      j = k;
      do {
        if (j >= NSLOTS - 5) break;  /* 2.34.10 invalid array access patch */
        xp = tools_expl(ep - em);
        xm = 1/xp;
        wp = xp*((ep + em)*h0);
        wm = xm*((ep + em)*h0);
        aw[j] = xm;
        aw[j + 1] = xp;
        aw[j + 2] = xm*(4*h0);
        aw[j + 3] = xp*(4*h0);
        aw[j + 4] = wm;
        aw[j + 5] = wp;
        ep *= ehp;
        em *= ehm;
        j += 6;
      } while (ep < tinyln && j <= lenaw - 6);
      t += h;
      k += nk;
    } while (t < 1);
    h *= 0.5;
    if (nk == 0) {
      if (j > lenaw - 12) j -= 6;
      nk = j - noff;
      k += nk;
      aw[1] = nk;
    }
  } while (2*k - noff - 6 <= lenaw);
  aw[0] = k - 6;
}


static int calc_intdei (lua_State *L) {
  int noff, lenawm, nk, k, j, jtmp, jm, m, klim;
  long double err, eps, r, a, epsh, ir, fp, fm, errt, errh, errd, h, iback, irback;
  /* "If there are fewer initializers in a brace-enclosed list than there are elements or members of an aggregate,
      or fewer characters in a string literal used to initialize an array of known size than there are elements
      in the array, the remainder of the aggregate shall be initialized implicitly the same as objects that have
      static storage duration." https://stackoverflow.com/questions/5636070/zero-an-array-in-c-code
      2.34.10 invalid array access patch */
  long double aw[NSLOTS] = {0};
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agnL_checknumber(L, 2);
  eps = agnL_optnonnegative(L, 3, 1.0e-15);  /* 2.13.0 */
  intdeiini(NSLOTS, 1.0e-307, eps, aw);
  noff = 5;
  lenawm = (int)(aw[0] + 0.5);
  nk = (int)(aw[1] + 0.5);
  epsh = aw[4];
  r = agnL_fnunicall(L, 1, a + aw[noff]);  /* 2.25.0 tweak */
  ir = r*aw[noff + 1];
  r *= aw[noff + 2];
  err = fabsl(r);
  k = nk + noff;
  j = noff;
  do {
    j += 6;
    if (j >= NSLOTS - 5) break;  /* 2.34.10 invalid array access patch */
    fm = agnL_fnunicall(L, 1, a + aw[j]);  /* 2.25.0 tweak */
    fp = agnL_fnunicall(L, 1, a + aw[j + 1]);  /* 2.25.0 tweak */
    ir += fm*aw[j + 2] + fp*aw[j + 3];
    fm *= aw[j + 4];
    fp *= aw[j + 5];
    r += fm + fp;
    err += fabsl(fm) + fabsl(fp);
  } while (aw[j] > epsh && j < k);
  errt = err*aw[3];
  errh = err*epsh;
  errd = 1 + 2*errh;
  jtmp = j;
  while (fabsl(fm) > errt && j < k) {
    j += 6;
    if (j >= NSLOTS - 4) break;  /* 2.34.10 invalid array access patch */
    fm = agnL_fnunicall(L, 1, a + aw[j]);  /* 2.25.0 tweak */
    ir += fm*aw[j + 2];
    fm *= aw[j + 4];
    r += fm;
  }
  jm = j;
  j = jtmp;
  while (fabsl(fp) > errt && j < k) {
    j += 6;
    if (j >= NSLOTS - 5) break;  /* 2.34.10 invalid array access patch */
    fp = agnL_fnunicall(L, 1, a + aw[j + 1]);  /* 2.25.0 tweak */
    ir += fp*aw[j + 3];
    fp *= aw[j + 5];
    r += fp;
  }
  if (j < jm) jm = j;
  jm -= noff + 6;
  h = 1;
  m = 1;
  klim = k + nk;
  while (errd > errh && klim <= lenawm) {
    iback = r;
    irback = ir;
    do {
      jtmp = k + jm;
      for (j = k + 6; j <= jtmp; j += 6) {
        if (j >= NSLOTS - 5) break;  /* 2.34.10 invalid array access patch */
        fm = agnL_fnunicall(L, 1, a + aw[j]);  /* 2.25.0 tweak */
        fp = agnL_fnunicall(L, 1, a + aw[j + 1]);  /* 2.25.0 tweak */
        ir += fm*aw[j + 2] + fp*aw[j + 3];
        r += fm*aw[j + 4] + fp*aw[j + 5];
      }
      k += nk;
      j = jtmp;
      do {
        j += 6;
        if (j >= NSLOTS - 4) break;  /* 2.34.10 invalid array access patch */
        fm = agnL_fnunicall(L, 1, a + aw[j]);  /* 2.25.0 tweak */
        ir += fm*aw[j + 2];
        fm *= aw[j + 4];
        r += fm;
      } while (fabsl(fm) > errt && j < k);
      j = jtmp;
      do {
        j += 6;
        if (j >= NSLOTS - 5) break;  /* 2.34.10 invalid array access patch */
        fp = agnL_fnunicall(L, 1, a + aw[j + 1]);  /* 2.25.0 tweak */
        ir += fp*aw[j + 3];
        fp *= aw[j + 5];
        r += fp;
      } while (fabsl(fp) > errt && j < k);
    } while (k < klim);
    errd = h*(fabsl(r - 2*iback) + fabsl(ir - 2*irback));
    h *= 0.5;
    m *= 2;
    klim = 2*klim - noff;
  }
  r *= h;
  if (errd > errh) {
    err = -errd*m;
  } else {
    err *= aw[2]*m;
  }
  if (err < 0)
    lua_pushfail(L);
  else
    lua_pushnumber(L, r);
  lua_pushnumber(L, err);
  return 2;
}


#define ONEoPI          (0.31830988618379067153)   /* = 1/Pi */
#define TWOoPI          (0.63661977236758134306)   /* = 2/Pi */
#define ONESIXTH        (0.16666666666666666667)
#define PI2SQUARED      (39.47841760435743447535)  /* = (2*Pi)^2 */

#define ONEoPIld        (0.318309886183790671537767526745L)  /* = 1/Pi */
#define TWOoPIld        (0.636619772367581343075535053490L)  /* = 2/Pi */
#define ONESIXTHld      (0.166666666666666666666666666667L)
#define PI2SQUAREDld    (39.4784176043574344753379639995L)   /* = (2*Pi)^2 */
#define ONEoPISQUAREDld (0.101321183642337771443879463210L)/* = 1/Pi^2 */

void intdeoini (int lenaw, long double tiny, long double eps, long double *aw) {
  /* ---- adjustable parameter ---- */
  int lmax = 5;
  long double efs = 0.1, enoff = 0.40, pqoff = 2.9, ppoff = -0.72;
  /* ------------------------------ */
  int noff0, nk0, noff, k, nk, j, q;
  long double tinyln, epsln, frq4, per2, pp, pq, ehp, ehm, h, t,
    ep, em, tk, xw, wg, xa;
  tinyln = -tools_logl(tiny);
  epsln = 1 - tools_logl(efs*eps);
  frq4 = TWOoPIld;
  per2 = M_PIld;
  pq = pqoff/epsln;
  pp = ppoff - tools_logl(pq*pq*frq4);
  ehp = tools_expl(2*pq);
  ehm = 1/ehp;
  aw[3] = lmax;
  aw[4] = eps;
  aw[5] = tools_sqrtl(efs*eps);
  noff0 = 6;
  nk0 = 1 + (int)(enoff*epsln);
  aw[1] = nk0;
  noff = 2*nk0 + noff0;
  wg = 0;
  xw = 1;
  for (k = 1; k <= nk0; k++) {
    wg += xw;
    q = noff - 2*k;
    if (q >= NSLOTS - 1) break;  /* 2.34.10 invalid array access patch */
    aw[q] = wg;
    aw[q + 1] = xw;
    xw = xw*(nk0 - k)/k;
  }
  wg = per2/wg;
  for (k = noff0; k < NSLOTS - 1 && k <= noff - 2; k += 2) {  /* 2.34.10 invalid array access patch */
    aw[k] *= wg;
    aw[k + 1] *= wg;
  }
  xw = tools_expl(pp - 2*PIO4);
  aw[noff] = tools_sqrtl(xw*(per2*0.5));
  aw[noff + 1] = xw*pq;
  aw[noff + 2] = per2*0.5;
  h = 2;
  nk = 0;
  k = noff + 3;
  do {
    t = h*0.5;
    do {
      em = tools_expl(2*pq*t);
      ep = M_PIO4ld*em;
      em = M_PIO4ld/em;
      tk = t;
      j = k;
      do {
        if (j >= NSLOTS - 2) break;  /* 2.34.10 invalid array access patch */
        xw = tools_expl(pp - ep - em);
        wg = tools_sqrtl(frq4*xw + tk*tk);
        xa = xw/(tk + wg);
        wg = (pq*xw*(ep - em) + xa)/wg;
        aw[j] = xa;
        aw[j + 1] = xw*pq;
        aw[j + 2] = wg;
        ep *= ehp;
        em *= ehm;
        tk += 1;
        j += 3;
      } while (ep < tinyln && j <= lenaw - 3);
      t += h;
      k += nk;
    } while (t < 1);
    h *= 0.5;
    if (nk == 0) {
      if (j > lenaw - 6) j -= 3;
      nk = j - noff;
      k += nk;
      aw[2] = nk;
    }
  } while (2*k - noff - 3 <= lenaw);
  aw[0] = k - 3;
}

static int calc_intdeo (lua_State *L) {
  int lenawm, nk0, noff0, nk, noff, lmax, m, k, j, jm, l;
  long double r, a, omega, err, eps, per, perw, w02, ir, h, iback, irback, t, tk,
    xa, fm, fp, errh, s0, s1, s2, errd;
  /* "If there are fewer initializers in a brace-enclosed list than there are elements or members of an aggregate,
      or fewer characters in a string literal used to initialize an array of known size than there are elements
      in the array, the remainder of the aggregate shall be initialized implicitly the same as objects that have
      static storage duration." https://stackoverflow.com/questions/5636070/zero-an-array-in-c-code
      2.34.10 invalid array access patch */
  long double aw[NSLOTS] = {0};
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agnL_checknumber(L, 2);
  omega = agnL_optnumber(L, 3, 1);
  eps = agnL_optnonnegative(L, 4, 1.0e-15);  /* 2.13.0 */
  intdeoini(NSLOTS, 1.0e-307, eps, aw);
  lenawm = (int)(aw[0] + 0.5);
  nk0 = (int)(aw[1] + 0.5);
  noff0 = 6;
  nk = (int)(aw[2] + 0.5);
  noff = 2*nk0 + noff0;
  lmax = (int)(aw[3] + 0.5);
  eps = aw[4];
  per = 1/fabsl(omega);
  w02 = 2*aw[noff + 2];
  perw = per*w02;
  r = agnL_fnunicall(L, 1, a + aw[noff]*per);  /* 2.25.0 tweak */
  ir = r*aw[noff + 1];
  r *= aw[noff + 2];
  err = fabsl(r);
  h = 2;
  m = 1;
  k = noff;
  jm = fm = fp = errh = 0;  /* to prevent compiler warnings */
  do {
    iback = r;
    irback = ir;
    t = h*0.5;
    do {
      if (k == noff) {
        tk = 1;
        k += nk;
        j = noff;
        do {
          j += 3;
          if (j >= NSLOTS - 2) break;  /* 2.34.10 invalid array access patch */
          xa = per*aw[j];
          fm = agnL_fnunicall(L, 1, a + xa);  /* 2.25.0 tweak */
          fp = agnL_fnunicall(L, 1, a + xa + perw*tk);  /* 2.25.0 tweak */
          ir += (fm + fp)*aw[j + 1];
          fm *= aw[j + 2];
          fp *= w02 - aw[j + 2];
          r += fm + fp;
          err += fabsl(fm) + fabsl(fp);
          tk += 1;
        } while (aw[j] > eps && j < k);
        errh = err*aw[5];
        err *= eps;
        jm = j - noff;
      } else {
        tk = t;
        for (j = k + 3; j < NSLOTS - 2 && j <= k + jm; j += 3) {  /* 2.34.10 invalid array access patch */
          xa = per*aw[j];
          fm = agnL_fnunicall(L, 1, a + xa);  /* 2.25.0 tweak */
          fp = agnL_fnunicall(L, 1, a + xa + perw*tk);  /* 2.25.0 tweak */
          ir += (fm + fp)*aw[j + 1];
          fm *= aw[j + 2];
          fp *= w02 - aw[j + 2];
          r += fm + fp;
          tk += 1;
        }
        j = k + jm;
        k += nk;
      }
      while (fabs(fm) > err && j < k) {
        j += 3;
        if (j >= NSLOTS - 2) break;
        fm = agnL_fnunicall(L, 1, a + per*aw[j]);  /* 2.25.0 tweak */
        ir += fm*aw[j + 1];
        fm *= aw[j + 2];
        r += fm;
      }
      fm = agnL_fnunicall(L, 1, a + perw*tk);  /* 2.25.0 tweak */
      s2 = w02*fm;
      r += s2;
      if (fabsl(fp) > err || fabsl(s2) > err) {
        l = 0;
        for (;;) {
          l++;
          s0 = 0;
          s1 = 0;
          s2 = fm*aw[noff0 + 1];
          for (j = noff0 + 2; j < NSLOTS - 1 && j <= noff - 2; j += 2) {  /* 2.34.10 invalid array access patch */
            tk += 1;
            fm = agnL_fnunicall(L, 1, a + perw*tk);  /* 2.25.0 tweak */
            s0 += fm;
            s1 += fm*aw[j];
            s2 += fm*aw[j + 1];
          }
          if (s2 <= err || l >= lmax) break;
          r += w02*s0;
        }
        r += s1;
        if (s2 > err) err = s2;
      }
      t += h;
    } while (t < 1);
    if (m == 1) {
      errd = 1 + 2*errh;
    } else {
      errd = h*(fabsl(r - 2*iback) + fabsl(ir - 2*irback));
    }
    h *= 0.5;
    m *= 2;
  } while (errd > errh && 2*k - noff <= lenawm);
  r *= h*per;
  if (errd > errh) {
    err = -errd*per;
  } else {
    err *= per*m*0.5;
  }
  if (err < 0)
    lua_pushfail(L);
  else
    lua_pushnumber(L, r);
  lua_pushnumber(L, err);
  return 2;
}


static int calc_Si (lua_State *L) {  /* 0.32.2 */
  lua_Number x, si, ci;
  x = agn_checknumber(L, 1);
  if (x == HUGE_VAL)
    lua_pushnumber(L, 0.5*PI);
  else if (-x == HUGE_VAL)
    lua_pushnumber(L, -0.5*PI);
  else {
    sici(x, &si, &ci);
    lua_pushnumber(L, si);
  }
  return 1;
}


static int calc_Ci (lua_State *L) {  /* 0.32.2 */
  lua_Number x, si, ci;
  x = agn_checknumber(L, 1);
  if (x == HUGE_VAL)
    lua_pushnumber(L, 0);
  else if (-x == HUGE_VAL)
    lua_pushfail(L);
  else if (x < 0)
    lua_pushundefined(L);
  else {
    sici(agn_checknumber(L, 1), &si, &ci);
    lua_pushnumber(L, ci);
  }
  return 1;
}


static int calc_Ssi (lua_State *L) {  /* 0.32.2 */
  lua_Number x, si, ci;
  x = agn_checknumber(L, 1);
  if (x == HUGE_VAL)
    lua_pushnumber(L, 0);
  else if (-x == HUGE_VAL)
    lua_pushnumber(L, -PI);
  else {
    sici(x, &si, &ci);
    lua_pushnumber(L, si - 0.5*PI);
  }
  return 1;
}


static int calc_Shi (lua_State *L) {  /* 0.32.2 */
  lua_Number x, shi, chi;
  x = agn_checknumber(L, 1);
  if (-x == HUGE_VAL)
    lua_pushnumber(L, x);
  else {
    shichi(x, &shi, &chi);
    lua_pushnumber(L, shi);
  }
  return 1;
}


static int calc_Chi (lua_State *L) {  /* 0.32.2 */
  lua_Number x, shi, chi;
  x = agn_checknumber(L, 1);
  if (-x == HUGE_VAL)
    lua_pushfail(L);
  else if (x < 0)
    lua_pushundefined(L);
  else {
    shichi(agn_checknumber(L, 1), &shi, &chi);
    lua_pushnumber(L, chi);
  }
  return 1;
}


static int calc_dawson (lua_State *L) {  /* 0.32.2 */
  lua_pushnumber(L, dawsn(agn_checknumber(L, 1)));
  return 1;
}


/* Scaled Dawson Integral w_im(x) = 2*calc.dawson(x)/sqrt(Pi), 2.21.6 */
static int calc_scaleddawson (lua_State *L) {
  lua_pushnumber(L, tools_w_im(agn_checknumber(L, 1)));
  return 1;
}


static int calc_Ei (lua_State *L) {  /* 0.32.2, extended 2.1.5 for negative arguments */
  lua_Number x;
  x = agn_checknumber(L, 1);
  if (x == HUGE_VAL)
    lua_pushnumber(L, HUGE_VAL);
  else if (-x == HUGE_VAL)
    lua_pushnumber(L, 0);
  else
    lua_pushnumber(L, (x > 0) ? ei(x) : Ei(x));
  return 1;
}


static int calc_Psi (lua_State *L) {  /* 0.32.2 */
  lua_Number x;
  x = agn_checknumber(L, 1);
  lua_pushnumber(L, psi(x));
  return 1;
}


static int calc_dilog (lua_State *L) {  /* 0.32.2 */
  lua_Number x;
  x = agn_checknumber(L, 1);
  if (x == HUGE_VAL)
    lua_pushnumber(L, -HUGE_VAL);
  else
    lua_pushnumber(L, spence(x));
  return 1;
}


static int calc_fresnelc (lua_State *L) {  /* 0.32.3 */
  lua_Number x, s, c;
  x = agn_checknumber(L, 1);
  fresnl(x, &s, &c);
  lua_pushnumber(L, c);
  return 1;
}


static int calc_fresnels (lua_State *L) {  /* 0.32.3 */
  lua_Number x, s, c;
  x = agn_checknumber(L, 1);
  fresnl(x, &s, &c);
  lua_pushnumber(L, s);
  return 1;
}


/* calc.sections (f, a, b, step)

   returns all intervals where a function has a change in sign. f must be a function, a the left border of the
   main interval, b its right border, and step the step size. The return is a sequence of pairs denoting the
   found subintervals.

   Maple versions: 1997, 1998; tuned Agena version January 21, 2008;
   C Version of Agena version 29.05.2010, 0.32.3 */

static int calc_sections (lua_State *L) {
  long double xleft, xright, step, fl, fr, total;
  volatile long double s, t, c, absstep, i, eps;
  size_t seqcounter, counter;
  int places, placesstart, placesstep;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  eps = agn_gethepsilon(L);
  xleft = agn_checknumber(L, 2);
  xright = agn_checknumber(L, 3) + eps;
  step = agn_checknumber(L, 4);
  if (fabsl(step) <= eps)
    luaL_error(L, "Error in " LUA_QS ": step size |%lf| <= %lf threshold.", "calc.sections", step, eps);
  agn_createseq(L, 4);  /* result */
  fl = agnL_fnunicall(L, 1, xleft);  /* 2.25.0 tweak */
  i = xleft;
  placesstart = tools_ndigplaces(i);
  placesstep = tools_ndigplaces(step);
  places = placesstart > placesstep ? placesstart : placesstep;
  s = xleft;
  seqcounter = counter = 0;
  total = truncl(fabsl(xright - xleft)/step + 1);  /* total number of iterations, do not use floor, (int), etc ! */
  while (i <= xright || total > counter) {
    fr = agnL_fnunicall(L, 1, i + step);  /* 2.25.0 tweak */
    if (fl*fr <= 0) {
      agn_createpairnumbers(L, i, i + step);  /* 2.34.1 optimisation */
      lua_seqseti(L, -2, ++seqcounter);
    }
    counter++;
    /* since right border in the curent iteration is equal to the left border in the next
      iteration, avoid calling f twice. */
    fl = fr;
    /* apply Adapted Neumaier summation to avoid roundoff errors; in the context of a running iterator it is better than
       Kahan-Babuka. 2.34.0 */
    t = s + step;
    absstep = fabsl(step);
    c += (fabsl(s) >= absstep) ? (s - t) + step : (step - t) + s;
    s = t;
    i = s + c;
    /* Neumaier and Kx summation are inaccuarate over -step - 1 .. step + 1, so we have to adjust */
    if (L->vmsettings & 1 && fabsl(i) <= absstep + 1 + eps) {
      i = tools_roundfl(i, places, 4);
      if (fabsl(i) < eps || i == -0) i = 0;
    }
    /* internal roundoff prevention variable c and uncorrected accumulator s will be updated with the next iteration. */
  }
  return 1;  /* return result */
}


/* Modified Regula Falsi method; Maple versions: 1996, 1997; Agena port January 20, 2008/July 02, 2008
   This algorithm, taken from an unknown FORTRAN book in the 1980s, is the most accurate and the
   fastest one, applicable to a large range of function types. All other methods tested
   including checking for a change of sign are slower and more inaccurate. */

static int calc_regulafalsi (lua_State *L) {  /* switched to long doubles 2.34.10 */
  long double x1, x2, u, v, z, a, b, eps;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  eps = agnL_optnonnegative(L, 4, agn_getepsilon(L));  /* 2.24.3, default: Agena's current Eps setting (L->Eps) */
  x1 = a;
  x2 = b;
  z = AGN_NAN;
  u = agnL_fnunicall(L, 1, x1);  /* 2.24.3 tweak */
  v = agnL_fnunicall(L, 1, x2);  /* dito */
  if (fabsl(u) < eps) {
    lua_pushnumber(L, x1);
    return 1;
  }
  if (fabsl(v) < eps) {
    lua_pushnumber(L, x2);
    return 1;
  }
  while (fabsl(u - v) > eps) {
    z = x2 - v*(x2 - x1)/(v - u);
    u = v;
    x1 = x2;
    x2 = z;
    v = agnL_fnunicall(L, 1, x2);  /* 2.24.3 tweak */
  }
  if (z >= a && z <= b)
    lua_pushnumber(L, z);
  else
    lua_pushnil(L);
  return 1;
}


/* Determines the root of the univariate function f in the borders a and b and returns a number if successful,
   and null otherwise.

   In general, the function will even return accurate results where `calc.regulafalsi` does not - or even cannot
   find a root at all*) -, but the runtime behaviour compared to `calc.regulafalsi` depends on the following conditions:
   a) the interval should not be too far from the origin,
   b) the width of the interval should not be too small.
   If both conditions are met, then `calc.zeroin` can be faster than `calc.regulafalsi`.

   The algorithm uses bisection combined with linear or quadric inverse interpolation, followed by applying
   Regula Falsi to the estimate done by the previous actions.

   Algorithm designed by G. Forsythe, M. Malcolm, C. Moler, Computer methods for mathematical computations.
   M., Mir, 1980, p.180 of the Russian edition, extended by awalz.

   *) Try exp(x)*sin(x) over [20*Pi -2, 20*Pi + 2].

   Source based upon http://www.bvds.be/bvds/mat350/zeroin.c. 2.24.3 */

/*
import calc;

f := << x -> exp(x)*sin(x) >>

st := 2
maxiter := 100k;
border := 20

t := time();
to maxiter do
   for i from -border to border do
      x := calc.regulafalsi(f, i*Pi - st, i*Pi + st, Eps)
   od
od;

time()-t:

t := time();
to maxiter do
   for i from -border to border do
      y := calc.zeroin(f, i*Pi - st, i*Pi + st, Eps)
   od
od;
time()-t:
*/

#define MAXITERS 10  /* ten interpolations suffice */
static int calc_zeroin (lua_State *L) {
  long double a, b, c, fa, fb, fc, p, q, previous, next, dbleps, eps, cureps, u, v, z;  /* changed 2.34.10 */
  int cnt = MAXITERS;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  c = a;
  eps = agnL_optpositive(L, 4, agn_getepsilon(L));  /* default: Agena's current Eps setting (L->Eps) */
  dbleps = agn_getdblepsilon(L);
  fa = agnL_fnunicall(L, 1, a);
  fb = agnL_fnunicall(L, 1, b);
  fc = fa;
  if (fabsl(fa) < eps) {  /* root at endpoints ? */
    lua_pushnumber(L, a);
    return 1;
  }
  if (fabsl(fb) < eps) {
    lua_pushnumber(L, b);
    return 1;
  }
  while (cnt--) {
    previous = b - a;
    if (fabsl(fc) < fabsl(fb)) {
      a = b; b = c; c = a;
      fa = fb; fb = fc; fc = fa;
    }
    cureps = 2.0L*DBL_EPSILON*fabsl(b) + 0.5L*eps;
    next = 0.5L*(c - b);
  /* if (fabs(next) <= cureps || fb == 0.0) break; */
  if (fabsl(next) <= cureps || tools_approxl(fb, 0, eps)) break;
  if (fabsl(previous) >= cureps && fabsl(fa) > fabsl(fb)) {  /* can interpolation be done ? */
      register long double cmb, ux, uy, uxm1, ptwice;
      cmb = c - b;
      if (a == c) {  /* do linear interpolation */
        ux = fb/fa;
        p = cmb*ux;
        q = 1.0L - ux;
      } else {       /* do quadric inverse interpolation*/
        q = fa/fc;
        ux = fb/fc;
        uy = fb/fa;
        uxm1 = ux - 1.0L;
        p = uy*(cmb*q*(q - ux) - (b - a)*uxm1);
        q = (q - 1.0L)*uxm1*(uy - 1.0L);
      }
    if (p > 0.0L)
        q = -q;
      else
        p = -p;
      ptwice = 2.0L*p;
      if (ptwice < (1.5L*cmb*q - fabsl(cureps*q)) && ptwice < fabsl(previous*q))
        next = p/q;
    }
    if (fabsl(next) < cureps) {
      next = (next > 0.0L) ? cureps : -cureps;
    }
    a = b; fa = fb;
    b += next;
    fb = agnL_fnunicall(L, 1, b);
    if ((fb > 0.0L && fc > 0.0L) || (fb < 0.0L && fc < 0.0L)) {  /* do not replace by fb*fc > 0 */
      c = a; fc = fa;
    }
  }
  /* now refine the estimate using regula falsi */
  z = AGN_NAN;
  u = fa; v = fb;
  while (fabsl(u - v) > dbleps) {  /* just comparing against eps produces inaccurate results */
    z = b - v*(b - a)/(v - u);
    u = v;
    a = b;
    b = z;
    v = agnL_fnunicall(L, 1, b);
  }
  if (z != z && tools_approxl(fb, 0.0L, eps))  /* regula falsi loop has not been run */
    lua_pushnumber(L, b);
  else if (z >= a && z <= b)
    lua_pushnumber(L, z);
  else
    lua_pushnil(L);
  return 1;
}


#define aux_geteps(L, x, nargs, procname) { \
  eps = 0; \
  delta = 0; \
  sided = 0; \
  deriv = 1; \
  while (lua_ispair(L, nargs)) { \
    agn_pairgeti(L, nargs, 1); \
    if (!agn_isstring(L, -1)) { \
      agn_poptop(L); \
      luaL_error(L, "Error in " LUA_QS ": left-hand side of option must be a string.", (procname)); \
    } \
    option = lua_tostring(L, -1); \
    if (tools_streq(option, "eps")) { \
      agn_pairgeti(L, nargs, 2); \
      if (agn_isnumber(L, -1)) { \
        eps = lua_tonumber(L, -1); \
        agn_poptop(L); \
        if (eps <= 0) { \
          agn_poptoptwo(L); \
          luaL_error(L, "Error in " LUA_QS ": epsilon must be positive.", (procname)); \
        }; \
      } else  { \
        agn_poptoptwo(L); \
        luaL_error(L, "Error in " LUA_QS ": right-hand side of epsilon option must be a number.", (procname)); \
      } \
    } else if (tools_streq(option, "delta")) { \
      agn_pairgeti(L, nargs, 2); \
      if (agn_isnumber(L, -1)) { \
        delta = lua_tonumber(L, -1); \
        agn_poptop(L); \
        if (delta <= 0) { \
          agn_poptoptwo(L); \
          luaL_error(L, "Error in " LUA_QS ": delta must be positive.", (procname)); \
        }; \
      } else  { \
        agn_poptoptwo(L); \
        luaL_error(L, "Error in " LUA_QS ": right-hand side of delta option must be a number.", (procname)); \
      } \
    } else if (tools_streq(option, "side")) { \
      agn_pairgeti(L, nargs, 2); \
      if (!agn_isstring(L, -1)) { \
        agn_poptoptwo(L); \
        luaL_error(L, "Error in " LUA_QS ": right-hand side of side option must be a string.", (procname)); \
      } \
      option = lua_tostring(L, -1); \
      if (tools_streq(option, "left")) sided = 1; \
      else if (tools_streq(option, "right")) sided = 2; \
      else if (tools_streq(option, "both")) sided = 3; \
      else if (tools_streq(option, "all")) sided = 4; \
      else { \
        agn_poptop(L); \
        luaL_error(L, "Error in " LUA_QS ": unknown right-hand side value for side option.", (procname)); \
      } \
      agn_poptop(L); \
    } else if (tools_streq(option, "deriv")) { \
      agn_pairgeti(L, nargs, 2); \
      if (agn_isnumber(L, -1)) { \
        deriv = lua_tointeger(L, -1); \
        agn_poptop(L); \
        if (deriv < 0 || deriv > 3) { \
          agn_poptoptwo(L); \
          luaL_error(L, "Error in " LUA_QS ": deriv must be 0, 1, 2 or 3.", (procname)); \
        }; \
      } else  { \
        agn_poptoptwo(L); \
        luaL_error(L, "Error in " LUA_QS ": right-hand side of deriv option must be a number.", (procname)); \
      } \
    } else { \
      agn_poptop(L); \
      luaL_error(L, "Error in " LUA_QS ": unknown option.", (procname)); \
    } \
    agn_poptop(L); \
    nargs--; \
  } \
  if (eps == 0) eps = tools_matheps((x)); \
  if (delta == 0) delta = tools_matheps((x)); \
}


/* Compute the symmetric derivative of a function f at a point x.
   The algorithm is based on Conte and de Boor's `Coefficients of Newton form of polynomial of degree 3`;
   written June 09, 2008, 0.11.2. The function must be at index 1.

   DO NOT MIGRATE TO LONG DOUBLES due to results less precise than with doubles (division by h) ! */
lua_Number aux_diff (lua_State *L, lua_Number x, lua_Number eps, int deriv, int nargs, lua_Number *abserr) {  /* 2.10.2 */
  lua_Number h, q, fph, fmh, a[4], b[4];
  int i, j;
  /* luaL_checkstack is called by agnL_fncall */
  if (deriv == 0) {  /* 2.14.1 */
    *abserr = 0;
    return agnL_fncall(L, 1, x, 3, nargs);
  }
  q = 0;
  h = eps;
  for (i=0; i < 4; i++) {  /* determine appropriate epsilon step size */
    a[i] = x + (i - 2)*h;  /* patched 2.10.1 */
    b[i] = agnL_fncall(L, 1, a[i], 3, nargs);  /* 2.10.1 */
  }
  for (j=0; j < 4; j++) {  /* compute divided difference table */
    for (i=0; i < 3 - j; i++) {
      b[i] = (b[i + 1] - b[i])/(a[i + j + 1] - a[i]);
    }
  }
  for (i=0; i < 4; i++) q += b[i];
  q = fabs(q);
  if (deriv < 2 && q < 100*eps) q = 100*eps;  /* for second and higher derivative, results would be imprecise */
  h = cbrt(eps/(2*q));
  if (deriv < 2 && h > 100*eps) h = 100*eps;  /* for second and higher derivative, results would be imprecise */
  *abserr = fabs(h*h*q);  /* absolute error, previously: fabs(h*h*q*100) */
  fph = agnL_fncall(L, 1, x + h, 3, nargs);  /* 2.10.1 */
  fmh = agnL_fncall(L, 1, x - h, 3, nargs);  /* 2.10.1 */
  if (deriv == 1) {
    return (fph - fmh)/(2*h);  /* symmetric difference quotient */
  } else if (deriv == 2) {  /* second derivative, new 2.10.2 */
    lua_Number fx = agnL_fncall(L, 1, x, 3, nargs);
    return (fph - 2*fx + fmh)/(h*h);  /* new 2.10.2 */
  } else {  /* third derivative, new 2.10.2 */
    lua_Number f2ph, f2mh, f3ph, f3mh;
    f2ph = agnL_fncall(L, 1, x + 2*h, 3, nargs);
    f2mh = agnL_fncall(L, 1, x - 2*h, 3, nargs);
    f3ph = agnL_fncall(L, 1, x + 3*h, 3, nargs);
    f3mh = agnL_fncall(L, 1, x - 3*h, 3, nargs);
    return (f3mh - 8*f2mh + 13*fmh - 13*fph + 8*f2ph - f3ph)/(8*h*h*h);
  }
}

static int calc_diff (lua_State *L) {
  lua_Number x, eps, delta, abserr;
  int sided, deriv, nargs;
  const char *option;
  (void)sided;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  option = NULL;
  x = agn_checknumber(L, 2);
  nargs = lua_gettop(L);
  aux_geteps(L, x, nargs, "calc.diff");  /* set eps & delta, changed 2.9.8 */
  lua_pushnumber(L, aux_diff(L, x, eps, deriv, nargs, &abserr));  /* externalised 2.10.2 */
  lua_pushnumber(L, abserr);  /* absolute error */
  return 2;
}


/*
Differentiation using Richardson`s extrapolation, taken from http://netlib.org/textbook/mathews/chap6.f

NUMERICAL METHODS: FORTRAN Programs, (c) John H. Mathews 1994, used by kind permission.

Author of the original FORTRAN routine:
Prof. John  H.  Mathews
Department of Mathematics
California State University Fullerton
Fullerton, CA  92634

The FORTRAN routine accompanies the book `NUMERICAL METHODS for Mathematics, Science and Engineering`,
2nd Ed, 1992, Prentice Hall, Englewood Cliffs, New Jersey, 07632, U.S.A.

Algorithm 6.2 (Differentiation Using Extrapolation).
Section 6.1, Approximating the Derivative, Page 327
Agena port January 01, 2010, 0.30.1

DO NOT MIGRATE TO LONG DOUBLES due to results less precise than with doubles (division by h) ! */

#define MAX_N 2*15  /* original value has been 15 */
static int calc_xpdiff (lua_State *L) {  /* 0.32.3 */
  lua_Number fx, fph, fmh, f2ph, f2mh, f3ph, f3mh, x, d[MAX_N + 1][MAX_N + 1],
    eps, delta, deriv, err, h, relerr, small, abserr;
  int j, k, n, sided, nargs, stencil7;
  const char *option = NULL;
  (void)sided;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  /* luaL_checkstack is called by agnL_fncall */
  x = agnL_checknumber(L, 2);
  nargs = lua_gettop(L);
  aux_geteps(L, x, nargs, "calc.xpdiff");  /* set eps & delta, changed 2.10.1 */
  fx = agnL_fncall(L, 1, x, 3, nargs);
  if (deriv == 0) {  /* 2.14.1 */
    lua_pushnumber(L, fx);
    return 1;
  } else if (!isfinite(fx)) {  /* infinity or undefined ? */
    lua_pushundefined(L);
    return 1;
  }
  stencil7 = 1;  /* default method for third derivative */
  small = 1e-7;
  h = 1.0;
  n = 0;
  err = 1.0;
  relerr = 1.0;
  for (j=0; j < 10; j++) {  /* 2.10.2, avoid `undefined` */
    fph = agnL_fncall(L, 1, x + h, 3, nargs);  /* 2.10.1; h and not eps */
    fmh = agnL_fncall(L, 1, x - h, 3, nargs);  /* 2.10.1; h and not eps */
    if (isfinite(fph) && isfinite(fmh)) break;
    h /= 2;
  }
  if ( (deriv != 3) && (!isfinite(fph) || !isfinite(fmh)) ) {  /* changes 2.16.1 due to GCC warning */
    /* fallback, as call to calc.diff often computes a reliable result */
    lua_pushnumber(L, aux_diff(L, x, eps, deriv, nargs, &abserr));
    lua_pushnumber(L, abserr);
    return 2;
  }
  if (deriv == 1)
    d[0][0] = 0.5*(fph - fmh)/h;
  else if (deriv == 2)
    d[0][0] = (fph - 2*fx + fmh)/(h*h);  /* new 2.10.2, formula taken from Wikipedia, 'Symmetric derivative' */
  else {
    f2ph = agnL_fncall(L, 1, x + 2*h, 3, nargs);
    f2mh = agnL_fncall(L, 1, x - 2*h, 3, nargs);
    f3ph = agnL_fncall(L, 1, x + 3*h, 3, nargs);
    f3mh = agnL_fncall(L, 1, x - 3*h, 3, nargs);
    if (isfinite(f2ph) && isfinite(f2mh) && (!isfinite(f3ph) || !isfinite(f3mh)))
      stencil7 = 0;
    else if (!isfinite(f2ph) || !isfinite(f2mh) || !isfinite(f3ph) || !isfinite(f3mh)) {
      /* fallback, as call to calc.diff often computes a reliable result */
      lua_pushnumber(L, aux_diff(L, x, eps, deriv, nargs, &abserr));
      lua_pushnumber(L, abserr);
      return 2;
    }
    /* new 2.10.2, 7-stencil, www.holoborodko.com/pavel/numerical-methods/numerical-derivative/central-differences/
       by Pavel Holoborodko */
    if (stencil7)
      d[0][0] = (f3mh - 8*f2mh + 13*fmh - 13*fph + 8*f2ph - f3ph)/(8*h*h*h);
    else  /* use 5-stencil method, also published on Pavel Holoborodko's site */
      d[0][0] = (-f2mh + 2*fmh - 2*fph + 2*f2ph)/(2*h*h*h);
  }
  for (j=1; (j <= MAX_N) && (relerr > eps) && (err > delta); j++) {
    h /= 2;
    fph = agnL_fncall(L, 1, x + h, 3, nargs);  /* 2.10.1 */
    fmh = agnL_fncall(L, 1, x - h, 3, nargs);  /* 2.10.1 */
    if (deriv == 1)
      d[j][0] = 0.5*(fph - fmh)/h;
    else if (deriv == 2)
      d[j][0] = (fph - 2*fx + fmh)/(h*h);  /* new 2.10.2 */
    else {  /* third derivative, new 2.10.2 */
      f2ph = agnL_fncall(L, 1, x + 2*h, 3, nargs);
      f2mh = agnL_fncall(L, 1, x - 2*h, 3, nargs);
      if (stencil7) {
        f3ph = agnL_fncall(L, 1, x + 3*h, 3, nargs);
        f3mh = agnL_fncall(L, 1, x - 3*h, 3, nargs);
        d[j][0] = (f3mh - 8*f2mh + 13*fmh - 13*fph + 8*f2ph - f3ph)/(8*h*h*h);
      } else {
        d[j][0] = (-f2mh + 2*fmh - 2*fph + 2*f2ph)/(2*h*h*h);
      }
    }
    for (k=1; k <= j; k++) {
      d[j][k] = d[j][k - 1] + (d[j][k - 1] - d[j - 1][k - 1])/(tools_intpow(4, k) - 1);  /* 2.29.5 tuning */
    }
    err = fabs(d[j][j] - d[j - 1][j - 1]);
    relerr = 2*err/(fabs(d[j][j]) + fabs(d[j - 1][j - 1]) + small);
    n = j;
  }
  lua_pushnumber(L, d[n][n]);
  lua_pushnumber(L, err);
  return 2;
}



/* Computes a mathematical epsilon value that takes into account the magnitude of the value of function
   f at point x. By default, r = 3 points around (and including) the centre x are used to calculate
   the result. You may choose any other positive integer for r with r < 100.

   The optional fourth argument `adjust` - a Boolean - controls whether the function shall adjust the result if the
   internal epsilon value computed has become too large to be probably of any use. By default no adjustment takes place.

   The second return, a Boolean, indicates whether the function adjusted the result (only if `adjust` has been set
   to true).

   The function is useful to analyse functions whose values are very close to zero and where floating point arithmetic
   may lead to fatal round-off errors for it returns an epsilon value that is larger than the one returned by
   `math.eps`.

   DO NOT MIGRATE TO LONG DOUBLES due to results less precise than with doubles (division by h) ! */

static int calc_eps (lua_State *L) {  /* 2.9.8 */
  lua_Number eps, h, q, x, a[100], b[100];
  int i, j, r, n, adjust, adjustment, error;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  x = agn_checknumber(L, 2);
  r = agnL_optposint(L, 3, 3);
  if (r > 99)
    luaL_error(L, "Error in " LUA_QS ": third argument %d too big.", "calc.eps", r);
  adjust = agnL_optboolean(L, 4, 0);
  adjustment = 0;
  eps = tools_matheps(x);  /* get mathematical epsilon */
  h = eps;
  q = 0;
  n = r + 1;
  for (i=0; i < n; i++) {
    a[i] = x + (i - (int)(r/2))*h;  /* compute n points around the centre */
    b[i] = aux_fnunivar(L, a[i], 1, &error);
  }
  for (j=0; j < n; j++) {
    for (i=0; i < r - j; i++) {
      b[i] = (b[i + 1] - b[i])/(a[i + j + 1] - a[i]);
    }
  }
  for (i=0; i < n; i++) q += b[i];
  q = fabs(q);
  if ((adjustment = adjust && q < 100*eps)) q = 100*eps;
  h = sun_pow(eps/((r - 1)*q), 1.0/r, 0);
  if ((adjustment = adjust && h > 100*eps)) h = 100*eps;
  lua_pushnumber(L, h);  /* `ordinate` epsilon */
  lua_pushboolean(L, adjustment);
  return 2;
}


/* Estimates the minimum location of a univariate function through one-dimensional search over a given range [a, b] using "Golden section"
   combined with parabolic interpolation. Returns the _abscissa_ (x-axis) value where a minimum has been found. The algorithm has been
   described in:

   G.Forsythe, M.Malcolm, C.Moler, `Computer Methods for Mathematical Computations`. M., Mir, 1980, p.202 of the
   Russian edition.

   f:  function under investigation
   a:  left border of the range the min is seeked
   b:    right border of the range the min is seeked
   tol:  (optional) acceptable tolerance */
static int calc_fminbr (lua_State *L) {
  long double x, v, w, fv, fx, fw, r, a, b, tol, eps;  /* Abscissae, descr. see above, f(x), f(v), f(w), r */
  int error;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  tol = agnL_optnumber(L, 4, AGN_EPSILON);
  r = PHIINVSQld; /* = (3 - sqrt(5))/2; Golden section ratio, 2.17.7 tweak; chenged from PHI-1 back to (3 - sqrt(5))/2 2.21.4 */
  if (!(tol > 0 && b > a))
    luaL_error(L, "Error in " LUA_QS ": tolerance <= 0 or a <= b.", "calc.fminbr");
  v = a + r*(b - a);
  fv = aux_fnunivar(L, v, 0, &error);  /* First step - always gold section */
  x = v; w = v;
  fx = fv; fw = fv;
  eps = agn_getepsilon(L);  /* 2.1.3 optimisation, a and b might be changed in this loop, 2.21.8 change, 2.21.9 change */
  while (1) {  /* main iteration loop */
    long double range, middle_range, tol_act, new_step, t, fa, fb, ft;
    range = b - a;           /* Range over which the minimum is seeked for */
    middle_range = 0.5*(a + b);  /* 2.17.7 tweak */
    tol_act = eps*fabsl(x) + tol/3;  /* Actual tolerance; new_step: Step at this iteration */
    if (fabsl(x - middle_range) + 0.5L*range <= 2.0L*tol_act) {  /* 2.17.7 tweak */
      fa = aux_fnunivar(L, a, 0, &error);
      fb = aux_fnunivar(L, b, 0, &error);
      fx = aux_fnunivar(L, x, 0, &error);
      if (fa < fx) {  /* the function is not that good if the minimum is exactly at the left or right border */
        lua_pushnumber(L, a);   /* 0.29.4 patch, 28.12.2009 */
      } else if (fb < fx) {
        lua_pushnumber(L, b);
      } else {
        lua_pushnumber(L, x);  /* acceptable approximation is found */
      }
      return 1;
    }
    /* Obtain the gold section step */
    new_step = r*(x < middle_range ? b - x : a - x);
    /* Decide if the interpolation can be tried */
    if (fabsl(x - w) >= tol_act) {  /* If x and w are distinct interpolatiom may be tried */
      long double p, q, t;       /* Interpolation step is calculated as p/q; division operation is delayed */
                                 /* until last moment */
      t = (x - w)*(fx - fv);
      q = (x - v)*(fx - fw);
      p = (x - v)*q - (x - w)*t;
      q = 2.0L*(q - t);
      if (q > 0.0L)     /* q was calculated with the op- */
        p = -p;         /* posite sign; make q positive */
      else              /* and assign possible minus to */
        q = -q;         /* p */
      if (fabsl(p) < fabsl(new_step*q) &&  /* If x+p/q falls in [a,b] */
        p > q*(a - x + 2.0L*tol_act) &&    /* not too close to a and */
        p < q*(b - x - 2.0L*tol_act)) {    /* b, and isn't too large */
        new_step = p/q;                    /* it is accepted */
      }
      /* If p/q is too large then the gold section procedure can reduce [a,b] range to more extent  */
    }
    if (fabsl(new_step) < tol_act) {  /* Adjust the step to be not less */
      if (new_step > 0.0L)            /* than tolerance */
        new_step = tol_act;
      else
        new_step = -tol_act;
    }
    /* Obtain the next approximation to min and reduce the enveloping range */
    t = x + new_step;  /* Tentative point for the min */
    ft = aux_fnunivar(L, t, 0, &error);
    if (ft <= fx) {    /* t is a better approximation */
      if (t < x)       /* Reduce the range so that */
        b = x;         /* t would fall within it */
      else
        a = x;
      v = w; w = x; x = t;  /* Assign the best approx to x */
      fv = fw; fw = fx; fx = ft;
    } else {      /* x remains the better approx */
      if (t < x)  /* Reduce the range enclosing x */
        a = t;
      else
        b = t;
      if (ft <= fw || w == x) {
        v = w; w = t;
        fv = fw; fw = ft;
      } else if (ft <= fv || v == x || v == w) {
        v = t;
        fv = ft;
      }
    }
  }
}


/* Performs Golden section search for the minimum of a univariate function.

   Given a function f with a single local minimum in the interval [a, b], with a, b numbers, returns the _abscissa_
   value (x-axis) where a minimum has been found.

   Algorithm taken from https://en.wikipedia.org/wiki/Golden-section_search */

static int calc_fmings (lua_State *L) {
  long double a, b, c, d, yc, yd, eps, h;
  int i, n, error;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  if (b <= a)
    luaL_error(L, "Error in " LUA_QS ": left border must be less than right border.", "calc.fmings");
  eps = agnL_optpositive(L, 4, tools_intpow(agn_getepsilon(L), 2));
  h = b - a;
  if (h <= eps) {
    lua_pushnumber(L, (aux_fnunivar(L, a, 0, &error) < aux_fnunivar(L, b, 0, &error)) ? a : b);
    return 1;
  }
  /* Required steps to achieve tolerance */
  n = sun_ceill(tools_logl(eps/h)/tools_logl(PHIINVld));
  c = a + h*PHIINVSQld;
  d = a + h*PHIINVld;
  yc = aux_fnunivar(L, c, 0, &error);
  yd = aux_fnunivar(L, d, 0, &error);
  for (i=0; i < n; i++) {
    if (yc < yd) {
      b = d;
      d = c;
      yd = yc;
      h = PHIINVld*h;
      c = a + PHIINVSQld*h;
      yc = aux_fnunivar(L, c, 0, &error);
    } else {
      a = c;
      c = d;
      yc = yd;
      h = PHIINVld*h;
      d = a + PHIINVld*h;
      yd = aux_fnunivar(L, d, 0, &error);
    }
  }
  if (yc < yd) {
    lua_pushnumber(L, (aux_fnunivar(L, a, 0, &error) <= aux_fnunivar(L, d, 0, &error)) ? a : d);
  } else {
    lua_pushnumber(L, (aux_fnunivar(L, b, 0, &error) <= aux_fnunivar(L, c, 0, &error)) ? b : c);
  }
  return 1;
}


/* takes a table or sequence of pairs of numbers and stores them to the arrays x and y */
static void agn_toarrays (lua_State *L, int idx, const char *procname, size_t n, long double *x, long double *y) {
  size_t i;
  if (lua_type(L, idx) == LUA_TTABLE) {
    i = 0;
    lua_pushnil(L);
    while (lua_next(L, idx)) {
      agnL_pairgetilongnumbers(L, procname, -1, &(x[i]), &(y[i])); i++;
    }
  }
  else {
    for (i=0; i < n; i++) {
      lua_seqrawgeti(L, idx, i + 1);  /* push object */
      agnL_pairgetilongnumbers(L, procname, -1, &(x[i]), &(y[i]));
    }
  }
}


int nevillef (lua_State *L) {  /* 2.2.0 RC 3/2.34.1 */
  int n;
  long double t;  /* changed 2.34.10 */
  LongDoubleArray *x, *y;
  t = agn_checknumber(L, 1);
  n = agn_tonumber(L, lua_upvalueindex(3));
  x = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(1));
  y = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(2));
  lua_pushnumber(L, dneville(n, x, y, t));
  return 1;
}

static int calc_neville (lua_State *L) {
  size_t n, i;
  long double t, *x, *y;  /* changed 2.34.10 */
  luaL_argcheck(L, lua_type(L, 1) == LUA_TTABLE || lua_type(L, 1) == LUA_TSEQ, 1, "table or sequence of pairs expected");  /* 2.35.2 fix */
  n = agn_nops(L, 1);
  if (n < 1)  /* 2.35.2 fix */
    luaL_error(L, "Error in " LUA_QS ": sample is empty.", "calc.neville");
  agn_createlongarray(x, n, "calc.neville");
  agn_createlongarray(y, n, "calc.neville");
  agn_toarrays(L, 1, "calc.neville", n, x, y);
  if (lua_gettop(L) > 1) {
    t = agn_checknumber(L, 2);
    lua_pushnumber(L, neville(n, x, y, t));
  } else {  /* 2.2.0 RC 3, tuned 2.34.1 */
    LongDoubleArray *a, *b;
    aux_pushud(L, a, n, "calc.neville");  /* push as ud */
    aux_pushud(L, b, n, "calc.neville");  /* push as ud */
    for (i=0; i < n; i++) {
      setdblarray(a, i, x[i]);
      setdblarray(b, i, y[i]);
    }
    lua_pushnumber(L, n);  /* push number of elements */
    lua_pushcclosure(L, &nevillef, 3);
  }
  xfreeall(x, y);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


/* calc.polyfit(obj, n);

   Returns a sequence of coefficients of an nth-degree polynomial of a sample, in order of descending degree fitting the
   input sequence or sequence obj of pairs x[k]:y[k], with x[k] and y[k] being numbers, and using polynomial regression.
   The degree n must be a positive integer.

   The return may be passed to `calc.polygen` to generate a polynomial function (use `unpack` when passing the coefficient
   vector), e.g. calc.polygen( unpack( calc.polyfit(seq( 1:0, 1:3, 2:1 ), 2 ) ) ).

   For the original C source file see: http://programbank4u.blogspot.de/2013/04/polynomial-regression-c-program.html,
   posted 11th April 2013 by Harika. */

static int calc_polyfit (lua_State *L) {  /* 2.1.3 */
  size_t n, c;
  int m, i, j, k;
  long double *x, *y, *a, *b, sum, temp, q;  /* changed 2.34.10 */
  luaL_argcheck(L, lua_type(L, 1) == LUA_TTABLE || lua_type(L, 1) == LUA_TSEQ, 1, "table or sequence of pairs expected");  /* 2.35.2 fix */
  n = agn_nops(L, 1);
  m = agn_checkinteger(L, 2);
  if (m >= n)  /* This if/else statement intercepts empty samples */
    luaL_error(L, "Error in " LUA_QS ": too few items in structure for given degree %d.", "calc.polyfit", m);
  else if (m < 1)
    luaL_error(L, "Error in " LUA_QS ": degree must be positive.", "calc.polyfit");
  agn_createlongarray(x, n, "calc.polyfit");
  agn_createlongarray(y, n, "calc.polyfit");
  agn_toarrays(L, 1, "calc.polyfit", n, x, y);
  agn_createlongarray(a, n*n, "calc.polyfit");
  agn_createlongarray(b, n, "calc.polyfit");
  q = 0;
  for (i=0; i <= m; i++) {
    for (j=0; j <= m; j++) {
      sum = 0;
      for (k=0; k < n; k++)
        /* sum = sum + sun_pow(x[k], i + j, 1); */
        sum = tools_koaddl(sum, tools_powil(x[k], i + j), &q);  /* 2.13.0, 2.29.5 tuning */
      a[i + n*j] = sum;
    }
  }
  q = 0;
  for (i=0; i <= m; i++) {
    sum = 0;
    for (k=0; k < n; k++)
      /* sum = sum + y[k] * sun_pow(x[k], i, 1); */
      sum = tools_koaddl(sum, y[k]*tools_powil(x[k], i), &q);  /* 2.13.0, 2.29.5 tuning */
    b[i] = sum;
  }
  for (k=0; k <= m; k++) {
    for (i=0; i <= m; i++) {
      if (i == k) continue;
      temp = a[i + n*k]/a[k + n*k];
      q = 0;
      for (j=k; j <= m; j++)
        /* a[i + n*j] = a[i + n*j] - temp*a[k + n*j]; */
        a[i + n*j] = tools_koaddl(a[i + n*j], -temp*a[k + n*j], &q);  /* 2.13.0 */
      /* b[i] = b[i] - temp*b[k]; */
      q = 0;
      b[i] = tools_koaddl(b[i], -temp*b[k], &q);  /* 2.13.0 */
    }
  }
  agn_createseq(L, m + 1);
  c = 1;
  for (i=m; i >= 0; i--) {
    agn_seqsetinumber(L, -1, c++, b[i]/a[i + n*i]);
  }
  xfreeall(x, y, a, b);  /* 2.9.8 */
  return 1;
}


int newton (lua_State *L) {
  int i, n;
  long double t, r;  /* changed 2.34.10 */
  LongDoubleArray *x, *nf;
  t = agn_checknumber(L, 1);
  x = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(1));
  nf = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(2));
  n = agn_tonumber(L, lua_upvalueindex(3));
  r = getdblarray(nf, n - 1);
  for (i=n - 2; i >= 0; i--)
    r = r*(t - getdblarray(x, i)) + getdblarray(nf, i);
  lua_pushnumber(L, r);
  return 1;
}

static int calc_interp (lua_State *L) {
  size_t n, i, nargs;
  long double *x, *y, *coeff;  /* changed 2.34.10 */
  luaL_argcheck(L, lua_type(L, 1) == LUA_TTABLE || lua_type(L, 1) == LUA_TSEQ, 1, "table or sequence of pairs expected");  /* 2.35.2 fix */
  n = agn_nops(L, 1);
  if (n < 1)  /* 2.35.2 fix */
    luaL_error(L, "Error in " LUA_QS ": sample is empty.", "calc.interp");
  nargs = lua_gettop(L);
  agn_createlongarray(x, n, "calc.interp");
  agn_createlongarray(y, n, "calc.interp");
  agn_toarrays(L, 1, "calc.interp", n, x, y);
  if (nargs == 3 && lua_isseq(L, 3)) {
    size_t i;
    coeff = NULL;
    if (agn_seqsize(L, 3) != n) {
      xfreeall(x, y);  /* 2.8.2 patch, 2.9.8 */
      luaL_error(L, "Error in " LUA_QS ": number of coefficients must be equal to number of points.", "calc.interp");
    }
    agn_createlongarray(coeff, n, "calc.interp");
    for (i=0; i < n; i++)
      coeff[i] = lua_seqrawgetinumber(L, 3, i + 1);
  } else
    coeff = divdiff(n, x, y);  /* compute the coefficients of the Newton form */
  if (coeff == NULL) {
    xfreeall(x, y);  /* 2.2.0 RC 3 patch, 2.9.8 */
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.interp");
  }
  if (nargs > 1)
    lua_pushnumber(L, nf_eval(n, x, coeff, agn_checknumber(L, 2)));
  else {  /* 2.34.1: tuned 3.5 times: use userdata instead of sequences */
    LongDoubleArray *a, *c;
    aux_pushud(L, a, n, "calc.interp");  /* push x as ud */
    aux_pushud(L, c, n, "calc.interp");  /* push coefficients as ud */
    for (i=0; i < n; i++) {
      setdblarray(a, i, x[i]);
      setdblarray(c, i, coeff[i]);
    }
    lua_pushnumber(L, n);  /* push number of elements */
    lua_pushcclosure(L, &newton, 3);
  }
  xfreeall(x, y, coeff);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


static int calc_newtoncoeffs (lua_State *L) {
  size_t i, n;
  long double *x, *y, *coeff;  /* changed 2.34.10 */
  luaL_argcheck(L, lua_type(L, 1) == LUA_TTABLE || lua_type(L, 1) == LUA_TSEQ, 1, "table or sequence of pairs expected");  /* 2.35.2 fix */
  n = agn_nops(L, 1);
  if (n < 1)  /* 2.35.2 fix */
    luaL_error(L, "Error in " LUA_QS ": sample is empty.", "calc.newtoncoeffs");
  agn_createlongarray(x, n, "calc.newtoncoeffs");
  agn_createlongarray(y, n, "calc.newtoncoeffs");
  agn_toarrays(L, 1, "calc.newtoncoeffs", n, x, y);
  coeff = divdiff(n, x, y);
  if (coeff == NULL) {
    xfreeall(x, y);  /* 2.8.2 patch, 2.9.8 */
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.newtoncoeffs");
  }
  agn_createseq(L, n);
  for (i=0; i < n; i++) {
    agn_seqsetinumber(L, -1, i + 1, coeff[i]);
  }
  xfreeall(x, y, coeff);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


static void agn_nakcoeffs (lua_State *L, int mainidx, int idx, long double *x, size_t n) {  /* copies numbers in a sequence to an array */
  size_t i;
  lua_seqrawgeti(L, mainidx, idx);
  if (lua_type(L, -1) != LUA_TSEQ)
    luaL_error(L, "Error in " LUA_QS ": expected a sequence, got %s.", "calc.nakspline", luaL_typename(L, -1));
  if (agn_seqsize(L, -1) != n)
    luaL_error(L, "Error in " LUA_QS ": number of coefficients not equal to number of points, got .", "calc.nakspline");
  for (i=0; i < n; i++) {
    x[i] = lua_seqrawgetinumber(L, -1, i + 1);
  }
  agn_poptop(L);
}


static int calc_naksplinecoeffs (lua_State *L) {
  size_t n;
  long double *x, *y, *b, *c, *d;  /* changed 2.34.10 */
  luaL_argcheck(L, lua_type(L, 1) == LUA_TTABLE || lua_type(L, 1) == LUA_TSEQ, 1, "table or sequence of pairs expected");  /* 2.35.2 fix */
  n = agn_nops(L, 1);
  if (n < 4) {  /* 1.12.9 fix to prevent invalid array access in cubic_nak */
    lua_pushfail(L);
    return 1;
  }
  agn_createlongarray(x, n, "calc.naksplinecoeffs");
  agn_createlongarray(y, n, "calc.naksplinecoeffs");
  agn_createlongarray(b, n, "calc.naksplinecoeffs");
  agn_createlongarray(c, n, "calc.naksplinecoeffs");
  agn_createlongarray(d, n, "calc.naksplinecoeffs");
  agn_toarrays(L, 1, "calc.naksplinecoeffs", n, x, y);
  if (cubic_nak(n, x, y, b, c, d) == -1) {
    xfreeall(x, y, b, c, d);  /* 1.12.9 / 2.2.0 RC 3 fixes, 2.9.8 */
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.naksplinecoeffs");
  }
  agn_createseq(L, 3);
  agn_longarraytoseq(L, b, n);
  lua_seqseti(L, -2, 1);
  agn_longarraytoseq(L, c, n);
  lua_seqseti(L, -2, 2);
  agn_longarraytoseq(L, d, n);
  lua_seqseti(L, -2, 3);
  xfreeall(x, y, b, c, d);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


static int calc_clampedsplinecoeffs (lua_State *L) {
  size_t n;
  long double *x, *y, *b, *c, *d, da, db;  /* changed 2.34.10 */
  luaL_argcheck(L, lua_type(L, 1) == LUA_TTABLE || lua_type(L, 1) == LUA_TSEQ, 1, "table or sequence of pairs expected");  /* 2.35.2 fix */
  n = agn_nops(L, 1);
  if (n < 2) {  /* 1.12.9 fix to prevent invalid array access in cubic_nak */
    lua_pushfail(L);
    return 1;
  }
  lua_pushvalue(L, 2);
  agnL_pairgetilongnumbers(L, "calc.clampedsplinecoeffs", -1, &da, &db);  /* removes pair pushed on the stack */
  agn_createlongarray(x, n, "calc.clampedsplinecoeffs");
  agn_createlongarray(y, n, "calc.clampedsplinecoeffs");
  agn_createlongarray(b, n, "calc.clampedsplinecoeffs");
  agn_createlongarray(c, n, "calc.clampedsplinecoeffs");
  agn_createlongarray(d, n, "calc.clampedsplinecoeffs");
  agn_toarrays(L, 1, "calc.clampedsplinecoeffs", n, x, y);
  if (cubic_clamped(n, x, y, b, c, d, da, db) == -1) {
    xfreeall(x, y, b, c, d);  /* 1.12.9 / 2.2.0 RC 3 fix, 2.9.8 */
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.clampedsplinecoeffs");
  }
  agn_createseq(L, 3);
  agn_longarraytoseq(L, b, n);
  lua_seqseti(L, -2, 1);
  agn_longarraytoseq(L, c, n);
  lua_seqseti(L, -2, 2);
  agn_longarraytoseq(L, d, n);
  lua_seqseti(L, -2, 3);
  xfreeall(x, y, b, c, d);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


int nakspline (lua_State *L) {
  size_t i, found, n;
  long double t;  /* changed 2.34.10 */
  LongDoubleArray *x, *y, *b, *c, *d;
  i = 1;
  found = 0;
  t = agn_checknumber(L, 1);
  n = agn_tonumber(L, lua_upvalueindex(6));
  x = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(1));
  y = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(2));
  b = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(3));
  c = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(4));
  d = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(5));
  while (!found && (i < n - 1)) {
    if (t < getdblarray(x, i))
      found = 1;
    else
      i++;
  }
  i--;
  t = y->v[i] + (t - x->v[i])*(b->v[i] + (t - x->v[i])*(c->v[i] + (t - x->v[i])*d->v[i]));
  lua_pushnumber(L, t);
  return 1;
}

static int calc_nakspline (lua_State *L) {
  size_t n, i, nargs;
  long double t, *x, *y, *b, *c, *d;  /* changed 2.34.10 */
  luaL_argcheck(L, lua_type(L, 1) == LUA_TTABLE || lua_type(L, 1) == LUA_TSEQ, 1, "table or sequence of pairs expected");  /* 2.35.2 fix */
  n = agn_nops(L, 1);
  if (n < 4) {  /* 1.12.9 fix to prevent invalid array access in cubic_nak */
    lua_pushfail(L);
    return 1;
  }
  nargs = lua_gettop(L);
  agn_createlongarray(x, n, "calc.nakspline");
  agn_createlongarray(y, n, "calc.nakspline");
  agn_createlongarray(b, n, "calc.nakspline");
  agn_createlongarray(c, n, "calc.nakspline");
  agn_createlongarray(d, n, "calc.nakspline");
  agn_toarrays(L, 1, "calc.nakspline", n, x, y);
  if (nargs == 3 && lua_type(L, 3) == LUA_TSEQ) {
    if (agn_seqsize(L, 3) != 3) {
      xfreeall(x, y, b, c, d);  /* 2.2.0 RC 3 fix, 2.9.8 */
      luaL_error(L, "Error in " LUA_QS ": expected a sequence of three sequences.", "calc.nakspline");
    }
    agn_nakcoeffs(L, 3, 1, b, n);
    agn_nakcoeffs(L, 3, 2, c, n);
    agn_nakcoeffs(L, 3, 3, d, n);
  } else {
    if (cubic_nak(n, x, y, b, c, d) == -1) {
      xfreeall(x, y, b, c, d);  /* 1.12.9 / 2.2.0 RC 3 fix, 2.9.8 */
      luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.nakspline");
    }
  }
  if (nargs > 1) {
    t = agn_checknumber(L, 2);
    lua_pushnumber(L, spline_eval(n, x, y, b, c, d, t));
  } else {  /* 2.34.1 tuning, 6.6 times faster than using seqs */
    LongDoubleArray *X, *Y, *B, *C, *D;
    aux_pushud(L, X, n, "calc.nakspline");  /* push x as ud */
    aux_pushud(L, Y, n, "calc.nakspline");  /* push y as ud */
    aux_pushud(L, B, n, "calc.nakspline");  /* push b as ud */
    aux_pushud(L, C, n, "calc.nakspline");  /* push c as ud */
    aux_pushud(L, D, n, "calc.nakspline");  /* push d as ud */
    for (i=0; i < n; i++) {
      setdblarray(X, i, x[i]);
      setdblarray(Y, i, y[i]);
      setdblarray(B, i, b[i]);
      setdblarray(C, i, c[i]);
      setdblarray(D, i, d[i]);
    }
    lua_pushnumber(L, n);  /* push number of elements */
    lua_pushcclosure(L, &nakspline, 6);
  }
  xfreeall(x, y, b, c, d);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


static int calc_clampedspline (lua_State *L) {
  size_t n, i, nargs;
  long double t, *x, *y, *b, *c, *d, da, db;  /* changed 2.34.10 */
  luaL_argcheck(L, lua_type(L, 1) == LUA_TTABLE || lua_type(L, 1) == LUA_TSEQ, 1, "table or sequence of pairs expected");  /* 2.35.2 fix */
  n = agn_nops(L, 1);
  if (n < 2) {  /* 1.12.9 fix to prevent invalid array access in cubic_nak */
    lua_pushfail(L);
    return 1;
  }
  nargs = lua_gettop(L);
  if (nargs < 2) {
    luaL_error(L, "Error in " LUA_QS ": at least two arguments expected.", "calc.clampedspline");
  }
  lua_pushvalue(L, 2);
  agnL_pairgetilongnumbers(L, "calc.clampedsplinecoeffs", -1, &da, &db);  /* removes pair pushed on the stack */
  agn_createlongarray(x, n, "calc.clampedspline");
  agn_createlongarray(y, n, "calc.clampedspline");
  agn_createlongarray(b, n, "calc.clampedspline");
  agn_createlongarray(c, n, "calc.clampedspline");
  agn_createlongarray(d, n, "calc.clampedspline");
  agn_toarrays(L, 1, "calc.clampedspline", n, x, y);
  if (nargs == 4 && lua_type(L, 4) == LUA_TSEQ) {
    if (agn_seqsize(L, 4) != 3) {
      xfreeall(x, y, b, c, d);  /* 2.2.0 RC 3 fix, 2.9.8 */
      luaL_error(L, "Error in " LUA_QS ": expected a sequence of three sequences.", "calc.clampedspline");
    }
    agn_nakcoeffs(L, 4, 1, b, n);
    agn_nakcoeffs(L, 4, 2, c, n);
    agn_nakcoeffs(L, 4, 3, d, n);
  } else {
    if (cubic_clamped(n, x, y, b, c, d, da, db) == -1) {
      xfreeall(x, y, b, c, d);  /* 2.2.0 RC 3 fix, 2.9.8 */
      luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.clampedspline");
    }
  }
  if (nargs > 2) {
    t = agn_checknumber(L, 3);
    lua_pushnumber(L, spline_eval(n, x, y, b, c, d, t));
  } else {
    LongDoubleArray *X, *Y, *B, *C, *D;
    aux_pushud(L, X, n, "calc.clampedspline");  /* push x as ud */
    aux_pushud(L, Y, n, "calc.clampedspline");  /* push y as ud */
    aux_pushud(L, B, n, "calc.clampedspline");  /* push b as ud */
    aux_pushud(L, C, n, "calc.clampedspline");  /* push c as ud */
    aux_pushud(L, D, n, "calc.clampedspline");  /* push d as ud */
    for (i=0; i < n; i++) {
      setdblarray(X, i, x[i]);
      setdblarray(Y, i, y[i]);
      setdblarray(B, i, b[i]);
      setdblarray(C, i, c[i]);
      setdblarray(D, i, d[i]);
    }
    lua_pushnumber(L, n);     /* push number of elements */
    lua_pushcclosure(L, &nakspline, 6);
  }
  xfreeall(x, y, b, c, d);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


void aux_pushcoeffs (lua_State *L, LongDoubleArray *a, size_t nops, const char *procname) {
  size_t i;
  int rc;
  switch (lua_type(L, 1)) {  /* 2.34.0 extension */
    case LUA_TSEQ: {
      nops = agn_seqsize(L, 1);
      aux_pushud(L, a, nops, procname);
      for (i=0; i < nops; i++) {
        setdblarray(a, i, agn_seqrawgetinumber(L, 1, i + 1, &rc));
        if (!rc) luaL_error(L, "Error in " LUA_QS ": sequence must contain numbers.", procname);
      }
      break;
    }
    case LUA_TREG: {
      nops = agn_regsize(L, 1);
      aux_pushud(L, a, nops, procname);
      for (i=0; i < nops; i++) {
        setdblarray(a, i, agn_regrawgetinumber(L, 1, i + 1, &rc));
        if (!rc) luaL_error(L, "Error in " LUA_QS ": register must contain numbers.", procname);
      }
      break;
    }
    case LUA_TTABLE : {
      nops = agn_asize(L, 1);
      aux_pushud(L, a, nops, procname);
      for (i=0; i < nops; i++) {
        setdblarray(a, i, agn_rawgetinumber(L, 1, i + 1, &rc));
        if (!rc) luaL_error(L, "Error in " LUA_QS ": table must contain numbers.", procname);
      }
      break;
    }
    default: {
      aux_pushud(L, a, nops, procname);
      for (i=0; i < nops; i++) {  /* register coefficients in *a */
        setdblarray(a, i, agn_checknumber(L, i + 1));
      }
    }
  }
}

static int polygenfma_generator (lua_State *L) {
  long double x;  /* changed 2.34.10 */
  LongDoubleArray *a = NULL;
  x = agn_checknumber(L, 1);  /* get argument */
  a = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(1));
  lua_pushnumber(L, tools_polyevalfma(x, a, a->size));
  return 1;
}

static int calc_polygen (lua_State *L) {
  size_t nops;
  LongDoubleArray *a = NULL;  /* changed 2.34.10 */
  nops = agnL_gettop(L, "expected at least one argument", "calc.polygen");
  luaL_checkstack(L, nops, "too many arguments");
  aux_pushcoeffs(L, a, nops, "calc.polygen");
  lua_pushcclosure(L, &polygenfma_generator, 1);  /* register userdata in C closure; 2.35.0 spped-up by at least 100 % */
  return 1;
}


/* source: http://programbank4u.blogspot.de/2013/04/lagrange-interpolation-c-program.html */
static int linterp (lua_State *L) {  /* 2.1.3 */
  size_t n;
  int i, j;
  long double x0, sum, prod;  /* changed 2.34.10 */
  LongDoubleArray *x, *y;
  x0 = agn_checknumber(L, 1);
  x = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(1));
  y = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(2));
  n = x->size;
  sum = 0;
  for (i=0; i < n; i++) {
    prod = 1;
    for (j=0; j < n; j++) {
      if (j == i) continue;
      prod *= (x0 - getdblarray(x, j))/(getdblarray(x, i) - getdblarray(x, j));
    }
    sum += getdblarray(y, i)*prod;
  }
  lua_pushnumber(L, sum);
  return 1;
}

/* calc.linterp
   returns a function that conducts a Lagrange interpolation for a given sequence or table o of numeric pairs
   x:y where x and y denote a point in the plane. It is often said that Lagrange interpolation is suited for
   theoretical purposes only, since it is also very slow. */

static int calc_linterp (lua_State *L) {
  size_t n, i;
  int type;
  long double x, y;  /* changed 2.34.10 */
  LongDoubleArray *a, *b;
  type = lua_type(L, 1);
  if ((type != LUA_TSEQ && type != LUA_TTABLE) || ( n = agn_nops(L, 1) ) == 0) { /* 2.35.2 fix */
    luaL_error(L, "Error in " LUA_QS ": expected a sequence or table of one or more pairs.", "calc.linterp");
    return 1;
  }
  aux_newdblarray(L, n, "calc.linterp");  /* push userdata a */
  aux_newdblarray(L, n, "calc.linterp");  /* push userdata b */
  a = (LongDoubleArray *)lua_touserdata(L, -2);
  b = (LongDoubleArray *)lua_touserdata(L, -1);
  if (type == LUA_TSEQ) {
    for (i=1; i <= n; i++) {
      lua_seqrawgeti(L, 1, i);
      agnL_pairgetilongnumbers(L, "calc.linterp", -1, &x, &y);  /* pops the pair after evaluation */
      setdblarray(a, i - 1, x);
      setdblarray(b, i - 1, y);
    }
  } else {
    for (i=1; i <= n; i++) {
      lua_rawgeti(L, 1, i);
      agnL_pairgetilongnumbers(L, "calc.linterp", -1, &x, &y);  /* pops the pair after evaluation */
      setdblarray(a, i - 1, x);
      setdblarray(b, i - 1, y);
    }
  }
  lua_pushcclosure(L, &linterp, 2);
  return 1;
}


/* 2.1.5, Simpson-Simpson Adaptive Quadrature, taken from:
   http://www.mymathlib.com/quadrature/adaptive.html, by RLH, Copyright  2004 RLH. All rights reserved.
   Adapted for Agena. */

struct Subinterval {
  long double upper_limit;
  long double lower_limit;
  long double function[5];
  struct Subinterval *interval;
};

/* Description:

   Starting at the left-end point, a, find the min power of 2, m, so that the difference between using
   Simpson's rule and the composite Simpson's rule on the interval [a, a+(b-a)/2^m] is less than twice
   the tolerance * (length of the subinterval)/(b-a).  Then repeat the process for integrating over the
   interval [(a+(b-a)/2^m, b] until the right end point, b, is finally reached.

   The integral is then the sum of the integrals of each subinterval. If at any time, the length of the
   subinterval for which the estimates based on Simpson's rule and the composite Simpson's rule is less
   than min_h the process is terminated.

  Arguments:
     double a          The lower limit of the integration interval.
     double b          The upper limit of integration.
     double tolerance  The acceptable error estimate of the integral. The integral of a subinterval is
                       accepted when the magnitude of the estimated error falls below the pro-rated tolerance
                       of the subinterval.
     double min_h      The minimum subinterval length. If no subinterval of length > min_h is found for which
                       the estimated error falls below the pro-rated tolerance, the process terminates.

  Return:
     The integral of f(x) from a to a +  h, or `fail' if no subinterval of length > min_h was found for which
     the estimated error was less that the pro-rated error.


  Source:
     http://www.mymathlib.com/c_source/quadrature/adaptive/simpson_simpson.c, adapted by alexander walz */

static void Simpsons_Rule_Update (lua_State *L, struct Subinterval *pinterval, long double *s1, long double *s2) {
  long double h, h4;
  int error;
  h = pinterval->upper_limit - pinterval->lower_limit;
  h4 = 0.25*h;
  pinterval->function[1] = aux_fnunivar(L, pinterval->lower_limit + h4, 1, &error);
  pinterval->function[3] = aux_fnunivar(L, pinterval->upper_limit - h4, 1, &error);
  *s1 = pinterval->function[0] + 4.0*pinterval->function[2]
       + pinterval->function[4];
  *s1 *= 0.166666666666666666666667*h;
  *s2 = pinterval->function[0] + 4.0*pinterval->function[1]
       + 2.0*pinterval->function[2] + 4.0*pinterval->function[3]
       + pinterval->function[4];
  *s2 *= 0.0833333333333333333333333*h;
}

static int calc_simaptive (lua_State *L) {
  struct Subinterval *pinterval, *qinterval;
  int err, error;
  long double integral, epsilon_density, epsilon, s1, s2, a, b, tolerance, min_h;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  min_h = agnL_optnumber(L, 4, 1e-7);
  tolerance = agnL_optnumber(L, 5, 0.5*agn_getepsilon(L));  /* 2.21.8/9 change */
  integral = 0.0;
  epsilon_density = 2.0*tolerance/(b - a);
  /* Create the initial level, with lower_limit = a, upper_limit = b, and f(x) evaluated at a, b, and (a + b) / 2. */
  pinterval = (struct Subinterval*)agn_malloc(L, sizeof(struct Subinterval), "calc.simaptive", NULL);
  pinterval->interval = NULL;
  pinterval->upper_limit = b;
  pinterval->lower_limit = a;
  pinterval->function[0] = aux_fnunivar(L, pinterval->lower_limit, 1, &error);
  pinterval->function[2] = aux_fnunivar(L, 0.5*(pinterval->upper_limit + pinterval->lower_limit), 1, &error);
  pinterval->function[4] = aux_fnunivar(L, pinterval->upper_limit, 1, &error);
  /* Calculate the tolerance for the current interval. Calculate the single subinterval Simpson rule,
     and the two subintervals composite Simpson rule. */
  err = 1;
  epsilon = epsilon_density*(b - a);
  Simpsons_Rule_Update(L, pinterval, &s1, &s2);
  while (pinterval->upper_limit - pinterval->lower_limit > min_h) {
    if (fabsl(s1 - s2) < epsilon) {
      /* If the two estimates are close, then increment the integral and if we are not at the right end,
         set the left end of the new interval to the right end of the old interval and the right end of
         the new interval remains the same (as the previous right end for this interval. */
      integral += s2;
      if (pinterval->interval == NULL) {
        err = 0;
        break;
      }
      qinterval = pinterval->interval;
      qinterval->lower_limit = pinterval->upper_limit;
      qinterval->function[0] = qinterval->function[2];
      qinterval->function[2] = qinterval->function[3];
      xfree(pinterval);
      pinterval = qinterval;
    }
    else {
      /* If the two estimates are not close, then create a new interval with same left end point and right
         end point at the midpoint of the current interval. */
      qinterval = (struct Subinterval*)agn_malloc(L, sizeof(struct Subinterval), "calc.simaptive", NULL);
      qinterval->interval = pinterval;
      qinterval->lower_limit = pinterval->lower_limit;
      qinterval->upper_limit = 0.5*(pinterval->upper_limit + pinterval->lower_limit);
      qinterval->function[0] = pinterval->function[0];
      qinterval->function[2] = pinterval->function[1];
      qinterval->function[4] = pinterval->function[2];
      pinterval = qinterval;
    }
    Simpsons_Rule_Update(L, pinterval, &s1, &s2);
    epsilon = epsilon_density*(pinterval->upper_limit - pinterval->lower_limit);
  }
  while (pinterval != NULL) {
    qinterval = pinterval->interval;
    xfree(pinterval);
    pinterval = qinterval;
  }
  lua_pushnumber(L, (err) ? AGN_NAN : integral);
  return 1;
}


/* Added January 09, 2008; 0.9.0 / July 12, 2008; 0.12.0; C version: 2.1.5, 04.03.2014.
   Integration of a function on an interval using a modified bisection method. Algorithm taken from
   `Fortran 90 Kurs` by G. Schmitt, Oldenbourg Verlag, 1996, p. 119 */
static int calc_gtrap (lua_State *L) {
  int k, n, error;
  long double delta, dold, h, sum, fnew, fold, a, b, eps, fa, fb, fah;  /* 2.34.10 change to longdoubles */
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  eps = agnL_optnonnegative(L, 4, agn_getepsilon(L));  /* 2.13.0, 2.21.8/9 change */
  dold = 1;
  n = 2;
  h = (b - a)/n;
  fa =  aux_fnunivar(L, a, 1, &error);
  fah = aux_fnunivar(L, a + h, 1, &error);
  fb =  aux_fnunivar(L, b, 1, &error);
  sum = fa + 2*fah + fb;
  fold = h*sum*0.5;  /* 2.17.7 tweak */
  do {
    n = 2*n;
    h *= 0.5;  /* 2.17.7 tweak */
    for (k=1; k < n; k = k + 2) {
      sum += 2*aux_fnunivar(L, a + k*h, 1, &error);
    }
    fnew = h*sum*0.5;    /* 2.17.7 tweak */
    delta = fabsl((fnew - fold)/fnew);
    fold = fnew;
    dold = delta;
  } while (delta > eps || delta < dold);
  lua_pushnumber(L, fnew);
  return 1;
}


/* The Airy wave function returns both the first independent solution to the differential equation y"(x) = x*y and its first derivative, for any real x. */
static int calc_Ai (lua_State *L) {  /* 2.8.1 */
  lua_Number x, ai, bi, dai, dbi;
  int r;
  x = agn_checknumber(L, 1);
  r = airy(x, &ai, &dai, &bi, &dbi);
  if (r == -1)
    lua_pushfail(L);
  else {
    lua_pushnumber(L, ai);
    lua_pushnumber(L, dai);
  }
  return 1 + (r == 0);
}


/* The Airy wave function returns both the second independent solution to the differential equation y"(x) = x*y and its first derivative, for any real x. */
static int calc_Bi (lua_State *L) {  /* 2.8.1 */
  lua_Number x, ai, bi, dai, dbi;
  int r;
  x = agn_checknumber(L, 1);
  r = airy(x, &ai, &dai, &bi, &dbi);
  if (r == -1)
    lua_pushfail(L);
  else {
    lua_pushnumber(L, bi);
    lua_pushnumber(L, dbi);
  }
  return 1 + (r == 0);
}

/* Computes the Riemann Zeta function in the real domain, using the zetac function.
 *
 *                inf.
 *                 -    -x
 *   zetac(x)  =   >   k   ,   x > 1,
 *                 -
 *                k=2
 *
 * is related to the Riemann zeta function by
 *
 *  Riemann zeta(x) = zetac(x) + 1. */

static int calc_zeta (lua_State *L) { /* 2.8.1 */
  lua_pushnumber(L, zetac(agn_checknumber(L, 1)) + 1);
  return 1;
}


/* Returns the polylogarithm of order n at a real point x. The return is a number, or `fail` if n < -1 for this situation is
   not implemented.

 * The polylogarithm of order n is defined by the series
 *
 *              inf   k
 *               -   x
 *  Li (x)  =    >   ---  .
 *    n          -     n
 *              k=1   k
*/

static int calc_polylog (lua_State *L) {  /* 2.8.1 */
  int flag;
  lua_Number r;
  r = polylog(agn_checkinteger(L, 1), agn_checknumber(L, 2), &flag);
  if (flag)  /* not implemented for n < -1 */
    lua_pushfail(L);
  else
    lua_pushnumber(L, r);
  return 1;
}


/* Evaluates the exponential integral
 *
 *                 inf.
 *                   -
 *                  | |   -xt
 *                  |    e
 *      E (x)  =    |    ----  dt.
 *       n          |      n
 *                | |     t
 *                 -
 *                  1

  for non-negative n and real x. The return is a number. */

static int calc_En (lua_State *L) {  /* 2.8.1 */
  lua_pushnumber(L, expn(agn_checkinteger(L, 1), agn_checknumber(L, 2)));
  return 1;
}


/* Evaluates the incomplete gamma integral defined by
 *
 *                           x
 *                            -
 *                   1       | |  -t  a-1
 *  igam(a,x)  =   -----     |   e   t   dt.
 *                  -      | |
 *                 | (a)    -
 *                           0
 *
 *
 * In this implementation both arguments must be positive. */

static int calc_igamma (lua_State *L) {  /* 2.8.1 */
  lua_Number a, x;
  x = agn_checknumber(L, 1);
  a = agn_checknumber(L, 2);
  if (x <= 0 || a <= 0)
    lua_pushfail(L);
  else
    lua_pushnumber(L, igam(a, x));
  return 1;
}


/* Evaluates the complemented incomplete gamma integral defined by
 *
 *  igamc(a,x)   =   1 - igam(a,x)
 *
 *                            inf.
 *                              -
 *                     1       | |  -t  a-1
 *               =   -----     |   e   t   dt.
 *                    -      | |
 *                   | (a)    -
 *                             x
 *
 * In this implementation both arguments must be positive. */
static int calc_igammac (lua_State *L) {  /* 2.8.1 */
  lua_Number a, x;
  x = agn_checknumber(L, 1);
  a = agn_checknumber(L, 2);
  if (x <= 0 || a <= 0)
    lua_pushfail(L);
  else
    lua_pushnumber(L, igamc(a, x));
  return 1;
}


/* Returns the incomplete beta integral of the arguments, evaluated
 * from zero to x.  The function is defined as
 *
 *                  x
 *     -            -
 *    | (a+b)      | |  a-1     b-1
 *  -----------    |   t   (1-t)   dt.
 *   -     -     | |
 *  | (a) | (b)   -
 *                 0
 *
 * The domain of definition is 0 <= x <= 1.  In this
 * implementation a and b are restricted to positive values. */

static int calc_ibeta (lua_State *L) {  /* 2.8.2 */
  lua_Number a, b, x;
  x = agn_checknumber(L, 1);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  if (a <= 0 || b <= 0)
    lua_pushfail(L);
  else if (x < 0 || x > 1)
    lua_pushundefined(L);
  else
    lua_pushnumber(L, incbet(a, b, x));
  return 1;
}


/* Inverse of incomplete beta integral
 *
 * SYNOPSIS:
 *
 * double a, b, x, y, incbi();
 *
 * x = incbi( a, b, y );
 *
 * DESCRIPTION:
 *
 * Given y, the function finds x such that
 *
 *  incbet( a, b, x ) = y . */

static int calc_invibeta (lua_State *L) {  /* 2.8.2 */
  lua_Number y, a, b;
  y = agn_checknumber(L, 1);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  lua_pushnumber(L, incbi(a, b, y));
  return 1;
}


/* Returns the limit of a function at a given point. If the left or right limit does not exist, `undefined`
   is returned. The function returns `undefined` instead of +/- infinity if the left and the right limit are either
   +infinity or -infinity. 2.7 times faster than the Agena version.

   f := << x -> (x^3+2*x^2-4*x-8)/(x^2+4*x+4) >>, reduced to: f := << x -> x - 2 >>, singularity at x = -2 */

static int calc_limit (lua_State *L) {  /* 2.9.7, 2.34.10 swirch to longdoubles */
  long double x, a, b, eps, delta, r;
  int nargs, sided, deriv;
  const char *option = NULL;
  nargs = lua_gettop(L);
  luaL_checktype(L, 1, LUA_TFUNCTION);
  /* luaL_checkstack is called by agnL_fncall */
  x = agn_checknumber(L, 2);
  (void)delta; (void)deriv;
  aux_geteps(L, x, nargs, "calc.limit");  /* set eps, changed 2.9.8 */
  /* sun_nextafter may return undefined or a wrong finite value in function calls within the vicinity of singularities, e.g.
     f := << x -> 1/x^2 >> at x = 0, or f := << x -> (x^2-3)/(x-3) >> at x = 3 */
  a = agnL_fncall(L, 1, x - eps, 3, nargs);  /* 2.9.8 */
  b = agnL_fncall(L, 1, x + eps, 3, nargs);  /* 2.9.8 */
  switch (sided) {  /* 2.9.8 */
    case 1: {  /* left-sided limit */
      lua_pushnumber(L, fabsl(a) < eps ? 0 : a);
      return 1;
    }
    case 2: {  /* right-sided limit */
      lua_pushnumber(L, fabsl(b) < eps ? 0 : b);
      return 1;
    }
    case 3: {  /* both left and right-sided limits */
      lua_pushnumber(L, fabsl(a) < eps ? 0 : a);
      lua_pushnumber(L, fabsl(b) < eps ? 0 : b);
      return 2;
    }
  }
  if (isunordered(a, b))  /* at least one function value is undefined ? */
    lua_pushundefined(L);
  else if ( (a == HUGE_VAL && b == HUGE_VAL) || (a == -HUGE_VAL && b == -HUGE_VAL) )
    lua_pushnumber(L, a);
  else {
    if (tools_approxl(a, b, 2*eps)) {  /* 2.9.8 fix, lim sin(x), x=0, returnd undefined for check failed */
      r = (a + b)*0.5;  /* 2.17.7 tweak */
      lua_pushnumber(L, (fabsl(r) < eps) ? 0 : r);  /* apply `math.zeroin` */
    } else
      lua_pushundefined(L);
  }
  if (sided == 4) {  /* 2.9.8, return left and right-sided limit, as well */
    lua_pushnumber(L, fabsl(a) < eps ? 0 : a);
    lua_pushnumber(L, fabsl(b) < eps ? 0 : b);
  }
  return 1 + 2 * (sided == 4);
}


/* Returns `true` if a real function f is continuous at the given point x (a number), and `false` otherwise.
  Three times faster than the Agena implementation. */
static int calc_iscont (lua_State *L) {  /* 2.9.7, 2.34.10 swirch to longdoubles */
  long double x, a, b, fx, eps, delta;
  int nargs, sided, deriv;
  (void)delta; (void)sided; (void)deriv;
  const char *option = NULL;
  nargs = lua_gettop(L);
  luaL_checktype(L, 1, LUA_TFUNCTION);
  /* luaL_checkstack is called by agnL_fncall */
  x = agn_checknumber(L, 2);
  aux_geteps(L, x, nargs, "calc.iscont");  /* set eps, changed 2.9.8 */
  fx = agnL_fncall(L, 1, x, 3, nargs);  /* 2.9.8 */
  if (tools_fpisnanl(fx))  /* 2.10.1, undefined ? */
    lua_pushfalse(L);
  else {
    a = agnL_fncall(L, 1, sun_nextafterl(x, -HUGE_VAL), 3, nargs);  /* 2.9.8 */
    b = agnL_fncall(L, 1, sun_nextafterl(x, HUGE_VAL), 3, nargs);  /* 2.9.8 */
    lua_pushboolean(L, tools_fpisfinitel(a) && tools_fpisfinitel(b) && tools_approxl(a, fx, eps) && tools_approxl(fx, b, eps));  /* 2.9.8 */
  }
  return 1;
}


/* Computes the Euclidian distance, i.e. the straight-line distance, of two points (a, f(a)) and
   (b, f(b)) on a curve defined by a function f in one real, in the Euclidean plane. a, b must be numbers.
   Around twice as fast than an Agena implementation. */
static int calc_eucliddist (lua_State *L) {  /* 2.9.7, 2.34.10 change to longdoubles */
  long double a, b, fa, fb;
  int nargs;
  nargs = lua_gettop(L);
  luaL_checktype(L, 1, LUA_TFUNCTION);
  /* luaL_checkstack is called by agnL_fncall */
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  /* determine f(a) */
  fa = agnL_fncall(L, 1, a, 4, nargs);  /* 2.9.8 */
  /* determine f(b) */
  fb = agnL_fncall(L, 1, b, 4, nargs);  /* 2.9.8 */
  /* splitting the formula into several separate steps is not faster */
  lua_pushnumber(L, tools_hypotl(a - b, fa - fb));  /* 2.9.8 improvement */
  return 1;
}


/* The function checks for the differentiability of a function f at a point x. It computes the left and right-sided
   difference quotients of f at x. If both are equal and the left and right limit of f at x is equal to f(x), the
   function returns `true`, and `false` otherwise. */
static int calc_isdiff (lua_State *L) {  /* 2.10.1, 2.34.10 change to longdoubles */
  long double x, fx, eps, delta, fph, fmh, f1ph, f1mh;
  int nargs, sided, deriv;
  (void)delta; (void)sided; (void)deriv;  /* to prevent compiler warnings */
  const char *option = NULL;
  nargs = lua_gettop(L);
  luaL_checktype(L, 1, LUA_TFUNCTION);
  /* luaL_checkstack is called by agnL_fncall */
  x = agnL_checknumber(L, 2);
  aux_geteps(L, x, nargs, "calc.isdiff");  /* set eps */
  /* check if f is defined at x */
  fx = agnL_fncall(L, 1, x, 3, nargs);
  if (!isfinite(fx)) {  /* infinity or undefined ? */
    lua_pushfalse(L);
    return 1;
  }
  fmh = agnL_fncall(L, 1, x - eps, 3, nargs);
  f1mh = (fx - fmh)/eps;  /* left-sided difference quotient */
  fph = agnL_fncall(L, 1, x + eps, 3, nargs);
  f1ph = (fph - fx)/eps;  /* right-sided difference quotient */
  lua_pushboolean(L, tools_approxl(f1mh, f1ph, eps) && tools_approxl(fmh, fx, eps) && tools_approxl(fph, fx, eps));
  return 1;
}


static long double chebytwobynops[101] = {
  0.0L,
  2.0000000000000000000000L,
  1.0000000000000000000000L,
  0.6666666666666666666667L,
  0.5000000000000000000000L,
  0.4000000000000000000000L,
  0.3333333333333333333333L,
  0.2857142857142857142857L,
  0.2500000000000000000000L,
  0.2222222222222222222222L,
  0.2000000000000000000000L,
  0.1818181818181818181818L,
  0.1666666666666666666667L,
  0.1538461538461538461538L,
  0.1428571428571428571429L,
  0.1333333333333333333333L,
  0.1250000000000000000000L,
  0.1176470588235294117647L,
  0.1111111111111111111111L,
  0.1052631578947368421053L,
  0.1000000000000000000000L,
  0.0952380952380952380952L,
  0.0909090909090909090909L,
  0.0869565217391304347826L,
  0.0833333333333333333333L,
  0.0800000000000000000000L,
  0.0769230769230769230769L,
  0.0740740740740740740741L,
  0.0714285714285714285714L,
  0.0689655172413793103448L,
  0.0666666666666666666667L,
  0.0645161290322580645161L,
  0.0625000000000000000000L,
  0.0606060606060606060606L,
  0.0588235294117647058824L,
  0.0571428571428571428571L,
  0.0555555555555555555556L,
  0.0540540540540540540541L,
  0.0526315789473684210526L,
  0.0512820512820512820513L,
  0.0500000000000000000000L,
  0.0487804878048780487805L,
  0.0476190476190476190476L,
  0.0465116279069767441860L,
  0.0454545454545454545455L,
  0.0444444444444444444444L,
  0.0434782608695652173913L,
  0.0425531914893617021277L,
  0.0416666666666666666667L,
  0.0408163265306122448980L,
  0.0400000000000000000000L,
  0.0392156862745098039216L,
  0.0384615384615384615385L,
  0.0377358490566037735849L,
  0.0370370370370370370370L,
  0.0363636363636363636364L,
  0.0357142857142857142857L,
  0.0350877192982456140351L,
  0.0344827586206896551724L,
  0.0338983050847457627119L,
  0.0333333333333333333333L,
  0.0327868852459016393443L,
  0.0322580645161290322581L,
  0.0317460317460317460317L,
  0.0312500000000000000000L,
  0.0307692307692307692308L,
  0.0303030303030303030303L,
  0.0298507462686567164179L,
  0.0294117647058823529412L,
  0.0289855072463768115942L,
  0.0285714285714285714286L,
  0.0281690140845070422535L,
  0.0277777777777777777778L,
  0.0273972602739726027397L,
  0.0270270270270270270270L,
  0.0266666666666666666667L,
  0.0263157894736842105263L,
  0.0259740259740259740260L,
  0.0256410256410256410256L,
  0.0253164556962025316456L,
  0.0250000000000000000000L,
  0.0246913580246913580247L,
  0.0243902439024390243902L,
  0.0240963855421686746988L,
  0.0238095238095238095238L,
  0.0235294117647058823529L,
  0.0232558139534883720930L,
  0.0229885057471264367816L,
  0.0227272727272727272727L,
  0.0224719101123595505618L,
  0.0222222222222222222222L,
  0.0219780219780219780220L,
  0.0217391304347826086957L,
  0.0215053763440860215054L,
  0.0212765957446808510638L,
  0.0210526315789473684211L,
  0.0208333333333333333333L,
  0.0206185567010309278351L,
  0.0204081632653061224490L,
  0.0202020202020202020202L,
  0.0200000000000000000000L
};

static int cheby (lua_State *L) {  /* 2.10.4, switched to longdoubles 2.34.10 */
  long double p, q, x, y, y2, d0, d1, d2, e0, e1, e2, f0, f1, f2, g0, g1, g2, h0, h1, h2, i0, i1, i2;
  int i, deriv;
  LongDoubleArray *a;  /* changed 2.34.10 */
  x = agn_checknumber(L, 1);  /* get argument */
  a = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(1));
  p = getdblarray(a, a->size);
  q = getdblarray(a, a->size + 1);
  deriv = (int)getdblarray(a, a->size + 2);
  d1 = d2 = e1 = e2 = f1 = f2 = g1 = g2 = h1 = h2 = i1 = i2 = 0;
  d0 = e0 = f0 = g0 = h0 = i0 = getdblarray(a, a->size - 1);
  y = (2.0*x - p - q)/(q - p);
  for (i=a->size - 2; i >= 0; i--) {
    d2 = d1; d1 = d0;
    y2 = 2.0*y;
    d0 = y2*d1 - d2 + getdblarray(a, i);  /* term 2*y*d1 with factor 2 necessary for deriv = 0 */
    if (deriv > 0 && i > 0) {  /* first derivative, `fall through` */
      e2 = e1; e1 = e0;
      e0 = d0 - e2 + y2*e1;
    }  /* continue statements slow down the function */
    if (deriv > 1 && i > 1) {  /* second derivative, `fall through` */
      f2 = f1; f1 = f0;
      f0 = e0 - f2 + y2*f1;
    }  /* continue statements slow down the function */
    if (deriv > 2 && i > 2) {  /* third derivative, `fall through` */
      g2 = g1; g1 = g0;
      g0 = f0 - g2 + y2*g1;
    }  /* continue statements slow down the function */
    if (deriv > 3 && i > 3) {  /* fourth derivative, `fall through` */
      h2 = h1; h1 = h0;
      h0 = g0 - h2 + y2*h1;
    }  /* continue statements slow down the function */
    if (deriv > 4 && i > 4) {  /* fifth derivative, `fall through` */
      i2 = i1; i1 = i0;
      i0 = h0 - i2 + y2*i1;
    }
  }
  switch (deriv) {  /* ? factor according to http://oeis.org/A038110 ? */
    /* substituting call to tools_intpow with z*z*.. is not faster */
    case 0:
      lua_pushnumber(L,   0.5*(d0 - d2)); break;  /* fabs(q - p)^0 = 1 */
    case 1:
      lua_pushnumber(L,     2*(e0 - e2)/fabsl(q - p)); break;  /* fabs(q - p)^1 = fabs(q - p) */
    case 2:
      lua_pushnumber(L,    16*(f0 - f2)/tools_powil(fabsl(q - p), 2)); break;
    case 3:
      lua_pushnumber(L,   192*(g0 - g2)/tools_powil(fabsl(q - p), 3)); break;
    case 4:
      lua_pushnumber(L,  3072*(h0 - h2)/tools_powil(fabsl(q - p), 4)); break;
    case 5:
      lua_pushnumber(L, 61440*(i0 - i2)/tools_powil(fabsl(q - p), 5)); break;
    default:  /* should not happen */
      luaL_error(L, "Error in " LUA_QS ": cannot compute derivative.", "calc.cheby");
  }
  return 1;
}

/* Returns a function computing the Chebyshev interpolant for a given point. `f` is the function to be interpolated,
   p and q represent the domain of the definition, `n` is the order of the interpolant. As a rule of thumb, the larger
   the domain, the larger n should be.

   You may optionally pass the `deriv=k` option as the very last argument to compute the first (k=1), second (k=2)
   or third (k=3) derivative, where k defaults to 0, i.e. the function itself.

   Using this function may speed up numeric computations significantly if the expression to be evaluated consists of
   many subexpressions - and if accuracy is not of primary concern. See also: calc.chebycoeffs.
   The function is thrice as fast as an Agena implementation. */
static int calc_cheby (lua_State *L) {  /* 2.10.4, switched to longdoubles 2.34.10 */
  long double p, q, x, k, f, *y;
  int i, j, nops, nargs, deriv;
  const char *option;
  LongDoubleArray *a;
  deriv = 0;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  /* luaL_checkstack is called by agnL_fncall */
  p = agn_checknumber(L, 2);
  q = agn_checknumber(L, 3);
  if (p > q)
    luaL_error(L, "Error in " LUA_QS ": invalid domain %LF .. %LF.", "calc.cheby", p, q);
  nops = agn_checkposint(L, 4);
  nargs = lua_gettop(L);
  if (lua_ispair(L, nargs)) {
    agn_pairgeti(L, nargs, 1);
    if (!agn_isstring(L, -1)) {
      agn_poptop(L);
      luaL_error(L, "Error in " LUA_QS ": left-hand side of option must be a string.", "calc.cheby");
    }
    option = lua_tostring(L, -1);
    if (tools_streq(option, "deriv")) {  /* 2.16.12 tweak */
      agn_pairgeti(L, nargs, 2);
      if (agn_isnumber(L, -1)) {
        deriv = lua_tointeger(L, -1);
        agn_poptop(L);
        if (deriv < 0 || deriv > 5) {
          agn_poptoptwo(L); \
          luaL_error(L, "Error in " LUA_QS ": deriv must be in 1 .. 5.", "calc.cheby");
        };
      } else  {
        agn_poptoptwo(L);
        luaL_error(L, "Error in " LUA_QS ": right-hand side of deriv option must be a number.", "calc.cheby");
      }
    } else {
      agn_poptop(L);
      luaL_error(L, "Error in " LUA_QS ": unknown option.", "calc.cheby");
    }
    nargs--;
  }
  /* pushes LongDoubleArray userdata of Chebyshev coefficient vector onto stack with three additional information, see below */
  aux_pushud(L, a, nops + 3, "calc.cheby");
  a->size = nops;
  if (a->size - deriv < 2)
    luaL_error(L, "Error in " LUA_QS ": vector would have too few coefficients.", "calc.cheby");
  y = malloc(nops*sizeof(long double));
  k = M_PIld/(2.0*nops);
  for (i=0; i < nops; i++) {
    /* x = (p + q)/2 + cos(k*(2*i + 1))*(q - p)/2.0; */
    x = 0.5*(p + q + sun_cosl(k*(2*i + 1))*(q - p));  /* changed 2.10.4, 2.34.10 */
    y[i] = agnL_fncall(L, 1, x, 5, nargs);
  }
  f = (nops <= 100) ? chebytwobynops[nops] : 2.0/nops;
  for (i=0; i < nops; i++) {
    setdblarray(a, i, 0.0);
    for (j=0; j < nops; j++) {
      a->v[i] += y[j]*sun_cosl(k*(i*(2*j + 1)));
    }
    a->v[i] *= f; /* 2.0/nops; */
  }
  xfree(y);
  setdblarray(a, nops++, p);  /* pack domain p .. q into LongDoubleArray just for speed ... */
  setdblarray(a, nops++, q);
  setdblarray(a, nops, deriv);  /* ... and derivative to be computed, as well */
  lua_pushcclosure(L, &cheby, 1);
  return 1;
}


#define DLong ieee_long_double_shape_type

/* Computes a SavitzkyGolay filter for the univariate function f to smooth its data by returning a factory interpolating f
   at a point.

   It fits successive subsets of neighbouring data with a low-degree polynomial using the linear least-square method. By
   default, 15 equally-spaced points to the left of x and 15 equally-spaced points points to the right of x are examined.
   You can change this `window` by passing another odd value with the 'points' option. All adjacent points are separated by
   distance eps which is 1e-5 by default. You can change the distance with the 'eps' option.

   Alternatively it can also compute derivatives of any degree n by passing the option deriv = <n>. The larger the degree n
   of the derivative, however, the least accurate the results will become.

   The degree d of the smoothing least-square polynomial is 3 by default and can be changed by the degree=<d> option. Recommended
   degrees are d = 2 or 4, with d not exceeding 6.

   The function automatically determines the most suitable settings for the window and the spacing eps of its points,
   but you can switch this off by passing the adaptive=false option.

   See also: calc.savgolcoeffs. */
static int savgol (lua_State *L) {  /* 2.10.4, switched to longdoubles 2.34.10 */
  long double x, signal, eps;
  int i, c, fact, nleft, nright, deriv;
  LongDoubleArray *a;
  /* luaL_checkstack is called by agnL_fncall */
  x = agn_checknumber(L, 1);  /* get argument */
  a = (LongDoubleArray *)lua_touserdata(L, lua_upvalueindex(1));
  nleft = getdblarray(a, a->size);
  nright = getdblarray(a, a->size + 1);
  deriv = getdblarray(a, a->size + 2);
  eps = getdblarray(a, a->size + 3);
  signal = 0;
  c = 0;
  for (i=nleft; i <= nright; i++) {
    signal += getdblarray(a, c++)*agnL_fncall(L, lua_upvalueindex(2), x + i*eps, 10, 9);  /* not a multi-variate function */
  };
  /* for formula below see: www.maths.lth.se/matstat/wafo/documentation/wafodoc/wafo/misc/savgol.html
     compute denominator separately to avoid overflows ((signal * n!) / ...) */
  fact = 1;
  for (i=1; i <= deriv; i++) fact *= i;
  lua_pushnumber(L, signal*(fact/tools_powil(eps, deriv)));  /* 2.34.10 improvement */
  return 1;
}

static int calc_savgol (lua_State *L) {  /* 2.10.4, 2.34.10 */
  int i, npoints, nleft, nright, polydeg, isadaptive, deriv, nops, nargs;
  long double eps, factor;
  const char *option;
  LongDoubleArray *a;
  eps = 1e-5;      /* distance between sample points */
  npoints = 31;    /* number of neighbouring points to be examined */
  deriv = 0;       /* n-th derivative, default: compute smoothed function */
  polydeg = 3;     /* polynomial degree for smoothing */
  isadaptive = 1;  /* use adaptive measures to boost precision depending on the degree of the derivative */
  luaL_checktype(L, 1, LUA_TFUNCTION);
  nargs = lua_gettop(L);
  while (lua_ispair(L, nargs)) {
    agn_pairgeti(L, nargs, 1);
    if (!agn_isstring(L, -1)) {
      agn_poptop(L);
      luaL_error(L, "Error in " LUA_QS ": left-hand side of option must be a string.", "calc.savgol");
    }
    option = lua_tostring(L, -1);
    if (tools_streq(option, "deriv")) {  /* 2.16.12 tweak */
      agn_pairgeti(L, nargs, 2);
      if (agn_isnumber(L, -1)) {
        deriv = lua_tointeger(L, -1);
        agn_poptop(L);
        if (deriv < 0) {
          agn_poptoptwo(L);
          luaL_error(L, "Error in " LUA_QS ": deriv must be non-negative.", "calc.savgol");
        };
      } else {
        agn_poptoptwo(L);
        luaL_error(L, "Error in " LUA_QS ": right-hand side of deriv option must be a number.", "calc.savgol");
      }
    } else if (tools_streq(option, "points")) {  /* 2.16.12 tweak */
       agn_pairgeti(L, nargs, 2);
       if (agn_isnumber(L, -1)) {
         npoints = lua_tointeger(L, -1);
         agn_poptop(L);
         if (npoints < 1 || npoints % 2 == 0) {
           agn_poptoptwo(L);
           luaL_error(L, "Error in " LUA_QS ": number of points must be positive and odd.", "calc.savgol");
         };
       } else {
         agn_poptoptwo(L);
         luaL_error(L, "Error in " LUA_QS ": right-hand side of points option must be a number.", "calc.savgol");
       }
     } else if (tools_streq(option, "eps")) {  /* 2.16.12 tweak */
       agn_pairgeti(L, nargs, 2);
       if (agn_isnumber(L, -1)) {
         eps = agn_tonumber(L, -1);
         agn_poptop(L);
         if (eps <= 0) {
           agn_poptoptwo(L);
           luaL_error(L, "Error in " LUA_QS ": eps must be positive.", "calc.savgol");
         }
       } else {
         agn_poptoptwo(L);
         luaL_error(L, "Error in " LUA_QS ": right-hand side of eps option must be a number.", "calc.savgol");
       }
     } else if (tools_streq(option, "degree")) {  /* 2.16.12 tweak */
       agn_pairgeti(L, nargs, 2);
       if (agn_isnumber(L, -1)) {
         polydeg = lua_tointeger(L, -1);
         agn_poptop(L);
         if (polydeg <= 0) {
           agn_poptoptwo(L); \
           luaL_error(L, "Error in " LUA_QS ": degree must be positive.", "calc.savgol");
         };
       } else {
         agn_poptoptwo(L);
         luaL_error(L, "Error in " LUA_QS ": right-hand side of degree option must be a number.", "calc.savgol");
       }
     } else if (tools_streq(option, "adaptive")) {  /* 2.16.12 tweak */
      agn_pairgeti(L, nargs, 2);
      if (agn_isnumber(L, -1)) {
        isadaptive = lua_toboolean(L, -1);
        agn_poptop(L);
      } else {
        agn_poptoptwo(L);
        luaL_error(L, "Error in " LUA_QS ": right-hand side of adaptive option must be a boolean.", "calc.savgol");
      }
    } else {
      agn_poptop(L);
      luaL_error(L, "Error in " LUA_QS ": unknown option.", "calc.savgol");
    }
    nargs--;
  }
  factor = (isadaptive == 1) ? deriv + 1 : 1;
  eps *= factor;
  nright = 0.5*npoints*factor;  /* 2.17.7 tweak */
  nleft = -nright;
  luaL_checkstack(L, 5, "not enough stack space");  /* 2.31.7 fix */
  if (agnL_gettablefield(L, "calc", "savgolcoeffs", "calc.savgol", 1) != LUA_TFUNCTION)
    luaL_error(L, "Error in " LUA_QS ": could not find `calc.savgolcoeffs`.", "calc.savgol");
  lua_pushinteger(L, -nleft);
  lua_pushinteger(L, nright);
  lua_pushinteger(L, deriv);
  lua_pushinteger(L, polydeg + deriv*isadaptive);
  lua_call(L, 4, 1);
  if (!lua_isreg(L, -1))
    luaL_error(L, "Error in " LUA_QS ": coefficient vector is invalid.", "calc.savgol");
  nops = agn_regsize(L, -1);
  /* pushes LongDoubleArray userdata of Savgol coefficient vector onto stack with four additional information, see below */
  aux_pushud(L, a, nops + 4, "calc.savgol");
  a->size = nops;
  for (i=0; i < nops; i++) {
    agn_reggeti(L, -2, i + 1);
    setdblarray(a, i, ((DLong *)luaL_checkudata(L, -1, "longdouble"))->value);
    agn_poptop(L);
  }
  lua_remove(L, -2);
  setdblarray(a, nops++, nleft);
  setdblarray(a, nops++, nright);
  setdblarray(a, nops++, deriv);
  setdblarray(a, nops, eps);
  lua_pushvalue(L, 1);  /* push function */
  lua_pushcclosure(L, &savgol, 2);
  return 1;
}


/* The function receives a non-negative integer n and any number x and returns 0 if x < 0, 1 if x > 1, and smoothly interpolates,
   using an (2*n+1)'th-degree Hermite polynomial, between 0 and 1, otherwise.

   The slope of the smoothstep function is zero at both edges, so the result is differentiable over the whole real domain.

   Wikipedia: `Smoothstep is a family of sigmoid-like interpolation and clamping functions commonly used in computer graphics and
   video game engines`, for example to naturally accelerate or decelerate an object.

   See also: http://sol.gfxile.net/interpolation, https://en.wikipedia.org/wiki/Smoothstep */
static int calc_smoothstep (lua_State *L) {  /* 2.10.4 */
  long double x, r;
  lua_Integer a;
  int n, KenPerlin;
  x = agn_checknumber(L, 1);
  if (agn_isstring(L, 2) && tools_streq(agn_tostring(L, 2), "perlin")) {  /* 2.14.1, 2.16.12 tweak */
    a = 0; KenPerlin = 1;
  } else {
    a = agn_checknonnegint(L, 2); KenPerlin = 0;
  }
  r = 0;
  if (x < 0) r = 0;
  else if (x > 1) r = 1;
  else {
    /* 2.14.1, see: https://en.wikipedia.org/wiki/Smoothstep: "[Kenneth] Perlin suggests an improved version of the smoothstep function,
       which has zero 1st- and 2nd-order derivatives at x = 0 and x = 1" */
    if (KenPerlin) {
      r = x*x*x*(x*(x*6 - 15) + 10);
    } else {
      for (n=0; n < a; n++) {
        /* for formula, see: https://stackoverflow.com/questions/41195063/general-smoothstep-equation */
        r += tools_powil(-1, n)*tools_powil(x, a + n) *
             tools_binomial(a + n - 1, n)*tools_binomial(2*a - 1, a - n - 1);
      }
    }
  }
  lua_pushnumber(L, r);
  return 1;
}


/* Computes the sigmoid function, having a characteristic `S`-shaped curve or sigmoid curve, for any number x.
   The result is a number between - but excluding - 0 and 1. See: https://en.wikipedia.org/wiki/Sigmoid_function */
static int calc_sigmoid (lua_State *L) {  /* 2.10.4, 2.34.10 switch to long doubles */
  long double x, t;
  x = agn_checknumber(L, 1);
  t = tools_expl(x);
  lua_pushnumber(L, t/(t + 1));
  return 1;
}


/* The logit function is the inverse of the sigmoid, or logistic function and is defined as: logit(x) := ln(p/(1-p)) =
   2*atanh(2*p-1). If x = 0, it will return -infinity, and if x = 1 the result will be +infinity. 2.30.2 */
static int calc_logit (lua_State *L) {
  long double x, r;
  x = agn_checknumber(L, 1);
  r = 2*sun_atanhl(2*x - 1);  /* 25 % faster than sun_log(x) - sun_log(1 - x) */
  if (tools_fpisnanl(r)) {
    /* extend domain to [0, 1]; See: https://calculus.subwiki.org/wiki/Inverse_logistic_function */
    if (x == 0) r = -HUGE_VAL;
    else if (x == 1) r = HUGE_VAL;
  }
  lua_pushnumber(L, r);
  return 1;
}


/* Computes the probit function, the inverse of the cumulative distribution function of the standard normal distribution:
   sqrt(2)*inverf(2p-1). 2.30.2 */
static int calc_probit (lua_State *L) {
  lua_Number x, r;
  x = agn_checknumber(L, 1);
  r = -SQRT2*tools_inverfc(2*x);
  if (r == -0) r = 0;
  if (isnan(r)) {
    /* extend domain to [0, 1]; See: https://calculus.subwiki.org/wiki/Inverse_logistic_function */
    if (x == 0) r = -HUGE_VAL;
    else if (x == 1) r = HUGE_VAL;
  }
  lua_pushnumber(L, r);
  return 1;
}


/* calc.logistic(x [, max [, k [, x0]]])

   Computes the logistic function, having a characteristic `S`-shaped curve or sigmoid curve, for any number x, according
   to the formula

   L(x) = max/(1 + exp(-k*(x - x0))),

   where max is the curve's maximum value, k its steepness, and x0 the x-value of the sigmoid's midpoint. By default,
   max = 1, k = 1, and x0 = 0, computing the sigmoid function. If only x is given, it works like `calc.sigmoid`.

   The result is a number between - but excluding - 0 and max. See: https://en.wikipedia.org/wiki/Logistic_function */
static int calc_logistic (lua_State *L) {  /* 2.10.4, 234.10 switch to long double */
  long double x;
  x = agn_checknumber(L, 1);
  if (lua_gettop(L) == 1) {  /* = sigmoid function */
    lua_Number t;
    t = tools_expl(x);
    lua_pushnumber(L, t/(t + 1));
  } else {
    long double x0, l, k;
    l = agnL_optnumber(L, 2, 1);   /* curve's maximum value */
    k = agnL_optnumber(L, 3, 1);   /* steepness of the curve */
    x0 = agnL_optnumber(L, 4, 0);  /* x-value of the sigmoid's midpoint */
    lua_pushnumber(L, l/(1 + tools_expl(-k*(x - x0))));
  }
  return 1;
}


/* Computes the Softsign function f(x) = x/(1 + |x|). 40 percent faster than an Agena implementation */
static int calc_softsign (lua_State *L) {  /* 2.14.1, 2.34.10 switch to long double */
  long double x;
  x = agn_checknumber(L, 1);
  lua_pushnumber(L, x/(1 + fabsl(x)));
  return 1;
}


/* Computes the Gaussian function a*exp(-(x-b)^2/(2*c^2)) where at a real or complex point x, with a, b, c
   being (real) numbers. By default, a = 1, b = 0, c = 1/sqrt(2). The return depends on the type of x.

   See also: expx2.

   See: https://en.wikipedia.org/wiki/Gaussian_function; formula computed using pencil, paper and Maple V Release 4. */
static int calc_gaussian (lua_State *L) {  /* 2.10.5 */
  lua_Number a, b, c;
  a = agnL_optnumber(L, 2, 1);
  b = agnL_optnumber(L, 3, 0);
  c = agnL_optnumber(L, 4, M_SQRT1_2);  /* = 1/sqrt(2) = sqrt(2)/2 */
  switch (lua_type(L, 1)) {
    case LUA_TNUMBER: {
      lua_Number x;
      x = agn_tonumber(L, 1);
      lua_pushnumber(L, a*tools_expx2(M_SQRT1_2*(x - b)/c, -1));
      break;
    }
    case LUA_TCOMPLEX: {
      lua_Number x, y, re, im;
#ifndef PROPCMPLX
      agn_Complex z = agn_tocomplex(L, 1);
      x = creal(z);
      y = cimag(z);
#else
      x = agn_complexreal(L, 1); y = agn_compleximag(L, 1);
#endif
      re = tools_cexpx2(0.5*(M_SQRT2*(x - b)/c), M_SQRT1_2*y/c , -1, &im);
#ifndef PROPCMPLX
      agn_createcomplex(L, a*(re + im*I));
#else
      agn_createcomplex(L, a*re, a*im);
#endif
      break;
    }
    default:
      luaL_nonumorcmplx(L, 1, "calc.gaussian");
  }
  return 1;
}


/* Computes the Weierstra function (formula) and returns a number. The precision is given by its fourth argument, eps, which is `Eps` by default. */
static int calc_weier (lua_State *L) {  /* 2.12.0 RC 4 */
  volatile long double a, b, x, r, eps, z, q;  /* changed 2.34.10 */
  int i;
  x = agn_checknumber(L, 1);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  eps = agnL_optpositive(L, 4, agn_getepsilon(L));  /* 2.13.0 change */
  if (luai_numiseven(b) || !(0 < a && a < 1)) {  /* b even or a out-of-range? -> infinite loops otherwise */
    lua_pushundefined(L);
    return 1;
  }
  i = r = q = 0;
  while (1) {
    z = tools_powil(a, i)*sun_cosl(tools_powil(b, i)*x);
    if (fabsl(z) < eps) break;
    r = tools_koaddl(r, z, &q); i++; /* 2.13.0/2.34.10 change */
  }
  lua_pushnumber(L, r + q);
  return 1;
}


/* Computes the first derivative of the univariate or multivariate function f at real point x, a number.

   If the option eps=h is given, the epsilon value h (a positive number preferably close to zero) is used
   internally for the computation, its default is math.epsilon(x).

   The second, etc. arguments to f may be given right after argument x.

   The return is the imaginary part of f(x + I*h)/h , or `fail` if f did not evaluate to the complex plane.

   This function as at least three times faster than `calc.xpdiff`.

   The idea has been taken from the Euler Math Toolbox, thus its name. See also:
   http://www.euler-math-toolbox.de by Rene Grothmann, version 2018-11-16, source file util/numerical.e,
   function `diffc`;
   http://adl.stanford.edu/hyperdual/Fike_AIAA-2011-886.pdf, page 4;
   https://www.sciencedirect.com/science/article/pii/S0377042712004207

   DO NOT MIGRATE TO LONG DOUBLES due to results less precise than with doubles (division by h) ! */

static int calc_eulerdiff (lua_State *L) {  /* 2.14.3 */
  lua_Number x, eps, delta;
  int sided, deriv, i, c, nargs;
#ifndef PROPCMPLX
  agn_Complex z;
#else
  lua_Number im;
#endif
  const char *option = NULL;
  (void)sided; (void)delta;
  c = 1;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  x = agn_checknumber(L, 2);
  nargs = lua_gettop(L);
  aux_geteps(L, x, nargs, "calc.eulerdiff");  /* set eps */
  if (deriv < 1 || deriv > 2)
    luaL_error(L, "Error in " LUA_QS ": deriv option must either be 1 or 2.", "calc.eulerdiff");
  /* first check whether f is defined at x, ..., otherwise the function would erroneously return a finite result.
     This is not bulletproof, e.g. sqrt(x)' with x = 0 does not exist. */
  if (isnan(agnL_fncall(L, 1, x, 3, nargs))) {
    lua_pushundefined(L);
    return 1;
  }
  luaL_checkstack(L, 2 + (nargs > 2 ? nargs - 3 + 1 : 0), "not enough stack space");  /* 2.31.7 fix */
  if (deriv == 1) {  /* first derivative: << f, x, h -> imag(f(x + I*h)) / h >> */
    lua_pushvalue(L, 1);  /* push function */
#ifndef PROPCMPLX
    agn_createcomplex(L, x + I*eps);  /* push `x + I*h` */
#else
    agn_createcomplex(L, x, eps);
#endif
    for (i=3; i <= nargs; i++, c++) lua_pushvalue(L, i);  /* push further arguments if they exist */
    if (lua_pcall(L, c, 1, 0) != 0 || !lua_iscomplex(L, -1)) {
      agn_poptop(L);  /* remove error message string */
      lua_pushfail(L);
      return 1;
    }
#ifndef PROPCMPLX
    z = agn_tocomplex(L, -1);
    agn_poptop(L);
    lua_pushnumber(L, cimag(z)/eps);
#else
    im = agn_compleximag(L, -1);
    agn_poptop(L);
    lua_pushnumber(L, im/eps);
#endif
  } else {  /* second derivative: << f, x, h -> 2*(f(x)-Real[f(x + I*h)])/h^2 >> */
    lua_Number z0, z1, h;
    /* part one */
    lua_pushvalue(L, 1);  /* push function */
    lua_pushnumber(L, x);
    for (i=3; i <= nargs; i++, c++) lua_pushvalue(L, i);  /* push further arguments if they exist */
    if (lua_pcall(L, c, 1, 0) != 0) {
      agn_poptop(L);  /* remove error message string */
      lua_pushfail(L);
      return 1;
    }
    z0 = agn_tonumber(L, -1);
    agn_poptop(L);
    h = eps;
    /* part two; to prevent catastrophic numerical cancellation, we iterate to get an acceptable result: */
    luaL_checkstack(L, 2 + (nargs > 2 ? nargs - 3 + 1 : 0), "not enough stack space");  /* 2.31.7 fix */
    do {
      c = 1;
      lua_pushvalue(L, 1);  /* push function */
#ifndef PROPCMPLX
      agn_createcomplex(L, x + I*h);  /* push `x + I*h` */
#else
      agn_createcomplex(L, x, h);
#endif
      for (i=3; i <= nargs; i++, c++) lua_pushvalue(L, i);  /* push further arguments if they exist */
      if (lua_pcall(L, c, 1, 0) != 0 || !lua_iscomplex(L, -1)) {
        agn_poptop(L);  /* remove error message string */
        lua_pushfail(L);
        return 1;
      }
#ifndef PROPCMPLX
      z = agn_tocomplex(L, -1);
      z1 = creal(z);
      agn_poptop(L);
#else
      z1 = agn_complexreal(L, -1);
      agn_poptop(L);
#endif
      h *= 10;
    } while (h < 1 && fabs(z0 - z1) < eps);
    h /= 10;
    lua_pushnumber(L, 2*(z0 - z1)/(h*h));
  }
  return 1;
}


/* Taken from the Fortran 90 source file:
   https://people.sc.fsu.edu/~jburkardt/f77_src/special_functions/special_functions.f, subroutine bernob

  cc BERNOB computes the Bernoulli number Bn.
  c
  c  Licensing:
  c
  c    This routine is copyrighted by Shanjie Zhang and Jianming Jin.  However,
  c    they give permission to incorporate this routine into a user program
  c    provided that the copyright is acknowledged.
  c
  c  Modified:
  c
  c    11 July 2012
  c
  c  Author:
  c
  c    Shanjie Zhang, Jianming Jin
  c
  c  Reference:
  c
  c    Shanjie Zhang, Jianming Jin,
  c    Computation of Special Functions,
  c    Wiley, 1996,
  c    ISBN: 0-471-11963-6,
  c    LC: QA351.C45.
*/

/* Computes the n'th Bernoulli number Bn. */

static int calc_bernoulli (lua_State *L) {  /* 2.20.2, 2.34.10 switched to long doubles */
  long double r1, r2, s;
  int k, m, n;
  n = agn_checknonnegint(L, 1);
  switch (n) {
    case 0:
      lua_pushnumber(L, 1);
      break;
    case 1:
      lua_pushnumber(L, -0.5);
      break;
    case 2:
      lua_pushnumber(L, ONESIXTHld);
      break;
    default: {
      lua_Number eps = agnL_optpositive(L, 2, agn_getdblepsilon(L));
      if (n & 1) {  /* bernoulli(odd integer) = 0 */
        lua_pushnumber(L, 0);
        return 1;
      }
      r1 = ONEoPISQUAREDld;
      r2 = 0.0;
      for (m=4; m <= n; m += 2) {
        r1 = -r1*(m - 1)*m/PI2SQUAREDld;  /* PI2SQUARED = PI2 * PI2 */
        r2 = 1.0;
        for (k=2; k <= 10000; k++) {
          s = tools_powil(1.0/k, m);  /* tuned 2.29.5 */
          r2 += s;
          if (s < eps) break;
        }
      }
      lua_pushnumber(L, r1*r2);  /* <=> bn(m) = r1 * r2 */
    }
  }
  return 1;
}


/* Taken from the Fortran 90 source file:
   https://people.sc.fsu.edu/~jburkardt/f77_src/special_functions/special_functions.f, subroutine eulerb

  cc EULERB computes the Euler number En.
  c
  c  Licensing:
  c
  c    This routine is copyrighted by Shanjie Zhang and Jianming Jin.  However,
  c    they give permission to incorporate this routine into a user program
  c    provided that the copyright is acknowledged.
  c
  c  Modified:
  c
  c    09 July 2012
  c
  c  Author:
  c
  c    Shanjie Zhang, Jianming Jin
  c
  c  Reference:
  c
  c    Shanjie Zhang, Jianming Jin,
  c    Computation of Special Functions,
  c    Wiley, 1996,
  c    ISBN: 0-471-11963-6,
  c    LC: QA351.C45.
*/

#define TWOoPISQUARED  0.40528473456935108576
#define TWOoPICUBED    0.25801227546559591346

#define TWOoPISQUAREDld  0.405284734569351085775517852840
#define TWOoPICUBEDld    0.258012275465595913475376421509

static int calc_euler (lua_State *L) {  /* 2.20.2, 2.34.10 switch to long doubles */
  long double r1, r2, s, isgn;
  int k, m, n;
  n = agn_checknonnegint(L, 1);
  switch (n) {
    case 0:
      lua_pushnumber(L, 1);
      break;
    case 1:
      lua_pushnumber(L, 0);
      break;
    case 2:
      lua_pushnumber(L, -1.0);
      break;
    default: {
      lua_Number eps = agnL_optpositive(L, 2, agn_getdblepsilon(L));
      if (n & 1) {  /* euler(odd integer) = 0 */
        lua_pushnumber(L, 0);
        return 1;
      }
      r1 = -4.0*TWOoPICUBEDld;
      r2 = 0.0;
      for (m=4; m <= n; m += 2) {
        r1 = - r1*(m - 1)*m*TWOoPISQUAREDld;  /* 0.405.. =  TWOoPI* TWOoPI */
        r2 = 1.0;
        isgn = 1.0;
        for (k=3; k <= 1000; k += 2) {
          isgn = -isgn;
          s = tools_powil(1.0/k, m + 1);  /* tuned 2.29.5 */
          r2 = r2 + isgn*s;
          if (s < eps) break;
        }
      }
      lua_pushnumber(L, r1*r2);  /* <=> bn(m) = r1 * r2 */
    }
  }
  return 1;
}


static long double gam0 (long double x) {
/*
c*********************************************************************
c
cc GAM0 computes the Gamma function for the LAMV function.
c
c  Licensing:
c
c    This routine is copyrighted by Shanjie Zhang and Jianming Jin.  However,
c    they give permission to incorporate this routine into a user program
c    provided that the copyright is acknowledged.
c
c  Modified:
c
c    09 July 2012
c
c  Author:
c
c    Shanjie Zhang, Jianming Jin
c
c  Reference:
c
c    Shanjie Zhang, Jianming Jin,
c    Computation of Special Functions,
c    Wiley, 1996,
c    ISBN: 0-471-11963-6,
c    LC: QA351.C45.
c
c  Parameters:
c
c    Input, double precision X, the argument.
c
c    Output, double precision GA, the function value. */
  int k;
  long double y;
  long double g[25] = {
     1.0,
     0.5772156649015329e+00,
    -0.6558780715202538e+00,
    -0.420026350340952e-01,
     0.1665386113822915e+00,
    -0.421977345555443e-01,
    -0.96219715278770e-02,
     0.72189432466630e-02,
    -0.11651675918591e-02,
    -0.2152416741149e-03,
     0.1280502823882e-03,
    -0.201348547807e-04,
    -0.12504934821e-05,
     0.11330272320e-05,
    -0.2056338417e-06,
     0.61160950e-08,
     0.50020075e-08,
    -0.11812746e-08,
     0.1043427e-09,
     0.77823e-11,
    -0.36968e-11,
     0.51e-12,
    -0.206e-13,
    -0.54e-14,
    0.14e-14 };
  y = g[24];
  for (k=23; k >= 0; k--) {
    y = fmal(y, x, g[k]);  /* 2.35.2 tweak */
  }
  return 1.0/(y*x);
}

/* Taken from the Fortran 90 source file:
   https://people.sc.fsu.edu/~jburkardt/f77_src/special_functions/special_functions.f, subroutine lamv

   See also: `Computation of Special Functions`, p. 182ff, by Shanjie Zhang and Jianming Jin.

   Computes:
   - calc.lamv[1] := << v, x -> 2^v*gamma(v + 1)*besselj(v, x)/x^v >> and its derivative
   - calc.lamv[2] := << v, x -> 2*v/x*(calc.lamv[1](v - 1, x) - calc.lamv[1](v, x)) >>
   - the actual order processed, may differ from the input !

   (Maple: lamv := (v, x) -> 2^v*GAMMA(v + 1)*BesselJ(v, x)/x^v; )

c*********************************************************************
c
cc LAMV computes lambda functions and derivatives of [integral] order. Fractional orders have been deliberately
c  switched off due to wrong results, compared with Maple V R4.
c
c  Licensing:
c
c    This routine is copyrighted by Shanjie Zhang and Jianming Jin.  However,
c    they give permission to incorporate this routine into a user program
c    provided that the copyright is acknowledged.
c
c  Modified:
c
c    31 July 2012
c
c  Author:
c
c    Shanjie Zhang, Jianming Jin
c
c  Reference:
c
c    Shanjie Zhang, Jianming Jin,
c    Computation of Special Functions,
c    Wiley, 1996,
c    ISBN: 0-471-11963-6,
c    LC: QA351.C45.
c
c  Parameters:
c
c    Input, double precision V, the order (a posint).
c
c    Input, double precision X, the argument.
c
c    Output, double precision VM, the highest order computed.
c
c    Output, double precision VL(0:*), DL(0:*), the Lambda function and
c    derivative, of orders N+V0. */
static int calc_lambda (lua_State *L) {  /* 2.20.2 */
  long double v, a0, bjv0, bjv1, bk, ck, cs, f, f0, f1, f2, fac, ga, px, qx, r, r0, rc, rp,
    rq, sk, uk, v0, vk, vm, vv, x, x2, xk, *dl, *vl;
  int i, j, k, k0, m, n;
  v = agn_checkposint(L, 1);  /* order; we won't accept fractional orders due to incorrect results */
  x = agn_checknumber(L, 2);  /* argument */
  bjv0 = 0.0; bjv1 = 0.0; f = 0.0;
  x = fabsl(x);
  x2 = x*x;
  n = (int)v;
  vl = malloc((n + 1)*sizeof(long double));  /* the Lambda function, of order N + V0. */
  if (vl == NULL)
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.lambda");
  dl = malloc((n + 1)*sizeof(long double));  /* the derivative, of order N + V0. */
  if (dl == NULL)
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.lambda");
  v0 = v - n;  /* v0 = frac(v), will always be 0 */
  vm = v;  /* highest order computed */
  if (x <= 12.0) {
    long double eps = agnL_optpositive(L, 3, agn_getdblepsilon(L));
    for (k=0; k <= n; k++) {
      vk = v0 + k;
      bk = 1.0;
      r = 1.0;
      for (i=1; i <= 50; i++) {
        r = -0.25*r*x2/(i*(i + vk));
        bk += r;
        if (fabsl(r) < fabsl(bk)*eps) break;
      }
      vl[k]= bk;
      uk = 1.0;
      r = 1.0;
      for (i=1; i <= 50; i++) {
        r = -0.25*r*x2/(i*(i + vk + 1.0));
        uk += r;
        if (fabsl(r) < fabsl(uk)*eps) break;
      }
      dl[k] = -0.5*x/(vk + 1.0)*uk;
    }
    goto leave;
  }
  if (x < 35.0)
    k0 = 11;
  else if (x < 50.0)
    k0 = 10;
  else
    k0 = 8;
  for (j=0; j <= 1; j++) {
    vv = 4.0*(j + v0)*(j + v0);
    px = 1.0;
    rp = 1.0;
    for (k=1; k <= k0; k++) {
      rp = -0.78125e-02*rp*tools_powil(vv - (4.0*k - 3.0), 2)*tools_powil(vv - (4.0*k - 1.0), 2)  /* 2.29.5 optimisation */
           / (k*(2.0*k - 1.0)*x2);
      px += rp;
    }
    qx = 1.0;
    rq = 1.0;
    for (k=1; k <= k0; k++) {
      rq = - 0.78125e-02*rq*tools_powil(vv - (4.0*k - 1.0), 2)*tools_powil(vv - (4.0*k + 1.0), 2)  /* 2.29.5 optimisation */
           / (k*(2.0*k + 1.0)*x2);
      qx += rq;
    }
    qx = 0.125*(vv - 1.0)*qx/x;
    xk = x - (0.5*(j + v0) + 0.25)*M_PIld;
    a0 = tools_sqrtl(TWOoPIld/x);
    sk = sun_sinl(xk);
    ck = sun_cosl(xk);
    if (j == 0)
      bjv0 = a0*(px*ck - qx*sk);
    else
      bjv1 = a0*(px*ck - qx*sk);
  }
  ga = (v0 == 0.0) ? 1.0 : v0*gam0(v0);   /* since we allow only integral orders, gam0 is never called */
  fac = tools_powl(2.0/x, v0)*ga;
  vl[0] = bjv0;
  dl[0] = -bjv1 + v0/x*bjv0;
  vl[1] = bjv1;
  dl[1] = bjv0 - (1.0 + v0)/x*bjv1;
  r0 = 2.0*(1.0 + v0)/x;
  if (n <= 1) {
    vl[0] = fac*vl[0];
    dl[0] = fac*dl[0] - v0/x*vl[0];
    vl[1] = fac*r0*vl[1];
    dl[1] = fac*r0*dl[1] - (1.0 + v0)/x*vl[1];
    goto leave;
  }
  if (0 && n >= 2 && n <= truncl(0.9*x)) {
    /* this is very imprecise with larger x, so do not run it: */
    f0 = bjv0;
    f1 = bjv1;
    for (k=2; k <= n; k++) {
      f = 2.0*(k + v0 - 1.0)/x*f1 - f0;
      f0 = f1;
      f1 = f;
      vl[k] = f;
    }
  } else if (n >= 2) {
    m = tools_msta1l(x, 200);
    if (m < n)
      n = m;
    else
      m = tools_msta2l(x, n, 15);
    f2 = 0.0;
    f1 = 1.0e-100;
    for (k=m; k >= 0; k--) {
      f = 2.0*(v0 + k + 1.0)/x*f1 - f2;
      if (k <= n) vl[k] = f;
      f2 = f1;
      f1 = f;
    }
    cs = (fabsl(bjv0) <= fabsl(bjv1)) ? bjv1/f2 : bjv0/f;
    for (k=0; k <= n; k++)
      vl[k] = cs*vl[k];
  }
  vl[0] = fac*vl[0];
  for (j=1; j <= n; j++) {
    rc = fac*r0;
    vl[j] = rc*vl[j];
    dl[j - 1] = -0.5*x/(j + v0)*vl[j];
    r0 = 2.0*(j + v0 + 1)/x*r0;
  }
  dl[n] = 2.0*(v0 + n)*(vl[n - 1] - vl[n])/x;
  vm = n + v0;  /* highest order computed, n = int(v) or modified integer, v0 = frac(v) */
leave:
  lua_pushnumber(L, vl[n]);
  lua_pushnumber(L, dl[n]);
  lua_pushnumber(L, vm);
  xfree(dl); xfree(vl);
  return 3;
}


/* The following two C functions have been taken from:
   https://github.com/CompPhysics/ComputationalPhysics/blob/master/doc/Programs/LecturePrograms/programs/NumericalIntegration/cpp/program2.cpp
   written by Morten Hjorth-Jensen */

/*
 ** The function gauleg() takes the lower and upper limits of integration x1, x2,
 ** calculates and return the abcissas in x[0, ..., n - 1] and the weights in
 ** w[0, ..., n - 1] of length n of the Gauss--Legendre n--point quadrature formulae.
*/

static void gauleg (long double x1, long double x2, long double x[], long double w[], int n, long double eps) {
  int m, j, i;
  long double z1, z, xm, xl, pp, p3, p2, p1, *x_low, *x_high, *w_low, *w_high;
  /* roots are symmetric in the interval */
  m  = 0.5*(n + 1);
  xm = 0.5*(x2 + x1);
  xl = 0.5*(x2 - x1);
  /* pointer initialisation */
  x_low  = x;
  x_high = x + n - 1;
  w_low  = w;
  w_high = w + n - 1;
  /* loops over desired roots */
  for (i=1; i <= m; i++) {
    z = sun_cosl(M_PIld*(i - 0.25)/(n + 0.5));
    /* Starting with the above approximation to the ith root we enter the mani loop of refinement bt Newtons method. */
    do {
      p1 = 1.0;
      p2 = 0.0;
      /* loop up recurrence relation to get the Legendre polynomial evaluated at x */
      for (j=1; j <= n; j++) {
        p3 = p2;
        p2 = p1;
        p1 = ((2.0*j - 1.0)*z*p2 - (j - 1.0)*p3)/j;
      }
      /* p1 is now the desired Legrendre polynomial. Next compute ppp its derivative by standard relation involving also p2,
         polynomial of one lower order. */
      pp = n*(z*p1 - p2)/(z*z - 1.0);
      z1 = z;
      z = z1 - p1/pp;  /* Newton's method */
    } while (fabsl(z - z1) > eps);
    /* Scale the root to the desired interval and put in its symmetric counterpart. Compute the weight and its symmetric
       counterpart */
    *(x_low++)  = xm - xl*z;
    *(x_high--) = xm + xl*z;
    *w_low      = 2.0*xl/((1.0 - z*z)*pp*pp);
    *(w_high--) = *(w_low++);
  }
}

/* Gauss-Legendre integration, 7 times faster than calc.integ with same precision */
static int calc_gauleg (lua_State *L) {  /* 2.21.4, 2.34.10 switch to long doubles */
  int i, n, error;
  long double a, b, r, eps, *x, *w;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  if (b <= a)
    luaL_error(L, "Error in " LUA_QS ": left border must be less than right border.", "calc.gauleg");
  n = agnL_optposint(L, 4, (b - a)*20);  /* total sample points, default is dependent on the range given */
  eps = agnL_optpositive(L, 5, tools_intpow(agn_getepsilon(L), 2));
  /* reserve space in memory for vectors containing the mesh points weights and function values for the use of the
     Gauss-Legendre method */
  x = malloc(n*sizeof(long double));
  w = malloc(n*sizeof(long double));
  /* set up the mesh points and weights */
  gauleg(a, b, x, w, n, eps);
  /* evaluate the integral with the Gauss-Legendre method; note that we initialise the sum */
  r = 0.0;
  for (i=0; i < n; i++) {
    r += w[i]*aux_fnunivar(L, x[i], 0, &error);
  }
  lua_pushnumber(L, r);
  xfree(x); xfree(w);
  return 1;
}


/* Determines the `variance` of a function f in a given interval [a, b] using trapezoidal rule. The return
   is an integer counting how often the trapezoidal rule has been applied. 2.21.4 */
static double trapezoidal_rule (lua_State *L, int idx, long double a, long double b, int n) {
  long double r, fa, fb, x, step;
  int j, error;
  step = (b - a)/((long double)n);
  fa = 0.5*aux_fnunivar(L, a, 0, &error);
  fb = 0.5*aux_fnunivar(L, b, 0, &error);
  r = 0.0;
  for (j=1; j <= n - 1; j++) {
    x = j*step + a;
    r += aux_fnunivar(L, x, 0, &error);
  }
  r = (r + fb + fa)*step;
  return r;
}

#define MAXRECURSIONS   50

static int adaptive_integration (lua_State *L, int idx, long double a, long double b, int n, int steps, long double eps) {
  long double c, i0, il, ir;  /* changed 2.34.10 */
  if (steps > MAXRECURSIONS) return steps;
  c = (a + b)*0.5;
  i0 = trapezoidal_rule(L, 1, a, b, n);  /* the whole integral */
  il = trapezoidal_rule(L, 1, a, c, n);  /* the left half */
  ir = trapezoidal_rule(L, 1, c, b, n);  /* the right half */
  if (fabsl(il + ir - i0) >= eps) {
    int stepsl, stepsr;  /* stepsr to prevent compiler warnings */
    stepsl = adaptive_integration(L, idx, a, c, n, ++steps, eps);
    stepsr = adaptive_integration(L, idx, c, b, n, ++steps, eps);
    steps = stepsl + stepsr;
  }  /* else do nothing, just return steps */
  return steps;
}

/* Returns a positive integer that indicates whether a function `f` in one real changes slowly or rapidly on the
   given interval [a, b]. The larger the result, the larger is its rate of change. By default,
   - the result is relative, i.e. given per unit on the abscissa (true for relative, false for absolute),
   - there are `n`=10 points to sample per unit, and
   - the bail-out value `eps` - a positive value close to zero - is `Eps`.
   Internally, the function uses adaptive integration with trapezoidal rule and counts the number of trapezoids
   evaluated. Note that the results are estimates. */
static int calc_variance (lua_State *L) {
  int n, relative, r, units;
  long double a, b, eps;  /* changed 2.34.10 */
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  if (b <= a)
    luaL_error(L, "Error in " LUA_QS ": left border must be less than right border.", "calc.variance");
  units = sun_roundl(b) - sun_roundl(a);  /* round to nearest integer away from zero */
  relative = agnL_optboolean(L, 4, 1);
  n = agnL_optposint(L, 5, units*10);  /* 10 sample points per unit */
  eps = agnL_optpositive(L, 6, agn_getepsilon(L));
  r = adaptive_integration(L, 1, a, b, n, 0, eps);
  lua_pushinteger(L, relative ? r/units : r);
  return 1;
}


/* Computes the scaled complex complementary error (Faddeeva) function w(z) = exp(-z^2)*erfc(-I*z), for numbers
   and complex numbers z. The return is a complex value. 2.21.6 */
#ifndef PROPCMPLX
#define cmplxreal(z) creal(z)
#define cmplximag(z) cimag(z)
#else
#define cmplxreal(z) (z).cmplx[0]
#define cmplximag(z) (z).cmplx[1]
#endif
static int calc_w (lua_State *L) {
  lua_Number eps;
  eps = agnL_optnonnegative(L, 2, 0);
  switch (lua_type(L, 1)) {
    case LUA_TNUMBER: {
      lua_Number x;
      x = agn_tonumber(L, 1);
#ifndef PROPCMPLX
      agn_Complex re;
      re = tools_w(x + I*0, eps);
#else
      PROPCMPLX_CMPLX z, re;
      z.cmplx[0] = x; z.cmplx[1] = 0;
      re = tools_w(z, eps);
#endif
      agn_pushcomplex(L, cmplxreal(re), cmplximag(re));
      break;
    }
    case LUA_TCOMPLEX: {
#ifndef PROPCMPLX
      agn_Complex z, re;
      z = agn_tocomplex(L, 1);
#else
      PROPCMPLX_CMPLX z, re;
      cmplxreal(z) = agn_complexreal(L, 1); cmplximag(z) = agn_compleximag(L, 1);
#endif
      re = tools_w(z, eps);
#ifndef PROPCMPLX
      agn_createcomplex(L, re);
#else
      agn_createcomplex(L, cmplxreal(re), cmplximag(re));
#endif
      break;
    }
    default:
      luaL_nonumorcmplx(L, 1, "calc.w");
  }
  return 1;
}


static int calc_aitken (lua_State *L) {
  /* Finds the limit of the sequence x[n+1] = f(x[n]) with initial x[0] and tolerance eps, with a maximum of iter iterations, using
     Aitken extrapolation. eps by default is DoubleEps and iter is 20.
     Returns either the approximated limit alpha and the first derivative at alpha f'(alpha) if successful, and `undefined` twice
     otherwise. The third return is the number of iterations taken to compute the result.
     For example calc.aitken(<< x -> 1/2*(x + 2/x) >>, 1) ~= sqrt(2).
     See https://en.wikipedia.org/wiki/Aitken%27s_delta-squared_process */
  long double x0, x1, x2, eps, aitken, d1, denom, l;  /* changed 2.34.10 */
  int nargs, iter, rc, i;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  nargs = lua_gettop(L);  /* 2.1.4 */
  /* luaL_checkstack is called by agnL_fncall */
  x0 = agn_checknumber(L, 2);  /* start value */
  eps = agnL_optpositive(L, 3, L->DoubleEps);
  iter = agnL_optposint(L, 4, 20);
  l = AGN_NAN;
  rc = 0;
  for (i=0; i < iter; i++) {
    x1 = agnL_fncall(L, 1, x0, 5, nargs);
    x2 = agnL_fncall(L, 1, x1, 5, nargs);
    d1 = x2 - x1;
    denom = d1 - (x1 - x0);
    if (fabsl(denom) < eps) break;
    aitken = x2 - (d1*d1)/denom;
    if (tools_approxl(aitken, x2, eps)) {
      l = fabsl(d1/(x1 - x0));
      rc = 1; i++;
      break;
    }
    x0 = aitken;
  }
  lua_pushnumber(L, rc ? aitken : AGN_NAN);
  lua_pushnumber(L, l);
  lua_pushinteger(L, i);
  return 3;
}


/* Computes the n-th Chebyshev polynomial of the first kind, evaluated at x, with n a non-negative integer and x a number. 2.33.1 */
static int calc_chebyt (lua_State *L) {  /* 2.34.10 switch to long doubles */
  long double a, t1, t4, t6, t9, t15, t17, t20;
  int n = agn_checknonnegint(L, 1);
  a = agn_checknumber(L, 2);
  t1 = a*a;
  t4 = sqrtl(t1 + 1.0 + 2.0*a);
  t6 = sqrtl(t1 + 1.0 - 2.0*a);
  t9 = n*sun_acosl(0.5*(t4 - t6));
  t15 = tools_powil(0.5*(t4 + t6), 2.0);
  t17 = sqrtl(t15 - 1.0);
  t20 = n*tools_signl(a)*tools_logl(0.5*(t4 + t6) + t17);
  lua_pushnumber(L, sun_cosl(t9)*tools_coshl(t20));
  return 1;
}


static const luaL_Reg calclib[] = {
  {"bernoulli", calc_bernoulli},           /* added on May 17, 2020 */
  {"Ai", calc_Ai},                         /* added on June 29, 2015 */
  {"aitken", calc_aitken},                 /* added on September 20, 2022 */
  {"Bi", calc_Bi},                         /* added on June 29, 2015 */
  {"cheby", calc_cheby},                   /* added on July 05, 2017 */
  {"chebyt", calc_chebyt},                 /* added on October 17, 2022 */
  {"Chi", calc_Chi},                       /* added on May 24, 2010 */
  {"Ci", calc_Ci},                         /* added on May 24, 2010 */
  {"clampedspline", calc_clampedspline},   /* added on November 04, 2012 */
  {"clampedsplinecoeffs", calc_clampedsplinecoeffs},  /* added on November 04, 2012 */
  {"dawson", calc_dawson},                 /* added on May 25, 2010 */
  {"diff", calc_diff},                     /* added on May 29, 2010 */
  {"dilog", calc_dilog},                   /* added on May 25, 2010 */
  {"Ei", calc_Ei},                         /* added on May 25, 2010 */
  {"En", calc_En},                         /* added on July 07, 2015 */
  {"eps", calc_eps},                       /* added on February 14, 2016 */
  {"eucliddist", calc_eucliddist},         /* added on January 25, 2016 */
  {"euler", calc_euler},                   /* added on May 17, 2020 */
  {"eulerdiff", calc_eulerdiff},           /* added on December 26, 2018 */
  {"fminbr", calc_fminbr},                 /* added on May 29, 2010 */
  {"fmings", calc_fmings},                 /* added on July 02, 2020 */
  {"fprod", calc_fprod},                   /* added on Nov 23, 2009 */
  {"fresnelc", calc_fresnelc},             /* added on May 29, 2010 */
  {"fresnels", calc_fresnels},             /* added on May 29, 2010 */
  {"fsum", calc_fsum},                     /* added on April 20, 2008 */
  {"gauleg", calc_gauleg},                 /* added on July 01, 2020 */
  {"gaussian", calc_gaussian},             /* added on August 06, 2017 */
  {"gtrap", calc_gtrap},                   /* added on March 04, 2014 */
  {"ibeta", calc_ibeta},                   /* added on July 08, 2015 */
  {"igamma", calc_igamma},                 /* added on July 07, 2015 */
  {"igammac", calc_igammac},               /* added on July 07, 2015 */
  {"intde", calc_intde},                   /* added on May 18, 2010 */
  {"intdei", calc_intdei},                 /* added on May 22, 2010 */
  {"intdeo", calc_intdeo},                 /* added on May 22, 2010 */
  {"interp", calc_interp},                 /* added on October 28, 2012 */
  {"invibeta", calc_invibeta},             /* added on July 08, 2015 */
  {"iscont", calc_iscont},                 /* added on January 21, 2016 */
  {"isdiff", calc_isdiff},                 /* added on May 04, 2017 */
  {"lambda", calc_lambda},                 /* added on May 17, 2020 */
  {"limit", calc_limit},                   /* added on January 21, 2016 */
  {"linterp", calc_linterp},               /* added on January 28, 2014 */
  {"logistic", calc_logistic},             /* added on August 04, 2017 */
  {"logit", calc_logit},                   /* added on August 01, 2022 */
  {"nakspline", calc_nakspline},           /* added on October 28, 2012 */
  {"naksplinecoeffs", calc_naksplinecoeffs}, /* added on October 28, 2012 */
  {"neville", calc_neville},               /* added on October 28, 2012 */
  {"newtoncoeffs", calc_newtoncoeffs},     /* added on October 28, 2012 */
  {"polyfit", calc_polyfit},               /* added on January 28, 2014 */
  {"polygen", calc_polygen},               /* added on July 21, 2013 */
  {"polylog", calc_polylog},               /* added on July 07, 2015 */
  {"probit", calc_probit},                 /* added on August 01, 2022 */
  {"Psi", calc_Psi},                       /* added on May 25, 2010 */
  {"regulafalsi", calc_regulafalsi},       /* added on May 29, 2010 */
  {"savgol", calc_savgol},                 /* added on July 16, 2017 */
  {"scaleddawson", calc_scaleddawson},     /* added on July 21, 2020 */
  {"sections", calc_sections},             /* added on May 29, 2010 */
  {"Shi", calc_Shi},                       /* added on May 24, 2010 */
  {"Si", calc_Si},                         /* added on May 24, 2010 */
  {"sigmoid", calc_sigmoid},               /* added on March 04, 2014 */
  {"simaptive", calc_simaptive},           /* added on March 04, 2014 */
  {"smoothstep", calc_smoothstep},         /* added on August 02, 2017 */
  {"softsign", calc_softsign},             /* added on October 18, 2018 */
  {"Ssi", calc_Ssi},                       /* added on May 24, 2010 */
  {"variance", calc_variance},             /* added on July 01, 2020 */
  {"w", calc_w},                           /* added on July 26, 2020 */
  {"weier", calc_weier},                   /* added on June 16, 2018 */
  {"xpdiff", calc_xpdiff},                 /* added on May 29, 2010 */
  {"zeroin", calc_zeroin},                 /* added on February 02, 2022 */
  {"zeta", calc_zeta},                     /* added on July 07, 2015 */
  {NULL, NULL}
};


/*
** Open calculus library
*/

LUALIB_API int luaopen_calc (lua_State *L) {
  luaL_newmetatable(L, "calc");
  luaL_register(L, NULL, numarray_arraylib);  /* associate __gc method */
  luaL_register(L, AGENA_CALCLIBNAME, calclib);
  return 1;
}

