/*
** $Id: fastmath.c,v 1.67 2005/08/26 17:36:32 roberto Exp $
** Approximations
** See Copyright Notice in agena.h
*/

/*
import fastmath, stats

s := seq();

for x from -100 to 100 by 0.001 do
   insert fastmath.dexp10(x) |- math.exp10(x) into s
od;

stats.median(stats.sorted(s)):

stats.amean(s):

import fastmath

gcc := fastmath.sunpow
fn := fastmath.dpow

watch();
for i from -10 to 10 by 0.001 do
   x := fn(i)
od;
watch():

watch();
for i from -10 to 10 by 0.001 do
   y := gcc(i)
od;
watch():
*/

#include <stdlib.h>
#include <math.h>
#include <stdint.h>  /* for UINT32_MAX */

#define fastmath_c
#define LUA_LIB

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"
#include "agnconf.h"
#include "agncmpt.h"  /* for trunc and isfinite, and FLT_EVAL_METHOD constant, and off64* types */

/* For unknown reasons I am too tired to clarify, we import the *GET_*, *SET_*, *INSERT*, *EXTRACT*,
   etc. Sun Microsystems macros from the following header file instead of agnhlps.h: */
#include "sunpro.h"

#define AGENA_LIBVERSION	"fastmath 1.1.1 for Agena as of May 13, 2019\n"

#if !(defined(LUA_DOS) || defined(__OS2__) || defined(LUA_ANSI))
#define AGENA_FASTMATHLIBNAME "fastmath"
LUALIB_API int (luaopen_fastmath) (lua_State *L);
#endif


/* Approximates the inverse root 1/root(x, degree) using the Quake III method. x is the radicant, degree the
   degree-th root which by default is 2. n is the number of iterations to be conducted and by default is 2^degree.
   xhalf is the internal equivalent of x, 0.5*x by default. The greater the degree, the less accurate is the result.

   See: https://stackoverflow.com/questions/11644441/fast-inverse-square-root-on-x64,
   https://www.xaymar.com/2016/12/17/what-is-the-fastest-way-to-get-an-inverse-square-root-part-2
   Modified. 2.10.4
   Must use a union instead of *(<type> *)&y, and the proposed int64_t type given in the web pages
   listed above is wrong, anyway. */
static int fastmath_invroot (lua_State *L) {
  lua_Number x, y, xhalf;
  double_cast d;
  int i, n, degree;
  x = agn_checknumber(L, 1);
  if (x ==0) y = 0;
  else if (x < 0) y = AGN_NAN;
  else {
    degree = agnL_optposint(L, 2, 2);
    n = agnL_optposint(L, 3, tools_intpow(2, degree));
    xhalf = agnL_optnumber(L, 4, 0.5*x);  /* internal equivalent for x, 0.5*x by default */
    /* int64_t i = *(int64_t *)&y; does not work with GCC 6.3.0, especially not with int64_t's */
    d.f = x;
    /* The magic number 0x5fe6eb50c7b537a9 is for doubles is from https://cs.uwaterloo.ca/~m32rober/rsqrt.pdf */
    /* d.i = 0x5fe6eb50c7b537a9 - (d.i >> 1);  */ /* first guess of 1/sqrt(x) */
    d.i = agnL_optnumber(L, 5, 0x5fe6eb50c7b537a9LL) - (d.i >> 1);  /* 0x5fe6eb50c7b537a9ll taken from:
      https://www.xaymar.com/2016/12/17/what-is-the-fastest-way-to-get-an-inverse-square-root-part-2/ */
    /* y = *(lua_Number *)&i;  does not work with GCC 6.3.0  */
    y = d.f;
    for (i=0; i < n; i++)  /* 4 Newton iterations to approximate 1/sqrt(x) */
      y = y*(1.5 - (xhalf*tools_intpow(y, degree)));  /* 1.5 = `threehalfs` */
    /* to compute the square root, now execute: y = ((1/y) + (x*y))*0.5; (1 additional Heron iteration has shown
       to maximise the precision), see: http://www.reactos.org/pipermail/ros-diffs/2015-May/057628.html */
  }
  lua_pushnumber(L, y);  /* square y to return 1/x */
  return 1;
}


/* Approximates 1/sqrt(x) using Quakes Fast Inverse Square Root method.
   5 % faster than the inverse of GCC's builtin sqrt.
   In the range -1000 .. 1000 with step size 0.1, mean error is 5.96994e-5, median error is 5.008699e-5.
   See: https://betterexplained.com/articles/understanding-quakes-fast-inverse-square-root */
static int fastmath_invsqrt (lua_State *L) {  /* new in 2.14.6 */
  float x, xhalf;
  ieee_float_shape_signedint u;
  x = agn_checknumber(L, 1);
  xhalf = 0.5f * x;
  u.f = x;                        /* store floating-point bits in integer */
  u.i = 0x5f3759df - (u.i >> 1);  /* initial guess for Newton's method */
  x = u.f;                        /* convert new bits into float */
  x = x*(1.5f - xhalf*x*x);       /* One round of Newton's method */
  lua_pushnumber(L, x);
  return 1;
}


/* Approximates of the reciprocal of its argument of type number. With x < 0 returns infinity.
   The return is a number. The function is purely experimental.
   54 % slower than GCC's `1/x`.
   In the range 0 .. 1000 with step size 0.001, mean error is 0.0006720364, median error is 0.00010009.
   See: https://stackoverflow.com/questions/9939322/fast-1-x-division-reciprocal/39714493 */
static int fastmath_reciprocal (lua_State *L) {
  double_cast d;
  d.f = agn_checknumber(L, 1);
  d.i = (0xbfcdd6a18f6a6f52ULL - d.i) >> 1;
  lua_pushnumber(L, d.f*d.f);
  return 1;
}


/* Computes the reciprocal 1/sqrt(x), taken from https://github.com/nickzman/hyperspace/blob/master/frsqrt.hh#L399
   24 % slower than GCC's 1/sqrt(x).
   In the range 0 .. 1000 with step size 0.001, mean error is 2.455737e-018, median error is 0. */
static int fastmath_rsqrt (lua_State *L) {
  double x = agn_checknumber(L, 1);
  const double xhalf = 0.5*x;
  long long i = *(long long*)&x;
  i = 0x5fe6ec85e7de30daLL - (i >> 1);
  x = *(double*)&i;
  x = x*(1.5 - xhalf*x*x);
  x = x*(1.5 - xhalf*x*x);
  x = x*(1.5 - xhalf*x*x);
  x = x*(1.5 - xhalf*x*x);
  lua_pushnumber(L, x);
  return 1;
}


/* Taken from `branchless scalar math routines`, by Phillip Trudeau.
   see: https://raw.githubusercontent.com/pmttavara/pt_math/master/pt_math.h

   Copyright (c) Phillip Trudeau-Tavara, 2017-2019. All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

   The float version is not faster. 17 % slower than GCC's sqrt. */
static INLINE double pt_sqrt (double x) {
  long long y;
  double z;
  if (x < 0) return AGN_NAN;
  { int j; for (j = 0; j < CHAR_BIT; j++) ((char *)&y)[j] = ((char *)&x)[j]; }
  y = ((y - 0x0010000000000000LL) >> 1) + 0x2000000000000000LL;
  { int j; for (j = 0; j < CHAR_BIT; j++) ((char *)&z)[j] = ((char *)&y)[j]; }
  z = 0.5*(x/z + z);
  return 0.5*(x/z + z);
}

/* Approximates sqrt(x), 2.32.1.
   17 % slower than GCC's sqrt.
   In the range 0 .. 1000 with step size 0.1, median error is 0.000000089006936 and mean error is 0.000031388319. */
static int fastmath_dsqrt (lua_State *L) {
  lua_pushnumber(L, pt_sqrt(agn_checknumber(L, 1)));
  return 1;
}


static int fastmath_drsqrt (lua_State *L) {  /* 2.32.1 */
  int j;
  long long y;
  double x, z;
  x = agn_checknumber(L, 1);
  if (x <= 0) {
    lua_pushundefined(L);
    return 1;
  }
  z = 0.5*x;
  for (j = 0; j < CHAR_BIT; j++) ((char *)&y)[j] = ((char *)&x)[j];
  y = 0x5fe6eb50c7b537a9LL - (y >> 1);
  for (j = 0; j < CHAR_BIT; j++) ((char *)&x)[j] = ((char *)&y)[j];
  x *= 1.5 - z*x*x;
  lua_pushnumber(L, x*(1.5 - z*x*x));
  return 1;
}


/* Approximates of the square root using two numeric methods. The return is a number. The returns are two
   numbers: guesses computed using C doubles and floats, in this order. The function is purely experimental.
   See: http://bits.stephan-brumme.com/squareRoot.html
   11 % slower than GCC's sqrt.
   In the range 0 .. 1000 with step size 0.001, median error is 0.476598445 and mean error is 0.48653319. */
static int fastmath_sqroot (lua_State *L) {
  ieee_float_shape_type d;
  double_cast e;
  double x = agn_checknumber(L, 1);
  if (x < 0) {  /* 2.12.0 RC 3 change */
    lua_pushundefined(L);
    return 1;
  }
  d.value = x;
  e.f = x;
  /* adjust bias */
  d.word += (127 << 23);
  /* approximation of square root */
  d.word >>= 1;
  /* see: http://www.reactos.org/pipermail/ros-diffs/2015-May/057628.html */
  e.i = 0x5fe6eb50c7b537a9LL - (e.i >> 1);
  lua_pushnumber(L, 1/e.f);    /* double approximation */
  lua_pushnumber(L, d.value);  /* float approximation */
  return 2;
}

/*
http://jsperf.com/mysin-vs-sin/9
https://www.desmos.com/calculator/8nkxlrmp7a
http://devmaster.net/posts

See: https://web.archive.org/web/20161223122133/http://http.developer.nvidia.com:80/Cg/asin.html
*/
static int fastmath_sincosfast (lua_State *L) {  /* 2.11.0, 15 percent faster than calling Agena's sin and cos operators */
  float x, s, c;
  x = agn_checknumber(L, 1);
  if (lua_gettop(L) == 1)
    tools_sincosfast(x, &s, &c);
  else {
    /* taken from http://lab.polygonal.de/2007/07/18/fast-and-accurate-sinecosine-approximation/ */
    /* always wrap input angle to -PI .. PI */
    x = tools_reducerange(x, -PI, PI);
    /* compute sine */
    s = INVPIO4 * x - (x)*INVPISQO4*x*x;
    s += 0.225 * (tools_signum(s)*s*s - s);           /* 2.15.0 change, 2.17.1 optimisation */
    /* compute cosine: sin(x + PI/2) = cos(x) */
    x += PIO2;
    if (x > PI) x -= PI2;
    c = INVPIO4 * x - tools_signum(x)*INVPISQO4*x*x;  /* 2.15.0 change, 2.17.1 optimisation */
    c += 0.225 * (tools_signum(c)*c*c - c);           /* 2.15.0 change, 2.17.1 optimisation */
  }
  lua_pushnumber(L, s);
  lua_pushnumber(L, c);
  return 2;
}


/* Taken from `branchless scalar math routines`, by Phillip Trudeau.
   see: https://raw.githubusercontent.com/pmttavara/pt_math/master/pt_math.h */
static double pt_round_unchecked (double x) {
  *(volatile double *)(char *)&x += 6755399441055744.0;
  *(volatile double *)(char *)&x -= 6755399441055744.0;
  return x;
}

#define pt_sin(x) ({ \
  x *= 0.15915494309189534; \
  x -= pt_round_unchecked(x); \
  x *= 0.5 - (x < 0 ? -x : x); \
  (x*(57.3460872862336*(x < 0 ? -x : x) + 12.4158695446104)); \
})

/* 30 % faster than sun_sin, 2.32.1
   In the range -1000 .. 1000 with step size 0.001, median error is 0.000574565 and mean error is 0.0005261111. */
static int fastmath_dsin (lua_State *L) {
  double x = agn_checknumber(L, 1);
  lua_pushnumber(L, pt_sin(x));
  return 1;
}


/* 28 % faster than sun_cos, 2.32.1
   In the range -1000 .. 1000 with step size 0.001, median error is 0.000574439 and mean error is 0.000526013. */
static int fastmath_dcos (lua_State *L) {
  double x = agn_checknumber(L, 1) + PIO2;
  lua_pushnumber(L, pt_sin(x));
  return 1;
}


/* 37 % faster than sun_tan, 2.32.2
   In the range -1000 .. 1000 with step size 0.001, median error is 0.00068536 and mean error is 0.136447. */
static int fastmath_dtan (lua_State *L) {
  double x, y, si, co;
  x = y = agn_checknumber(L, 1);
  si = pt_sin(x);
  y += PIO2;
  co = pt_sin(y);
  lua_pushnumber(L, co != 0 ? si/co : AGN_NAN);
  return 1;
}


/* 17 % faster than gcc's asin
   In the range -1 .. 1 with step size 0.000001, median error is 0.008815365 and mean error is 0.007884382. */
static int fastmath_dasin (lua_State *L) {  /* 2.32.1 */
  double x = agn_checknumber(L, 1);
  if (fabs(x) > 1) {
    lua_pushundefined(L);
    return 1;
  }
  x = pt_sqrt(1 + x) - pt_sqrt(1 - x);
  lua_pushnumber(L, x * (0.131754508171 * (x < 0 ? -x : x) + 0.924391722181));
  return 1;
}


/* 26 % faster than gcc's acos
   In the range -1 .. 1 with step size 0.000001, median error is 0.00881536 and mean error is 0.00788438. */
static int fastmath_dacos (lua_State *L) {  /* 2.32.1 */
  double x = agn_checknumber(L, 1);
  if (fabs(x) > 1) {
    lua_pushundefined(L);
    return 1;
  }
  x = pt_sqrt(1 + x) - pt_sqrt(1 - x);
  lua_pushnumber(L, PIO2 - x*(0.131754508171*(x < 0 ? -x : x) + 0.924391722181));
  return 1;
}


/* 39 % faster than gcc's atan
   In the range -1 .. 1 with step size 0.000001, median error is 0.0042227 and mean error is 0.00382986. */
static int fastmath_datan (lua_State *L) {  /* 2.32.1 */
  double x, y;
  x = agn_checknumber(L, 1);
  x /= (x < 0 ? -x : x) + 1;
  y = (x < 0 ? -x : x);
  lua_pushnumber(L, x*(y*(-1.45667498914*y + 2.18501248371) + 0.842458832225));
  return 1;
}


/* Returns an approximation of the sine of the given number x, in radians; 2.14.6
   7 % slower than Sun's sun_sin
   In the range -1000 .. 1000 with step size 0.001, median error is 0.000474045 and mean error is 0.00044432. */
static int fastmath_sinfast (lua_State *L) {
  lua_pushnumber(L, tools_sinfast(agn_checknumber(L, 1)));
  return 1;
}


/* Returns an approximation of the cosine of the given number x, in radians. Depending on the CPU used, it is up to 40 percent faster
   than calling the `cos` operator. 2.14.6
   8 % slower than Sun's sun_cos
   In the range -1000 .. 1000 with step size 0.001, median error is 0.000473941 and mean error is 0.000444221. */
static int fastmath_cosfast (lua_State *L) {
  lua_pushnumber(L, tools_cosfast(agn_checknumber(L, 1)));
  return 1;
}


/* Returns an approximation of the tangent of the given number x, in radians. 2.14.6
   11 % slower than Sun's sun_tan
   In the range -1000 .. 1000 with step size 0.001, median error is 0.000467639 and mean error is 0.1244864. */
static int fastmath_tanfast (lua_State *L) {
  lua_Number x, si, co;
  x = agn_checknumber(L, 1);
  si = tools_sinfast(x);
  co = tools_cosfast(x);
  lua_pushnumber(L, co == 0 ? AGN_NAN : si/co);  /* 2.32.2 patch */
  return 1;
}


/* Approximates sqrt(x) but is much slower than the sqrt operator. If x < 0, returns `undefined`.
   In the range 0 .. 1000 with step size 0.1, median error is 0.00791 and mean error is 0.00829. */
static int fastmath_sqrtfast (lua_State *L) {  /* 2.11.0 */
  lua_pushnumber(L, tools_sqrtfast(agn_checknumber(L, 1)));
  return 1;
}


/* The function approximates sqrt(x).
   1 % slower than GCC's sqrt.
   In the range 0 .. 1000 with step size 0.1, median error is 0.2600843 and mean error is 0.3900767. */
static int fastmath_sqrtapx (lua_State *L) {  /* 2.15.1 */
  lua_pushnumber(L, tools_sqrtapx(agn_checknumber(L, 1)));
  return 1;
}


/* The function approximates log2(x) for number x, and returns a number. If x <= 0, the result is wrong.
   18 % faster than GCC's log2.
   In the range 0 .. 1000 with step size 0.1, median error is 0.00008242754665 and mean error is 0.0000764132614. */
static int fastmath_lbfast (lua_State *L) {  /* new in 2.14.6 */
  lua_pushnumber(L, tools_lbfast(agn_checknumber(L, 1)));
  return 1;
}


/* The function approximates ln(x) for number x, and returns a number. If x <= 0, the result is wrong.
   18 % faster than GCC's log2.
   In the range 0 .. 1000 with step size 0.001, median error is 0.000057118 and mean error is 0.000052952. */
static int fastmath_lnfast (lua_State *L) {  /* 2.14.6 */
  lua_pushnumber(L, (float)LN2 * tools_lbfast(agn_checknumber(L, 1)));
  return 1;
}


/* Returns the hypotenuse of the two numbers x and y; the return is a number. The function is sixty percent faster than `hypot`,
   but prone two round-off errors. 60 percent faster than `hypot', 27 percent faster than the Agena implementation sqrt(x**2 + y**2). */
static int fastmath_hypotfast (lua_State *L) {  /* 2.14.4 */
  lua_Number x, y;
  x = agn_checknumber(L, 1);
  y = agn_checknumber(L, 2);
  lua_pushnumber(L, sqrt(x*x + y*y));
  return 1;
}


/* Approximates 2^x, 2.14.6.
   9 % faster than tools_exp2.
   In the range -1000 .. 10 with step size 0.001, median error is 1.1755178e-038 and mean error is 0.00003296572. */
static int fastmath_exp2fast (lua_State *L) {
  lua_pushnumber(L, tools_pow2fast(agn_checknumber(L, 1)));
  return 1;
}


/* Approximates e^x with enough precision.
   43% _faster_ than GCC's exp.
   In the range -1000 .. 10 with step size 0.001, median error is 1.1755178-038 and mean error is 0.0005315205.
   For x > 10 accuracy becomes quite bad. 2.14.6 */
static int fastmath_expfast (lua_State *L) {
  lua_pushnumber(L, tools_pow2fast((float)agn_checknumber(L, 1) * (float)INVLN2));
  return 1;
}


/* Taken from `branchless scalar math routines`,
   by Phillip Trudeau
   see: https://raw.githubusercontent.com/pmttavara/pt_math/master/pt_math.h */
static double pt_exp2 (double x) {
  double y;
  long long exponent;
  if (x <= -1022) return 0;
  if (x > 1024) return HUGE_VAL;
  exponent = (long long)(x + 1023);
  x += 1023 - exponent;
  exponent <<= 52;
  { int j; for (j=0; j < CHAR_BIT; j++) ((char *)&y)[j] = ((char *)&exponent)[j]; }
  x *= x*0.339766027260413688582 + 0.660233972739586311418;
  return y*(x + 1);
}

/* Approximates E^x, 2.32.1
   43 % faster than Sun's sun_exp.
   In the range -1000 .. 10 with step size 0.001, median error is 1.446234407e-218 and mean error is 0.0384171. */
static int fastmath_dexp (lua_State *L) {
  lua_pushnumber(L, pt_exp2(INVLN2*agn_checknumber(L, 1)));
  return 1;
}

/* Approximates 2^x, 2.32.1
   5 % faster than tools_exp2.
   In the range -1000 .. 10 with step size 0.001, median error is 1.32809424e-152 and mean error is 0.00254373. */
static int fastmath_dexp2 (lua_State *L) {
  lua_pushnumber(L, pt_exp2(agn_checknumber(L, 1)));
  return 1;
}


/* Approximates 10^x, 2.32.2
   79 % faster than tools_exp2.
   In the range -1000 .. 5 with step size 0.001, median error is 0 and mean error is 0.0717718. */
static int fastmath_dexp10 (lua_State *L) {
  lua_pushnumber(L, pt_exp2(LOG2_10*agn_checknumber(L, 1)));
  return 1;
}


/* Taken from `branchless scalar math routines`, by Phillip Trudeau.
   see: https://raw.githubusercontent.com/pmttavara/pt_math/master/pt_math.h */
static INLINE double pt_log2 (double x) {
  long long y;
  double result;
  if (x == 0) return -HUGE_VAL;
  if (x < 0) return AGN_NAN;
  { int j; for (j=0; j < CHAR_BIT; j++) ((char *)&y)[j] = ((char *)&x)[j]; }
  y >>= 52;
  result = (double)y;
  { int j; for (j=0; j < CHAR_BIT; j++) ((char *)&y)[j] = ((char *)&x)[j]; }
  y = (y & 0x000fffffffffffffLL) | 0x3ff0000000000000LL;
  { int j; for (j=0; j < CHAR_BIT; j++) ((char *)&x)[j] = ((char *)&y)[j]; }
  result += -1024 + x*(-0.33333333333333333*x + 2) - 0.66666666666666666;
  return result;
}

/* Approximates the logarithm for base exp(1).
   22 % faster than sun_log; 2.32.1
   In the range 0 .. 1000 with step size 0.001, median error is 0.003392382 and mean error is 0.003499538.*/
static int fastmath_dlog (lua_State *L) {
  lua_pushnumber(L, LN2*pt_log2(agn_checknumber(L, 1)));
  return 1;
}

/* Approximates the logarithm for base 2.
   26 % faster than sun_log2; 2.32.1
   In the range 0 .. 1000 with step size 0.001, median error is 0.00489417 and mean error is 0.0050487668. */
static int fastmath_dlog2 (lua_State *L) {
  lua_pushnumber(L, pt_log2(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_dlog10 (lua_State *L) {
  lua_pushnumber(L, INVLOG2_10*pt_log2(agn_checknumber(L, 1)));
  return 1;
}


/* Taken from `branchless scalar math routines`, by Phillip Trudeau.
   see: https://raw.githubusercontent.com/pmttavara/pt_math/master/pt_math.h */
static double pt_pow (double a, double b) {
  unsigned long long i;
  double sign = 1;
  if (b < 0) {
    if (a == 0) return AGN_NAN;  /* added to avoid infinity */
    a = 1/a; b = -b;  /* changed, do not call pt_pow recursively */
  }
  i = (unsigned long long)b;
  b -= i;
  if (a < 0) {
    if (b) return AGN_NAN;
    else {
      a = -a;
      if (i & 1) sign = -1;  /* changed, do not call pt_pow recursively; return -pt_pow(a, (double)i);  */
    }
  }
  if (a == 0 && b == 0) {  /* added so that 0^b, b > 0, is 0 and not undefined */
    return (b + i == 0) ? AGN_NAN : 0;
  }
  b = pt_exp2(b*pt_log2(a));
  for (; i; a *= a, i >>= 1) {
    if (i & 1) b *= a;
  }
  return sign*b;
}

/* 48 % faster than Sun's sun_pow.
   In the range -10 .. 10 with step size 0.01, median error is 0.00151075 and mean error is 1782116753241 [sic !]. */
static int fastmath_dpow (lua_State *L) {
  lua_pushnumber(L, pt_pow(agn_checknumber(L, 1), agn_checknumber(L, 2)));
  return 1;
}


static double pt_trunc (double x) {
  if (!((x < 0 ? -x : x) <= 4503599627370495.5)) return x;
  return (double)(long long)x;
}

static INLINE double pt_frac (double x) { return x - pt_trunc(x); }

/* 1 % slower than GCC's trunc.
   In the range -1000 .. 1000 with step size 0.000001, median error is 0 and mean error is 0. */
static int fastmath_dtrunc (lua_State *L) {
  lua_pushnumber(L, pt_trunc(agn_checknumber(L, 1)));
  return 1;
}

/* As fast as Sun's sun_frac.
   In the range -1000 .. 1000 with step size 0.000001, median error is 0 and mean error is 0. */
static int fastmath_dfrac (lua_State *L) {
  lua_pushnumber(L, pt_frac(agn_checknumber(L, 1)));
  return 1;
}


/* ***************************************************************************************************** */
/* For reference tests only:                                                                             */
/* ***************************************************************************************************** */

static int fastmath_sunsin (lua_State *L) {
  lua_pushnumber(L, sun_sin(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccsin (lua_State *L) {
  lua_pushnumber(L, sin(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_suncos (lua_State *L) {
  lua_pushnumber(L, sun_cos(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gcccos (lua_State *L) {
  lua_pushnumber(L, cos(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_suntan (lua_State *L) {
  lua_pushnumber(L, sun_tan(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gcctan (lua_State *L) {
  lua_pushnumber(L, tan(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_sunlog (lua_State *L) {
  lua_pushnumber(L, sun_log(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gcclog (lua_State *L) {
  lua_pushnumber(L, log(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_sunexp (lua_State *L) {
  lua_pushnumber(L, sun_exp(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_sunexp2 (lua_State *L) {
  lua_pushnumber(L, sun_exp2(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_exp2_9 (lua_State *L) {
  lua_pushnumber(L, tools_exp2_9(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_sunexp10 (lua_State *L) {
  lua_pushnumber(L, sun_exp10(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_exp10_9 (lua_State *L) {
  lua_pushnumber(L, tools_exp10_9(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccexp (lua_State *L) {
  lua_pushnumber(L, exp(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_sunlog2 (lua_State *L) {
  lua_pushnumber(L, sun_log2(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gcclog2 (lua_State *L) {
  lua_pushnumber(L, log2(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccsqrt (lua_State *L) {
  lua_pushnumber(L, sqrt(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccrsqrt (lua_State *L) {
  lua_pushnumber(L, 1.0/sqrt(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_sunasinh (lua_State *L) {
  lua_pushnumber(L, sun_asinh(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccasinh (lua_State *L) {
  lua_pushnumber(L, asinh(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccacosh (lua_State *L) {
  lua_pushnumber(L, acosh(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccasin (lua_State *L) {
  lua_pushnumber(L, asin(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccacos (lua_State *L) {
  lua_pushnumber(L, acos(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccatan (lua_State *L) {
  lua_pushnumber(L, atan(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_sunpow (lua_State *L) {
  lua_pushnumber(L, sun_pow(agn_checknumber(L, 1), agn_checknumber(L, 2), 1));
  return 1;
}

static int fastmath_gccatanh (lua_State *L) {
  lua_pushnumber(L, atanh(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccrecip (lua_State *L) {
  lua_pushnumber(L, 1/agn_checknumber(L, 1));
  return 1;
}

static int fastmath_gcctrunc (lua_State *L) {
  lua_pushnumber(L, trunc(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_sunfrac (lua_State *L) {
  lua_pushnumber(L, sun_frac(agn_checknumber(L, 1)));
  return 1;
}


static const luaL_Reg fastmathlib[] = {
  {"expfast", fastmath_expfast},              /* added on January 21, 2019 */
  {"dexp", fastmath_dexp},                    /* added on September 27, 2022 */
  {"dexp2", fastmath_dexp2},                  /* added on September 27, 2022 */
  {"dexp10", fastmath_dexp10},                /* added on September 30, 2022 */
  {"dpow", fastmath_dpow},                    /* added on September 28, 2022 */
  {"hypotfast", fastmath_hypotfast},          /* added on January 01, 2019 */
  {"invroot", fastmath_invroot},              /* added on August 01, 2017 */
  {"invsqrt", fastmath_invsqrt},              /* added on January 21, 2019 */
  {"rsqrt", fastmath_rsqrt},                  /* added on September 28, 2022 */
  {"lbfast", fastmath_lbfast},                /* added on June 16, 2018 */
  {"lnfast", fastmath_lnfast},                /* added on January 21, 2019 */
  {"dlog", fastmath_dlog},                    /* added on September 27, 2022 */
  {"dlog2", fastmath_dlog2},                  /* added on September 27, 2022 */
  {"dlog10", fastmath_dlog10},                /* added on September 30, 2022 */
  {"exp2fast", fastmath_exp2fast},            /* added on January 22, 2019 */
  {"pow2fast", fastmath_exp2fast},            /* BACKWARD compatibility; added on January 22, 2019 */
  {"reciprocal", fastmath_reciprocal},        /* added on August 01, 2017 */
  {"cosfast", fastmath_cosfast},              /* added on January 18, 2019 */
  {"sinfast", fastmath_sinfast},              /* added on January 18, 2019 */
  {"sincosfast", fastmath_sincosfast},        /* added on October 17, 2017 */
  {"dasin", fastmath_dasin},                  /* added on September 27, 2022 */
  {"dacos", fastmath_dacos},                  /* added on September 27, 2022 */
  {"datan", fastmath_datan},                  /* added on September 27, 2022 */
  {"dsin", fastmath_dsin},                    /* added on September 27, 2022 */
  {"dcos", fastmath_dcos},                    /* added on September 27, 2022 */
  {"dtan", fastmath_dtan},                    /* added on September 27, 2022 */
  {"sqroot", fastmath_sqroot},                /* added on August 07, 2017 */
  {"sqrtfast", fastmath_sqrtfast},            /* added on October 17, 2017 */
  {"dsqrt", fastmath_dsqrt},                  /* added on September 27, 2022 */
  {"drsqrt", fastmath_drsqrt},                /* added on September 27, 2022 */
  {"sqrtapx", fastmath_sqrtapx},              /* added on May 23, 2019 */
  {"tanfast", fastmath_tanfast},              /* added on January 18, 2019 */
  {"dtrunc", fastmath_dtrunc},                /* added on September 30, 2022 */
  {"dfrac", fastmath_dfrac},                  /* added on September 30, 2022 */
  /* reference functions */
  {"gccsqrt", fastmath_gccsqrt},
  {"gccacos", fastmath_gccacos},
  {"gccacosh", fastmath_gccacosh},
  {"gccasin", fastmath_gccasin},
  {"gccasinh", fastmath_gccasinh},
  {"gccatanh", fastmath_gccatanh},
  {"gccatan", fastmath_gccatan},
  {"gcccos", fastmath_gcccos},
  {"gccexp", fastmath_gccexp},
  {"gcclog", fastmath_gcclog},
  {"gcclog2", fastmath_gcclog2},
  {"gccsin", fastmath_gccsin},
  {"gccsqrt", fastmath_gccsqrt},
  {"gccrecip", fastmath_gccrecip},
  {"gccrsqrt", fastmath_gccrsqrt},
  {"gcctan", fastmath_gcctan},
  {"gccrsqrt", fastmath_gccrsqrt},
  {"sunasinh", fastmath_sunasinh},
  {"suncos", fastmath_suncos},
  {"sunexp", fastmath_sunexp},
  {"sunexp2", fastmath_sunexp2},
  {"sunexp10", fastmath_sunexp10},
  {"exp2_9", fastmath_exp2_9},
  {"exp10_9", fastmath_exp10_9},
  {"sunlog", fastmath_sunlog},
  {"sunlog2", fastmath_sunlog2},
  {"sunpow", fastmath_sunpow},
  {"sunsin", fastmath_sunsin},
  {"suntan", fastmath_suntan},
  {"gcctrunc", fastmath_gcctrunc},
  {"sunfrac", fastmath_sunfrac},
  {NULL, NULL}
};


/*
** Open fastmath library
*/
LUALIB_API int luaopen_fastmath (lua_State *L) {
  luaL_register(L, AGENA_FASTMATHLIBNAME, fastmathlib);
  return 1;
}

