/*
** $Id: lapi.c,v 2.55 2006/06/07 12:37:17 roberto Exp $
** Lua API
** See Copyright Notice in agena.h
*/

#include <assert.h>
#include "prepdefs.h"

#if (defined(__SOLARIS) || defined(__linux__) || defined(__APPLE__))  /* 2.8.4, trunc, log2, exp2, fma, fdim */
#include <tgmath.h>
#endif

#include <math.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>  /* for isdigit */

#ifndef LUA_DOS
#include <fenv.h>   /* for feg/setround */
#endif

#define lapi_c
#define LUA_CORE

#include "agena.h"
#include "agnconf.h"
#include "agncmpt.h"
#include "sofa.h"
#include "agnhlps.h"  /* for tools_isfloat */

#include "agnxlib.h"
#include "agenalib.h"
#include "lcomplex.h"
#include "ldebug.h"
#include "ldo.h"
#include "lfunc.h"
#include "lgc.h"
#include "lmem.h"
#include "lobject.h"
#include "lpair.h"
#include "lseq.h"
#include "lreg.h"
#include "lset.h"
#include "lstate.h"
#include "lstring.h"
#include "ltable.h"
#include "ltm.h"
#include "lundump.h"
#include "lvm.h"
#include "lopcodes.h"
#include "llex.h"
#include "lstrlib.h"

const char lua_ident[] =
  "$Agena: " AGENA_RELEASE " " AGENA_COPYRIGHT " $\n"
  "$Authors: " LUA_AUTHORS " $\n"
  "$URL: agena.sourceforge.net $\n";

#define api_checknelems(L, n)   api_check(L, (n) <= (L->top - L->base))

#define api_checkvalidindex(L, i)   api_check(L, (i) != luaO_nilobject)

#define api_incr_top(L)   {api_check(L, L->top < L->ci->top); L->top++;}

#define api_incr_top_by_three(L)   {api_check(L, L->top + 1 < L->ci->top); L->top += 3;}

#define api_incr_top_by_two(L)   {api_check(L, L->top + 1 < L->ci->top); L->top += 2;}


LUA_API TValue *index2adr (lua_State *L, int idx) {
  if (idx > 0) {
    TValue *o = L->base + (idx - 1);
    api_check(L, idx <= L->ci->top - L->base);
    if (o >= L->top) return cast(TValue *, luaO_nilobject);
    else return o;
  }
  else if (idx > LUA_REGISTRYINDEX) {
    api_check(L, idx != 0 && -idx <= L->top - L->base);
    return L->top + idx;
  }
  else switch (idx) {  /* pseudo-indices */
    case LUA_REGISTRYINDEX: return registry(L);
    case LUA_ENVIRONINDEX: {
      Closure *func = curr_func(L);
      sethvalue(L, &L->env, func->c.env);
      return &L->env;
    }
    case LUA_GLOBALSINDEX: return gt(L);
    case LUA_CONSTANTSINDEX: return constants(L);
    default: {  /* upvalues */
      Closure *func = curr_func(L);
      /* idx = LUA_CONSTANTSINDEX - idx; */
      idx = lua_upvalueindex(idx);  /* changed 2.20.0 */
      /* idx now is positive */
      return (idx <= func->c.nupvalues)
                ? &func->c.upvalue[idx - 1]
                : cast(TValue *, luaO_nilobject);
    }
  }
}


LUA_API GCObject *index2gco (lua_State *L, int idx) {  /* 2.9.1 */
  return obj2gco(index2adr(L, idx));
}


static Table *getcurrenv (lua_State *L) {
  if (L->ci == L->base_ci)  /* no enclosing function? */
    return hvalue(gt(L));  /* use global table as environment */
  else {
    Closure *func = curr_func(L);
    return func->c.env;
  }
}


void luaA_pushobject (lua_State *L, const TValue *o) {
  setobj2s(L, L->top, o);
  api_incr_top(L);
}


LUA_API int lua_checkstack (lua_State *L, int size) {
  /* 5.1.4 patch 3 */
  int res = 1;
  lua_lock(L);
  if (size > LUAI_MAXCSTACK || (L->top - L->base + size) > LUAI_MAXCSTACK)
    res = 0;  /* stack overflow */
  else if (size > 0) {
    luaD_checkstack(L, size);
    if (L->ci->top < L->top + size)
      L->ci->top = L->top + size;
  }
  lua_unlock(L);
  return res;
}


LUA_API void lua_xmove (lua_State *from, lua_State *to, int n) {
  int i;
  if (from == to) return;
  lua_lock(to);
  api_checknelems(from, n);
  api_check(from, G(from) == G(to));
  api_check(from, to->ci->top - to->top >= n);
  from->top -= n;
  for (i = 0; i < n; i++) {
    setobj2s(to, to->top++, from->top + i);
  }
  lua_unlock(to);
}


LUA_API lua_CFunction lua_atpanic (lua_State *L, lua_CFunction panicf) {
  lua_CFunction old;
  lua_lock(L);
  old = G(L)->panic;
  G(L)->panic = panicf;
  lua_unlock(L);
  return old;
}


LUA_API lua_State *lua_newthread (lua_State *L) {
  lua_State *L1;
  lua_lock(L);
  luaC_checkGC(L);
  L1 = luaE_newthread(L);
  setthvalue(L, L->top, L1);
  api_incr_top(L);
  lua_unlock(L);
  luai_userstatethread(L, L1);
  return L1;
}


/*
** basic stack manipulation
*/

/*
** convert an acceptable stack index into an absolute index; taken from Lua 5.4.0 RC 2, 2.21.2
*/
LUA_API int lua_absindex (lua_State *L, int idx) {
  return (idx > 0 || ispseudo(idx))
         ? idx
         : cast_int(L->top - L->ci->func) + idx;
}


LUA_API int lua_gettop (lua_State *L) {
  return cast_int(L->top - L->base);
}


LUA_API void lua_settop (lua_State *L, int idx) {
  lua_lock(L);
  if (idx >= 0) {
    api_check(L, idx <= L->stack_last - L->base);
    while (L->top < L->base + idx)
      setnilvalue(L->top++);
    L->top = L->base + idx;
  }
  else {
    api_check(L, -(idx + 1) <= (L->top - L->base));
    L->top += idx + 1;  /* `subtract' index (index is negative) */
  }
  lua_unlock(L);
}


LUA_API void lua_remove (lua_State *L, int idx) {
  StkId p;
  lua_lock(L);
  p = index2adr(L, idx);
  api_checkvalidindex(L, p);
  while (++p < L->top) setobjs2s(L, p - 1, p);
  L->top--;
  lua_unlock(L);
}


LUA_API void lua_insert (lua_State *L, int idx) {
  StkId p, q;
  lua_lock(L);
  p = index2adr(L, idx);
  api_checkvalidindex(L, p);
  for (q = L->top; q > p; q--) setobjs2s(L, q, q - 1);
  setobjs2s(L, p, L->top);
  lua_unlock(L);
}


LUA_API void lua_replace (lua_State *L, int idx) {
  StkId o;
  lua_lock(L);
  /* explicit test for incompatible code */
  if (idx == LUA_ENVIRONINDEX && L->ci == L->base_ci)
    luaG_runerror(L, "Error: no calling environment.");
  api_checknelems(L, 1);
  o = index2adr(L, idx);
  api_checkvalidindex(L, o);
  if (idx == LUA_ENVIRONINDEX) {
    Closure *func = curr_func(L);
    api_check(L, ttistable(L->top - 1));
    func->c.env = hvalue(L->top - 1);
    luaC_barrier(L, func, L->top - 1);
  } else {
    setobj(L, o, L->top - 1);
    if (idx < LUA_CONSTANTSINDEX)  /* function upvalue ?  Fix 2.20.3 */
      luaC_barrier(L, curr_func(L), L->top - 1);
  }
  L->top--;
  lua_unlock(L);
}


LUA_API void lua_pushvalue (lua_State *L, int idx) {
  lua_lock(L);
  setobj2s(L, L->top, index2adr(L, idx));
  api_incr_top(L);
  lua_unlock(L);
}



/*
** access functions (stack -> C)
*/


LUA_API int lua_type (lua_State *L, int idx) {
  StkId o = index2adr(L, idx);
  return (o == luaO_nilobject) ? LUA_TNONE : ttype(o);
}


LUA_API const char *lua_typename (lua_State *L, int t) {
  UNUSED(L);
  return (t == LUA_TNONE) ? "no value" : luaT_typenames[t];
}


LUA_API int lua_iscfunction (lua_State *L, int idx) {
  StkId o = index2adr(L, idx);
  return iscfunction(o);
}


LUA_API int lua_isnumber (lua_State *L, int idx) {  /* tuned by 40 %, 2.21.1 */
  lua_Number n;
  const TValue *o = index2adr(L, idx);
  return tonumberx(o, &n);
}


LUA_API int lua_isstring (lua_State *L, int idx) {
  int t = lua_type(L, idx);
  return (t == LUA_TSTRING || t == LUA_TNUMBER);
}


LUA_API int lua_isuserdata (lua_State *L, int idx) {
  const TValue *o = index2adr(L, idx);
  return (ttisuserdata(o) || ttislightuserdata(o));
}


LUA_API int lua_rawequal (lua_State *L, int index1, int index2) {
  StkId o1 = index2adr(L, index1);
  StkId o2 = index2adr(L, index2);
  return (o1 == luaO_nilobject || o2 == luaO_nilobject) ? 0
         : luaO_rawequalObj(o1, o2);
}


LUA_API int lua_rawaequal (lua_State *L, int index1, int index2) {  /* 2.3.0 RC 3 */
  StkId o1 = index2adr(L, index1);
  StkId o2 = index2adr(L, index2);
  return (o1 == luaO_nilobject || o2 == luaO_nilobject) ? 0
         : luaV_approxequalvalonebyone(L, o1, o2);
}


LUA_API int lua_equal (lua_State *L, int index1, int index2) {
  StkId o1, o2;
  int i;
  lua_lock(L);  /* may call tag method */
  o1 = index2adr(L, index1);
  o2 = index2adr(L, index2);
  i = (o1 == luaO_nilobject || o2 == luaO_nilobject) ? 0 : equalobj(L, o1, o2);
  lua_unlock(L);
  return i;
}


LUA_API int agn_equalref (lua_State *L, int index1, int index2) {  /* 2.12.3, Lua 5.1.4's lua_equal API function */
  StkId o1, o2;
  int i;
  lua_lock(L);  /* may call tag method */
  o1 = index2adr(L, index1);
  o2 = index2adr(L, index2);
  i = (o1 == luaO_nilobject || o2 == luaO_nilobject) ? 0 : equalref(L, o1, o2);
  lua_unlock(L);
  return i;
}


LUA_API int lua_lessthan (lua_State *L, int index1, int index2) {
  StkId o1, o2;
  int i;
  lua_lock(L);  /* may call tag method */
  o1 = index2adr(L, index1);
  o2 = index2adr(L, index2);
  i = (o1 == luaO_nilobject || o2 == luaO_nilobject) ? 0
       : luaV_lessthan(L, o1, o2, 1);
  lua_unlock(L);
  return i;
}


LUA_API lua_Number lua_tonumber (lua_State *L, int idx) {  /* tuned by 40 %, 2.21.1 */
  int r;
  lua_Number n;
  const TValue *o = index2adr(L, idx);
  r = tonumberx(o, &n);
  return (r) ? n : 0;  /* 0 = failure, could not convert value */
}


LUA_API lua_Number agn_tonumber (lua_State *L, int idx) {
  const TValue *o = index2adr(L, idx);
  return nvalue(o);
}


/* do not delete this function since it is absolutely needed by `toNumber` to convert strings to numbers
   (agn_tonumber does not do this) */
LUA_API lua_Number agn_tonumberx (lua_State *L, int idx, int *exception) {  /* added October 18, 2008 */
  int r;  /* tuned 2.21.1 */
  lua_Number n;
  const TValue *o;
  o = index2adr(L, idx);
  r = tonumberx(o, &n);
  *exception = r == 0;
  return (r) ? n : 0;  /* 0 = failure, could not convert value */
}


#ifndef PROPCMPLX
LUA_API agn_Complex agn_tocomplexx (lua_State *L, int idx, int *exception) {  /* added 10.08.2009, 0.26.1 */
  TValue n;
  const TValue *o = index2adr(L, idx);
  if (tocomplex(o, &n)) {
    *exception = 0;
    return cvalue(o);
  }
  else {  /* failure, could not convert value */
    *exception = 1;
    return 0;
  }
}

LUA_API agn_Complex agn_tocomplex (lua_State *L, int idx) {
  StkId o = index2adr(L, idx);
  return cvalue(o);
}

LUA_API lua_Number agn_complexreal (lua_State *L, int idx) {  /* 2.21.8 */
  StkId o = index2adr(L, idx);
  return creal(cvalue(o));
}

LUA_API lua_Number agn_compleximag (lua_State *L, int idx) {  /* 2.21.8 */
  StkId o = index2adr(L, idx);
  return cimag(cvalue(o));
}

#else

LUA_API void agn_tocomplexx (lua_State *L, int idx, int *exception, lua_Number *a) {  /* added 10.08.2009, 0.26.1 */
  TValue n;
  const TValue *o = index2adr(L, idx);
  if (tocomplex(o, &n)) {
    *exception = 0;
    a[0] = cvalue(o)[0];
    a[1] = cvalue(o)[1];
  }
  else {  /* failure, could not convert value */
    a[0] = a[1] = 0;
    *exception = 1;
  }
}


LUA_API lua_Number agn_complexreal (lua_State *L, int idx) {
  StkId o = index2adr(L, idx);
  return cvalue(o)[0];
}

LUA_API lua_Number agn_compleximag (lua_State *L, int idx) {
  StkId o = index2adr(L, idx);
  return cvalue(o)[1];
}
#endif


/* Expects a complex number at stack position idx and returns its real and imaginary part in re and im.
   If the value at idx is a complex number, returns 1 and 0 otherwise. The function can be used both with
   Agena versions using standard complex.h complex functions and those using proprietary complex arithmetic.
   2.17.6 */
LUA_API int agn_getcmplxparts (lua_State *L, int idx, double *re, double *im) {
  if (lua_iscomplex(L, idx)) {
#ifndef PROPCMPLX
    agn_Complex z;
    StkId o = index2adr(L, idx);
    z = cvalue(o);
    *re = creal(z); *im = cimag(z);
#else
    StkId o = index2adr(L, idx);
    *re = cvalue(o)[0]; *im = cvalue(o)[1];
#endif
    return 1;
  } else if (agn_isnumber(L, idx)) {  /* 2.31.6 */
    *re = agn_tonumber(L, idx);
    *im = 0;
    return 1;
  }
  return 0;
}


LUA_API lua_Integer lua_tointeger (lua_State *L, int idx) {  /* tuned 2.21.1 */
  lua_Number n;
  const TValue *o = index2adr(L, idx);
  if (tonumberx(o, &n)) {
    lua_Integer res;
    lua_number2integer(res, n);
    return res;
  }
  else
    return 0;
}


LUA_API lua_Integer agn_tointeger (lua_State *L, int idx) {  /* 2.27.8 */
  const TValue *o = index2adr(L, idx);
  return (lua_Integer)(nvalue(o));
}


LUA_API int32_t lua_toint32_t (lua_State *L, int idx) {  /* tuned 2.21.2 */
  lua_Number n;
  const TValue *o = index2adr(L, idx);
  if (tonumberx(o, &n)) {
#if defined(__APPLE__) || defined(__ARMCPU) || defined(__PPCCPU)
    n = tools_reducerange(n, -2147483648.0, 2147483648.0);
#endif
    return (int32_t)n;
  }
  else
    return 0;
}


LUA_API uint32_t lua_touint32_t (lua_State *L, int idx) {  /* 2.14.5, tuned 2.21.2, patched 2.31.1 for PPC, ARM, Apple */
  lua_Number n;
  const TValue *o = index2adr(L, idx);
  if (tonumberx(o, &n)) {
#if defined(__APPLE__) || defined(__ARMCPU) || defined(__PPCCPU)
    /* 2.31.1/2.31.2 patch to work as on `normal` x86 platforms: force wrap-around for large values. 4294967296.0 = 2^32 */
    if (!(n >= 0 && n < 4294967296.0))  /* 80 percent faster than calling tools_reducerange(n, 0, 4294967296.0); 2.31.2 tweak */
      n = luai_nummod(n, 4294967296.0);
#endif
    return (uint32_t)n;
  }
  else
    return 0;
}


LUA_API off64_t lua_tooff64_t (lua_State *L, int idx) {  /* tuned 2.21.2 */
  lua_Number n;
  const TValue *o = index2adr(L, idx);
  if (tonumberx(o, &n)) {
    off64_t res;
    res = (off64_t)n;
    return res;
  }
  else
    return 0;
}


LUA_API int agn_isinteger (lua_State *L, int idx) {  /* 2.16.6 */
  StkId o = index2adr(L, idx);
  if (o == luaO_nilobject || ttype(o) != LUA_TNUMBER)
    return 0;
  else
    return tools_isint(nvalue(o));
}


LUA_API int agn_isfloat (lua_State *L, int idx) {  /* 2.21.6 */
  StkId o = index2adr(L, idx);
  if (o == luaO_nilobject || ttype(o) != LUA_TNUMBER)
    return 0;
  else
    return tools_isfloat(nvalue(o));
}


LUA_API int lua_toboolean (lua_State *L, int idx) {
  const TValue *o = index2adr(L, idx);
  if l_isfail(o)
    return LUA_TFAIL;  /* = -1 */
  else
    return !l_isfalse(o);  /* with non-booleans, returns !0 == 1 */
}


LUA_API const char *lua_tolstring (lua_State *L, int idx, size_t *len) {
  StkId o = index2adr(L, idx);
  if (!ttisstring(o)) {
    lua_lock(L);  /* `luaV_tostring' may create a new string */
    if (!luaV_tostring(L, o)) {  /* conversion failed? */
      if (len != NULL) *len = 0;
      lua_unlock(L);
      return NULL;
    }
    luaC_checkGC(L);
    o = index2adr(L, idx);  /* previous call may reallocate the stack */
    lua_unlock(L);
  }
  if (len != NULL) *len = tsvalue(o)->len;
  return svalue(o);
}


LUA_API const char *agn_tostring (lua_State *L, int idx) {  /* 0.24.0 */
  StkId o = index2adr(L, idx);
  return svalue(o);
}


LUA_API size_t lua_objlen (lua_State *L, int idx) {
  StkId o = index2adr(L, idx);
  switch (ttype(o)) {
    case LUA_TSTRING: return tsvalue(o)->len;
    case LUA_TUSERDATA: return uvalue(o)->len;
    case LUA_TTABLE: return luaH_getn(hvalue(o));
    case LUA_TNUMBER: {
      size_t l;
      lua_lock(L);  /* `luaV_tostring' may create a new string */
      l = (luaV_tostring(L, o) ? tsvalue(o)->len : 0);
      lua_unlock(L);
      return l;
    }
    default: return 0;
  }
}


LUA_API lua_CFunction lua_tocfunction (lua_State *L, int idx) {
  StkId o = index2adr(L, idx);
  return (!iscfunction(o)) ? NULL : clvalue(o)->c.f;
}


LUA_API void *lua_touserdata (lua_State *L, int idx) {
  StkId o = index2adr(L, idx);
  switch (ttype(o)) {
    case LUA_TUSERDATA: return (rawuvalue(o) + 1);
    case LUA_TLIGHTUSERDATA: return pvalue(o);
    default: return NULL;
  }
}


LUA_API lua_State *lua_tothread (lua_State *L, int idx) {
  StkId o = index2adr(L, idx);
  return (!ttisthread(o)) ? NULL : thvalue(o);
}


LUA_API const void *lua_topointer (lua_State *L, int idx) {
  StkId o = index2adr(L, idx);
  switch (ttype(o)) {
    case LUA_TTABLE: return hvalue(o);
    case LUA_TFUNCTION: return clvalue(o);
    case LUA_TTHREAD: return thvalue(o);
    case LUA_TUSERDATA:
    case LUA_TLIGHTUSERDATA:
      return lua_touserdata(L, idx);
    case LUA_TSET: return usvalue(o);      /* added 0.10.0 */
    case LUA_TSEQ: return seqvalue(o);     /* added 0.11.0 */
    case LUA_TREG: return regvalue(o);     /* added 2.3.0 RC 3 */
    case LUA_TPAIR: return pairvalue(o);   /* added 0.11.1 */
    default: return NULL;
  }
}



/*
** push functions (C -> stack)
*/


LUA_API void lua_pushnil (lua_State *L) {
  lua_lock(L);
  setnilvalue(L->top);
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API void lua_pushnumber (lua_State *L, lua_Number n) {
  lua_lock(L);
  setnvalue(L->top, n);
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API void lua_pushundefined (lua_State *L) {
  lua_lock(L);
  setnvalue(L->top, AGN_NAN);
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API void lua_pushinteger (lua_State *L, lua_Integer n) {
  lua_lock(L);
  setnvalue(L->top, cast_num(n));
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API void lua_pushlstring (lua_State *L, const char *s, size_t len) {
  lua_lock(L);
  luaC_checkGC(L);
  setsvalue2s(L, L->top, luaS_newlstr(L, s, len));
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API void lua_pushstring (lua_State *L, const char *s) {
  if (s == NULL)
    lua_pushnil(L);
  else
    lua_pushlstring(L, s, tools_strlen(s));  /* 2.17.8 tweak */
}


LUA_API void lua_pushchar (lua_State *L, char s) {  /* 2.12.0 RC 2 */
  char str[2];
  str[0] = s;
  str[1] = '\0';
  lua_pushstring(L, str);
}


LUA_API const char *lua_pushvfstring (lua_State *L, const char *fmt,
                                      va_list argp) {
  const char *ret;
  lua_lock(L);
  luaC_checkGC(L);
  ret = luaO_pushvfstring(L, fmt, argp);
  lua_unlock(L);
  return ret;
}


LUA_API const char *lua_pushfstring (lua_State *L, const char *fmt, ...) {
  const char *ret;
  va_list argp;
  lua_lock(L);
  luaC_checkGC(L);
  va_start(argp, fmt);
  ret = luaO_pushvfstring(L, fmt, argp);
  va_end(argp);
  lua_unlock(L);
  return ret;
}


LUA_API void lua_pushcclosure (lua_State *L, lua_CFunction fn, int n) {
  Closure *cl;
  lua_lock(L);
  luaC_checkGC(L);
  api_checknelems(L, n);
  cl = luaF_newCclosure(L, n, getcurrenv(L));
  cl->c.f = fn;
  L->top -= n;
  while (n--)
    setobj2n(L, &cl->c.upvalue[n], L->top + n);
  setclvalue(L, L->top, cl);
  lua_assert(iswhite(obj2gco(cl)));
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API void lua_pushboolean (lua_State *L, int b) {
  lua_lock(L);
  setbvalue(L->top, b != 0);  /* ensure that true is 1 */
  api_incr_top(L);
  lua_unlock(L);
}


/* b: -1 = fail, 0 = false, 1 = true */
LUA_API void agn_pushboolean (lua_State *L, int b) {  /* 16.01.2013 */
  lua_lock(L);
  setbvalue(L->top, (b == -1) + (b != 0));  /* ensure that true is 1, no branching 2.17.4 */
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API void lua_pushlightuserdata (lua_State *L, void *p) {
  lua_lock(L);
  setpvalue(L->top, p);
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API int lua_pushthread (lua_State *L) {
  lua_lock(L);
  setthvalue(L, L->top, L);
  api_incr_top(L);
  lua_unlock(L);
  return (G(L)->mainthread == L);
}



/*
** get functions (Lua -> stack)
*/


/* taken from Lua 5.4.0 RC 4, 2.21.2, 5 % faster */
LUA_API int lua_gettable (lua_State *L, int idx) {
  const TValue *slot;
  TValue *t;
  lua_lock(L);
  t = index2value(L, idx);
  if (luaV_fastget(L, t, s2v(L->top - 1), slot, luaH_get)) {
    setobj2s(L, L->top - 1, slot);
  } else
    luaV_finishget(L, t, s2v(L->top - 1), L->top - 1, slot);
  lua_unlock(L);
  return ttype(s2v(L->top - 1));
}


/* taken from Lua 5.4.0 RC 4, 2.21.2, 10 % faster */
LUA_API int lua_getfield (lua_State *L, int idx, const char *k) {
  const TValue *slot, *t;
  TString *str;
  lua_lock(L);
  t = index2value(L, idx);
  str = luaS_new(L, k);
  if (luaV_fastget(L, t, str, slot, luaH_getstr)) {
    setobj2s(L, L->top, slot);
    api_incr_top(L);
  } else {
    setsvalue2s(L, L->top, str);
    api_incr_top(L);
    luaV_finishget(L, t, s2v(L->top - 1), L->top - 1, slot);
  }
  lua_unlock(L);
  return ttype(s2v(L->top - 1));
}


/* Checks whether a table at index idx includes a given field (a string key) and returns 0 (false) or 1 (true).
   Metamethods are ignored. The stack is not changed. 2.21.0 */
LUA_API int lua_hasfield (lua_State *L, int idx, const char *k) {
  StkId t;
  TValue key;
  int r;
  lua_lock(L);
  t = index2adr(L, idx);
  api_checkvalidindex(L, t);
  setsvalue(L, &key, luaS_new(L, k));
  r = ttisnotnil(luaH_get(hvalue(t), &key));  /* do a primitive get */
  lua_unlock(L);
  return r;
}


/* Pushes onto the stack the value t[i], where t is the value at the given index. As in Lua, this function may trigger a
   metamethod for the "index" event. Returns the type of the pushed value. 2.21.1, taken from Lua 5.4.0 RC 4 */
LUA_API int lua_geti (lua_State *L, int idx, lua_Integer n) {
  TValue *t;
  const TValue *slot;
  lua_lock(L);
  t = index2value(L, idx);
  if (luaV_fastgeti(L, t, n, slot)) {
    setobj2s(L, L->top, slot);
  }
  else {
    TValue aux;
    setivalue(&aux, n);
    luaV_finishget(L, t, &aux, L->top, slot);
  }
  api_incr_top(L);
  lua_unlock(L);
  return ttype(s2v(L->top - 1));
}


/* Does the equivalent to t[n] = v, where t is the value at the given index and v is the value at the top of the stack.
   This function pops the value from the stack. As in Lua, this function may trigger a metamethod for the "newindex" event.
   2.21.1, taken from Lua 5.4.0 RC 4 */
LUA_API void lua_seti (lua_State *L, int idx, lua_Integer n) {
  TValue *t;
  const TValue *slot;
  lua_lock(L);
  api_checknelems(L, 1);
  t = index2value(L, idx);
  if (luaV_fastgeti(L, t, n, slot)) {
    luaV_finishfastset(L, t, slot, s2v(L->top - 1));
  }
  else {
    TValue aux;
    setivalue(&aux, n);
    luaV_finishset(L, t, &aux, s2v(L->top - 1), slot);
  }
  L->top--;  /* pop value */
  lua_unlock(L);
}


LUA_API int lua_rawget (lua_State *L, int idx) {
  StkId t;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttistable(t));
  setobj2s(L, L->top - 1, luaH_get(hvalue(t), L->top - 1));
  lua_unlock(L);
  return ttype(s2v(L->top - 1));  /* from Lua 5.4.0 RC 4; 2.21.1 */
}


/* check whether the value at the stack top exists in the set at stack index idx. Pops the value and puts true or false in its place. */
LUA_API int lua_srawget (lua_State *L, int idx) {  /* 0.10.0 */
  StkId t;
  int r;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttisuset(t));
  r = agnUS_get(usvalue(t), L->top - 1);
  setbvalue(L->top - 1, r);
  lua_unlock(L);
  return r;  /* 2.12.3 */
}


/* Checks whether the value at the stack top exists in the set at stack index idx. Returns 0 or 1. If pop is 1, pops the value, otherwise leaves
   the stack unchanged. */
LUA_API int lua_shas (lua_State *L, int idx, int pop) {  /* 2.31.5 */
  StkId t;
  int r;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttisuset(t));
  r = agnUS_get(usvalue(t), L->top - 1);
  L->top -= pop;
  lua_unlock(L);
  return r;  /* 2.12.3 */
}


LUA_API void lua_seqrawget (lua_State *L, int idx, int pushnil) {
  StkId t;
  const TValue *r;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttisseq(t));
  r = agnSeq_rawgeti(seqvalue(t), L->top - 1);
  if (r != NULL) {
    setobj2s(L, L->top - 1, r);
  } else if (pushnil) {
    setobj2s(L, L->top - 1, luaO_nilobject);
  } else {
    luaG_runerror(L, "Error: index out of range.");
  }
  lua_unlock(L);
}


LUA_API void agn_reggeti (lua_State *L, int idx, size_t n) {  /* 2.3.0 RC 3 */
  StkId t;
  const TValue *r;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttisreg(t));
  r = agnReg_geti(regvalue(t), n);
  if (r != NULL) {
    setobj2s(L, L->top, r);
    api_incr_top(L);
  }
  else
    luaG_runerror(L, "Error: index out of %s range.", (regvalue(t)->top < regvalue(t)->maxsize) ? "current" : "\b");
  lua_unlock(L);
}


LUA_API lua_Number agn_reggetinumber (lua_State *L, int idx, int n) {  /* 2.3.0 RC 3, tuned by 4 % 2.21.2 */
  StkId t;
  const TValue *o;
  lua_Number p;
  t = index2adr(L, idx);
  api_check(L, ttisreg(t));
  o = agnReg_geti(regvalue(t), n);
  if (tonumberx(o, &p))
    return nvalue(o);
  else
    return 0;  /* changed 2.75.5 */
}


LUA_API lua_Number agn_regrawgetinumber (lua_State *L, int idx, int n, int *rc) {  /* 2.27.5 */
  StkId t;
  const TValue *o;
  lua_Number p;
  t = index2adr(L, idx);
  api_check(L, ttisreg(t));
  o = agnReg_geti(regvalue(t), n);
  if ( (*rc = tonumberx(o, &p)) )
    return nvalue(o);
  else
    return HUGE_VAL;
}


LUA_API void agn_regrawget (lua_State *L, int idx, int pushnil) {  /* 2.3.0 RC 3 */
  StkId t;
  const TValue *r;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttisreg(t));
  r = agnReg_rawgeti(regvalue(t), L->top - 1);
  if (r != NULL) {
    setobj2s(L, L->top - 1, r);
  } else if (pushnil) {
    setobj2s(L, L->top - 1, luaO_nilobject);
  } else
    luaG_runerror(L, "Error: index out of %s range.", (regvalue(t)->top < regvalue(t)->maxsize) ? "current" : "\b");
  lua_unlock(L);
}


LUA_API int agn_regreduce (lua_State *L, int idx, size_t newsize, int nil) {  /* 2.3.0 RC 3 */
  int r;
  StkId t;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttisreg(t));
  r = agnReg_reduce(L, regvalue(t), newsize, nil);
  lua_unlock(L);
  return r;
}


LUA_API int agn_regextend (lua_State *L, int idx, size_t newsize) {  /* 2.3.0 RC 3 */
  int r;
  StkId t;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttisreg(t));
  r = agnReg_extend(L, regvalue(t), newsize);
  lua_unlock(L);
  return r;
}


LUA_API void agn_seqresize (lua_State *L, int idx, size_t newsize) {  /* 2.3.0 RC 4 */
  StkId t;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttisseq(t));
  agnSeq_resize(L, seqvalue(t), newsize);
  lua_unlock(L);
}


#define actnodesize(t)    ( ((t)->lsizenode) == 0 ? 0 : (1 << ((t)->lsizenode)))  /* patched Agena 1.0.3 */

/* Resizes the set at stack index `idx` to `newsize` pre-allocated slots. `protect` controls whether to allow only a set to be shrunk
   without dropping any elements (protect = 1), shrunk or enlarged without dropping any elements (protect == 2), or whether to have
   full control what may happen: shrinking or expanding, dropping or not dropping any elements (protect == 0). With protect == 0, it is
   advised that the set is empty. 2.33.4 */
LUA_API void agn_setresize (lua_State *L, int idx, size_t newsize, int protect) {
  StkId t;
  UltraSet *us;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttisuset(t));
  us = usvalue(t);
  if ( protect == 0 ||  /* resize whatever might happen, should be chosen only if set is empty */
      (protect == 1 && newsize >= us->size && newsize < actnodesize(us)) ||  /* only shrink if possible, never drop anything */
      (protect == 2 && newsize >= us->size && newsize != actnodesize(us)) )  /* enlarge or shrink if possible, never drop anything */
    agnUS_resize(L, us, newsize);
  lua_unlock(L);
}


/* Resizes the array part of a table to the given number newsize of pre-allocated slots.
   If you do not know whether there are still non-null values beyond the new size in the array part,
   then set checkholes to 1 instead of 0 as otherwise Agena will crash in such situations. If set
   to 1, the function will not resize the table if there are any excess non-null values. */
LUA_API void agn_tabresize (lua_State *L, int idx, size_t newsize, int checkholes) {  /* 2.29.0 */
  StkId t;
  Table *h;
  int lastnonnil;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttistable(t));
  h = hvalue(t);
  if (newsize == h->sizearray || newsize == 0) {
    lua_unlock(L);
    return;
  }
  if (checkholes) {
    agnH_hasholes(L, h, &lastnonnil);
    /* if (r || lastnonnil >= newsize) { */
    if (lastnonnil >= newsize) {
      lua_unlock(L);
      return;
    }
    /* int i; for (i=h->sizearray; i > newsize; i--) {
      setnilvalue(&h->array[i - 1]);
    } */
  }
  luaH_resizearray(L, h, newsize);
  lua_unlock(L);
}


/* agn_pairrawget: idx should be a stack value k with k=1 for the left hand side, or
   k=2 for the right hand side */
LUA_API void agn_pairrawget (lua_State *L, int idx) {
  StkId t;
  const TValue *r;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttispair(t));
  r = agnPair_rawgeti(pairvalue(t), L->top - 1);
  if (r != NULL) {
    setobj2s(L, L->top - 1, r);
  }
  else
    luaG_runerror(L, "Error, index out of range.");
  lua_unlock(L);
}


/* get a value at index n from a sequence which resides at stack pos */
LUA_API void lua_seqrawgeti (lua_State *L, int idx, size_t n) {  /* patched 0.31.6 */
  StkId t;
  const TValue *r;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttisseq(t));
  r = agnSeq_geti(seqvalue(t), n);
  if (r != NULL) {
    setobj2s(L, L->top, r);
    api_incr_top(L);
  }
  else
    luaG_runerror(L, "Error, index out of range.");
  lua_unlock(L);
}


LUA_API lua_Number lua_seqrawgetinumber (lua_State *L, int idx, int n) {  /* 0.12.0, tuned by 4 % 2.21.2 */
  StkId t;
  const TValue *o;
  lua_Number p;
  t = index2adr(L, idx);
  api_check(L, ttisseq(t));
  o = agnSeq_geti(seqvalue(t), n);
  if (tonumberx(o, &p))
    return p;
  else
    return HUGE_VAL;
}


LUA_API lua_Number agn_seqgetinumber (lua_State *L, int idx, int n) {  /* 2.2.0 RC 2, tuned by 4 % 2.21.2 */
  StkId t;
  const TValue *o;
  lua_Number p;
  t = index2adr(L, idx);
  api_check(L, ttisseq(t));
  o = agnSeq_geti(seqvalue(t), n);
  if (tonumberx(o, &p))
    return p;
  else
    return 0;
}


LUA_API lua_Number agn_seqrawgetinumber (lua_State *L, int idx, int n, int *rc) {  /* 2.27.5 */
  StkId t;
  const TValue *o;
  lua_Number p;
  t = index2adr(L, idx);
  api_check(L, ttisseq(t));
  o = agnSeq_geti(seqvalue(t), n);
  if ( (*rc = tonumberx(o, &p)) )
    return p;
  else
    return HUGE_VAL;
}


LUA_API void agn_createpair (lua_State *L, int idxleft, int idxright) {  /* 0.11.1; tuned 0.12.2 */
  StkId l, r;
  lua_lock(L);
  l = index2adr(L, idxleft);
  r = index2adr(L, idxright);
  luaC_checkGC(L);  /* 2.34.4, call GC before pair creation */
  setpairvalue(L, L->top, agnPair_create(L, l, r));
  api_incr_top(L);
  lua_unlock(L);
}


/* Sets the two values at indices idxleft and idxright to the pair at index idx. The function does not pop the values. */
LUA_API void agn_pairset (lua_State *L, int idx, int idxleft, int idxright) {  /* 2.29.0 */
  Pair *h;
  StkId t, l, r;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttispair(t));
  h = pairvalue(t);
  l = index2adr(L, idxleft);
  r = index2adr(L, idxright);
  agnPair_seti(L, h, 1, l);
  agnPair_seti(L, h, 2, r);
  lua_unlock(L);
}


/* Sets the value at the stack top to the pair at index idx and pops the value. If pos is 1, then the value is put into the left part of the pair,
   if pos is 2, then the right part is set. */
LUA_API void agn_pairseti (lua_State *L, int idx, int pos) {  /* 2.29.0 */
  StkId t;
  lua_lock(L);
  api_checknelems(L, 1);
  t = index2adr(L, idx);
  api_check(L, ttispair(t));
  agnPair_seti(L, pairvalue(t), pos, index2adr(L, -1));
  L->top--;
  lua_unlock(L);
}


LUA_API void agn_createpairnumbers (lua_State *L, lua_Number x, lua_Number y) {  /* Reintroduced 2.9.8, modified 2.14.2 & 2.34.1 */
  StkId l, r;
  Pair *p;
  luaL_checkstack(L, 2, "not enough stack space");
  lua_lock(L);
  setnvalue(L->top, x);
  api_incr_top(L);
  setnvalue(L->top, y);
  api_incr_top(L);
  l = index2adr(L, -2);
  r = index2adr(L, -1);
  luaC_checkGC(L);  /* 2.34.4, call GC before pair creation */
  p = agnPair_create(L, l, r);
  L->top -= 2;
  setpairvalue(L, L->top, p);
  api_incr_top(L);
  luaC_checkGC(L);
  lua_unlock(L);
}


#ifndef PROPCMPLX
LUA_API void agn_createcomplex (lua_State *L, agn_Complex c) {  /* 0.12.0, changed 0.22.1 */
  setcvalue(L->top, c);
  api_incr_top(L);
}
#else
LUA_API void agn_createcomplex (lua_State *L, lua_Number a, lua_Number b) {  /* 0.12.0 */
  lua_Number z[2];
  z[0] = a;
  z[1] = b;
  setcvalue(L->top, z);
  api_incr_top(L);
}
#endif


/* get left (n=1) or right (n=2) value in a pair, FIXME: make this consistent */
LUA_API void agn_pairgeti (lua_State *L, int idx, int n) {  /* 0.11.1 */
  StkId t;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttispair(t));
  setobj2s(L, L->top, agnPair_geti(pairvalue(t), n));
  api_incr_top(L);
  lua_unlock(L);
}


/* get real part of complex number */
LUA_API void agn_complexgetreal (lua_State *L, int idx) {  /* 0.12.0 */
  StkId t;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttiscomplex(t));
  setnvalue(L->top, complexreal(t));
  api_incr_top(L);
  lua_unlock(L);
}


/* get imaginary part of complex number */
LUA_API void agn_complexgetimag (lua_State *L, int idx) {  /* 0.12.0 */
  StkId t;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttiscomplex(t));
  setnvalue(L->top, compleximag(t));
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API void lua_rawgeti (lua_State *L, int idx, int n) {
  StkId o;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  setobj2s(L, L->top, luaH_getnum(hvalue(o), n));
  api_incr_top(L);
  lua_unlock(L);
}


/* Pushes onto the stack the value t[k], where t is the table at the given index and k is the pointer p represented as a
   light userdata. The access is raw; that is, it does not use the __index metavalue.

   Returns the type of the pushed value. Taken from Lua 5.4.0 RC 5, 2.21.2 */

static int finishrawget (lua_State *L, const TValue *val) {
  if (isempty(val))  /* avoid copying empty items to the stack */
    setnilvalue(s2v(L->top));
  else
    setobj2s(L, L->top, val);
  api_incr_top(L);
  lua_unlock(L);
  return ttype(s2v(L->top - 1));
}

static Table *gettable (lua_State *L, int idx) {
  TValue *t = index2value(L, idx);
  api_check(L, ttistable(t));
  return hvalue(t);
}

LUA_API int lua_rawgetp (lua_State *L, int idx, const void *p) {
  Table *t;
  TValue k;
  lua_lock(L);
  t = gettable(L, idx);
  setpvalue(&k, cast_voidp(p));
  return finishrawget(L, luaH_get(t, &k));
}


LUA_API void lua_createtable (lua_State *L, int narray, int nrec) {
  lua_lock(L);
  luaC_checkGC(L);
  sethvalue(L, L->top, luaH_new(L, narray, nrec));  /* also sets hasnil field to 0 */
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API void agn_createtable (lua_State *L, int narray, int nrec) {
  Table *newtable;
  lua_lock(L);
  luaC_checkGC(L);
  newtable = luaH_new(L, narray, nrec);  /* Agena 1.7.10, avoid trouble with `size` */
  newtable->hasnil = 1;
  sethvalue(L, L->top, newtable);
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API void agn_createset (lua_State *L, int nrec) {  /* 0.10.0 */
  lua_lock(L);
  luaC_checkGC(L);
  setusvalue(L, L->top, agnUS_new(L, nrec));
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API void agn_createseq (lua_State *L, int nrec) {  /* 0.11.0 */
  lua_lock(L);
  luaC_checkGC(L);
  setseqvalue(L, L->top, agnSeq_new(L, nrec));
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API void agn_createreg (lua_State *L, int nrec) {  /* 2.3.0 RC 3 */
  lua_lock(L);
  luaC_checkGC(L);
  setregvalue(L, L->top, agnReg_new(L, nrec));
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API int lua_getmetatable (lua_State *L, int objindex) {
  const TValue *obj;
  Table *mt = NULL;
  int res;
  lua_lock(L);
  obj = index2adr(L, objindex);
  switch (ttype(obj)) {
    case LUA_TTABLE:
      mt = hvalue(obj)->metatable;
      break;
    case LUA_TUSERDATA:
      mt = uvalue(obj)->metatable;
      break;
    case LUA_TSET:  /* 0.11.0 */
      mt = usvalue(obj)->metatable;
      break;
    case LUA_TSEQ:  /* 0.11.0 */
      mt = seqvalue(obj)->metatable;
      break;
    case LUA_TPAIR:  /* 0.11.0 */
      mt = pairvalue(obj)->metatable;
      break;
    case LUA_TREG:  /* 0.11.0 */
      mt = regvalue(obj)->metatable;
      break;
    default:
      mt = G(L)->mt[ttype(obj)];
      break;
  }
  if (mt == NULL)
    res = 0;
  else {
    sethvalue(L, L->top, mt);
    api_incr_top(L);
    res = 1;
  }
  lua_unlock(L);
  return res;
}


LUA_API void lua_getfenv (lua_State *L, int idx) {
  StkId o;
  lua_lock(L);
  o = index2adr(L, idx);
  api_checkvalidindex(L, o);
  switch (ttype(o)) {
    case LUA_TFUNCTION:
      sethvalue(L, L->top, clvalue(o)->c.env);
      break;
    case LUA_TUSERDATA:
      sethvalue(L, L->top, uvalue(o)->env);
      break;
    case LUA_TTHREAD:
      setobj2s(L, L->top,  gt(thvalue(o)));
      break;
    default:
      setnilvalue(L->top);
      break;
  }
  api_incr_top(L);
  lua_unlock(L);
}


/* Pushes onto the stack the n-th user value associated with the full userdata at the given index and returns
   the type of the pushed value.

   If the userdata does not have that value or the value at idx is not a userdata, pushes null and returns LUA_TNONE.
   n should always be 1 as the function would always return LUA_TNONE otherwise.

   Taken from Lua 5.4.0 RC 5, 2.21.2 */

LUA_API int lua_getiuservalue (lua_State *L, int idx, int n) {
  TValue *o;
  int t;
  lua_lock(L);
  o = index2value(L, idx);
  api_check(L, ttisfulluserdata(o));
  if (!ttisuserdata(o) || (n <= 0 || n > uvalue(o)->nuvalue)) {
    setnilvalue(s2v(L->top));
    t = LUA_TNONE;
  }
  else {
    setobj2s(L, L->top, o);
    t = ttype(s2v(L->top));
  }
  api_incr_top(L);
  lua_unlock(L);
  return t;
}

/*
** set functions (stack -> Lua)
**
** t[k] = value at the top of the stack (where 'k' is a string)
*/

/* taken from Lua 5.4.0 RC 4, 2.21.2. Compared to the 5.1 implementation, a little bit faster with string keys,
   not faster with number keys. */
LUA_API void lua_settable (lua_State *L, int idx) {
  TValue *t;
  const TValue *slot;
  lua_lock(L);
  api_checknelems(L, 2);
  t = index2value(L, idx);
  if (luaV_fastget(L, t, s2v(L->top - 2), slot, luaH_get)) {
    luaV_finishfastset(L, t, slot, s2v(L->top - 1));
  }
  else
    luaV_finishset(L, t, s2v(L->top - 2), s2v(L->top - 1), slot);
  L->top -= 2;  /* pop index and value */
  lua_unlock(L);
}


/* taken from Lua 5.4.0 RC 4, 2.21.2 */
LUA_API void lua_setfield (lua_State *L, int idx, const char *k) {
  const TValue *slot, *t;
  TString *str;
  lua_lock(L);  /* unlock done in 'auxsetstr' */
  t = index2value(L, idx);
  str = luaS_new(L, k);
  api_checknelems(L, 1);
  if (luaV_fastget(L, t, str, slot, luaH_getstr)) {
    luaV_finishfastset(L, t, slot, s2v(L->top - 1));
    L->top--;  /* pop value */
  }
  else {
    setsvalue2s(L, L->top, str);  /* push 'str' (to make it a TValue) */
    api_incr_top(L);
    luaV_finishset(L, t, s2v(L->top - 1), s2v(L->top - 2), slot);
    L->top -= 2;  /* pop value and key */
  }
  lua_unlock(L);  /* lock done by caller */
}


LUA_API void agn_rawsetfield (lua_State *L, int idx, const char *k) {  /* 2.12.1, do not invoke metamethods */
  StkId t;
  Table *h;
  TString *str;
  lua_lock(L);
  api_checknelems(L, 1);
  t = index2adr(L, idx);
  api_checkvalidindex(L, t);
  h = hvalue(t);
  str = luaS_new(L, k);
  setobj2t(L, luaH_setstr(L, h, str), L->top - 1);
  luaC_barriert(L, h, L->top - 1);
  L->top--;  /* pop value */
  lua_unlock(L);
}


LUA_API void agn_deletefield (lua_State *L, int idx, const char *k) {  /* 2.31.8, do not invoke metamethods */
  StkId t;
  Table *h;
  TString *str;
  lua_lock(L);
  api_checknelems(L, 0);
  t = index2adr(L, idx);
  api_checkvalidindex(L, t);
  h = hvalue(t);
  str = luaS_new(L, k);
  setnilvalue(L->top);
  api_incr_top(L);
  setobj2t(L, luaH_setstr(L, h, str), L->top - 1);
  luaC_barriert(L, h, L->top - 1);
  L->top--;  /* pop value */
  lua_unlock(L);
}


LUA_API void agn_rawgetfield (lua_State *L, int idx, const char *k) {  /* 2.12.1, do not invoke metamethods */
  StkId t;
  TString *str;
  lua_lock(L);
  api_checknelems(L, 1);
  t = index2adr(L, idx);
  api_checkvalidindex(L, t);
  str = luaS_new(L, k);
  setobj2s(L, L->top, luaH_getstr(hvalue(t), str));
  api_incr_top(L);  /* put table value at top of stack */
  lua_unlock(L);
}


LUA_API void lua_rawset (lua_State *L, int idx) {
  StkId t;
  lua_lock(L);
  api_checknelems(L, 2);
  t = index2adr(L, idx);
  api_check(L, ttistable(t));
  setobj2t(L, luaH_set(L, hvalue(t), L->top - 2), L->top - 1);
  luaC_barriert(L, hvalue(t), L->top - 1);
  /* never remove the following line [sic !], otherwise `size' will return wrong results in case of deletions */
  if (ttisnil(L->top - 1)) hvalue(t)->hasnil = 1;
  L->top -= 2;
  lua_unlock(L);
}


LUA_API void lua_rawseti (lua_State *L, int idx, int n) {
  StkId o;
  lua_lock(L);
  api_checknelems(L, 1);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  setobj2t(L, luaH_setnum(L, hvalue(o), n), L->top - 1);
  luaC_barriert(L, hvalue(o), L->top - 1);
  /* 1.6.9, never remove the following line [sic !], otherwise `size' will return wrong results in case of deletions */
  if (ttisnil(L->top - 1)) hvalue(o)->hasnil = 1;
  L->top--;
  lua_unlock(L);
}


/* Does the equivalent of t[p] = v, where t is the table at the given index, p is encoded as a light userdata, and v
   is the value on the top of the stack.

   This function pops the value from the stack. The assignment is raw, that is, it does not use the __writeindex metavalue.

   Taken from Lua 5.4.0 RC 5, 2.21.2 */
LUA_API void lua_rawsetp (lua_State *L, int idx, const void *p) {
  Table *t;
  TValue *slot, key;
  int n = 1;
  setpvalue(&key, cast_voidp(p));
  lua_lock(L);
  api_checknelems(L, n);
  t = gettable(L, idx);
  slot = luaH_set(L, t, &key);
  setobj2t(L, slot, s2v(L->top - 1));
  invalidateTMcache(t);
  luaC_barriert(L, t, s2v(L->top - 1));
  L->top -= n;
}


/* 0.22.1: returns a true, deep copy of the structure at stack index idx. The copy is put on top of the stack, but the
   original structure is not removed. Extended 2.34.4.
   what = 1: array only
   what = 2: hash only
   what = 3: array and hash part;
   what = 7: do not copy metatable and user-defined type */
LUA_API void agn_copy (lua_State *L, int idx, int what) {
  lua_lock(L);
  luaC_checkGC(L);  /* 2.34.4: collect garbage _before_ calling "VM", see lua_concat */
  agenaV_copy(L, index2value(L, idx), L->top, NULL, what);
  api_incr_top(L);
  lua_unlock(L);
}


LUA_API void agn_unique (lua_State *L, int idx) {  /* 2.34.4 */
  StkId base = index2adr(L, idx);
  lua_lock(L);
  luaC_checkGC(L);  /* 2.34.4: collect garbage _before_ calling "VM", see lua_concat */
  agenaV_unique(L, base);
  L->top = base + 1;
  lua_unlock(L);
}


LUA_API void agn_join (lua_State *L, int idx, int nargs) {  /* 2.34.5 */
  StkId base = index2adr(L, idx);
  lua_lock(L);
  luaC_checkGC(L);  /* collect garbage _before_ calling "VM", see lua_concat */
  agenaV_join(L, base, nargs);
  L->top = base + 1;
  lua_unlock(L);
}


LUA_API void agn_replace (lua_State *L, int idx, int nargs) {  /* 2.34.5 */
  StkId base = index2adr(L, idx);
  lua_lock(L);
  luaC_checkGC(L);  /* collect garbage _before_ calling "VM", see lua_concat */
  agenaV_replace(L, base, nargs);
  L->top = base + 1;
  lua_unlock(L);
}


LUA_API void agn_instr (lua_State *L, int idx, int nargs) {  /* 2.34.5 */
  StkId base = index2adr(L, idx);
  lua_lock(L);
  luaC_checkGC(L);  /* collect garbage _before_ calling "VM", see lua_concat */
  agenaV_instr(L, base, nargs);
  L->top = base + 1;
  lua_unlock(L);
}


/* nargs must be at least two; 2.34.5 */
LUA_API void agn_values (lua_State *L, int idx, int nargs) {
  StkId base = index2adr(L, idx);
  lua_lock(L);
  luaC_checkGC(L);  /* collect garbage _before_ calling "VM", see lua_concat */
  agenaV_values(L, base, nargs);
  L->top = base + 1;
  lua_unlock(L);
}


/* nargs must be at least three; 2.34.5 */
LUA_API void agn_times (lua_State *L, int idx, int nargs) {
  StkId base = index2adr(L, idx);
  lua_lock(L);
  luaC_checkGC(L);  /* collect garbage _before_ calling "VM", see lua_concat */
  agenaV_times(L, base, nargs);
  L->top = base + 1;
  lua_unlock(L);
}

/* Empties a table at index idx and leaves the stack unchanged. The function also frees allocated, previously occupied memory
   in the array and hash parts and optionally conducts a garbage collection if gc is set to 1. 2.33.2 */
LUA_API void agn_cleanse (lua_State *L, int idx, int gc) {
  if (!lua_istable(L, idx)) return;
  lua_pushvalue(L, idx);
  lua_pushnil(L);
  while (lua_next(L, -2)) {
    agn_poptop(L);  /* pop value */
    lua_pushvalue(L, -1);  /* duplicate key */
    lua_pushnil(L);
    lua_rawset(L, -4);
  }
  agn_reorder(L, -1);  /* shrink array and hash part, i.e. free unused memory */
  agn_poptop(L);  /* pop table from stack */
  if (gc) lua_gc(L, LUA_GCCOLLECT, 0);
}


/* Empties a set at index idx and leaves the stack unchanged. The function optionally conducts a garbage collection if gc is
   set to 1. 2.33.2 */
LUA_API void agn_cleanseset (lua_State *L, int idx, int gc) {
  if (!lua_isset(L, idx)) return;
  lua_pushvalue(L, idx);
  lua_pushnil(L);
  while (lua_usnext(L, -2)) {
    lua_sdelete(L, -3); /* delete value in the set and pop it */
    agn_poptop(L);  /* pop the key from stack, too, since it cannot be used for next iteration as it has been purged from the set */
    lua_pushnil(L); /* restart iteration */
  }
  agn_poptop(L);  /* pop set from stack */
  agn_setresize(L, idx, 1, 0);
  if (gc) lua_gc(L, LUA_GCCOLLECT, 0);
}


LUA_API void lua_srawset (lua_State *L, int idx) {
  StkId t;
  UltraSet *s;
  lua_lock(L);
  api_checknelems(L, 1);
  t = index2adr(L, idx);
  s = usvalue(t);
  api_check(L, ttisuset(t));
  agnUS_set(L, s, L->top - 1);
  luaC_barrierset(L, s, L->top - 1);
  L->top--;
  lua_unlock(L);
}


LUA_API void lua_sdelete (lua_State *L, int idx) {  /* 0.10.0 */
  StkId t;
  UltraSet *s;
  lua_lock(L);
  api_checknelems(L, 1);
  t = index2adr(L, idx);
  s = usvalue(t);
  api_check(L, ttisuset(t));
  agnUS_delete(L, s, L->top - 1);
  L->top--;  /* pop value */
  lua_unlock(L);
}


LUA_API void agn_pairrawset (lua_State *L, int idx) {
  StkId t;
  int r;
  Pair *p;
  lua_lock(L);
  api_checknelems(L, 2);
  t = index2adr(L, idx);
  p = pairvalue(t);  /* 2.2.0 */
  api_check(L, ttispair(t));
  r = agnPair_rawseti(L, p, L->top - 2, L->top - 1);
  if (r != 1)
    luaG_runerror(L, "Error in rawset (agn_pairrawset).");
  luaC_barrierpair(L, p, L->top - 1);
  L->top -= 2;
  lua_unlock(L);
}


LUA_API void lua_seqinsert (lua_State *L, int idx) {  /* 0.11.0 */
  StkId t;
  TValue *val;
  Seq *s;
  lua_lock(L);
  api_checknelems(L, 1);
  t = index2adr(L, idx);
  api_check(L, ttisseq(t));
  s = seqvalue(t);
  val = L->top - 1;
  if (agnSeq_seti(L, s, s->size + 1, val) == -1)
    luaG_runerror(L, "Error in rawset (lua_seqinsert): memory allocation failed.");
  luaC_barrierseq(L, s, val);
  L->top--;  /* pop value */
  lua_unlock(L);
}


LUA_API void lua_reginsert (lua_State *L, int idx) {  /* 2.34.2 */
  int i, flag;
  StkId t;
  TValue *val;
  Reg *s;
  lua_lock(L);
  api_checknelems(L, 1);
  t = index2adr(L, idx);
  api_check(L, ttisseq(t));
  s = regvalue(t);
  val = L->top - 1;
  flag = 1;
  for (i=1; i <= s->top; i++) {
    if (ttisnil(agnReg_geti(s, i))) {
      agnReg_seti(L, s, i, val);
      luaC_barrierreg(L, s, val);
      flag = 0;
      break;
    }
  }
  if (flag)  /* XXX: can be improved */
    luaG_runerror(L, "Error while inserting: no free slot in register found.");
  L->top--;  /* pop value */
  lua_unlock(L);
}


/* Inserts the object at stack index idxv into the table, set, sequence or register at stack index idxs. The function
   does not change the stack. */
LUA_API void agn_structinsert (lua_State *L, int idxs, int idxv) {  /* 2.34.2 */
  StkId o;
  TValue *val;
  lua_lock(L);
  api_checknelems(L, 1);
  o = index2adr(L, idxs);
  val = index2value(L, idxv);
  switch (ttype(o)) {
    case LUA_TTABLE: {
      Table *h = hvalue(o);
      setobj2t(L, luaH_setnum(L, h, luaH_getn(h) + 1), val);
      luaC_barriert(L, h, val);
      if (ttisnil(val)) h->hasnil = 1;
      break;
    }
    case LUA_TSET: {
      UltraSet *s = usvalue(o);
      agnUS_set(L, s, val);
      luaC_barrierset(L, s, val);
      break;
    }
    case LUA_TSEQ: {
      Seq *s = seqvalue(o);
      if (agnSeq_seti(L, s, s->size + 1, val) == -1)
        luaG_runerror(L, "Error while inserting: memory allocation failed.");
      luaC_barrierseq(L, s, val);
      break;
    }
    case LUA_TREG: {
      int i, flag;
      Reg *s = regvalue(o);
      flag = 1;
      for (i=1; i <= s->top; i++) {
        if (ttisnil(agnReg_geti(s, i))) {
          agnReg_seti(L, s, i, val);
          luaC_barrierreg(L, s, val);
          flag = 0;
          break;
        }
      }
      if (flag)  /* XXX: can be improved */
        luaG_runerror(L, "Error while inserting: no free slot in register found.");
      break;
    }
    default:
      luaG_runerror(L, "Error while inserting: structure expected, got %s.", luaL_typename(L, idxv));
  }
  lua_unlock(L);
}


LUA_API void lua_seqrawset (lua_State *L, int idx) {
  StkId t;
  int r;
  Seq *h;
  lua_lock(L);
  api_checknelems(L, 2);
  t = index2adr(L, idx);
  h = seqvalue(t);  /* 2.2.0 */
  api_check(L, ttisseq(t));
  r = agnSeq_seti(L, h, (int)nvalue(L->top - 2), L->top - 1);
  switch (r) {  /* 2.29.0 improved error handling */
    case 0:
      luaG_runerror(L, "Error in internal " LUA_QS " with sequence, given index is not next free index.", "rawset");
      break;
    case -1:
      luaG_runerror(L, "Error in internal " LUA_QS " with sequence, memory allcoation failed.", "rawset");
      break;
    case -2:
      luaG_runerror(L, "Error in internal " LUA_QS " with sequence, null cannot be added.", "rawset");
      break;
  }
  luaC_barrierseq(L, h, L->top - 1);
  L->top -= 2;
  lua_unlock(L);
}


LUA_API void lua_seqseti (lua_State *L, int idx, size_t n) {  /* 0.11.0 */
  StkId t;
  Seq *h;
  lua_lock(L);
  api_checknelems(L, 1);
  t = index2adr(L, idx);
  h = seqvalue(t);  /* 2.2.0 */
  api_check(L, ttisseq(t));
  agnSeq_seti(L, h, n, L->top - 1);
  luaC_barrierseq(L, h, L->top - 1);
  L->top--;  /* pop inserted value */
  lua_unlock(L);
}


LUA_API void agn_seqsetinumber (lua_State *L, int idx, int n, lua_Number x) {  /* 2.14.2 */
  StkId t;
  Seq *h;
  lua_lock(L);
  setnvalue(L->top, x);
  api_incr_top(L);
  t = index2adr(L, idx - 1);
  h = seqvalue(t);
  api_check(L, ttisseq(t));
  agnSeq_seti(L, h, n, L->top - 1);
  luaC_barrierseq(L, h, L->top - 1);
  L->top--;  /* pop inserted value */
  lua_unlock(L);
}


/* assumes that the value is `at the top` of the stack and the numeric key just below the stack */
LUA_API void agn_regset (lua_State *L, int idx) {  /* 2.3.0 RC 3 */
  StkId t;
  int r;
  Reg *h;
  lua_lock(L);
  api_checknelems(L, 2);
  t = index2adr(L, idx);
  h = regvalue(t);
  api_check(L, ttisreg(t));
  r = agnReg_seti(L, h, (int)nvalue(L->top - 2), L->top - 1);
  if (r != 1)
    luaG_runerror(L, "Error in internal " LUA_QS " with register, given index is not next free index.", "rawset");
  luaC_barrierreg(L, h, L->top - 1);
  L->top -= 2;
  lua_unlock(L);
}


LUA_API void agn_regpurge (lua_State *L, int idx, int n) {  /* 2.3.0 RC 3, n must be positive ! */
  StkId t;
  Reg *h;
  lua_lock(L);
  api_checknelems(L, 1);
  t = index2adr(L, idx);
  h = regvalue(t);  /* 2.2.0 */
  api_check(L, ttisreg(t));
  agnReg_purge(h, n);
  lua_unlock(L);
}


LUA_API void agn_regseti (lua_State *L, int idx, size_t n) {  /* 2.3.0 RC 3 */
  StkId t;
  Reg *h;
  lua_lock(L);
  api_checknelems(L, 1);
  t = index2adr(L, idx);
  h = regvalue(t);  /* 2.2.0 */
  api_check(L, ttisreg(t));
  agnReg_seti(L, h, n, L->top - 1);
  luaC_barrierreg(L, h, L->top - 1);
  L->top--;  /* pop inserted value */
  lua_unlock(L);
}


LUA_API void agn_regsetinumber (lua_State *L, int idx, int n, lua_Number x) {  /* 2.14.2 */
  StkId t;
  Reg *h;
  lua_lock(L);
  setnvalue(L->top, x);
  api_incr_top(L);
  t = index2adr(L, idx - 1);
  h = regvalue(t);  /* 2.2.0 */
  api_check(L, ttisreg(t));
  agnReg_seti(L, h, n, L->top - 1);
  luaC_barrierreg(L, h, L->top - 1);
  L->top--;  /* pop inserted value */
  lua_unlock(L);
}


LUA_API void agn_setutype (lua_State *L, int idxobj, int idxtype) {  /* 0.11.1; extended 0.14.0, patched 0.28.0,
  patched 1.2.1, extended 2.3.0 RC 3 */
  StkId t, ty;
  TString *str;
  lua_lock(L);
  t = index2adr(L, idxobj);
  ty = index2adr(L, idxtype);
  if (ttisstring(ty))
    str = luaS_new(L, svalue(ty));
  else if (ttisnil(ty))
    str = NULL;
  else {
    luaG_runerror(L, "Error: string or null expected.");
    return;
  }
  agenaV_setutype(L, t, str);
  lua_unlock(L);
}


LUA_API void agn_setutypestring (lua_State *L, int idxobj, const char *type) {  /* 2.14.2, based on agn_setutype */
  lua_lock(L);
  if (type == NULL)  /* 2.15.1 extension */
    agenaV_setutype(L, index2adr(L, idxobj), NULL);
  else
    agenaV_setutype(L, index2adr(L, idxobj), luaS_new(L, type));
  lua_unlock(L);
}


TString *getutype (lua_State *L, int idx) {
  StkId t;
  TString *r;
  lua_lock(L);
  t = index2adr(L, idx);
  if (ttisseq(t))
    r = seqvalue(t)->type;
  else if (ttispair(t))
    r = pairvalue(t)->type;
  else if (ttistable(t))
    r = hvalue(t)->type;
  else if (ttisuset(t))
    r = usvalue(t)->type;
  else if (ttisuserdata(t)) { /* 2.3.0 RC 3 */
    r = uvalue(t)->type;
  } else if (ttisreg(t))  /* 2.5.3 */
    r = regvalue(t)->type;
  else if (ttisfunction(t)) {  /* 2.15.1 hardening */
    Closure *fcl = clvalue(t);
    if (fcl->c.isC && fcl->c.type != NULL)
      r = fcl->c.type;
    else if (!fcl->c.isC && fcl->l.type != NULL)
      r = fcl->l.type;
    else
      r = NULL;
  } else  /* 2.15.1 fix */
    r = NULL;
  lua_unlock(L);
  return r;
}


LUA_API int agn_getutype (lua_State *L, int idx) {  /* 0.11.1; extended 0.14.0, modified 0.28.0 */
  TString *r = getutype(L, idx);
  if (r == NULL)
    return 0;
  else {
    setsvalue(L, L->top, r);
    api_incr_top(L);
  }
  lua_unlock(L);
  return 1;
}


LUA_API int agn_isutype (lua_State *L, int idx, const char *str) {  /* based on agn_getutype, 1.9.1, 2.2.6 */
  int result;
  TString *r = getutype(L, idx);
  if (r == NULL)
    result = 0;
  else
    result = (tools_streq(getstr(r), str));  /* 2.16.12 tweak */
  return result;
}


LUA_API int agn_isutypeset (lua_State *L, int idx) {  /* 0.12.2; extended 0.14.0, modified 0.28.0 */
  StkId t;
  TString *r;
  lua_lock(L);
  t = index2adr(L, idx);
  if (ttisseq(t))
    r = seqvalue(t)->type;
  else if (ttisreg(t))
    r = regvalue(t)->type;
  else if (ttispair(t))
    r = pairvalue(t)->type;
  else if (ttistable(t))
    r = hvalue(t)->type;
  else if (ttisuset(t))  /* 1.9.1 */
    r = usvalue(t)->type;
  else if (ttisfunction(t)) {
    Closure *fcl = clvalue(t);
    if (fcl->c.isC && fcl->c.type != NULL)
      r = fcl->c.type;
    else if (!fcl->c.isC && fcl->l.type != NULL)
      r = fcl->l.type;
    else
      r = NULL;
  }
  else if (ttisuset(t))
    r = usvalue(t)->type;
  else
    r = NULL;
  lua_unlock(L);
  return (r != NULL);
}


LUA_API int agn_issequtype (lua_State *L, int idx, const char *str) {  /* 0.12.1, modified 0.28.0 */
  StkId t;
  TString *r;
  lua_lock(L);
  t = index2adr(L, idx);
  if (ttisseq(t))
    r = seqvalue(t)->type;
  else
    r = NULL;
  lua_unlock(L);
  if (r == NULL)
    return 0;
  else
    return (tools_streq(getstr(r), str));  /* 2.16.12 tweak */
}


LUA_API int agn_istableutype (lua_State *L, int idx, const char *str) {  /* 0.12.2, modified 0.28.0 */
  StkId t;
  TString *r;
  lua_lock(L);
  t = index2adr(L, idx);
  if (ttistable(t))
    r = hvalue(t)->type;
  else
    r = NULL;
  lua_unlock(L);
  if (r == NULL)
    return 0;
  else
    return (tools_streq(getstr(r), str));  /* 2.16.12 tweak */
}


LUA_API int agn_issetutype (lua_State *L, int idx, const char *str) {  /* 0.14.0, modified 0.28.0 */
  StkId t;
  TString *r;
  lua_lock(L);
  t = index2adr(L, idx);
  if (ttisuset(t))
    r = usvalue(t)->type;
  else
    r = NULL;
  lua_unlock(L);
  if (r == NULL)
    return 0;
  else
    return (tools_streq(getstr(r), str));  /* 2.16.12 tweak */
}


LUA_API int lua_setmetatable (lua_State *L, int objindex) {
  TValue *obj;
  Table *mt;
  lua_lock(L);
  api_checknelems(L, 1);
  obj = index2adr(L, objindex);
  api_checkvalidindex(L, obj);
  if (ttisnil(L->top - 1))
    mt = NULL;
  else {
    api_check(L, ttistable(L->top - 1));
    mt = hvalue(L->top - 1);
  }
  agenaV_setmetatable(L, obj, mt);
  L->top--;
  lua_unlock(L);
  return 1;
}


/* If k is non-NULL, sets metatable `k' to the structure at index position idx. The function does not change the stack.
   If k is NULL, deletes the metatable from the object at idx and if settype == 1, the user-defined type. */
LUA_API void lua_setmetatabletoobject (lua_State *L, int idx, const char *k, int settype) {
  StkId t;
  TValue *obj;
  lua_lock(L);
  t = index2adr(L, LUA_REGISTRYINDEX);
  api_checkvalidindex(L, t);
  if (k == NULL) {
    obj = index2adr(L, idx);  /* address of object at `idx` */
    api_checkvalidindex(L, obj);
    agenaV_setmetatable(L, obj, NULL);  /* delete metatable to object at `idx` */
    if (settype) agenaV_setutype(L, obj, NULL);
  } else {
    Table *mt;
    TValue key;
    TString *kstr;
    kstr = luaS_new(L, k);
    setsvalue(L, &key, kstr);
    luaV_gettable(L, t, &key, L->top);
    api_incr_top(L);
    obj = index2adr(L, idx - 1);
    api_checkvalidindex(L, obj);
    if (settype) agenaV_setutype(L, obj, kstr);  /* set user-defined type to object at `idx`, 2.21.9 optimisation */
    if (ttisnil(L->top - 1))  /* no metatable found, invalid string index ? */
      mt = NULL;
    else {
      api_check(L, ttistable(L->top - 1));
      mt = hvalue(L->top - 1);
    }
    agenaV_setmetatable(L, obj, mt);  /* set metatable to object at `idx` */
    L->top--;  /* drop metatable */
  }
  lua_unlock(L);
}


LUA_API int lua_setfenv (lua_State *L, int idx) {
  StkId o;
  int res = 1;
  lua_lock(L);
  api_checknelems(L, 1);
  o = index2adr(L, idx);
  api_checkvalidindex(L, o);
  api_check(L, ttistable(L->top - 1));
  switch (ttype(o)) {
    case LUA_TFUNCTION:
      clvalue(o)->c.env = hvalue(L->top - 1);
      break;
    case LUA_TUSERDATA:
      uvalue(o)->env = hvalue(L->top - 1);
      break;
    case LUA_TTHREAD:
      sethvalue(L, gt(thvalue(o)), hvalue(L->top - 1));
      break;
    default:
      res = 0;
      break;
  }
  if (res) luaC_objbarrier(L, gcvalue(o), hvalue(L->top - 1));  /* Lua 5.1.3 patch */
  L->top--;
  lua_unlock(L);
  return res;
}


/*
** `load' and `call' functions (run Lua code)
*/

#define adjustresults(L,nres) \
    { if (nres == LUA_MULTRET && L->top >= L->ci->top) L->ci->top = L->top; }

#define checkresults(L,na,nr) \
     api_check(L, (nr) == LUA_MULTRET || (L->ci->top - L->top >= (nr) - (na)))


LUA_API void lua_call (lua_State *L, int nargs, int nresults) {
  StkId func;
  lua_lock(L);
  api_checknelems(L, nargs + 1);
  checkresults(L, nargs, nresults);
  func = L->top - (nargs + 1);
  luaD_call(L, func, nresults, 1);
  adjustresults(L, nresults);
  lua_unlock(L);
}


/* Calls a function and returns a numeric result, it leaves nothing on the stack. Always returns the first result of
   the function call. Patched 2.31.10 to really leave nothing on the stack. */
LUA_API lua_Number agn_ncall (lua_State *L, int nargs, int *error, int quit) {
  StkId func;
  const TValue *o;
  lua_Number r;
  lua_lock(L);
  *error = 0;
  api_checknelems(L, nargs + 1);
  checkresults(L, nargs, 1);
  func = L->top - (nargs + 1);
  luaD_call(L, func, 1, 0);  /* changed to exactly one result to avoid problems with incorrect results and stack corruption, 2.31.10 */
  o = L->top - 1;
  if (!ttisnumber(o)) {
    *error = 1;
    lua_unlock(L);
    if (quit) luaG_runerror(L, "Error: return of function call is not a number, but a %s.", lua_typename(L, ttype(o)));
    return 0.0;
  }
  r = nvalue(o);
  L->top--;   /* delete result from stack (it has already been stored to r) */
  lua_unlock(L);
  return r;
}


/* Calls a function and returns a complex result; used by fractals package. Always returns the first result of
   the function call. Patched 2.31.10 to really leave nothing on the stack. */
#ifndef PROPCMPLX
LUA_API agn_Complex agn_ccall (lua_State *L, int nargs) {
  StkId func;
  const TValue *o;
  agn_Complex r;
  lua_lock(L);
  api_checknelems(L, nargs + 1);
  checkresults(L, nargs, 1);
  func = L->top - (nargs + 1);
  luaD_call(L, func, 1, 0);  /* changed to exactly one result to avoid problems with incorrect results and stack corruption, 2.31.10 */
  o = L->top - 1;
  if (!ttiscomplex(o))
    luaG_runerror(L, "Error: return of function call is not complex, but a %s.", lua_typename(L, ttype(o)));  /* new in 0.27.1, 1.9.3 */
  r = cvalue(o);
  L->top--;   /* delete result from stack (it has already been stored to r) */
  lua_unlock(L);
  return r;
}
#else
LUA_API void agn_ccall (lua_State *L, int nargs, lua_Number *real, lua_Number *imag) {  /* 0.27.1 */
  StkId func;
  const TValue *o;
  lua_lock(L);
  api_checknelems(L, nargs + 1);
  checkresults(L, nargs, 1);
  func = L->top - (nargs + 1);
  luaD_call(L, func, 1, 0);  /* changed to exactly one result to avoid problems with incorrect results and stack corruption, 2.31.10 */
  o = L->top - 1;
  if (!ttiscomplex(o))
    luaG_runerror(L, "Error: return of function call is not complex, but a %s.", lua_typename(L, ttype(o)));  /* new in 0.27.1, 1.9.3 */
  *real = complexreal(o);
  *imag = compleximag(o);
  L->top--;   /* delete result from stack (it has already been stored to r) */
  lua_unlock(L);
}
#endif


/*
** Execute a protected call.
*/
struct CallS {  /* data to `f_call' */
  StkId func;
  int nresults;
};


static void f_call (lua_State *L, void *ud) {
  struct CallS *c = cast(struct CallS *, ud);
  luaD_call(L, c->func, c->nresults, 1);
}



LUA_API int lua_pcall (lua_State *L, int nargs, int nresults, int errfunc) {
  struct CallS c;
  int status;
  ptrdiff_t func;
  lua_lock(L);
  api_checknelems(L, nargs + 1);
  checkresults(L, nargs, nresults);
  if (errfunc == 0)
    func = 0;
  else {
    StkId o = index2adr(L, errfunc);
    api_checkvalidindex(L, o);
    func = savestack(L, o);
  }
  c.func = L->top - (nargs + 1);  /* function to be called */
  c.nresults = nresults;
  status = luaD_pcall(L, f_call, &c, savestack(L, c.func), func);
  adjustresults(L, nresults);
  lua_unlock(L);
  return status;  /* returns 0 on success */
}


/*
** Execute a protected C call.
*/
struct CCallS {  /* data to `f_Ccall' */
  lua_CFunction func;
  void *ud;
};


static void f_Ccall (lua_State *L, void *ud) {
  struct CCallS *c = cast(struct CCallS *, ud);
  Closure *cl;
  cl = luaF_newCclosure(L, 0, getcurrenv(L));
  cl->c.f = c->func;
  setclvalue(L, L->top, cl);  /* push function */
  api_incr_top(L);
  setpvalue(L->top, c->ud);  /* push only argument */
  api_incr_top(L);
  luaD_call(L, L->top - 2, 0, 1);
}


LUA_API int lua_cpcall (lua_State *L, lua_CFunction func, void *ud) {
  struct CCallS c;
  int status;
  lua_lock(L);
  c.func = func;
  c.ud = ud;
  status = luaD_pcall(L, f_Ccall, &c, savestack(L, L->top), 0);
  lua_unlock(L);
  return status;
}


LUA_API int lua_dump (lua_State *L, lua_Writer writer, void *data, int strip) {
  int status;
  TValue *o;
  lua_lock(L);
  api_checknelems(L, 1);
  o = s2v(L->top - 1);
  if (isLfunction(o))
    status = luaU_dump(L, getproto(o), writer, data, strip);
  else
    status = 1;
  lua_unlock(L);
  return status;
}


LUA_API int lua_load (lua_State *L, lua_Reader reader, void *data,
                      const char *chunkname) {
  ZIO z;
  int status;
  lua_lock(L);
  if (!chunkname) chunkname = "?";
  luaZ_init(L, &z, reader, data);
  status = luaD_protectedparser(L, &z, chunkname);
  lua_unlock(L);
  return status;
}


LUA_API int  lua_status (lua_State *L) {
  return L->status;
}


/*
** Garbage-collection function
*/

LUA_API int lua_gc (lua_State *L, int what, int data) {
  int res = 0;
  global_State *g;
  lua_lock(L);
  g = G(L);
  switch (what) {
    case LUA_GCSTOP: {
      g->GCthreshold = MAX_LUMEM;
      break;
    }
    case LUA_GCRESTART: {
      g->GCthreshold = g->totalbytes;
      break;
    }
    case LUA_GCCOLLECT: {
      luaC_fullgc(L);
      break;
    }
    case LUA_GCCOUNT: {
      /* GC values are expressed in Kbytes: #bytes/2^10 */
      res = cast_int(g->totalbytes >> 10);
      break;
    }
    case LUA_GCCOUNTB: {
      res = cast_int(g->totalbytes & 0x3ff);
      break;
    }
    case LUA_GCSTEP: {
      lu_mem a = (cast(lu_mem, data) << 10);
      if (a <= g->totalbytes)
        g->GCthreshold = g->totalbytes - a;
      else
        g->GCthreshold = 0;
      while (g->GCthreshold <= g->totalbytes) {  /* Lua 5.1.3 patch 10 */
        luaC_step(L);
        if (g->gcstate == GCSpause) {  /* end of cycle? */
          res = 1;  /* signal it */
          break;
        }
      }
      break;
    }
    case LUA_GCSETPAUSE: {
      res = g->gcpause;
      g->gcpause = data;
      break;
    }
    case LUA_GCSETSTEPMUL: {
      res = g->gcstepmul;
      g->gcstepmul = data;
      break;
    }
    case LUA_GCSTATUS: {  /* 2.2.5 */
      res = !(g->GCthreshold == MAX_LUMEM);
      break;
    }
    default: res = -1;  /* invalid option */
  }
  lua_unlock(L);
  return res;
}


LUA_API LUAI_UMEM agn_usedbytes (lua_State *L) {
  global_State *g;
  LUAI_UMEM res;
  lua_lock(L);
  g = G(L);
  res = (LUAI_UMEM)g->totalbytes;
  lua_unlock(L);
  return res;
}


/*
** miscellaneous functions
*/


LUA_API int lua_error (lua_State *L) {
  lua_lock(L);
  api_checknelems(L, 1);
  luaG_errormsg(L);
  lua_unlock(L);
  return 0;  /* to avoid warnings */
}


/* The traversal is undefined if you assign a value to a non-existent field. You may change values, hewever, or even set them to `null`. */
LUA_API int lua_next (lua_State *L, int idx) {
  StkId t;
  int more;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttistable(t));
  more = luaH_next(L, hvalue(t), L->top - 1);
  if (more) {
    api_incr_top(L);
  }
  else  /* no more elements */
    L->top--;  /* remove key */
  lua_unlock(L);
  return more;
}


LUA_API int lua_usnext (lua_State *L, int idx) {  /* 0.10.0 */
  StkId t;
  int more;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttisuset(t));
  more = agnUS_next(L, usvalue(t), L->top - 1);
  if (more) {
    api_incr_top(L);
  }
  else  /* no more elements */
    L->top--;  /* remove key */
  lua_unlock(L);
  return more;
}


LUA_API int lua_seqnext (lua_State *L, int idx) {  /* 0.11.2 */
  StkId t;
  int more;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttisseq(t));
  more = agnSeq_next(L, seqvalue(t), L->top - 1);
  if (more) {
    api_incr_top(L);
  }
  else  /* no more elements */
    L->top--;  /* remove key */
  lua_unlock(L);
  return more;
}


LUA_API int lua_regnext (lua_State *L, int idx) {  /* 2.3.0 RC 3 */
  StkId t;
  int more;
  lua_lock(L);
  t = index2adr(L, idx);
  api_check(L, ttisreg(t));
  more = agnReg_next(L, regvalue(t), L->top - 1);
  if (more) {
    api_incr_top(L);
  }
  else  /* no more elements */
    L->top--;  /* remove key */
  lua_unlock(L);
  return more;
}


/* 0.12.0, used by for i in string loop; optimized 0.24.1, improved 2.16.1 */
LUA_API int lua_strnext (lua_State *L, int idx) {
  StkId t, below;
  int i, more;
  lua_lock(L);
  below = L->top - 1;
  t = index2adr(L, idx);
  api_check(L, ttisstring(t));
  if (ttisnil(below))  /* first iteration */
    i = 0;
  else if (ttisnumber(below))
    i = nvalue(below);  /* get current string position */
  else {
    i = -1;
    luaG_runerror(L, "Error: invalid key to " LUA_QL("next") ".");
  }
  if (i < tsvalue(t)->len) {  /* 2.16.1 patch, do not use strlen since it quits with the first embedded \0 */
    more = 1;
    setnvalue(below, cast_num(i + 1));  /* set new key */
    setsvalue(L, L->top, luaS_newchar(L, svalue(t) + i));  /* set character */
    api_incr_top(L);
  }
  else {  /* no more elements */
    more = 0;
    L->top--;  /* remove key */
  }
  lua_unlock(L);
  return more;
}


LUA_API void lua_concat (lua_State *L, int n) {
  lua_lock(L);
  api_checknelems(L, n);
  if (n >= 2) {
    luaC_checkGC(L);
    luaV_concat(L, n, cast_int(L->top - L->base) - 1);
    L->top -= (n - 1);
  }
  else if (n == 0) {  /* push empty string */
    setsvalue2s(L, L->top, luaS_newlstr(L, "", 0));
    api_incr_top(L);
  }
  /* else n == 1; nothing to do */
  lua_unlock(L);
}


LUA_API lua_Alloc lua_getallocf (lua_State *L, void **ud) {
  lua_Alloc f;
  lua_lock(L);
  if (ud) *ud = G(L)->ud;
  f = G(L)->frealloc;
  lua_unlock(L);
  return f;
}


LUA_API void lua_setallocf (lua_State *L, lua_Alloc f, void *ud) {
  lua_lock(L);
  G(L)->ud = ud;
  G(L)->frealloc = f;
  lua_unlock(L);
}


LUA_API void *lua_newuserdata (lua_State *L, size_t size) {
  Udata *u;
  lua_lock(L);
  luaC_checkGC(L);
  u = luaS_newudata(L, size, getcurrenv(L));
  setuvalue(L, L->top, u);
  api_incr_top(L);
  lua_unlock(L);
  return u + 1;
}


/* This function creates and pushes on the stack a new full userdata, with nuvalue associated Agena values, called user values,
   plus an associated block of raw memory with size bytes. (The user values can be set and read with the functions lua_setiuservalue
   and lua_getiuservalue.) In Agena, nuvalue should always be 1, otherwise the function issues an error to avoid segmentation faults.

   The function returns the address of the block of memory. 2.21.2 */
LUA_API void *lua_newuserdatauv (lua_State *L, size_t size, int nuvalue) {
  if (nuvalue != 1)
    luaL_error(L, "third argument to lua_newuserdatauv must always be 1.");
  return lua_newuserdata(L, size);
}


static const char *aux_upvalue (StkId fi, int n, TValue **val) {
  Closure *f;
  if (!ttisfunction(fi)) return NULL;
  f = clvalue(fi);
  if (f->c.isC) {
    if (!(1 <= n && n <= f->c.nupvalues)) return NULL;
    *val = &f->c.upvalue[n - 1];
    return "";
  }
  else {
    Proto *p = f->l.p;
    if (!(1 <= n && n <= p->sizeupvalues)) return NULL;
    *val = f->l.upvals[n - 1]->v;
    return getstr(p->upvalues[n - 1]);
  }
}


LUA_API const char *lua_getupvalue (lua_State *L, int funcindex, int n) {
  const char *name;
  TValue *val;
  lua_lock(L);
  name = aux_upvalue(index2adr(L, funcindex), n, &val);
  if (name) {
    setobj2s(L, L->top, val);
    api_incr_top(L);
  }
  lua_unlock(L);
  return name;
}


LUA_API const char *lua_setupvalue (lua_State *L, int funcindex, int n) {
  const char *name;
  TValue *val;
  StkId fi;
  lua_lock(L);
  fi = index2adr(L, funcindex);
  api_checknelems(L, 1);
  name = aux_upvalue(fi, n, &val);
  if (name) {
    L->top--;
    setobj(L, val, L->top);
    luaC_barrier(L, clvalue(fi), L->top);
  }
  lua_unlock(L);
  return name;
}


LUA_API int lua_nupvalues (lua_State *L, int funcindex) {  /* 2.12.3 */
  StkId idx;
  Closure *f;
  lua_lock(L);
  idx = index2adr(L, funcindex);
  if (!ttisfunction(idx)) return -1;
  f = clvalue(idx);
  return ( (f->c.isC) ? f->c.nupvalues : f->l.p->sizeupvalues );
}


/* changed/additions */


/* agn_poptop: pops the value at the top of the stack; written July 12, 2007 */
LUA_API void agn_poptop (lua_State *L) {
  lua_lock(L);
  api_check(L, 1 <= (L->top - L->base));
  L->top--;  /* decrease top */
  lua_unlock(L);
}


/* agn_poptop: pops the value at the top of the stack; written April 2, 2008 */
LUA_API void agn_poptoptwo (lua_State *L) {
  lua_lock(L);
  api_check(L, 1 <= (L->top - L->base));
  L->top -= 2;  /* decrease top */
  lua_unlock(L);
}


/* gets the i-th number from the table at the given index idx and returns it as a lua_Number;
   written July 12, 2007 */
LUA_API lua_Number agn_getinumber (lua_State *L, int idx, int n) {
  StkId o;
  const TValue *p;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  p = luaH_getnum(hvalue(o), n);
  lua_unlock(L);
  if (ttype(p) == LUA_TNUMBER)
    return nvalue(p);
  else
    return 0;
}


LUA_API lua_Number agn_rawgetinumber (lua_State *L, int idx, int n, int *rc) {  /* 2.27.5 */
  StkId o;
  const TValue *p;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  p = luaH_getnum(hvalue(o), n);
  lua_unlock(L);
  if ( (*rc = (ttype(p) == LUA_TNUMBER)) )
    return nvalue(p);
  else
    return 0;
}

/* gets the i-th string from the table at the given index idx and returns it;
   written July 21, 2007; used by baselib `concat` function for high-speed to table elements */
LUA_API const char *agn_getistring (lua_State *L, int idx, int n) {
  StkId o;
  const TValue *p;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  p = luaH_getnum(hvalue(o), n);
  lua_unlock(L);
  if (ttype(p) == LUA_TSTRING)
    return svalue(p);
  else
    return NULL;
}


/* for fast determination of indices in tables.indices; sets the value just below the top of the stack to the table at idx and pops
   the topmost value. Used by tables.indices. */
LUA_API void lua_rawsetikey (lua_State *L, int idx, int n) {
  StkId o;
  lua_lock(L);
  api_checknelems(L, 2);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  setobj2t(L, luaH_setnum(L, hvalue(o), n), L->top - 2);
  luaC_barriert(L, hvalue(o), L->top - 2);
  if (ttisnil(L->top - 2)) hvalue(o)->hasnil = 1;
  L->top--;
  lua_unlock(L);
}


/* Inserts the value at the top of the stack to the table at index idx, more precisely, it is added to the end of the array part of the table.
   The value is popped from the stack. */
LUA_API void agn_rawinsert (lua_State *L, int idx) {  /* 2.11.4 */
  StkId o;
  Table *h;
  lua_lock(L);
  api_checknelems(L, 1);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  h = hvalue(o);
  setobj2t(L, luaH_setnum(L, h, luaH_getn(h) + 1), L->top - 1);  /* works like INSERT statement */
  luaC_barriert(L, h, L->top - 1);
  if (ttisnil(L->top - 1)) h->hasnil = 1;
  L->top--;
  lua_unlock(L);
}

/* Inserts the value at stack index vidx to the table residing at index idx, more precisely, it is added to the end of the array part of the table.
   If vidx = -1, then the value is popped from the stack, otherwise the stack is left untouched. */
LUA_API void agn_rawinsertfrom (lua_State *L, int tidx, int vidx) {  /* 2.11.4 */
  StkId o, p;
  Table *h;
  lua_lock(L);
  o = index2adr(L, tidx);
  api_check(L, ttistable(o));
  h = hvalue(o);
  p = index2adr(L, vidx);
  setobj2t(L, luaH_setnum(L, h, luaH_getn(h) + 1), p);  /* works like INSERT statement */
  luaC_barriert(L, h, p);
  if (ttisnil(p)) h->hasnil = 1;
  if (vidx == -1) L->top--;  /* if value is at the top of the stack, remove it, otherwise leave it in the stack */
  lua_unlock(L);
}


/* same as lua_rawset, but only deleting the value, and not the key; August 04, 2007 */
LUA_API void lua_rawset2 (lua_State *L, int idx) {
  StkId t;
  lua_lock(L);
  api_checknelems(L, 2);
  t = index2adr(L, idx);
  api_check(L, ttistable(t));
  setobj2t(L, luaH_set(L, hvalue(t), L->top - 2), L->top - 1);
  luaC_barriert(L, hvalue(t), L->top - 1);
  if (ttisnil(L->top - 1)) hvalue(t)->hasnil = 1;
  L->top--;  /* pop value from the stack */
  lua_unlock(L);
}


/* fast Boolean check, 2.28.6; June 17, 2022, returns 0 for `false` and `fail`, 1 for `true`, and issues an error otherwise.
   This also means that the function will reject `null`. */
LUALIB_API int agn_checkboolean (lua_State *L, int idx) {
  StkId o;
  int t;
  lua_lock(L);
  o = index2adr(L, idx);
  t = ttype(o);
  if (t != LUA_TBOOLEAN && t != LUA_TFAIL)
    tag_error(L, idx, LUA_TBOOLEAN);
  lua_unlock(L);
  return bvalue(o);
}


/* fast number check, 0.6.0; October 25, 2007 */
LUALIB_API lua_Number agn_checknumber (lua_State *L, int idx) {
  StkId o;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) != LUA_TNUMBER)
    tag_error(L, idx, LUA_TNUMBER);
  lua_unlock(L);
  return nvalue(o);
}


LUALIB_API lua_Integer agn_checkinteger (lua_State *L, int idx) {  /* 1.12.9 */
  StkId o;
  lua_Number x;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) != LUA_TNUMBER)
    tag_error(L, idx, LUA_TNUMBER);
  x = nvalue(o);
  if (tools_isfloat(x))  /* 2.14.13 change */
    luaL_error(L, "Wrong argument #%d: expected an integer, got float.", idx);
  lua_unlock(L);
  return x;
}


LUALIB_API lua_Number agn_checkpositive (lua_State *L, int idx) {  /* 2.14.3 */
  StkId o;
  lua_Number x;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) != LUA_TNUMBER)
    tag_error(L, idx, LUA_TNUMBER);
  x = nvalue(o);
  if (x <= 0)
    luaL_error(L, "Wrong argument #%d: expected a positive number.", idx);
  lua_unlock(L);
  return x;
}


LUALIB_API lua_Number agn_checknonnegative (lua_State *L, int idx) {  /* 2.14.3 */
  StkId o;
  lua_Number x;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) != LUA_TNUMBER)
    tag_error(L, idx, LUA_TNUMBER);
  x = nvalue(o);
  if (x < 0)
    luaL_error(L, "Wrong argument #%d: expected a non-negative number.", idx);
  lua_unlock(L);
  return x;
}


LUALIB_API lua_Integer agn_checkposint (lua_State *L, int idx) {  /* 2.7.0 */
  StkId o;
  lua_Number x;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) != LUA_TNUMBER)
    tag_error(L, idx, LUA_TNUMBER);
  x = nvalue(o);
  if (!tools_isposint(x))  /* 2.14.13 change */
    luaL_error(L, "Wrong argument #%d: expected a positive integer.", idx);
  lua_unlock(L);
  return x;
}


LUALIB_API lua_Integer agn_checknonnegint (lua_State *L, int idx) {  /* 2.9.8 */
  StkId o;
  lua_Number x;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) != LUA_TNUMBER)
    tag_error(L, idx, LUA_TNUMBER);
  x = nvalue(o);
  if (!tools_isnonnegint(x))  /* 2.14.13 change */
    luaL_error(L, "Wrong argument #%d: expected a non-negative integer.", idx);
  lua_unlock(L);
  return x;
}


LUALIB_API uint32_t agn_checkuint32_t (lua_State *L, int idx) {  /* 2.16.11 */
  StkId o;
  lua_Number x;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) != LUA_TNUMBER)
    tag_error(L, idx, LUA_TNUMBER);
  x = nvalue(o);
  if (!tools_isuint(x, 32))  /* 0xffffffff = 2^32 - 1 */
    luaL_error(L, "Wrong argument #%d: expected a %d-byte unsigned integer.", idx, sizeof(long int));
  lua_unlock(L);
  return x;
}


LUALIB_API uint16_t agn_checkuint16_t (lua_State *L, int idx) {  /* 2.16.11 */
  StkId o;
  lua_Number x;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) != LUA_TNUMBER)
    tag_error(L, idx, LUA_TNUMBER);
  x = nvalue(o);
  if (!tools_isuint(x, 16))  /* 0xffff = 2^16 - 1 */
    luaL_error(L, "Wrong argument #%d: expected 2-byte unsigned integer.", idx);
  lua_unlock(L);
  return x;
}


#ifndef PROPCMPLX
LUA_API agn_Complex agn_checkcomplex (lua_State *L, int idx) {
  StkId o;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) != LUA_TCOMPLEX)
    tag_error(L, idx, LUA_TCOMPLEX);
  lua_unlock(L);
  return cvalue(o);
}


LUA_API agn_Complex agn_optcomplex (lua_State *L, int narg, agn_Complex def) {
  return luaL_opt(L, agn_checkcomplex, narg, def);
}
#else
LUA_API lua_Number *agn_checkcomplex (lua_State *L, int idx) {
  StkId o;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) != LUA_TCOMPLEX)
    tag_error(L, idx, LUA_TCOMPLEX);
  lua_unlock(L);
  return cvalue(o);
}
#endif


/* string check without conversion of numbers, 0.9.1, 20.01.2008 */
LUALIB_API const char *agn_checklstring (lua_State *L, int idx, size_t *len) {
  StkId o;
  lua_lock(L);
  o = index2adr(L, idx);
  if (!ttisstring(o)) {
    tag_error(L, idx, LUA_TSTRING);
    if (len != NULL) *len = 0;
    lua_unlock(L);
    return NULL;
  }
  if (len != NULL) *len = tsvalue(o)->len;
  lua_unlock(L);
  return svalue(o);
}


/* string check without conversion of numbers, 0.9.1, 20.01.2008 */
LUALIB_API const char *agn_checkstring (lua_State *L, int idx) {
  StkId o;
  lua_lock(L);
  o = index2adr(L, idx);
  if (!ttisstring(o)) {
    tag_error(L, idx, LUA_TSTRING);
    lua_unlock(L);
    return NULL;
  }
  lua_unlock(L);
  return svalue(o);
}


/* 0.5.2, June 01, 2007; tuned April 02, 2008; FIXME: define as macros ? */
LUA_API int agn_istrue (lua_State *L, int idx) {
  return bvalue(index2adr(L, idx)) == 1;  /* 1 is true, 2 is fail, 0 is false */
}


/* 0.10.0, April 02, 2008; FIXME: define as macro ? */
LUA_API int agn_isfalse (lua_State *L, int idx) {
  return bvalue(index2adr(L, idx)) == 0;  /* 1 is true, 2 is fail, 0 is false */
}


/* 0.10.0, April 02, 2008; FIXME: define as macro ? */
LUA_API int agn_isfail (lua_State *L, int idx) {
  return l_isfail(index2adr(L, idx));
}


/* 0.9.1, January 12, 2008; extended 0.10.0 to handle sets and sequences */
LUA_API size_t agn_nops (lua_State *L, int idx) {
  StkId o;
  o = index2adr(L, idx);
  switch (ttype(o)) {
    case LUA_TTABLE:
      return agenaV_nops(L, hvalue(o));
    case LUA_TSET:
      return usvalue(o)->size;
    case LUA_TSEQ:
      return seqvalue(o)->size;
    case LUA_TREG:
      return regvalue(o)->top;
    case LUA_TPAIR:
      return 2;  /* 2.10.0, for unpack */
    case LUA_TSTRING:  /* 2.12.1 */
      return tsvalue(o)->len;
    default:
      return 0;
  }
}


LUA_API int agn_fnext (lua_State *L, int idxt, int idxf, int mode) {
  StkId t, f;
  int more;
  lua_lock(L);
  t = index2adr(L, idxt);
  api_check(L, ttistable(t));
  f = index2adr(L, idxf);
  api_check(L, ttisfunction(f));
  more = luaHF_next(L, hvalue(t), f, L->top - 1, mode);
  if (more) {
    if (mode == 1) {
      api_incr_top_by_three(L);
    }
    else {
      api_incr_top_by_two(L);
    }
  }
  else  /* no more elements */
    L->top--;  /* remove key */
  lua_unlock(L);
  return more;
}


/* extended 0.22.0 to report holes in an array */
LUA_API void agn_tablestate (lua_State *L, int idx, size_t a[], int mode) {
  StkId o;
  Table *t;
  int lowestidx;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  t = hvalue(o);
  if (mode)
    agnH_nops(L, t, a);  /* a[0] = assigned elements in array, a[1] = assigned elements in hash part,
                            a[2] = info whether array contains at least one hole */
  else {
    a[0] = a[1] = a[2] = 0;
  }
  a[3] = t->sizearray;  /* number of allocated elements in array */
  /* number of allocated elements in hash part: */
  a[4] = (a[1] == 1 && actnodesize(t) == 0) ? 1 : actnodesize(t);  /* Agena 1.0.5 patch */
  a[5] = t->hasnil;
  a[6] = t->metatable != NULL;
  a[7] = luaH_isdummy(t->node);  /* 2.3.0 RC 4 */
  a[8] = luaH_getn(t);  /* 2.3.0 RC 4 */
  a[10] = agnH_aborders(L, t, &lowestidx);
  a[9] = lowestidx;
  lua_unlock(L);
}


LUA_API void agn_tablesize (lua_State *L, int idx, size_t a[]) {  /* 2.3.0 RC 4 */
  StkId o;
  Table *t;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  t = hvalue(o);
  a[0] = luaH_getn(t);  /* rough guess on the size of a table using a logarithmic method */
  a[1] = luaH_isdummy(t->node);  /* hash part empty ? */
  a[2] = t->hasnil;     /* estimate on whether `null` has been assigned to a table */
  lua_unlock(L);
}


LUA_API void agn_arrayborders (lua_State *L, int idx, size_t a[]) {  /* 2.14.10 */
  StkId o;
  Table *t;
  int low;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  t = hvalue(o);
  a[1] = agnH_aborders(L, t, &low);
  a[0] = low;
  lua_unlock(L);
}


/* Returns the smallest and largest assigned index - in this order - in the array and hash part of a table, in array a.
   If zeros are returned, the array and hash parts of the table are empty. 2.30.1 */
LUA_API void agn_borders (lua_State *L, int idx, size_t a[]) {
  StkId o;
  Table *t;
  int low;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  t = hvalue(o);
  a[1] = agnH_borders(L, t, &low);
  a[0] = low;
  lua_unlock(L);
}


/* Returns all integer indices of the table t at stack index idx in a new table and puts it onto the top of the stack.
   It also sets flag to 0 if there are no integer indices in hash part of t, and flag to 1 if there is at least one
   integer index in the hash part. 2.30.1 */
LUA_API void agn_integerindices (lua_State *L, int idx, int *flag) {
  StkId o;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  agenaV_tintegerindices(L, o, flag);  /* increases the stack top by one */
  lua_unlock(L);
}


/* Returns all the values stored to the table at stack index idx in a new table and sets it to the top of the stack.
   flag is set to 0 if no value are residing in the hash part, and to 1 if there is at least one element in the hash
   part. 2.30.1 */
LUA_API void agn_entries (lua_State *L, int idx, int *flag) {
  StkId o;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  agenaV_tentries(L, o, flag);  /* increases the stack top by one */
  lua_unlock(L);
}


/* Pushes two tables onto the top of the stack: one with all the values in the array part of the table at index idx,
  and one with the values in its hash part. 2.30.1 */
LUA_API void agn_parts (lua_State *L, int idx) {
  StkId o;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  agenaV_tparts(L, o, 0);  /* increases the stack top by one */
  agenaV_tparts(L, o, 1);  /* increases the stack top by one, as well */
  lua_unlock(L);
}


/* Pushes a table with all the values in the array part of the table at index idx onto the top of the stack. 2.30.1 */
LUA_API void agn_arraypart (lua_State *L, int idx) {
  StkId o;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  agenaV_tparts(L, o, 0);  /* increases the stack top by one */
  lua_unlock(L);
}


/* Pushes a table with all the values in the hash part of the table at index idx onto the top of the stack. 2.30.1 */
LUA_API void agn_hashpart (lua_State *L, int idx) {
  StkId o;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  agenaV_tparts(L, o, 1);  /* increases the stack top by one */
  lua_unlock(L);
}


/* Moves all values in the hash part of the table at stack index idx to the end of its array part, emptying the hash part.
   The function works destructively and in-place. The function does not change the stack. 2.30.1 */
LUA_API void agn_reorder (lua_State *L, int idx) {
  StkId o;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  agnH_reorder(L, hvalue(o));
  lua_unlock(L);
}


/* Converts a table, sequence or register into a set; 2.30.2; grep "TEMPLATE to call the VM via the API" */
LUA_API void agn_toset (lua_State *L, int idx) {
  lua_lock(L);
  luaC_checkGC(L);  /* 2.34.4: collect garbage _before_ calling "VM", see lua_concat */
  agenaV_toset(L, index2adr(L, idx), L->top);  /* increases the stack top by one */
  api_incr_top(L);  /* increase the stack top by one */
  lua_unlock(L);
}


/* Determines the product of all the numbers in a table, sequence or register; DEFUNCT for calling an
   operator is 38 % faster. However, this demonstrator works flawlessly without Valgrind complaining;

   From a C library, call the function like this:

   static int agnB_mulup (lua_State *L) {
     agn_mulup(L, 1);
     return 1;
   };  2.30.2 */
LUA_API void agn_mulup (lua_State *L, int idx) {
  lua_lock(L);
  luaC_checkGC(L);  /* 2.34.4: collect garbage _before_ calling "VM", see lua_concat */
  agenaV_mulup(L, index2value(L, idx), L->top);
  api_incr_top(L);  /* increase the stack top by one */
  lua_unlock(L);
}


LUA_API void agn_sstate (lua_State *L, int idx, size_t a[]) {
  StkId o;
  UltraSet *t;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttisuset(o));
  t = usvalue(o);
  a[0] = t->size;
  a[1] = actnodesize(t);
  a[2] = t->metatable != NULL;
  lua_unlock(L);
}


LUA_API void agn_seqstate (lua_State *L, int idx, size_t a[]) {
  StkId o;
  Seq *t;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttisseq(o));
  t = seqvalue(o);
  a[0] = t->size;
  a[1] = t->maxsize;
  a[2] = t->metatable != NULL;
  lua_unlock(L);
}


/* Returns the current top pointer, the total number of items, and a flag indicating whether a metatable has been assigned to the register at index idx in a, a C array with three entries. The position of the top pointer is stored to a[0], the total number of entries to a[1]. The metatable flag is stored to a[2], where 1 indicates that the sequence features a metatable, and 0 means it does not. */

LUA_API void agn_regstate (lua_State *L, int idx, size_t a[]) {  /* 2.3.0 RC 3 */
  StkId o;
  Reg *t;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttisreg(o));
  t = regvalue(o);
  a[0] = t->top;
  a[1] = t->maxsize;
  a[2] = t->metatable != NULL;
  lua_unlock(L);
}


LUA_API void agn_pairstate (lua_State *L, int idx, size_t a[]) {  /* 2.3.0 RC 3 */
  StkId o;
  Reg *t;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttisreg(o));
  t = regvalue(o);
  a[0] = t->metatable != NULL;
  lua_unlock(L);
}


LUA_API int agn_regsize (lua_State *L, int idx) {  /* 2.3.0 RC 3 */
  StkId o;
  Reg *t;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttisreg(o));
  t = regvalue(o);
  return t->maxsize;
}


LUA_API size_t agn_reggettop (lua_State *L, int idx) {  /* 2.3.0 RC 3 */
  StkId o;
  Reg *t;
  o = index2adr(L, idx);
  api_checkvalidindex(L, o);
  lua_lock(L);
  api_check(L, ttisreg(o));
  t = regvalue(o);
  return t->top;
}


LUA_API int agn_regsettop (lua_State *L, int idx) {  /* 2.3.0 RC 3 */
  StkId o;
  Reg *t;
  int r;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttisreg(o));
  t = regvalue(o);
  r = agnReg_settop(t, L->top - 1);
  lua_unlock(L);
  L->top--;
  return r;
}


LUA_API size_t agn_seqsize (lua_State *L, int idx) {
  StkId o;
  Seq *t;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttisseq(o));
  t = seqvalue(o);
  lua_unlock(L);
  return t->size;
}


LUA_API size_t agn_size (lua_State *L, int idx) {
  StkId o;
  Table *t;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  t = hvalue(o);
  lua_unlock(L);
  return agenaV_nops(L, t);
}


/* returns number of assigned (not allocated) elements in the array part of a table */
LUA_API size_t agn_asize (lua_State *L, int idx) {
  StkId o;
  Table *t;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttistable(o));
  t = hvalue(o);
  lua_unlock(L);
  return agnH_asize(L, t);  /* tuned 2.11.4 */
}


LUA_API size_t agn_ssize (lua_State *L, int idx) {
  StkId o;
  UltraSet *t;
  lua_lock(L);
  o = index2adr(L, idx);
  api_check(L, ttisuset(o));
  t = usvalue(o);
  lua_unlock(L);
  return t->size;
}


/* 0.22.0: returns 1 if the function at idx is a C function, 0 if the function at idx is an Agena function,
   and -1 if it is no function at all. */
LUA_API int agn_getfunctiontype (lua_State *L, int idx) {
  const TValue *o;
  int fn;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) == LUA_TFUNCTION)
    fn = clvalue(o)->l.isC;
  else
    fn = -1;
  lua_unlock(L);
  return fn;
}


LUA_API void agn_creatertable (lua_State *L, int idx, int writemode) {
  TValue *o;
  Table *nullt;
  lua_lock(L);
  api_checknelems(L, 1);
  o = index2adr(L, idx);
  if (isLfunction(o)) {
    nullt = luaH_new(L, 0, 0);
    if (nullt) {
      clvalue(o)->l.rtable = nullt;
      luaC_objbarrier(L, clvalue(o), nullt);
    } else
      if (writemode)
        luaG_runerror(L, "Error: rtable initialisation failure.");
      else
        luaG_runerror(L, "Error: rotable initialisation failure.");
    if (writemode) clvalue(o)->l.updatertable = 1;  /* 0.22.0 */
  }
  else if (iscfunction(o) && !writemode) {
    nullt = luaH_new(L, 0, 0);
    if (nullt) {
      clvalue(o)->c.rtable = nullt;
      luaC_objbarrier(L, clvalue(o), nullt);
    }
    else
      luaG_runerror(L, "Error: rotable initialisation failure.");
  }
  else
    luaG_runerror(L, "Error: rtables are available for Agena procedures only.");
  lua_unlock(L);
}


LUA_API int agn_getrtable (lua_State *L, int idx) {
  const TValue *o;
  Table *mt = NULL;
  int res;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) == LUA_TFUNCTION)
    mt = clvalue(o)->l.rtable;
  else
    luaG_runerror(L, "Error: remember tables are available for procedures only.");
  if (mt == NULL) {
    res = 0; }
  else {
    sethvalue(L, L->top, mt);
    api_incr_top(L);
    res = 1;
  }
  lua_unlock(L);
  return res;
}


/* 0.22.0: returns 0 if rtable cannot be updated by the RETURN statement, 1 if it can,
   2 if `idx` has no remember table at all, and -1 if `idx` is not a function. */
LUA_API int agn_getrtablewritemode (lua_State *L, int idx) {
  const TValue *o;
  int mode = 0;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) == LUA_TFUNCTION)
    mode = (clvalue(o)->l.rtable == NULL) ? 2 : clvalue(o)->l.updatertable;
  else
    mode = -1;
  lua_unlock(L);
  return mode;
}


LUA_API void agn_setrtable (lua_State *L, int find, int kind, int vind) {
  TValue *o;
  Table *rt, *dict;
  int hashval, newone, flag;
  const TValue *entry;
  lua_lock(L);
  if (!lua_isfunction(L, find))
    luaG_runerror(L, "Error: must get a function as first argument.");
  if (!lua_istable(L, kind))
    luaG_runerror(L, "Error: must get a table as second argument.");
  if (!lua_istable(L, vind))
    luaG_runerror(L, "Error: must get a table as third argument.");
  o = index2adr(L, find);
  api_checkvalidindex(L, o);
  rt = clvalue(o)->l.rtable;
  if (rt == NULL)
    luaG_runerror(L, "Error: rtable not initialised.");
  if (!lua_checkstack(L, 2))
    luaG_runerror(L, "Error: not enough memory for stack available.");
  lua_pushnil(L);
  /* compute the respective hash value for the argument list given */
  hashval = 0;
  while (lua_next(L, kind) != 0) {
    hashval += hash(L->top - 1);  /* compute hash for all values in argument table */
    agn_poptop(L);
  }
  /* look whether the hash value already exists */
  entry = luaH_getnum(rt, hashval);
  if ((newone = (ttisnil(entry)))) {
    dict = luaH_new(L, 0, 1);  /* create a new dictionary */
  } else {
    dict = hvalue(entry);      /* use the existing dictionary */
  }
  flag = 0;
  if (!newone) {
    /* search for existing equal keys (arguments) in dictionary assigned to the respective hash value, 0.22.0 */
    Node *node;
    int i;
    Table *rk, *tk;
    tk = hvalue(index2adr(L, kind));  /* get argument table given by rset */
    for (i = 0; i < sizenode(dict); i++) {
      node = gnode(dict, i);
      if (ttisnotnil(gval(node))) {  /* key (the table of arguments) found */
        rk = hvalue(key2tval(node));  /* assign key (table of arguments) to Table object rk */
        if ((flag = agenaV_comptablesonebyone(L, rk, tk))) {  /* there is already in entry in the rtable */
          sethvalue(L, L->top, rk);
          setobjt2t(L, luaH_set(L, dict, L->top), index2adr(L, vind));  /* get results table given by rset */
          luaC_barriert(L, dict, index2adr(L, vind));
          break;
        }
      }
    }
  }
  if (!flag) {  /* arguments not yet included in rtable ? */
    /* first enter arguments~results pair into the dictionary */
    setobjt2t(L, luaH_set(L, dict, index2adr(L, kind)), index2adr(L, vind));
    luaC_barriert(L, dict, index2adr(L, vind));
  }
  /* set table to stack top */
  sethvalue(L, L->top, dict);
  /* then enter key~value pair (arguments~results) into rtable and assign it to key `hashvalue` */
  setobjt2t(L, luaH_setnum(L, rt, hashval), L->top);
  luaC_barriert(L, rt, L->top);
  lua_unlock(L);
}


LUA_API void agn_deletertable (lua_State *L, int objindex) {
  TValue *o;
  lua_lock(L);
  api_checknelems(L, 1);
  o = index2adr(L, objindex);
  api_checkvalidindex(L, o);  /* Agena 1.0.3 fix */
  /* do not luaH_free remember table, leave this to gc; Valgrind does not complain */
  if (isLfunction(o)) {
    clvalue(o)->l.rtable = NULL;
    clvalue(o)->l.updatertable = 0; }
  else if (iscfunction(o)) {
    clvalue(o)->c.rtable = NULL;
    clvalue(o)->c.updatertable = 0; }
  else
    luaG_runerror(L, "Error: remember tables are available for procedures only.");
  lua_unlock(L);
}


LUA_API int agn_getstorage (lua_State *L, int idx) {  /* 2.33.1 */
  const TValue *o;
  Table *mt = NULL;
  int res;
  lua_lock(L);
  o = index2adr(L, idx);
  if (ttype(o) == LUA_TFUNCTION)
    mt = clvalue(o)->l.storage;
  else
    luaG_runerror(L, "Error: storage tables are available for procedures only.");
  if (mt == NULL) {
    res = 0;
  } else {
    sethvalue(L, L->top, mt);
    api_incr_top(L);
    res = 1;
  }
  lua_unlock(L);
  return res;
}



/* sets the name of a readlibbed package to the registry, leaves nothing on the stack */
LUA_API int agn_setreadlibbed (lua_State *L, const char *libname) {  /* rewritten 2.9.1 */
  int rc = 0;
  lua_lock(L);
  lua_getfield(L, LUA_REGISTRYINDEX, "_READLIBBED");
  if (!lua_isset(L, -1)) {
    fprintf(stderr, "Warning: internal registry _READLIBBED set not found.\n");
  } else if (!luaL_isstandardlib(L, libname)) {  /* 2.31.9 fix especially for OS/2, DOS and ANSI */
    lua_pushstring(L, libname);  /* insert package name into _READLIBBED set */
    lua_srawset(L, -2);
    rc = 1;
  }
  agn_poptop(L);  /* pop _READLIBBED (or anything) from stack */
  lua_unlock(L);
  return rc;
}


LUA_API void agn_setclosetozero (lua_State *L, lua_Number x) {  /* 2.32.0 */
  if (x <= 0)
    luaG_runerror(L, "Error in " LUA_QS ": value must be positive.", "environ.kernel/closetozero");
  L->div_closetozero = x;
}


LUA_API lua_Number agn_getclosetozero (lua_State *L) {  /* 2.32.0 */
  return L->div_closetozero;
}


LUA_API void agn_setforadjust (lua_State *L, int value) {  /* 2.31.0 */
  L->vmsettings ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->vmsettings) & 1;
}


LUA_API int agn_getforadjust (lua_State *L) {  /* 2.31.0 */
  return L->vmsettings & 1;
}


LUA_API void agn_setkahanozawa (lua_State *L, int value) {  /* 2.4.2 */
  L->vmsettings ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->vmsettings) & 2;
  L->vmsettings ^= (-((lu_byte)0) ^ L->vmsettings) & 4;  /* 2.30.5 */
}


LUA_API int agn_getkahanozawa (lua_State *L) {  /* 2.4.2 */
  return L->vmsettings & 2;
}


LUA_API void agn_setkahanbabuska (lua_State *L, int value) {  /* 2.30.5 */
  L->vmsettings ^= (-((lu_byte)0) ^ L->vmsettings) & 2;  /* 2.30.5 */
  L->vmsettings ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->vmsettings) & 4;
}


LUA_API int agn_getkahanbabuska (lua_State *L) {  /* 2.30.5 */
  return L->vmsettings & 4;
}


/* see: http://graphics.stanford.edu/~seander/bithacks.html on how to modify bits, 0.32.0 */
LUA_API void agn_setbitwise (lua_State *L, int value) {
  L->settings ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings) & 1;
}


LUA_API int agn_getbitwise (lua_State *L) {  /* 0.32.0 */
  return L->settings & 1;
}


LUA_API void agn_setemptyline (lua_State *L, int value) {  /* 0.32.0 */
  L->settings ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings) & 2;
}


LUA_API int agn_getemptyline (lua_State *L) {  /* 0.32.0 */
  return L->settings & 2;
}


LUA_API void agn_setlibnamereset (lua_State *L, int value) {  /* 0.32.0 */
  L->settings ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings) & 4;
}


LUA_API int agn_getlibnamereset (lua_State *L) {  /* 0.32.0 */
  return L->settings & 4;
}


LUA_API void agn_setlongtable (lua_State *L, int value) {  /* 0.32.0 */
  L->settings ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings) & 8;
}


LUA_API int agn_getlongtable (lua_State *L) {  /* 0.32.0 */
  return L->settings & 8;
}


LUA_API void agn_setdebug (lua_State *L, int value) {  /* 0.32.2a */
  L->settings ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings) & 16;
}


LUA_API int agn_getdebug (lua_State *L) {  /* 0.32.0 */
  return L->settings & 16;
}


LUA_API void agn_setgui (lua_State *L, int value) {  /* 0.33.3 */
  L->settings ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings) & 32;
}


LUA_API int agn_getgui (lua_State *L) {  /* 0.33.3 */
  return L->settings & 32;
}


LUA_API void agn_setzeroedcomplex (lua_State *L, int value) {  /* 1.7.6 */
  L->settings ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings) & 64;
}


LUA_API int agn_getzeroedcomplex (lua_State *L) {  /* 1.7.6 */
  return L->settings & 64;
}


LUA_API void agn_setpromptnewline (lua_State *L, int value) {  /* 1.7.6 */
  L->settings ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings) & 128;
}


LUA_API int agn_getpromptnewline (lua_State *L) {  /* 1.7.6 */
  return L->settings & 128;
}


LUA_API void agn_setnoini (lua_State *L, int value) {  /* 2.8.6 */
  L->settings2 ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings2) & 1;
}


LUA_API int agn_getnoini (lua_State *L) {  /* 2.8.6 */
  return L->settings2 & 1;
}


LUA_API void agn_setnomainlib (lua_State *L, int value) {  /* 2.8.6 */
  L->settings2 ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings2) & 2;
}


LUA_API int agn_getnomainlib (lua_State *L) {  /* 2.8.6 */
  return L->settings2 & 2;
}


LUA_API void agn_setiso8601 (lua_State *L, int value) {  /* 2.9.8 */
  L->settings2 ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings2) & 4;
}


LUA_API int agn_getiso8601 (lua_State *L) {  /* 2.9.8 */
  return L->settings2 & 4;
}


LUA_API int agn_getconstants (lua_State *L) {  /* 2.20.0 */
  return L->settings2 & 8;
}


LUA_API void agn_setconstants (lua_State *L, int value) {  /* 2.20.0 */
  L->settings2 ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings2) & 8;
}


LUA_API int agn_getduplicates (lua_State *L) {  /* 2.20.1 */
  return L->settings2 & 16;
}


LUA_API void agn_setduplicates (lua_State *L, int value) {  /* 2.20.1 */
  L->settings2 ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings2) & 16;
}


LUA_API int agn_getseqautoshrink (lua_State *L) {  /* 2.29.0 */
  return L->settings2 & 32;
}


LUA_API void agn_setseqautoshrink (lua_State *L, int value) {  /* 2.29.0 */
  L->settings2 ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings2) & 32;
}


LUA_API int agn_getskipagenapath (lua_State *L) {  /* 2.35.4 */
  return L->settings2 & 64;
}


LUA_API void agn_setskipagenapath (lua_State *L, int value) {  /* 2.35.4 */
  L->settings2 ^= (-(value > 0 ? (lu_byte)1 : (lu_byte)0) ^ L->settings2) & 64;
}


LUA_API void agn_setbuffersize (lua_State *L, int value) {  /* 2.2.0 */
  if (value > 1073741824 || value < 512)
    luaG_runerror(L, "Error in " LUA_QS ": value must be in the range [512, 1024^3].", "environ.kernel/buffersize");
  else
    L->buffersize = value;
}


LUA_API int agn_getbuffersize (lua_State *L) {  /* 2.2.0 */
  return L->buffersize;
}


LUA_API void agn_setdigits (lua_State *L, ptrdiff_t x) {  /* x must be a number in [1, 17]; rewritten 2.8.6 */
  char *num, *oldpos;
  if (x < 1 || x > 17) luaG_runerror(L, LUA_QS "Error: setting must be an integer in [1, 17].", "environ.kernel/digits");
  num = L->numberformat;
  oldpos = L->numberformat;
  while (!isdigit(*num)) { num++; }
  if (x < 10) { *num = '0'; num++; *num = '0' + x; }
  else { *num = '1'; num++; *num = '0' + x - 10; }
  L->numberformat = oldpos;
}


LUA_API void agn_setregsize (lua_State *L, size_t x) {  /* 2.3.0 RC 3, set default register size */
  if (x < 0 || x > MAX_INT)
    luaG_runerror(L, LUA_QS "Error: setting must be an integer in [1, &ld].", "environ.kernel/regsize", MAX_INT);
  L->regsize = x;
}


LUA_API int agn_getdigits (lua_State *L) {  /* rewritten 2.8.6 */
  int n;
  sscanf(L->numberformat, "%*[^0123456789]%d", &n);
  return n;
}


LUA_API int agn_geterrmlinebreak (lua_State *L) {  /* 2.11.4 */
  return L->errmlinebreak;
}


LUA_API void agn_seterrmlinebreak (lua_State *L, int x) {  /* 2.11.4, set number of chars per line in syntax error messages. */
  if (x < 1 || tools_isfloat(x))
    luaG_runerror(L, LUA_QS "Error: setting must be a positive integer.", "environ.kernel/errmlinebreak");
  L->errmlinebreak = x;
}



LUA_API void agn_setepsilon (lua_State *L, lua_Number eps) {  /* 2.1.4 */
  if (eps < 0) luaG_runerror(L, "Error in " LUA_QS ": value must not be negative.", "environ.kernel/eps");
  L->Eps = eps;
  lua_pushnumber(L, eps);  /* 2.21.8 fix: change "Eps" in environment, too */
  lua_setglobal(L, "Eps");
}


LUA_API void agn_setdblepsilon (lua_State *L, lua_Number dbleps) {  /* 2.21.8 */
  if (dbleps < 0) luaG_runerror(L, "Error in " LUA_QS ": value must be non-negative.", "environ.kernel/doubleeps");
  L->DoubleEps = dbleps;
  lua_pushnumber(L, dbleps);  /* change "DoubleEps" in environment, too */
  lua_setglobal(L, "DoubleEps");
}


LUA_API void agn_sethepsilon (lua_State *L, lua_Number heps) {  /* 2.31.1 */
  if (heps < 0) luaG_runerror(L, "Error in " LUA_QS ": value must be non-negative.", "environ.kernel/heps");
  L->hEps = heps;
}


LUA_API size_t agn_getregsize (lua_State *L) {  /* default register size, 2.3.0 RC 3 */
  return L->regsize;
}


LUA_API lua_Number agn_getstructuresize (lua_State *L, int idx) {
  const TValue *o;
  o = index2adr(L, idx);
  api_checkvalidindex(L, o);  /* Agena 1.0.3 fix */
  switch (ttype(o)) {
    case LUA_TTABLE: {
      Table *h = hvalue(o);
      return (lua_Number)(sizeof(Table) + sizeof(TValue) * h->sizearray + sizeof(Node) * actnodesize(h));
    }
    case LUA_TSET: {
      UltraSet *s = usvalue(o);
      return (lua_Number)(sizeof(UltraSet) + sizeof(LNode)*actnodesize(s));
    }
    case LUA_TSEQ: {
      Seq *s = seqvalue(o);
      return (lua_Number)(sizeof(Seq) + sizeof(TValue)*s->maxsize);
    }
    case LUA_TPAIR: {
      return (lua_Number)(sizeof(Pair) + sizeof(TValue)*2);
    }
    case LUA_TFUNCTION: {
      Closure *cl = clvalue(o);
      return (lua_Number)((cl->c.isC) ? sizeCclosure(cl->c.nupvalues) :
                           sizeLclosure(cl->l.nupvalues));
    }
    case LUA_TREG: {
      Reg *s = regvalue(o);
      return (lua_Number)(sizeof(Reg) + sizeof(TValue)*s->maxsize);
    }
  }
  return 0;
}


/* Allocates size bytes of memory and returns a pointer to the newly allocated block. In case memory could not be
   allocated, it returns an error message including procname that called agn_malloc. The function optionally can
   free one or more other objects referenced by their pointers in case memory allocation failed.

   In all cases, the last argument must be NULL. */
LUA_API void *agn_malloc (lua_State *L, size_t size, const char *procname, ...) {  /* 1.9.1 */
  void *ptr;
  ptr = tools_malloc(size, 0);  /* 2.27.0 */
  if (ptr == NULL) {
    void *s;
    va_list args;
    va_start(args, procname);
    while ((s = va_arg(args, void *))) {
      xfree(s);
    }
    va_end(args);
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", procname);
  }
  return ptr;
}


/* Allocates a string buffer by internally determining its most efficient size, aligned along the "long" boundary. The
   return is a char* pointer to the beginning of the string. The function zeros only the last few bytes and assumes that
   the `leading` rest will be filled by real characters later on. Just pass l as the number of characters, excluding the
   terminating \0, and do not multiply it by sizeof(char). The function automatically adds a terminating \0.

   The function can optionally free variables passed after `procname` in case memory allocation fails internally. In any case,
   the last argument must always be NULL.

   The function is around six percent faster than agn_calloc. See also tools_stralloc in agnhlps.c. */
LUA_API char *agn_stralloc (lua_State *L, size_t len, const char *procname, ...) {  /* 2.16.5 */
  char *buf;
  size_t l, rem;
  /* determine optimal size, including terminating \0 */
  l = len;
  rem = (++l) % AGN_BLOCKSIZE;  /* length including \0 */
  l += (rem != 0)*(AGN_BLOCKSIZE - rem);
  buf = (char *)malloc(l*sizeof(char));
  if (buf == NULL) {
    void *s;
    va_list args;
    va_start(args, procname);             /* initialize the argument list */
    while ((s = va_arg(args, void *))) {  /* get the next argument value */
      xfree(s);
    }
    va_end(args);                         /* clean up */
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", procname);
  }
  memset(buf + len, 0, l - len);  /* only zero the bytes that will not be filled by characters later on */
  return buf;
}


LUA_API void agn_arraytoseq (lua_State *L, lua_Number *a, size_t n) {  /* converts an array to a sequence and pushes it on the top of the stack */
  size_t i;
  agn_createseq(L, n);
  for (i=0; i < n; i++)
    agn_seqsetinumber(L, -1, i + 1, a[i]);
}


LUA_API void agn_longarraytoseq (lua_State *L, long double *a, size_t n) {  /* converts an array to a sequence and pushes it on the top of the stack */
  size_t i;
  agn_createseq(L, n);
  for (i=0; i < n; i++)
    agn_seqsetinumber(L, -1, i + 1, a[i]);
}


LUA_API lua_Number agn_geteps (lua_State *L, const char *varname) {  /* 2.1.3, extended 2.21.8 */
  lua_Number Eps;
  lua_getglobal(L, varname);
  Eps = agn_isnumber(L, -1) ? agn_tonumber(L, -1) : AGN_EPSILON;
  agn_poptop(L);
  if (Eps < 0) luaG_runerror(L, "Error: Eps must not be negative.");
  return Eps;
}


/* Expects a valid userdata metatable at the top of the stack, assigns it to the userdata residing at stack
   index idx, and pops the value at the top of the stack thereafter. If the value at the top of the
   stack is null, then a metatable assigned to a userdatum is deleted, and null is popped from the stack. */
LUA_API void agn_setudmetatable (lua_State *L, int idx) {  /* 2.3.0 RC 3 */
  TValue *obj;
  Table *mt;
  lua_lock(L);
  obj = index2adr(L, idx);
  api_checkvalidindex(L, obj);
  if (ttisnil(L->top - 1))
    mt = NULL;
  else {
    api_check(L, ttistable(L->top - 1));
    mt = hvalue(L->top - 1);
  }
  uvalue(obj)->metatable = mt;
  if (mt)
    luaC_objbarrier(L, rawuvalue(obj), mt);
  L->top--;
  lua_unlock(L);
}


LUA_API lua_Integer lua_tointegerx (lua_State *L, int idx, int *isnum) {  /* taken from Lua 5.2.4, tuned 2.21.2 */
  lua_Number n;
  const TValue *o = index2adr(L, idx);
  if (tonumberx(o, &n)) {
    lua_Integer res;
    lua_number2integer(res, n);
    if (isnum) *isnum = 1;
    return res;
  }
  else {
    if (isnum) *isnum = 0;
    return 0;
  }
}


LUA_API lua_Unsigned lua_tounsignedx (lua_State *L, int idx, int *isnum) {  /* taken from Lua 5.2.4, tuned 2.21.2 */
  lua_Number n;
  const TValue *o = index2adr(L, idx);
  if (tonumberx(o, &n)) {
    lua_Unsigned res;
    lua_number2unsigned(res, n);
    if (isnum) *isnum = 1;
    return res;
  }
  else {
    if (isnum) *isnum = 0;
    return 0;
  }
}


/* Pushes an unsigned int onto the stack. */
LUA_API void lua_pushunsigned (lua_State *L, lua_Unsigned u) {  /* taken from Lua 5.2.4 */
  lua_Number n;
  lua_lock(L);
  n = lua_unsigned2number(u);
  setnvalue(L->top, n);
  api_incr_top(L);
  lua_unlock(L);
}


/* Gets the current rounding mode. Returns "downward" for FE_DOWNWARD, "upward" for FE_UPWARD, "nearest" for FE_TONEAREST,
   and "zero" for FE_TOWARDZERO. Not available in DOS. */
#ifndef LUA_DOS
LUA_API void agn_getround (lua_State *L) {  /* 2.8.6 */
  int r = fegetround();
  if (r < 0) {
    setnvalue(L->top, AGN_NAN);
  } else {
    switch (r) {
      case FE_DOWNWARD: {
        setsvalue(L, L->top, luaS_new(L, "downward"));
        break; /* rounding direction is set to the next lower integer. */
      }
      case FE_UPWARD: {
        setsvalue(L, L->top, luaS_new(L, "upward"));
        break;  /* rounding direction is set to the next greater integer. */
      }
      case FE_TONEAREST: {
        setsvalue(L, L->top, luaS_new(L, "nearest"));
        break;  /* round up or down toward whichever integer is nearest */
      }
      case FE_TOWARDZERO: {
        setsvalue(L, L->top, luaS_new(L, "towardzero"));
        break;  /* round positive values downward and negative values upward */
      }
      default: {
        setbvalue(L->top, 2);  /* push fail */
      }
    }
  }
  api_incr_top(L);
  return;
}


/* Sets the rounding mode. `what` may be "downward" for DE_DOWNWARD, "upward" for FE_UPWARD, "nearest" for FE_TONEAREST,
   and "zero" for FE_TOWARDZERO. Not available in DOS. */
LUA_API int agn_setround (lua_State *L, const char *what) {  /* 2.8.6 */
  int old, r;
  old = fegetround();
  if (tools_streq(what, "downward"))  /* 2.16.12 tweak */
    r = fesetround(FE_DOWNWARD);
  else if (tools_streq(what, "upward"))  /* 2.16.12 tweak */
    r = fesetround(FE_UPWARD);
  else if (tools_streq(what, "nearest"))  /* 2.16.12 tweak */
    r = fesetround(FE_TONEAREST);
  else if (tools_streq(what, "towardzero"))  /* 2.16.12 tweak */
    r = fesetround(FE_TOWARDZERO);
  else r = -1;
  if (r != 0)
    fesetround(old);
  return (r == 0);
}
#endif


LUA_API int agn_int2fb (lua_State *L, unsigned int x) {  /* 2.13.0 */
  return luaO_int2fb(x);
}


LUA_API int agn_fb2int (lua_State *L, int x) {  /* 2.13.0 */
  return luaO_fb2int(x);
}


LUA_API int agn_tofileno (lua_State *L, int n) {  /* 2.17.8 */
  int hnd;
  FILE **p = topnfile(L, n);
  hnd = fileno(*p);
  /* do not NULL p */
  return hnd;
}


/* start, end, init starting from 1, not 0. */
LUA_API const char *agn_strmatch (lua_State *L, const char *s, size_t s_len, 
    const char *p, ptrdiff_t init, ptrdiff_t *start, ptrdiff_t *end) {  /* 2.36.1 */
  int anchor;
  const char *s1, *res;
  MatchState ms;
  init = tools_posrelat(init, s_len) - 1;
  if (init < 0) init = 0;
  *start = *end = 0;
  if ((size_t)init >= s_len) return NULL;
  /* checking with strpbrk for special characters gives no gain in speed */
  anchor = (*p == '^') ? (p++, 1) : 0;
  s1 = s + init;
  ms.L = L;
  ms.src_init = s;
  ms.src_end = s + s_len;
  do {
    ms.level = 0;
    if ((res=match(&ms, s1, p, 0)) != NULL) {
      *start = s1 - s + 1; *end = res - s;
      return s1;
    }
  } while (s1++ < ms.src_end && !anchor);
  return NULL;
}


/* +++++++++++++++++++++++++++++++++++++++++++++++++ LUA 5.4 API FUNCTIONS +++++++++++++++++++++++++++++++++++++++++++++++ */

/* Returns the length of the value at the given index. It is equivalent to the 'size' operator in Lua and may trigger a
   metamethod for the "length" event. The result is pushed on the stack. Taken from Lua 5.4.0 RC 5, 2.21.2. */
LUA_API void lua_len (lua_State *L, int idx) {
  TValue *t;
  lua_lock(L);
  t = index2value(L, idx);
  luaV_objlen(L, L->top, t);
  api_incr_top(L);
  lua_unlock(L);
}


/* Performs an arithmetic operation over the two values (or one, in the case of negations) at the top of the stack, with the
   value at the top being the second operand, pops these values, and pushes the result of the operation. The function follows
   the semantics of the corresponding Agena operator (that is, it may call metamethods).

   The value of op must be one of the following constants:

   OP_ADD: performs addition (+)
   OP_SUB: performs subtraction (-)
   OP_MUL: performs multiplication (*)
   OP_DIV: performs float division (/)
   OP_INTDIV: performs integer division (\)
   OP_MOD: performs modulo (%)
   OP_POW: performs exponentiation (^)
   OP_IPOW: performs integer exponentiation (**)
   OP_UNM: performs mathematical negation (unary -)

   Taken from Lua 5.4.4, 2.21.1, 2.26.2 */
LUA_API void lua_arith (lua_State *L, int op) {
  lua_lock(L);
  if (op != LUA_OPUNM && op != LUA_OPBNOT) {
    api_checknelems(L, 2);  /* all other operations expect two operands */
  } else {  /* for unary operations, add fake 2nd operand */
    api_checknelems(L, 1);
    setobjs2s(L, L->top, L->top - 1);
    api_incr_top(L);
  }
  /* first operand at top - 2, second at top - 1; result go to top - 2 */
  luaO_arith(L, op, s2v(L->top - 2), s2v(L->top - 1), L->top - 2);
  L->top--;  /* remove second operand */
  lua_unlock(L);
}


/* Compares two Agena values. Returns 1 if the value at index index1 satisfies op when compared with the value at index index2,
   following the semantics of the corresponding Lua operator (that is, it may call metamethods). Otherwise returns 0. Also returns
   0 if any of the indices is not valid.

   The value of op must be one of the following constants:

     OP_EQ: compares for equality (==)
     OP_LT: compares for less than (<)
     OP_LE: compares for less or equal (<=)

   Taken from Lua 5.4.4, 2.21.1 */
LUA_API int lua_compare (lua_State *L, int index1, int index2, int op) {
  const TValue *o1;
  const TValue *o2;
  int i = 0;
  lua_lock(L);  /* may call tag method */
  o1 = index2value(L, index1);
  o2 = index2value(L, index2);
  if (isvalid(L, o1) && isvalid(L, o2)) {
    switch (op) {
      case OP_EQ: i = luaV_equalobj(L, o1, o2); break;
      case OP_LT: i = luaV_lessthan(L, o1, o2, 1); break;
      case OP_LE: i = luaV_lessequal(L, o1, o2, 1); break;
      default: api_check(L, "invalid option");
    }
  }
  lua_unlock(L);
  return i;
}


/* Sets the warning function to be used by Lua to emit warnings (see lua_WarnFunction). The ud parameter sets
   the value ud passed to the warning function. Taken from Lua 5.4.4, 2.21.1 */
LUA_API void lua_setwarnf (lua_State *L, lua_WarnFunction f, void *ud) {
  lua_lock(L);
  G(L)->ud_warn = ud;
  G(L)->warnf = f;
  lua_unlock(L);
}


/*
** 0 - warning system is off;
** 1 - ready to start a new message;
** 2 - previous message is to be continued.
*/
LUA_API int lua_getwarnf (lua_State *L) {
  int *warnstate = (int *)(G(L)->ud_warn);
  return *warnstate;
}


/* Emits a warning with the given message. A message in a call with tocont true should be continued in another call to this function.
   Taken from Lua 5.4.4, 2.21.1 */
LUA_API void lua_warning (lua_State *L, const char *msg, int tocont) {
  lua_lock(L);
  luaE_warning(L, msg, tocont);
  lua_unlock(L);
}


/* Taken from Lua 5.4.4, 2.21.1 */
/*
** Reverse the stack segment from 'from' to 'to'
** (auxiliary to 'lua_rotate')
** Note that we move(copy) only the value inside the stack.
** (We do not move additional fields that may exist.)
*/
static void reverse (lua_State *L, StkId from, StkId to) {
  for (; from < to; from++, to--) {
    TValue temp;
    setobj(L, &temp, s2v(from));
    setobjs2s(L, from, to);
    setobj2s(L, to, &temp);
  }
}


/* Taken from Lua 5.4.4, 2.21.1 */
static StkId index2stack (lua_State *L, int idx) {
  CallInfo *ci = L->ci;
  if (idx > 0) {
    StkId o = ci->func + idx;
    api_check(L, o < L->top);
    return o;
  }
  else {    /* non-positive index */
    api_check(L, idx != 0 && -idx <= L->top - (ci->func + 1));
    api_check(L, !ispseudo(idx));
    return L->top + idx;
  }
}

/* Rotates the stack elements between the valid index idx and the top of the stack. The elements are rotated n positions
   in the direction of the top, for a positive n, or -n positions in the direction of the bottom, for a negative n.
   The absolute value of n must not be greater than the size of the slice being rotated. This function cannot be called
   with a pseudo-index, because a pseudo-index is not an actual stack position. */
/* Taken from Lua 5.4.4, 2.21.1 */
/*
** Let x = AB, where A is a prefix of length 'n'. Then,
** rotate x n == BA. But BA == (A^r . B^r)^r.
*/
LUA_API void lua_rotate (lua_State *L, int idx, int n) {
  StkId p, t, m;
  lua_lock(L);
  t = L->top - 1;  /* end of stack segment being rotated */
  p = index2stack(L, idx);  /* start of segment */
  api_check(L, (n >= 0 ? n : -n) <= (t - p + 1));
  m = (n >= 0 ? t - n : p - n - 1);  /* end of prefix */
  reverse(L, p, m);  /* reverse the prefix with length 'n' */
  reverse(L, m + 1, t);  /* reverse the suffix */
  reverse(L, p, t);  /* reverse the entire segment */
  lua_unlock(L);
}


/* Copies the element at index fromidx into the valid index toidx, replacing the value at that position. Values at other
   positions are not affected. Taken from Lua 5.4.4, 2.21.1 */
LUA_API void lua_copy (lua_State *L, int fromidx, int toidx) {
  TValue *fr, *to;
  lua_lock(L);
  fr = index2value(L, fromidx);
  to = index2value(L, toidx);
  api_check(L, isvalid(L, to));
  setobj(L, to, fr);
  if (isupvalue(toidx))  /* function upvalue? */
    luaC_barrier(L, clvalue(s2v(L->ci->func)), fr);
  /* LUA_REGISTRYINDEX does not need gc barrier
     (collector revisits it before finishing collection) */
  lua_unlock(L);
}


/* Converts the zero-terminated string s to a number, pushes that number into the stack, and returns the total size of
   the string, that is, its length plus one. The conversion can result in an integer or a float, according to the lexical
   conventions of Agena. The string may have leading and trailing whitespaces and a sign. If the string is not a valid
   numeral, returns 0 and pushes nothing. (Note that the result can be used as a boolean, true if the conversion succeeds.)
   Taken from Lua 5.4.4, 2.21.1 */
LUA_API size_t lua_stringtonumber (lua_State *L, const char *s) {
  size_t sz = luaO_str2num(s, s2v(L->top));
  if (sz != 0)
    api_incr_top(L);
  return sz;
}


LUA_API const char *agn_gettoken (lua_State *L, int token) {
  if (token < FIRST_RESERVED) return NULL;
  return luaX_tokens[token - FIRST_RESERVED];
}


/***** Lua 5.2-compat functions, taken from lcurses-9.0.0.1 library, files /ext/include/compat-5.2.*; Agena 2.27.1 *****/

LUA_API lua_Number lua_tonumberx (lua_State *L, int i, int *isnum) {
  lua_Number n = lua_tonumber(L, i);
  if (isnum != NULL) {
    *isnum = (n != 0 || lua_isnumber(L, i));
  }
  return n;
}


#define PACKAGE_KEY "_COMPAT52_PACKAGE"

static void push_package_table (lua_State *L) {
  lua_pushliteral(L, PACKAGE_KEY);
  lua_rawget(L, LUA_REGISTRYINDEX);
  if (!lua_istable(L, -1)) {
    lua_pop(L, 1);
    /* try to get package table from globals */
    lua_pushliteral(L, "package");
    lua_rawget(L, LUA_GLOBALSINDEX);
    if (lua_istable(L, -1)) {
      lua_pushliteral(L, PACKAGE_KEY);
      lua_pushvalue(L, -2);
      lua_rawset(L, LUA_REGISTRYINDEX);
    }
  }
}

LUA_API void lua_getuservalue (lua_State *L, int i) {
  luaL_checktype(L, i, LUA_TUSERDATA);
  luaL_checkstack(L, 2, "not enough stack space");
  lua_getfenv(L, i);
  lua_pushvalue(L, LUA_GLOBALSINDEX);
  if (lua_rawequal(L, -1, -2)) {
    lua_pop(L, 1);
    lua_pushnil(L);
    lua_replace(L, -2);
  } else {
    lua_pop(L, 1);
    push_package_table(L);
    if (lua_rawequal(L, -1, -2)) {
      lua_pop(L, 1);
      lua_pushnil(L);
      lua_replace(L, -2);
    } else
      lua_pop(L, 1);
  }
}


LUA_API void lua_setuservalue (lua_State *L, int i) {
  luaL_checktype(L, i, LUA_TUSERDATA);
  if (lua_isnil(L, -1)) {
    luaL_checkstack(L, 1, "not enough stack space");
    lua_pushvalue(L, LUA_GLOBALSINDEX);
    lua_replace(L, -2);
  }
  lua_setfenv(L, i);
}


/*********************************************************************
* This file contains parts of Lua 5.2's source code:
*
* Copyright (C) 1994-2013 Lua.org, PUC-Rio.
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*********************************************************************/
