/*
** $Id: lreglib.c,v 1.38 2005/10/23 17:38:15 roberto Exp $
** Library for Register Manipulation
** See Copyright Notice in agena.h
*/


#include <stddef.h>

#define lreglib_c
#define LUA_LIB

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"


static int reg_settop (lua_State *L) {  /* 2.3.0 RC 3 */
  int r;
  luaL_argcheck(L, lua_isreg(L, 1), 1, "register expected");
  lua_pushvalue(L, 2);
  /* agnReg_settop checks the 2nd argument automatically and would return false in case of a wrong type */
  r = agn_regsettop(L, 1);
  lua_pushboolean(L, r);
  return 1;
}


static int reg_reduce (lua_State *L) {  /* 2.3.0 RC 3 */
  int nil;
  luaL_argcheck(L, lua_isreg(L, 1), 1, "register expected");
  nil = agnL_optboolean(L, 3, 1);
  lua_pushboolean(L, agn_regreduce(L, 1, agn_checkinteger(L, 2), nil));
  return 1;
}


static int reg_extend (lua_State *L) {  /* 2.3.0 RC 3 */
  luaL_argcheck(L, lua_isreg(L, 1), 1, "register expected");
  lua_pushboolean(L, agn_regextend(L, 1, agn_checkinteger(L, 2)));
  return 1;
}


/* creates a register with values a, a+step ..., b-step, b; based on calc.fseq
   uses a modified version of Kahan's summation algorithm to avoid roundoff errors
   0.30.0, December 30, 2009; merged with calc.fseq 1.3.3, February 01, 2011

   calc.fseq comments: creates a sequence with values f(a), ..., f(b); June 30, 2008;
   20 % faster than an Agena implementation; extended August 20, 2009, 0.27.0
   extended 0.30.0, December 29, 2009 with Kahan summation algorithm to avoid roundoff errors;
   extended 1.8.8, November 07, 2012, to also return non-numeric sequences; extended 2.4.1,
   January 15, 2015, to process the number of elements to be returned.

   For the idea on the Matlab version please see:
   http://de.mathworks.com/help/matlab/ref/linspace.html
   http://www.che.utah.edu/~sutherland/wiki/index.php/Matlab_Arrays#Linspace_and_Logspace */

static int reg_new (lua_State *L) {
  /* use `volatile` so that the compiler does not render the Kahan code effectless */
  /* volatile lua_Number idx, step, c, y, t; formerly nreg, 2.28.1; created October 15, 2014 */
  int isfunc, isint, isdefault;
  size_t counter, i, j, nargs, total, offset;
  lua_Number a, b, eps;
  volatile lua_Number s, c, cs, ccs, cc, t, idx, step, x;
  luaL_aux_nstructure(L, "registers.new", &nargs, &offset, &a, &b, &step, &eps, &total, &isfunc, &isdefault, &isint);
  luaL_checkstack(L, nargs, "too many arguments");
  agn_createreg(L, total);
  counter = 0;
  cs = ccs = 0;
  s = idx = a;
  if (isdefault) {  /* create a sequence of n slots and fill it with one and the same default of any type */
    agn_pairgeti(L, 2, 2);
  }
  /* total > counter: prevents that the last element is inserted even if a roundoff error occurred. */
  if (isfunc) {  /* function passed ? */
    while (idx <= b || tools_approx(idx, b, eps)) {
      lua_pushvalue(L, offset);  /* push function */
      lua_pushnumber(L, (fabs(idx) < AGN_HEPSILON) ? 0 : idx);  /* quite dirty hack to avoid roundoff errors with 0 */
      for (i=4 + offset, j = 0; i <= nargs; i++, j++) lua_pushvalue(L, i);
      lua_call(L, 1 + j, 1);
      agn_regseti(L, -2, ++counter);
      if (isint) {
        idx += step;
      } else {
        x = step;  /* Kahan-Babuska */
        t = s + x;
        c = (fabs(s) >= fabs(x)) ? (s - t) + x : (x - t) + s;
        s = t;
        t = cs + c;
        cc = (fabs(cs) >= fabs(c)) ? (cs - t) + c : (c - t) + cs;
        cs = t;
        ccs += cc;
        idx = s + cs + ccs;
      }
    }
  } else {
    while (idx <= b || tools_approx(idx, b, AGN_HEPSILON)) {
      if (isdefault) {  /* fill with default value  */
        lua_pushvalue(L, -1);
        agn_regseti(L, -3, idx);
      } else
        agn_regsetinumber(L, -1, ++counter, (fabs(idx) < AGN_HEPSILON) ? 0 : idx);
      if (isint) {  /* 2.12.2 */
        idx += step;
      } else {
        x = step;  /* Kahan-Babuska */
        t = s + x;
        c = (fabs(s) >= fabs(x)) ? (s - t) + x : (x - t) + s;
        s = t;
        t = cs + c;
        cc = (fabs(cs) >= fabs(c)) ? (cs - t) + c : (c - t) + cs;
        cs = t;
        ccs += cc;
        idx = s + cs + ccs;
      }
    }
  }
  if (isdefault) agn_poptop(L);
  return 1;
}


/* }====================================================== */

static const luaL_Reg reg_funcs[] = {
  {"extend", reg_extend},                 /* added October 15, 2014 */
  {"new", reg_new},                       /* added May 30, 2022 */
  {"reduce", reg_reduce},                 /* added October 15, 2014 */
  {"settop", reg_settop},                 /* added October 14, 2014 */
  {NULL, NULL}
};


LUALIB_API int luaopen_registers (lua_State *L) {
  luaL_register(L, AGENA_REGLIBNAME, reg_funcs);
  return 1;
}

