/*
** $Id: lseqlib.c,v 1.38 2005/10/23 17:38:15 roberto Exp $
** Library for Sequence Manipulation
** See Copyright Notice in agena.h
*/


#include <stddef.h>

#define lseqlib_c
#define LUA_LIB

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"
#include "lseq.h"


/* creates a sequence with values a, a+step ..., b-step, b; based on calc.fseq
   uses a modified version of Kahan's summation algorithm to avoid roundoff errors
   0.30.0, December 30, 2009; merged with calc.fseq 1.3.3, February 01, 2011

   calc.fseq comments: creates a sequence with values f(a), ..., f(b); June 30, 2008;
   20 % faster than an Agena implementation; extended August 20, 2009, 0.27.0
   extended 0.30.0, December 29, 2009 with Kahan summation algorithm to avoid roundoff errors;
   extended 1.8.8, November 07, 2012, to also return non-numeric sequences; extended 2.4.1,
   January 15, 2015, to process the number of elements to be returned.

   For the idea on the Matlab version please see:
   http://de.mathworks.com/help/matlab/ref/linspace.html
   http://www.che.utah.edu/~sutherland/wiki/index.php/Matlab_Arrays#Linspace_and_Logspace */

static int seq_new (lua_State *L) {
  /* use `volatile` so that the compiler does not render the Kahan code effectless */
  /* volatile lua_Number idx, step, c, y, t; formerly nseq, 2.28.1; created October 15, 2014 */
  int isfunc, isint, isdefault;
  size_t counter, i, j, nargs, total, offset;
  lua_Number a, b, eps;
  volatile lua_Number s, c, cs, ccs, cc, t, idx, step, x;
  luaL_aux_nstructure(L, "sequences.new", &nargs, &offset, &a, &b, &step, &eps, &total, &isfunc, &isdefault, &isint);
  luaL_checkstack(L, nargs, "too many arguments");
  agn_createseq(L, total);
  counter = 0;
  cs = ccs = 0;
  s = idx = a;
  if (isdefault) {  /* create a sequence of n slots and fill it with one and the same default of any type */
    agn_pairgeti(L, 2, 2);
  }
  /* total > counter: prevents that the last element is inserted even if a roundoff error occurred. */
  if (isfunc) {  /* function passed ? */
    while (idx <= b || tools_approx(idx, b, eps)) {
      lua_pushvalue(L, offset);  /* push function */
      lua_pushnumber(L, (fabs(idx) < AGN_HEPSILON) ? 0 : idx);  /* quite dirty hack to avoid roundoff errors with 0 */
      for (i=4 + offset, j = 0; i <= nargs; i++, j++) lua_pushvalue(L, i);
      lua_call(L, 1 + j, 1);
      lua_seqseti(L, -2, ++counter);
      if (isint) {
        idx += step;
      } else {
        x = step;  /* Kahan-Babuska, 2.12.1 */
        t = s + x;
        c = (fabs(s) >= fabs(x)) ? (s - t) + x : (x - t) + s;
        s = t;
        t = cs + c;
        cc = (fabs(cs) >= fabs(c)) ? (cs - t) + c : (c - t) + cs;
        cs = t;
        ccs += cc;
        idx = s + cs + ccs;
      }
    }
  } else {
    while (idx <= b || tools_approx(idx, b, eps)) {
      if (isdefault) {  /* fill with default value  */
        lua_pushvalue(L, -1);
        lua_seqseti(L, -3, idx);
      } else
        agn_seqsetinumber(L, -1, ++counter, (fabs(idx) < AGN_HEPSILON) ? 0 : idx);
      if (isint) {  /* 2.12.2 */
        idx += step;
      } else {
        x = step;  /* Kahan-Babuska */
        t = s + x;
        c = (fabs(s) >= fabs(x)) ? (s - t) + x : (x - t) + s;
        s = t;
        t = cs + c;
        cc = (fabs(cs) >= fabs(c)) ? (cs - t) + c : (c - t) + cs;
        cs = t;
        ccs += cc;
        idx = s + cs + ccs;
      }
    }
  }
  if (isdefault) agn_poptop(L);
  return 1;
}


/* Resizes a sequence to the given number of pre-allocated slots. If you actually shrink a sequence, then it discards any surplus elements.
   The function returns the number of allocated elements and the number of pre-allocated slots, which may be vacant.

   If newsize is 0 or newsize is less than the current size, then the function also purges all surplus values in the sequence.

   If the optional third argument is `true` and newsize is non-zero, then the function sets the optimum number of pre-allocated slots to the
   smallest power of 2 greater than or equal to newsize.

   Note that Agena automatically enlarges and shrinks a sequence if necessary when adding new or purging existing values, see environ.kernel/
   seqautoshrink.

   See also: math.nextpower, size, environ.attrib maxsize and size values. */
static int seq_resize (lua_State *L) {  /* 2.29.0 */
  int newsize, optimal;
  size_t size, a[3];
  luaL_argcheck(L, lua_type(L, 1) == LUA_TSEQ, 1, "argument is not a sequence");
  agn_seqstate(L, 1, a);
  if (lua_gettop(L) == 1) {  /* return optimum number of pre-allocated slots only, but do not resize. */
    agn_seqresize(L, 1, agnSeq_bestsize(a[0]));
  } else {
    newsize = agn_checknonnegint(L, 2);
    optimal = agnL_optboolean(L, 3, 0);
    size = a[0];
    if (optimal) {
      newsize = (newsize != 0) ? agnSeq_bestsize(newsize) : 0;
    }
    agn_seqresize(L, 1, newsize);
    /* agn_seqresize never actually shrinks a sequence to zero to avoid crashes, so now purge the last element if requested. */
    if (size > 0 && newsize == 0) {
      lua_pushnil(L);
      lua_seqseti(L, 1, 1);
    }
  }
  agn_seqstate(L, 1, a);
  lua_pushinteger(L, a[0]);  /* t->size */
  lua_pushinteger(L, a[1]);  /* t->maxsize */
  return 2;
}


static int seq_bestsize (lua_State *L) {  /* 2.29.0, UNDOC */
  int n = agn_checknonnegint(L, 1);
  lua_pushinteger(L, agnSeq_bestsize(n));
  return 1;
}

/* }====================================================== */

static const luaL_Reg seq_funcs[] = {
  {"bestsize", seq_bestsize},     /* added June 25, 2022 */
  {"new", seq_new},               /* added May 30, 2022 */
  {"resize", seq_resize},         /* added June 24, 2022 */
  {NULL, NULL}
};


LUALIB_API int luaopen_sequences (lua_State *L) {
  luaL_register(L, AGENA_SEQLIBNAME, seq_funcs);
  return 1;
}

