/*
** $Id: lsetlib.c,v 1.38 2005/10/23 17:38:15 roberto Exp $
** Library for UltraSet Manipulation
** See Copyright Notice in agena.h
*/

#include <stddef.h>

#define lsetlib_c
#define LUA_LIB

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"
#include "lseq.h"
#include "lset.h"


/* Creates a new set. If the first argument n is a positive integer, it pre-allocates n slots for future insertion, the default is 4. If the last
   argument is `true` the number of pre-allocated slots is adjusted to an optimum of the smallest power of 2 greater than or equal n. 2.33.4 */
static int set_newset (lua_State *L) {
  int total = agnL_optposint(L, 1, 4);
  if (agnL_optboolean(L, 2, 0)) total = tools_nextpower(total, 2, 1);
  agn_createset(L, total);
  return 1;
}


/* Resizes set `s` to store at least `n` elements. If the last argument is `true` the number of pre-allocated slots is adjusted to an optimum of the
   smallest power of 2 greater than or equal n.

   If only `s` is given, the number of pre-allocated slots will be changed to the smallest power of 2 greater than or equal the current size, usually
   freeing formerly occupied space.

   If n < size(s) or the number of pre-allocated slots would not change, the function does nothing and returns without modifying the set. */
static int set_resize (lua_State *L) {  /* 2.33.4 */
  size_t a[3];
  luaL_argcheck(L, lua_type(L, 1) == LUA_TSET, 1, "argument is not a set");
  agn_sstate(L, 1, a);
  if (lua_gettop(L) == 1) {  /* resize to optimum size, but do not drop elements. */
    agn_setresize(L, 1, tools_nextpower(a[0], 2, 1), 1);  /* only shrink if possible, never enlarge */
  } else {
    int newsize = agn_checkposint(L, 2);
    if (agnL_optboolean(L, 3, 0)) newsize = tools_nextpower(newsize, 2, 1);
    agn_setresize(L, 1, newsize, 2);  /* only resize if size < newsize and allocated size != newsize, shrink or enlarge */
  }
  agn_sstate(L, 1, a);
  lua_pushinteger(L, a[0]);  /* us->size */
  lua_pushinteger(L, (a[0] == 1 && a[1] == 0) ? 1 : a[1]);  /* actnodesize; lsizenode is set to 0 even if there is exactly one element in a set */
  return 2;
}


/* }====================================================== */

static const luaL_Reg set_funcs[] = {
  {"newset", set_newset},         /* added October 25, 2022 */
  {"resize", set_resize},         /* added October 25, 2022 */
  {NULL, NULL}
};


LUALIB_API int luaopen_sets (lua_State *L) {
  luaL_register(L, AGENA_SETSLIBNAME, set_funcs);
  return 1;
}

