/*
** $Id: ltablib.c,v 1.38 2005/10/23 17:38:15 roberto Exp $
** Library for Table Manipulation
** See Copyright Notice in agena.h
*/


#include <stddef.h>
#include <math.h>  /* for fmod */

#define ltablib_c
#define LUA_LIB

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"
#include "lobject.h"
#include "lopcodes.h"


#define aux_getn(L,n)   (luaL_checktype(L, n, LUA_TTABLE), luaL_getn(L, n))


static int tbl_maxn (lua_State *L) {
  lua_Number max = 0;
  luaL_checktype(L, 1, LUA_TTABLE);
  lua_pushnil(L);  /* first key */
  while (lua_next(L, 1)) {
    agn_poptop(L);  /* remove value */
    if (lua_type(L, -1) == LUA_TNUMBER) {
      lua_Number v = agn_tonumber(L, -1);  /* Agena 1.4.3/1.5.0 */
      if (v > max) max = v;
    }
  }
  lua_pushnumber(L, max);
  return 1;
}


static int tbl_allocate (lua_State *L) {  /* 22.12.2009 - 0.29.3 */
  int i, nargs;
  luaL_checktype(L, 1, LUA_TTABLE);
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  if (nargs == 1)
    luaL_error(L, "Error in " LUA_QS ": at least three arguments expected.", "tables.allocate");
  if ((nargs & 1) == 0)  /* 2.9.0 */
    luaL_error(L, "Error in " LUA_QS ": odd number of arguments expected.", "tables.allocate");
  lua_settop(L, nargs);
  for (i=3; i <= nargs; i = i + 2) {
    lua_settable(L, 1);
  }
  return 0;
}


/*************************************************************************/
/* functions added for agena 0.5.3                                       */
/*************************************************************************/

/* tbl_indices: returns all keys of a table in a new table */

static int tbl_indices (lua_State *L) {
  int i, nrets;
  i = 0;
  luaL_checktype(L, 1, LUA_TTABLE);
  if (lua_gettop(L) == 1) {
    lua_createtable(L, agn_size(L, 1), 0);
    lua_pushnil(L);
    while (lua_next(L, 1) != 0)
      lua_rawsetikey(L, -3, ++i);
    nrets = 1;
  } else {
    int flag;
    agn_integerindices(L, 1, &flag);
    lua_pushboolean(L, flag);
    nrets = 2;
  }
  return nrets;
}


static int tbl_entries (lua_State *L) {  /* 0.30.1/2.30.1 */
  int flag;
  luaL_checktype(L, 1, LUA_TTABLE);
  agn_entries(L, 1, &flag);
  lua_pushboolean(L, flag);
  return 2;
}


static int tbl_array (lua_State *L) {  /* 2.30.1 */
  luaL_checktype(L, 1, LUA_TTABLE);
  agn_arraypart(L, 1);
  return 1;
}


static int tbl_hash (lua_State *L) {  /* 2.30.1 */
  luaL_checktype(L, 1, LUA_TTABLE);
  agn_hashpart(L, 1);
  return 1;
}


static int tbl_parts (lua_State *L) {  /* 2.30.1 */
  luaL_checktype(L, 1, LUA_TTABLE);
  agn_parts(L, 1);
  return 2;
}


static int tbl_reshuffle (lua_State *L) {  /* 2.30.1 */
  luaL_checktype(L, 1, LUA_TTABLE);
  agn_reorder(L, 1);
  return 0;
}


static int tbl_getsize (lua_State *L) {  /* 2.3.0 RC 4 */
  int option;
  size_t a[3];
  luaL_checktype(L, 1, LUA_TTABLE);
  option = (lua_gettop(L) != 1);
  agn_tablesize(L, 1, a);
  lua_pushnumber(L, a[0]);
  if (option) {
    lua_pushboolean(L, a[1]);
    lua_pushboolean(L, a[2]);
  }
  return 1 + 2*option;
}


/* If any `option` is given, returns the actual number of elements currently stored in the array and hash part. If no option
   is given, then an estimate of the number of elements in the array part is returned, and 0 for the hash part as this
   cannot be estimated.
   Returns two integers: the first for the array part, the second for the hash part.
   See also: `size`, `tables.getsize`. */
static int tbl_getsizes (lua_State *L) {  /* 2.21.4 RC 4 */
  int option;
  luaL_checktype(L, 1, LUA_TTABLE);
  option = (lua_gettop(L) != 1);
  if (option) {
    size_t a[11];
    agn_tablestate(L, 1, a, 1);
    lua_pushnumber(L, a[0]);  /* actual number of elements currently stored in array part */
    lua_pushnumber(L, a[1]);  /* actual number of elements currently stored in hash part */
  } else {
    size_t a[3];
    agn_tablesize(L, 1, a);
    lua_pushnumber(L, a[0]);  /* estimated number of elements currently stored in array part */
    lua_pushnumber(L, 0);     /* (cannot estimate number of elements stored in hash part) */
  }
  return 2;
}


/* Returns the smallest and largest assigned index - in this order - in the array part of a table (default)
   or in the array and hash part if any option has been given. If zeros are returned, the array part or 
   the array and hash part of the table are empty. */
static int tbl_borders (lua_State *L) {  /* 2.14.10, lowest and highest index in array part */
  size_t a[2];
  luaL_checktype(L, 1, LUA_TTABLE);
  if (lua_gettop(L) == 1)  /* only traverse array part ? */
    agn_arrayborders(L, 1, a);  /* if zeros are returned, the array part is empty */
  else
    agn_borders(L, 1, a);  /* traverse both array and hash part, slower; 2.30.1 */
  lua_pushnumber(L, a[0]);
  lua_pushnumber(L, a[1]);
  return 2;
}


/* Inserts values into a table t of tables. If t[key] represents a table, value is added to the end of its array part. If t[key] is
   unassigned, then t[key] := [value]. */
static int tbl_include (lua_State *L) {  /* 2.11.4 */
  int i, nargs;
  luaL_checktype(L, 1, LUA_TTABLE);
  luaL_checkany(L, 3);
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  lua_pushvalue(L, 2);  /* 2.20.3 change */
  lua_rawget(L, 1);     /* dito */
  if (lua_isnil(L, -1)) {
    agn_poptop(L);
    lua_pushvalue(L, 2);  /* push key */
    lua_createtable(L, 1, 1);
    for (i=3; i <= nargs; i++) {
      lua_pushvalue(L, i);  /* push value */
      lua_rawseti(L, -2, i - 2);  /* insert value into new subtable, new subtable at the stack top */
    }
    lua_rawset(L, 1);  /* assign new subtable to table */
  } else if (lua_istable(L, -1)) {
    for (i=3; i <= nargs; i++)
      agn_rawinsertfrom(L, -1, i);
    agn_poptop(L);  /* pop subtable */
  } else
    luaL_error(L, "Error in " LUA_QS ": wrong kind of table.", "tables.include");
  return 0;
}


/* Checks whether a table includes a given field (a string key) amd returns true or false. Metamethods are ignored.
   50 % slower than the expression `(assigned) tbl[field]`. UNDOC, 2.21.0 */
static int tbl_hasfield (lua_State *L) {
  luaL_checktype(L, 1, LUA_TTABLE);
  luaL_checktype(L, 2, LUA_TSTRING);
  lua_pushboolean(L, lua_hasfield(L, 1, agn_tostring(L, 2)));
  return 1;
}


/*
** Operations that an object must define to mimic a table
** (some functions only need some of them)
*/
#define TAB_R  1      /* read */
#define TAB_W  2      /* write */
#define TAB_L  4      /* length */
#define TAB_RW  (TAB_R | TAB_W)    /* read/write */

static int checkfield (lua_State *L, const char *key, int n) {
  lua_pushstring(L, key);
  return (lua_rawget(L, -n) != LUA_TNIL);
}

/*
** Check that 'arg' either is a table or can behave like one (that is,
** has a metatable with the required metamethods)
*/
static void checktab (lua_State *L, int arg, int what) {
  if (lua_type(L, arg) != LUA_TTABLE) {  /* is it not a table? */
    int n = 1;  /* number of elements to pop */
    if (lua_getmetatable(L, arg) &&  /* must have metatable */
        (!(what & TAB_R) || checkfield(L, "__index", ++n)) &&
        (!(what & TAB_W) || checkfield(L, "__writeindex", ++n)) &&
        (!(what & TAB_L) || checkfield(L, "__size", ++n))) {
      lua_pop(L, n);  /* pop metatable and tested metamethods */
    } else
      luaL_checktype(L, arg, LUA_TTABLE);  /* force an error */
  }
}

/* Copies elements from the table a1 to the table a2, performing the equivalent to the following multiple assignment:
   a2[t], = a1[f],,a1[e]. The default for a2 is a1. The destination range can overlap with the source range.

   Returns the destination table a2.

   Example: The following statement copies four elements in table a from position 3 up to and including 6 to a new table b,
   starting with index 1:

   > a := ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'];

   > b := tables.move(a, 3, 6, 1, []);

   > b:
   [c, d, e, f]

   The next statement copies four elements in a to its beginning:

   > tables.move(a, 3, 6, 1);

   > a:
   [c, d, e, f, e, f, g, h]

** Copy elements (1[f], ..., 1[e]) into (tt[t], tt[t+1], ...). Whenever possible, copy in increasing order,
** which is better for rehashing. "possible" means destination after original range, or smaller than origin,
** or copying to another table. 2.21.1; taken from Lua 5.4.0. RC 4/5.4.4. */
static int tbl_move (lua_State *L) {
  lua_Integer f = luaL_checkinteger(L, 2);  /* source start position */
  lua_Integer e = luaL_checkinteger(L, 3);  /* source end position */
  lua_Integer t = luaL_checkinteger(L, 4);  /* destination start position */
  int tt = !lua_isnoneornil(L, 5) ? 5 : 1;  /* index of destination structure */
  checktab(L, 1, TAB_R);
  checktab(L, tt, TAB_W);
  if (e >= f) {  /* otherwise, nothing to move */
    lua_Integer n, i;
    luaL_argcheck(L, f > 0 || e < LUA_MAXINTEGER + f, 3, "too many elements to move");
    n = e - f + 1;  /* number of elements to move */
    luaL_argcheck(L, t <= LUA_MAXINTEGER - n + 1, 4, "destination wrap around");
    /* if APPEND: source end pos < dest start pos ||
          SHIFT LEFT: source start pos >= dest start pos ||
          source structure different from destination structure */
    if (t > e || t <= f || (tt != 1 && !lua_compare(L, 1, tt, LUA_OPEQ))) {
      for (i=0; i < n; i++) {  /* from left to right, optionally using metamethods */
        lua_geti(L, 1, f + i);
        lua_seti(L, tt, t + i);
      }
    } else {  /* SHIFT RIGHT in same structure */
      for (i=n - 1; i >= 0; i--) {  /* from right to left, optionally using metamethods */
        lua_geti(L, 1, f + i);
        lua_seti(L, tt, t + i);
      }
    }
  }
  lua_pushvalue(L, tt);  /* return destination table */
  return 1;
}


static int tbl_new (lua_State *L) {
  /* use `volatile` so that the compiler does not render the Kahan code effectless */
  /* volatile lua_Number idx, step, c, y, t; */
  int isfunc, isint, isdefault;
  size_t counter, i, j, nargs, total, offset;
  lua_Number a, b, eps;
  volatile lua_Number s, c, cs, ccs, cc, t, idx, step, x;
  luaL_aux_nstructure(L, "table.new", &nargs, &offset, &a, &b, &step, &eps, &total, &isfunc, &isdefault, &isint);
  luaL_checkstack(L, nargs, "too many arguments");
  agn_createtable(L, total, 0);
  counter = 0;
  cs = ccs = 0;
  s = idx = a;
  if (isdefault) {  /* create a sequence of n slots and fill it with one and the same default of any type */
    agn_pairgeti(L, 2, 2);
  }
  /* total > counter: prevents that the last element is inserted even if a roundoff error occurred. */
  if (isfunc) {  /* function passed ? */
    while (idx <= b || tools_approx(idx, b, eps)) {
      lua_pushvalue(L, offset);  /* push function */
      lua_pushnumber(L, (fabs(idx) < AGN_HEPSILON) ? 0 : idx);  /* quite dirty hack to avoid roundoff errors with 0 */
      for (i=4 + offset, j = 0; i <= nargs; i++, j++) lua_pushvalue(L, i);
      lua_call(L, 1 + j, 1);
      lua_seti(L, -2, ++counter);
      if (isint) {
        idx += step;
      } else {
        x = step;  /* Kahan-Babuska */
        t = s + x;
        c = (fabs(s) >= fabs(x)) ? (s - t) + x : (x - t) + s;
        s = t;
        t = cs + c;
        cc = (fabs(cs) >= fabs(c)) ? (cs - t) + c : (c - t) + cs;
        cs = t;
        ccs += cc;
        idx = s + cs + ccs;
      }
    }
  } else {
    while (idx <= b || tools_approx(idx, b, eps)) {
      if (isdefault) {  /* fill with default value  */
        lua_pushvalue(L, -1);
        lua_seti(L, -3, idx);
      } else
        lua_rawsetinumber(L, -1, ++counter, (fabs(idx) < AGN_HEPSILON) ? 0 : idx);
      if (isint) {  /* 2.12.2 */
        idx += step;
      } else {
        x = step;  /* Kahan-Babuska */
        t = s + x;
        c = (fabs(s) >= fabs(x)) ? (s - t) + x : (x - t) + s;
        s = t;
        t = cs + c;
        cc = (fabs(cs) >= fabs(c)) ? (cs - t) + c : (c - t) + cs;
        cs = t;
        ccs += cc;
        idx = s + cs + ccs;
      }
    }
  }
  if (isdefault) agn_poptop(L);
  return 1;
}


static int tbl_geti (lua_State *L) {  /* 2.21.1, UNDOC */
  lua_Integer idx = luaL_checkinteger(L, 2);
  luaL_checktype(L, 1, LUA_TTABLE);
  lua_geti(L, 1, idx);
  return 1;
}


static int tbl_getfield (lua_State *L) {  /* 2.21.2, UNDOC */
  luaL_checktype(L, 1, LUA_TTABLE);
  lua_getfield(L, 1, agn_checkstring(L, 2));
  return 1;
}


static int tbl_gettable (lua_State *L) {  /* 2.21.2, UNDOC */
  luaL_checktype(L, 1, LUA_TTABLE);
  lua_pushvalue(L, 2);
  lua_gettable(L, 1);
  return 1;
}


static int tbl_setfield (lua_State *L) {  /* 2.21.2, UNDOC */
  luaL_checktype(L, 1, LUA_TTABLE);
  lua_pushvalue(L, 3);
  lua_setfield(L, 1, agn_checkstring(L, 2));
  return 1;
}


static int tbl_settable (lua_State *L) {  /* 2.21.2, UNDOC */
  luaL_checkany(L, 3);
  luaL_checktype(L, 1, LUA_TTABLE);
  lua_pushvalue(L, 2);  /* key */
  lua_pushvalue(L, 3);  /* value */
  lua_settable(L, 1);
  return 0;
}


/* }====================================================== */


static const luaL_Reg tab_funcs[] = {
  {"allocate", tbl_allocate},  /* added 0.29.2, 22.12.2009 */
  {"array", tbl_array},        /* added 2.30.1, July 30, 2022 */
  {"borders", tbl_borders},    /* added 2.14.10, 22/04/2019 */
  {"entries", tbl_entries},    /* added 0.30.1, 03.01.2010 */
  {"geti", tbl_geti},          /* added 2.21.1, June 12, 2020 */
  {"getfield", tbl_getfield},  /* added 2.21.2, June 12, 2020 */
  {"getsize", tbl_getsize},    /* added 2.3.0 RC 4, November 08, 2014 */
  {"getsizes", tbl_getsizes},  /* added 2.21.4, July 02, 2020 */
  {"gettable", tbl_gettable},  /* added 2.21.2, June 12, 2020 */
  {"hasfield", tbl_hasfield},  /* added 2.21.0, June 04, 2020 */
  {"hash", tbl_hash},          /* added 2.30.1, July 30, 2022 */
  {"include", tbl_include},
  {"indices", tbl_indices},
  {"maxn", tbl_maxn},
  {"move", tbl_move},          /* added 2.21.1, June 12, 2020 */
  {"new", tbl_new},            /* added 2.28.1, May 30, 2022 */
  {"parts", tbl_parts},        /* added 2.30.1, July 30, 2022 */
  {"reshuffle", tbl_reshuffle}, /* added 2.30.1, July 30, 2022 */
  {"setfield", tbl_setfield},  /* added 2.21.2, June 12, 2020 */
  {"settable", tbl_settable},  /* added 2.21.2, June 12, 2020 */
  {NULL, NULL}
};


LUALIB_API int luaopen_table (lua_State *L) {
  luaL_register(L, LUA_TABLIBNAME, tab_funcs);
  return 1;
}


