/*
* lmapm.c
*
* FOR ANSI-VERSIONS ONLY
*
* big-number library for Lua 5.1 based on the MAPM library
* Luiz Henrique de Figueiredo <lhf@tecgraf.puc-rio.br>
* 03 Apr 2009 00:11:32
* This code is hereby placed in the public domain.
*
* MAPM binding for Agena initiated January 17, 2010
*
*/

#include <stdlib.h>
#include "m_apm.h"
#include "m_apm_lc.h"

#define mapm_c
#define LUA_LIB

#include "agena.h"
#include "agnxlib.h"

#include "agenalib.h"

#define AGENA_LIBVERSION	"MAPM 4.9.6 binding, version 1.2.0 as of October 17, 2022\n"


#if !(defined(LUA_DOS) || defined(__OS2__) || defined(__ANSI__))
#define AGENA_MAPMLIBNAME "mapm"
LUALIB_API int (luaopen_mapm) (lua_State *L);
#endif

#ifndef lua_boxpointer
#define lua_boxpointer(L, u) \
	(*(void **)(lua_newuserdata(L,  sizeof(void *))) = (u))
#endif

#define MYNAME		"mapm"
#define MYVERSION	MYNAME " library for " AGENA_VERSION " / Apr 2009 / "\
			"using MAPM " MAPM_LIB_SHORT_VERSION
#define MYTYPE		MYNAME " mapmnumber"


static int DIGITS = 17;
static lua_State *LL = NULL;

void M_apm_log_error_msg (int fatal, char *message) {
#ifdef IGNORE_MAPM_WARNINGS
  if (fatal)
#endif
  luaL_error(LL, "(MAPM) %s.", message);
}


static M_APM Bnew (lua_State *L) {
  M_APM x = m_apm_init();
  lua_boxpointer(L, x);
  luaL_getmetatable(L, MYTYPE);
  lua_setmetatable(L, -2);
  return x;
}


static M_APM Bget (lua_State *L, int i) {
  LL = L;
  switch (lua_type(L, i)) {
    case LUA_TNUMBER: {
      M_APM x = Bnew(L);
      m_apm_set_double(x, lua_tonumber(L, i));  /* Agena 1.4.3/1.5.0, 2.9.4 change */
      lua_replace(L, i);
      return x;
    }
    case LUA_TSTRING: {
      M_APM x = Bnew(L);
      m_apm_set_string(x, (char*)lua_tostring(L, i));
      lua_replace(L, i);
      return x;
    }
    default:
      return *((void**)luaL_checkudata(L, i, MYTYPE));
  }
  return NULL;
}


/* Pushes a MAPM userdata onto the stack and also returns double x converted to M_APM type */
static M_APM Bcreatemapm (lua_State *L, double y) {
  LL = L;
  M_APM x = Bnew(L);
  m_apm_set_double(x, y);
  return x;
}


static int Bdo0 (lua_State *L, void (*f)(M_APM y, M_APM x)) {
  M_APM a, c;
  a = Bget(L, 1);
  c = Bnew(L);
  f(c, a);
  return 1;
}


static int Bdo1 (lua_State *L, void (*f)(M_APM y, int n, M_APM x)) {
  M_APM a, c;
  int n = agnL_optnonnegint(L, 2, DIGITS);
  a = Bget(L, 1);
  c = Bnew(L);
  f(c, n, a);
  return 1;
}


static int Bdo2 (lua_State *L, void (*f)(M_APM z, M_APM x, M_APM y)) {
  M_APM a, b, c;
  a = Bget(L, 1);
  b = Bget(L, 2);
  c = Bnew(L);
  f(c, a, b);
  return 1;
}


static int Bdo3 (lua_State *L, void (*f)(M_APM z, int n, M_APM x, M_APM y)) {
  M_APM a, b, c;
  int n;
  n = agnL_optnonnegint(L, 3, DIGITS);
  a = Bget(L, 1);
  b = Bget(L, 2);
  c = Bnew(L);
  f(c, n, a, b);
  return 1;
}


static int Bdigits (lua_State *L) {  /* digits([n]) */
  DIGITS = agnL_optnonnegint(L, 1, DIGITS);
  lua_pushinteger(L, DIGITS);  /* changed Agena 0.30.3 */
  return 1;
}


static int Btostring (lua_State *L) {  /* tostring(x,[n,exp]) */
  char *s;
  int n;
  M_APM a;
  a = Bget(L, 1);
  n = agnL_optinteger(L, 2, DIGITS);
  if (lua_toboolean(L, 3)) {
    int m = (n < 0) ? m_apm_significant_digits(a) : n;
    s = malloc(m + 16);
    if (s != NULL) m_apm_to_string(s, n, a);
  } else
    s = m_apm_to_fixpt_stringexp(n, a, '.', 0, 0);
  lua_pushstring(L, s);
  if (s != NULL) { xfree(s); }
  return 1;
}


static int Btonumber (lua_State *L) {  /* tonumber(x), extended 2.31.12 */
  int n = agnL_optposint(L, 2, 20);    /* 20 is enough for IEEE doubles */
  lua_settop(L, 1);
  lua_pushinteger(L, n);
  lua_pushboolean(L, 1);
  Btostring(L);
  lua_pushnumber(L, lua_tonumber(L, -1));  /* Agena 1.4.3/1.5.0, patched 2.9.4 */
  return 1;
}


static int Bnumber (lua_State *L) {  /* number(x) */
  Bget(L, 1);
  lua_settop(L, 1);
  return 1;
}


static int Bround (lua_State *L) {   /* round(x) */
  return Bdo1(L, m_apm_round);
}


static int Binv (lua_State *L) {     /* inv(x) */
  return Bdo1(L, m_apm_reciprocal);
}

static int mt_recip (lua_State *L) {  /* recip(x) */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_reciprocal);
}


static int Bsqrt (lua_State *L) {     /* sqrt(x) */
  return Bdo1(L, m_apm_sqrt);
}

static int mt_sqrt (lua_State *L) {   /* sqrt(x) */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_sqrt);
}


static int Bcbrt (lua_State *L) {    /* cbrt(x) */
  return Bdo1(L, m_apm_cbrt);
}


static int Blog (lua_State *L) {     /* ln(x) */
  return Bdo1(L, m_apm_log);
}

static int mt_ln (lua_State *L) {   /* ln(x) */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_log);
}


static int Blog10 (lua_State *L) {   /* log10(x) */
  return Bdo1(L, m_apm_log10);
}


static int Bexp (lua_State *L) {     /* exp(x) */
  return Bdo1(L, m_apm_exp);
}

static int mt_exp (lua_State *L) {   /* exp(x) */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_exp);
}


static int Bsin (lua_State *L) {     /* sin(x) */
  return Bdo1(L, m_apm_sin);
}

static int mt_sin (lua_State *L) {   /* sin(x) */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_sin);
}


static int Bcos (lua_State *L) {     /* cos(x) */
  return Bdo1(L, m_apm_cos);
}

static int mt_cos (lua_State *L) {   /* cos(x) */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_cos);
}


static int Btan (lua_State *L) {     /* tan(x) */
  return Bdo1(L, m_apm_tan);
}

static int mt_tan (lua_State *L) {   /* tan(x) */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_tan);
}


static int Basin (lua_State *L) {    /* asin(x) */
  return Bdo1(L, m_apm_asin);
}

static int mt_arcsin (lua_State *L) {   /* arcsin(x) */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_asin);
}


static int Bacos (lua_State *L) {    /* acos(x) */
  return Bdo1(L, m_apm_acos);
}

static int mt_arccos (lua_State *L) {   /* arccos(x) */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_acos);
}


static int Batan (lua_State *L) {    /* atan(x) */
  return Bdo1(L, m_apm_atan);
}

static int mt_arctan (lua_State *L) {   /* arctan(x) */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_atan);
}


static int Bsincos (lua_State *L) {  /* sincos(x) */
  M_APM a, s, c;
  int n = agnL_optinteger(L, 2, DIGITS);
  a = Bget(L, 1);
  s = Bnew(L);
  c = Bnew(L);
  m_apm_sin_cos(s, c, n, a);
  return 2;
}


static int Batan2 (lua_State *L) {  /* atan2(y, x) */
  return Bdo3(L, m_apm_atan2);
}


static int Bsinh (lua_State *L) {   /* sinh(x) */
  return Bdo1(L, m_apm_sinh);
}

static int mt_sinh (lua_State *L) {   /* sinh(x) */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_sinh);
}


static int Bcosh (lua_State *L) {   /* cosh(x) */
  return Bdo1(L, m_apm_cosh);
}

static int mt_cosh (lua_State *L) {   /* cosh(x) */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_cosh);
}


static int Btanh (lua_State *L) {   /* tanh(x) */
  return Bdo1(L, m_apm_tanh);
}

static int mt_tanh (lua_State *L) {   /* tanh(x) */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_tanh);
}


static int Basinh (lua_State *L) {  /* asinh(x) */
  return Bdo1(L, m_apm_asinh);
}


static int Bacosh (lua_State *L) {  /* acosh(x) */
  return Bdo1(L, m_apm_acosh);
}


static int Batanh (lua_State *L) {  /* atanh(x) */
  return Bdo1(L, m_apm_atanh);
}


static int Babs (lua_State *L) {   /* abs(x) */
  return Bdo0(L, m_apm_absolute_value);
}


static int Bneg (lua_State *L) {   /* neg(x) */
  return Bdo0(L, m_apm_negate);
}


static int Bfactorial (lua_State *L) {  /* factorial(x) */
  return Bdo0(L, m_apm_factorial);
}


static int Bfloor (lua_State *L) {  /* floor(x) */
  return Bdo0(L, m_apm_floor);
}


static int Bceil (lua_State *L) {   /* ceil(x) */
  return Bdo0(L, m_apm_ceil);
}


static int Badd (lua_State *L) {    /* add(x, y) */
  return Bdo2(L, m_apm_add);
}


static int Bsub (lua_State *L) {    /* sub(x, y) */
  return Bdo2(L, m_apm_subtract);
}


static int Bmul (lua_State *L) {    /* mul(x, y) */
  return Bdo2(L, m_apm_multiply);
}


static int Bdiv (lua_State *L) {    /* div(x, y) */
  return Bdo3(L, m_apm_divide);
}


static int Bidiv (lua_State *L) {   /* idiv(x, y) */
  M_APM a, b, p, q;
  a = Bget(L, 1);
  b = Bget(L, 2);
  p = Bnew(L);
  q = Bnew(L);
  m_apm_integer_div_rem(p, q, a, b);
  return 2;
}


static int Bmod (lua_State *L) {    /* __mod(x, y) */
  Bidiv(L);
  return 1;
}


static int Bpow (lua_State *L) {    /* pow(x, y) */
  return Bdo3(L, m_apm_pow);
}


static int mt_ipow (lua_State *L) {   /* x ** n, 2.33.1 */
  M_APM x, r;
  int n;
  x = Bget(L, 1);
  n = agn_checkinteger(L, 2);
  lua_settop(L, 1);
  r = Bnew(L);
  m_apm_integer_pow(r, DIGITS, x, n);
  return 1;
}


static int Bcompare (lua_State *L) {  /* compare(x, y) */
  M_APM a, b;
  a = Bget(L, 1);
  b = Bget(L, 2);
  lua_pushinteger(L, m_apm_compare(a, b));
  return 1;
}


static int Beq (lua_State *L) {
  M_APM a, b;
  a = Bget(L, 1);
  b = Bget(L, 2);
  lua_pushboolean(L, m_apm_compare(a, b) == 0);
  return 1;
}


static int Blt (lua_State *L) {
  M_APM a, b;
  a = Bget(L, 1);
  b = Bget(L, 2);
  lua_pushboolean(L, m_apm_compare(a, b) < 0);
  return 1;
}


static int Bsign (lua_State *L) {       /* sign(x) */
  M_APM a = Bget(L, 1);
  lua_pushinteger(L, m_apm_sign(a));
  return 1;
}

static int mt_sign (lua_State *L) {   /* sign mt */
  M_APM a = Bget(L, 1);
  lua_settop(L, 1);
  Bcreatemapm(L, (double)m_apm_sign(a));
  return 1;
}


static int mt_abs (lua_State *L) {   /* abs mt */
  lua_settop(L, 1);
  return Bdo0(L, m_apm_absolute_value);
}


static int Bexponent (lua_State *L) {   /* exponent(x) */
  M_APM a = Bget(L, 1);
  lua_pushinteger(L, m_apm_exponent(a));
  return 1;
}


static int Bisint (lua_State *L) {      /* isint(x) */
  M_APM a = Bget(L, 1);
  lua_pushboolean(L, m_apm_is_integer(a));
  return 1;
}


static int Biseven (lua_State *L) {     /* iseven(x) */
  M_APM a = Bget(L, 1);
  lua_pushboolean(L, m_apm_is_integer(a) && m_apm_is_even(a));
  return 1;
}


static int Bisodd (lua_State *L) {      /* isodd(x) */
  M_APM a = Bget(L, 1);
  lua_pushboolean(L, m_apm_is_integer(a) && m_apm_is_odd(a));
  return 1;
}


static int Bdigitsin (lua_State *L) {   /* digitsin(x) */
  M_APM a = Bget(L, 1);
  lua_pushinteger(L, m_apm_significant_digits(a));
  return 1;
}


static int Bgc (lua_State *L) {
  M_APM a = Bget(L, 1);
  m_apm_free(a);
  lua_pushnil(L);
  lua_setmetatable(L, 1);
  m_apm_trim_mem_usage();  /* 2.9.6 */
  return 0;
}


/* Computes the n-th Chebyshev polynomial of the first kind, evaluated at x, with n a non-negative integer and x a number. 2.33.1
   42 percent faster than the Agena implementation which returns cos(n*arccos(x)):
   mapm.xchebyt := proc(n, x) is
     local t1, t4, t6, t9, t15, t17, t20
     if n :: nonnegint then
        n := mapm.xnumber(n)
     fi;
     if x :: number then
        x := mapm.xnumber(x)
     fi
     t1  := x*x;
     t4  := sqrt(t1 + mapm.one + mapm.two*x);
     t6  := sqrt(t1 + mapm.one - mapm.two*x);
     t9  := n*arccos(mapm.half*(t4 - t6));
     t15 := (mapm.half*(t4 + t6))^mapm.two;
     t17 := sqrt(t15 - mapm.one);
     t20 := n*sign(x)*ln(mapm.half*(t4 + t6) + t17);
     return cos(t9)*cosh(t20)
  end; */
static void m_apm_chebyt (M_APM rr, int places, M_APM n, M_APM x) {  /* 2.33.1, result is in n */
  M_APM a, b, c, d, t4, t6, t9, t15, t17;
  a   = M_get_stack_var();
  b   = M_get_stack_var();
  c   = M_get_stack_var();
  d   = M_get_stack_var();
  t4  = M_get_stack_var();
  t6  = M_get_stack_var();
  t9  = M_get_stack_var();
  t15 = M_get_stack_var();
  t17 = M_get_stack_var();
  /* t1(=a) = x*x; */
  m_apm_multiply(a, x, x);
  /* t4 = sqrt(t1 + 1.0 + 2.0*x);
     t6 = sqrt(t1 + 1.0 - 2.0*x); */
  m_apm_add(c, a, MM_One);
  m_apm_multiply(d, x, MM_Two);
  m_apm_add(a, c, d);
  m_apm_subtract(b, c, d);
  m_apm_sqrt(t4, places, a);
  m_apm_sqrt(t6, places, b);
  /* t9 = n*sun_acos(0.5*(t4 - t6)); */
  m_apm_subtract(a, t4, t6);
  m_apm_divide(b, places, a, MM_Two);
  m_apm_arccos(c, places, b);
  m_apm_multiply(t9, n, c);
  /* t15 = sun_pow(0.5*(t4 + t6), 2.0, 1); */
  m_apm_add(a, t4, t6);
  m_apm_divide(b, places, a, MM_Two);
  m_apm_multiply(t15, b, b);
  /* t17 = sqrt(t15 - 1.0); */
  m_apm_subtract(a, t15, MM_One);
  m_apm_sqrt(t17, places, a);
  /* t20 = n*tools_sign(a)*sun_log(0.5*(t4 + t6) + t17); */
  m_apm_set_double(a, (double)m_apm_sign(x));
  m_apm_multiply(b, n, a);
  m_apm_add(c, t4, t6);
  m_apm_divide(d, places, c, MM_Two);
  m_apm_add(d, d, t17);
  m_apm_log(d, places, d);
  m_apm_multiply(rr, b, d);
  /* sun_cos(t9)*sun_cosh(t20=rr) */
  m_apm_cos(a, places, t9);
  m_apm_cosh(b, places, rr);
  m_apm_multiply(rr, a, b);  /* result is in rr */
  M_restore_stack(9);  /* restore the 9 locals we used here */
}

static int Bchebyt (lua_State *L) {
  return Bdo3(L, m_apm_chebyt);
}


static void m_apm_hypot (M_APM rr, int places, M_APM x, M_APM y) {  /* 2.33.1 */
  M_APM a, b, c;
  int olddigits = DIGITS;
  DIGITS += 3;
  a = m_apm_init();  /* don't use M_get_stack_var or Agena will crash */
  b = m_apm_init();
  c = m_apm_init();
  m_apm_multiply(a, x, x);
  m_apm_multiply(b, y, y);
  m_apm_add(c, a, b);
  m_apm_sqrt(rr, places + 3, c);
  DIGITS = olddigits;
  m_apm_free(a); m_apm_free(b); m_apm_free(c);
}

static int Bhypot (lua_State *L) {
  return Bdo3(L, m_apm_hypot);
}


static void m_apm_square (M_APM rr, int places, M_APM x) {  /* 2.33.1 */
  int olddigits = DIGITS;
  DIGITS += 3;
  m_apm_multiply(rr, x, x);
  DIGITS = olddigits;
}

static int mt_square (lua_State *L) {  /* square mt, 2.33.1 */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_square);
}


static void m_apm_cube (M_APM rr, int places, M_APM x) {  /* 2.33.1 */
  m_apm_integer_pow(rr, places + 3, x, 3);
}

static int mt_cube (lua_State *L) {  /* cube mt, 2.33.1 */
  lua_settop(L, 1);
  return Bdo1(L, m_apm_cube);
}


static int mapm_xfma (lua_State *L) {  /* x*y + z, 2.33.1 */
  M_APM a, x, y, z, r;
  a = m_apm_init();  /* don't use M_get_stack_var or Agena will crash */
  x = Bget(L, 1);
  y = Bget(L, 2);
  z = Bget(L, 3);
  r = Bnew(L);
  int olddigits = DIGITS;
  DIGITS += 3;
  m_apm_multiply(a, x, y);
  m_apm_add(r, a, z);
  DIGITS = olddigits;
  m_apm_free(a);
  return 1;
}


static int mapm_xterm (lua_State *L) {  /* c*x**n, 2.33.1 */
  M_APM r, a, c, x, n;
  int olddigits = DIGITS;
  a = m_apm_init();  /* don't use M_get_stack_var or Agena will crash */
  c = Bget(L, 1);
  x = Bget(L, 2);
  n = Bget(L, 3);
  r = Bnew(L);
  DIGITS += 3;
  m_apm_pow(a, DIGITS, x, n);
  m_apm_multiply(r, c, a);
  DIGITS = olddigits;
  m_apm_free(a);
  return 1;
}


static int mapm_xlog2 (lua_State *L) {  /* 2.33.2 */
  int places;
  M_APM r, a, b, x;
  a = m_apm_init();  /* don't use M_get_stack_var or Agena will crash */
  b = m_apm_init();
  x = Bget(L, 1);
  places = agnL_optinteger(L, 2, DIGITS);
  r = Bnew(L);
  m_apm_log(a, places, x);
  m_apm_log(b, places, MM_Two);
  m_apm_divide(r, places, a, b);
  m_apm_free(a); m_apm_free(b);
  return 1;
}


static int mapm_xlog (lua_State *L) {  /* 2.33.2 */
  int places;
  M_APM r, a, b, x, n;
  a = m_apm_init();  /* don't use M_get_stack_var or Agena will crash */
  b = m_apm_init();
  x = Bget(L, 1);
  n = Bget(L, 2);
  places = agnL_optinteger(L, 3, DIGITS);
  r = Bnew(L);
  m_apm_log(a, places, x);
  m_apm_log(b, places, n);
  m_apm_divide(r, places, a, b);
  m_apm_free(a); m_apm_free(b);
  return 1;
}


static int mapm_xexp2 (lua_State *L) {  /* 2.33.2 */
  int places;
  M_APM r, x;
  x = Bget(L, 1);
  places = agnL_optinteger(L, 2, DIGITS);
  r = Bnew(L);
  m_apm_pow(r, places, MM_Two, x);
  return 1;
}


static int mapm_xexp10 (lua_State *L) {  /* 2.33.2 */
  int places;
  M_APM r, x;
  x = Bget(L, 1);
  places = agnL_optinteger(L, 2, DIGITS);
  r = Bnew(L);
  m_apm_pow(r, places, MM_Ten, x);
  return 1;
}


static const struct luaL_Reg mapm_lib [] = {  /* metamethods for userdata `n` */
  {"__abs", mt_abs},
  {"__sign", mt_sign},
  {"__add", Badd},  /* __add(x, y) */
  {"__div", Bdiv},  /* __div(x, y) */
  {"__eq",	Beq},   /* __eq(x, y) */
  {"__gc",	Bgc},
  {"__lt",	Blt},   /* __lt(x, y) */
  {"__mod", Bmod},  /* __mod(x, y) */
  {"__mul", Bmul},  /* __mul(x, y) */
  {"__pow", Bpow},  /* __pow(x, y) */
  {"__ipow", mt_ipow},  /* __ipow(x, y) */
  {"__square", mt_square},
  {"__cube", mt_cube},
  {"__sub", Bsub},  /* __sub(x, y) */
  {"__recip", mt_recip},
  {"__sqrt", mt_sqrt},
  {"__ln", mt_ln},
  {"__exp", mt_exp},
  {"__sin", mt_sin},
  {"__cos", mt_cos},
  {"__tan", mt_tan},
  {"__arcsin", mt_arcsin},
  {"__arccos", mt_arccos},
  {"__arctan", mt_arctan},
  {"__sinh", mt_sinh},
  {"__cosh", mt_cosh},
  {"__tanh", mt_tanh},
  {"__tostring", Btostring},  /* __tostring(x) */
  {"__unm", Bneg},  /* __unm(x) */
  {NULL, NULL}
};


static const luaL_Reg mapmlib[] = {
  {"xabs", Babs},
  {"xarccos", Bacos},
  {"xarccosh", Bacosh},
  {"xadd", Badd},
  {"xarcsin", Basin},
  {"xarcsinh", Basinh},
  {"xarctan", Batan},
  {"xarctan2", Batan2},
  {"xarctanh", Batanh},
  {"xcbrt", Bcbrt},
  {"xceil", Bceil},
  {"xchebyt", Bchebyt},
  {"xcompare", Bcompare},
  {"xcos", Bcos},
  {"xcosh", Bcosh},
  {"xdigits", Bdigits},
  {"xdigitsin", Bdigitsin},
  {"xdiv", Bdiv},
  {"xexp", Bexp},
  {"xexp2", mapm_xexp2},
  {"xexp10", mapm_xexp10},
  {"xexponent", Bexponent},
  {"xfactorial", Bfactorial},
  {"xfloor", Bfloor},
  {"xfma", mapm_xfma},
  {"xhypot", Bhypot},
  {"xidiv", Bidiv},
  {"xinv", Binv},
  {"xiseven", Biseven},
  {"xisint", Bisint},
  {"xisodd", Bisodd},
  {"xln", Blog},  /* changed 0.30.3 */
  {"xlog", mapm_xlog},
  {"xlog2", mapm_xlog2},
  {"xlog10", Blog10},
  {"xmod", Bmod},
  {"xmul", Bmul},
  {"xneg", Bneg},
  {"xnumber", Bnumber},
  {"xpow",	Bpow},
  {"xround", Bround},
  {"xsign", Bsign},
  {"xsin", Bsin},
  {"xsincos", Bsincos},
  {"xsinh", Bsinh},
  {"xsqrt", Bsqrt},
  {"xsub", Bsub},
  {"xtan", Btan},
  {"xtanh", Btanh},
  {"xterm", mapm_xterm},
  {"xtonumber", Btonumber},
  {"xtostring", Btostring},
  {NULL, NULL}
};


static void clearmapm (void) {  /* 2.9.6 */
  m_apm_free_all_mem();
}


LUALIB_API int luaopen_mapm (lua_State *L) {
  atexit(clearmapm);
  luaL_newmetatable(L, MYTYPE);         /* metatable for long userdata, adds it to the registry with key 'long' */
  luaL_register(L, NULL, mapm_lib);     /* assign C metamethods to this metatable */
  lua_setglobal(L, MYNAME);
  luaL_register(L, AGENA_MAPMLIBNAME, mapmlib);
  lua_rawsetstringstring(L, -1, "initstring", AGENA_LIBVERSION);
  lua_pushliteral(L, "version");
  lua_pushliteral(L, MYVERSION);
  lua_settable(L, -3);
  lua_pushliteral(L, "__index");
  lua_pushvalue(L, -2);
  lua_settable(L, -3);
  return 1;
}

