/*
** $Id: numarray.c,v 0.1 12.10.2015 $
** numarray - Numarray Library
** See Copyright Notice in agena.h
*/

/* CHANGE VERSION NUMBER in numarray.agn ! */

/* The numarray package implements arrays of the C data types of either double, unsigned char, and signed 4-byte integers.

The arrays implemented by this package are called numarrays for short.

Since numbers stored to numarrays consume less space, numarrays may be useful if a large amount of numbers have to be processed, but the amount of random-access memory of your system is limited. Operations on numarrays, however, are usually
slower than those on Agena's native structures: tables, pairs, sequences, or registers - so you trade speed for memory.

Internally, numarrays are userdata structures that also support various metamethods.

You can create numarrays, assign and read numbers, resize arrays, store them to binary files, and read from files.
Functions to convert arrays to Agena's native structures, and vice versa, are provided, as well.

To create an array of unsigned chars, use numarray.uchar, of doubles use numarray.double, and of integers use numarray.integer. The number of entries to be stored must be given when calling these three procedures. When
creating arrays, all slots are automatically filled with zeros. All array indices start with number 1.

Agena's standard indexing functions save and read numbers. So, for example, a[1] := -1 stores the number -1
to index 1 of the array a. a[1] reads the value stored at index 1 of the array a. Alternatively, numarray.put
and numarray.get save and read numbers, respectively. Furthermore, numarray.include bulk-assigns numbers.

Arrays can be shrunk or extended with the numarray.resize function.

Arrays can be converted to sequences and registers with numarray.toseq and numarray.toreg. numarray.toarray
creates arrays from sequences and registers.

numarray.whereis searches for numbers, and numarray.iterate can sequentially traverse arrays.

numarray.write writes the contents of any array to a binary file. numarray.readuchars reads a complete file of unsigned
chars, numarray.readdoubles of doubles, and numarray.readintegers of signed integers with only one call. To open
and close these files, use binio.open and binio.close. Most other binio function, such as binio.sync binio.rewind,
and binio.filepos, is supported, as well, with the exception of the binio.read* procedures. The low-level numarray.read
function is used by the above mentioned numarray.read* functions.

The following metamethods exist: standard read and write indexing (see above), in operator, strict and approximate
equality (=, ==, <>, ~=, ~<> operators), size and `tostring`. To easily add further metamethods, have a look at the end
of the lib/numarray.agn source file.

*/


#define numarray_c
#define LUA_LIB

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <unistd.h>

#include "agena.h"
#include "agnxlib.h"
#include "agenalib.h"
#include "lstate.h"
#include "agnhlps.h"

/* for numarray.used: */
#include "lstring.h"
#include "lapi.h"
#include "numarray.h"
#include "long.h"

#if !(defined(LUA_DOS) || defined(__OS2__) || defined(LUA_ANSI))
#define AGENA_NUMARRAYLIBNAME "numarray"
LUALIB_API int (luaopen_numarray) (lua_State *L);
#endif

/* numarray.[uchar|double|integer](n)

   Creates a userdata array of unsigned chars or (signed) doubles with the given number of entries n, with n an integer,
   and with each slot set to the number 0. Initially, the number of elements can be zero or more, use numarray.resize to
   extend the array before assigning values.

   Due to advice by Luiz Henrique de Figueiredo the implementation uses explicit malloc's of the arrays, since
   resizing userdata via lmem functionality seems to be tricky:
   http://stackoverflow.com/questions/29806478/reallocating-resizing-lua-5-1-userdata-in-c */

/* leaves a numarray on top of the stack */
#define createnumarrayfileio(t,ut,what,n) { \
  a = (NumArray *)lua_newuserdata(L, sizeof(NumArray)); \
  a->size = (n); \
  a->datatype = (what); \
  if (what == 1) { \
    a->data.n = malloc((n) * sizeof(t)); \
    if (a->data.n == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else if (what == 2) { \
    a->data.i = malloc((n) * sizeof(t)); \
    if (a->data.i == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else if (what == 3) { \
    a->data.i = malloc((n) * sizeof(t)); \
    if (a->data.us == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else if (what == 4) { \
    a->data.ui = malloc((n) * sizeof(t)); \
    if (a->data.ui == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else if (what == 5)  { \
    a->data.ld = malloc((n) * sizeof(t)); \
    if (a->data.ld == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else { \
    a->data.c = malloc((n) * sizeof(t)); \
    if (a->data.c == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } \
  lua_setmetatabletoobject(L, -1, "numarray", 0); \
  agn_setutypestring(L, -1, ut); \
  lua_createtable(L, 0, 0); \
  a->registry = luaL_ref(L, LUA_REGISTRYINDEX); \
}

/* leaves a numarray on top of the stack */
#define buildfn(t,ut,what) { \
  int64_t n; \
  NumArray *a; \
  n = luaL_checkint(L, 1); \
  if (n > MAX_INT) \
    luaL_error(L, "Error in " LUA_QS ": array too large, max. of 2147483647 elements accepted.", ut); \
  luaL_argcheck(L, n >= 0, 1, "invalid size"); \
  a = (NumArray *)lua_newuserdata(L, sizeof(NumArray)); \
  a->size = n; \
  a->datatype = what; \
  if (what == 1) { \
    a->data.n = calloc(n, sizeof(t)); \
    if (a->data.n == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else if (what == 2) { \
    a->data.i = calloc(n, sizeof(t)); \
    if (a->data.i == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else if (what == 3) { \
    a->data.us = calloc(n, sizeof(t)); \
    if (a->data.us == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else if (what == 4) { \
    a->data.ui = calloc(n, sizeof(t)); \
    if (a->data.ui == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else if (what == 5) { \
    a->data.ld = calloc(n, sizeof(t)); \
    if (a->data.ld == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else { \
    a->data.c = calloc(n, sizeof(t)); \
    if (a->data.c == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } \
  lua_setmetatabletoobject(L, -1, "numarray", 0); \
  agn_setutypestring(L, -1, ut); \
  lua_createtable(L, 0, 0); \
  a->registry = luaL_ref(L, LUA_REGISTRYINDEX); \
}

#define createarray(L, a, what, n) { \
  a = (NumArray *)lua_newuserdata(L, sizeof(NumArray)); \
  a->size = (n); \
  a->datatype = what; \
  if (what == 1) { \
    a->data.n = calloc(n, sizeof(lua_Number)); \
    if (a->data.n == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else if (what == 2) { \
    a->data.i = calloc(n, sizeof(int32_t)); \
    if (a->data.i == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else if (what == 3) { \
    a->data.us = calloc(n, sizeof(uint16_t)); \
    if (a->data.us == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else if (what == 4) { \
    a->data.ui = calloc(n, sizeof(uint32_t)); \
    if (a->data.ui == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else if (what == 5) { \
    a->data.ld = calloc(n, SIZEOFLDBL); \
    if (a->data.ld == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } else { \
    a->data.c = calloc(n, sizeof(unsigned char)); \
    if (a->data.c == NULL) \
      luaL_error(L, "Error: memory allocation failed."); \
  } \
  lua_setmetatabletoobject(L, -1, "numarray", 0); \
  switch (what) { \
    case 0: agn_setutypestring(L, -1, "uchar"); break; \
    case 1: agn_setutypestring(L, -1, "double"); break; \
    case 2: agn_setutypestring(L, -1, "int32"); break; \
    case 3: agn_setutypestring(L, -1, "ushort"); break; \
    case 4: agn_setutypestring(L, -1, "uint32"); break; \
    case 5: agn_setutypestring(L, -1, "longdouble"); break; \
  } \
  lua_createtable(L, 0, 0); \
  a->registry = luaL_ref(L, LUA_REGISTRYINDEX); \
}


static void numarray_checkoptions (lua_State *L, int *nargs, int *newstruct, int *equality, const char *procname) {
  int checkoptions;
  *newstruct = 1;  /* returns a new structure, do not work in-place */
  *equality = 1;   /* 1 = strict equality, 0 = approximate equality */
  /* check for options, here `map in-place` */
  if (*nargs > 2 && lua_isboolean(L, *nargs) && agn_istrue(L, *nargs)) {
    (*nargs)--; *newstruct = 0;
  }
  checkoptions = 2;  /* check n options, 2.29.0 */
  while (checkoptions-- && *nargs > 2 && lua_ispair(L, *nargs)) {
    agn_pairgeti(L, *nargs, 1);  /* get left value, set to stack index -2 */
    agn_pairgeti(L, *nargs, 2);  /* get right value, set to stack index  -1 */
    if (agn_isstring(L, -2)) {
      const char *option = agn_tostring(L, -2);
      if (tools_streq("inplace", option)) {
        *newstruct = !agn_checkboolean(L, -1);
      } else if (tools_streq("strict", option)) {
        *equality = agn_checkboolean(L, -1);
      } else {
        agn_poptoptwo(L);
        luaL_error(L, "Error in " LUA_QS ": unknown option " LUA_QS ".", procname, option);
      }
    }
    /* do not call lua_settop as it would corrupt the argument stack since we have already pushed values */
    (*nargs)--;
    agn_poptoptwo(L);
  }
}


static int numarray_uchar (lua_State *L) {
  buildfn(unsigned char, "uchar", 0);
  return 1;
}


static int numarray_double (lua_State *L) {
  buildfn(lua_Number, "double", 1);
  return 1;
}


static int numarray_int32 (lua_State *L) {  /* name changed 2.12.1 due to new integer op */
  buildfn(int32_t, "int32", 2);
  return 1;
}


static int numarray_ushort (lua_State *L) {  /* 2.18.2 */
  buildfn(uint16_t, "ushort", 3);
  return 1;
}


static int numarray_uint32 (lua_State *L) {  /* 2.22.1 */
  buildfn(uint32_t, "uint32", 4);
  return 1;
}


static int numarray_longdouble (lua_State *L) {  /* 2.35.0 */
  buildfn(long double, "longdouble", 5);
  return 1;
}


/* numarray.setitem (a, i, v)

   With a any numarray, sets number v to a[i], where i, the index, is an integer counting from 1. */

static int numarray_setitem (lua_State *L) {  /* set a number to the given slot. */
  NumArray *a = checknumarray(L, 1);
  union ldshape d;
  int64_t index = tools_posrelat(agn_checkinteger(L, 2), a->size) - 1;  /* realign Lua/Agena index to C index, starting from 0, extended 2.15.1 */
  if (index < 0 || index >= a->size) {  /* 2.16.7 fix */
    if (index == -1) {
      lua_settop(L, 3);
      lua_rawseti(L, LUA_REGISTRYINDEX, a->registry);
      return 0;
    } else
      luaL_error(L, "Error in " LUA_QS ": index %d out of range.", "numarray.setitem", index + 1);
  }
  getdlongnum(L, d, 3, "numarray.setitem");
  switch (a->datatype) {
    case 0: a->data.c[index]  = d.d; break;
    case 1: a->data.n[index]  = d.d; break;
    case 2: a->data.i[index]  = d.d; break;
    case 3: a->data.us[index] = d.d; break;  /* 2.18.2 */
    case 4: a->data.ui[index] = d.d; break;  /* 2.22.1 */
    case 5: a->data.ld[index] = tolong(L, 3, d); break;  /* 2.35.0 */
    default: lua_assert(0);
  }
  return 0;
}


/* numarray.getitem (a, i)

  With a any numarray, returns the value stored at a[i], where i, the index, is an integer counting from 1.

  2.15.2, returns succeeding values from a numeric array. It is at least twice as fast as individually reading the items. */
static int numarray_getitem (lua_State *L) {
  int64_t i, n, idx;
  NumArray *a = checknumarray(L, 1);
  idx = tools_posrelat(agn_checkinteger(L, 2), a->size) - 1;  /* realign Lua/Agena index to C index, starting from 0, extended 2.15.1 */
  n = agnL_optposint(L, 3, 1);
  luaL_checkstack(L, n, "too many values");  /* 2.31.7 fix */
  if (idx < 0 || idx + n - 1 >= a->size) {
    if (idx == -1) {  /* 2.15.1 */
      lua_rawgeti(L, LUA_REGISTRYINDEX, a->registry);
      return 1;
    }
    luaL_error(L, "Error in " LUA_QS ": index %d with given %d items out of range.", "numarray.getitem", (int)idx + 1, (int)n);  /* 2.17.9 fix */
  }
  switch (a->datatype) {
    case 0:
      for (i=0; i < n; i++) {
        lua_pushnumber(L, a->data.c[idx + i]);
      }
      break;
    case 1:
      for (i=0; i < n; i++) {
        lua_pushnumber(L, a->data.n[idx + i]);
      }
      break;
    case 2:
      for (i=0; i < n; i++) {
        lua_pushnumber(L, a->data.i[idx + i]);
      }
      break;
    case 3:  /* 2.18.2 */
      for (i=0; i < n; i++) {
        lua_pushnumber(L, a->data.us[idx + i]);
      }
      break;
    case 4:  /* 2.22.1 */
      for (i=0; i < n; i++) {
        lua_pushnumber(L, a->data.ui[idx + i]);
      }
      break;
    case 5:  /* 2.35.0 */
      for (i=0; i < n; i++) {
        createdlong(L, a->data.ld[idx + i]);
      }
      break;
    default: lua_assert(0);
  }
  return n;
}


static int numarray_subarray (lua_State *L) {  /* 2.17.9 */
  size_t nargs;
  int64_t i, n, start, start0, stop;
  NumArray *b = checknumarray(L, 1);
  nargs = lua_gettop(L);
  start0 = agn_checkinteger(L, 2);  /* 2.19.1 fix: avoid returning registry table with index < 0 and -index >= size */
  start = tools_posrelat(start0, b->size);  /* realign Lua/Agena index to C index, starting from 0, extended 2.15.1 */
  stop = tools_posrelat(agnL_optinteger(L, 3, start), b->size);
  n = stop - start + 1;
  start--;
  luaL_checkstack(L, n, "too many value");  /* 2.31.7 fix */
  if (nargs == 2) {  /* simple indexing a[idx] */
    if (start < 0 || start + n - 1 >= b->size) {
      if (start0 == 0) {  /* 2.15.1; 2.19.1 fix */
        lua_rawgeti(L, LUA_REGISTRYINDEX, b->registry);
        return 1;
      }
      luaL_error(L, "Error in " LUA_QS ": index %d out of range.", "numarray.getitem", (int)start0);
    }
    switch (b->datatype) {
      case 0:
        for (i=0; i < n; i++) {
          lua_pushnumber(L, b->data.c[start + i]);
        }
        break;
      case 1:
        for (i=0; i < n; i++) {
          lua_pushnumber(L, b->data.n[start + i]);
        }
        break;
      case 2:
        for (i=0; i < n; i++) {
          lua_pushnumber(L, b->data.i[start + i]);
        }
        break;
      case 3:  /* 2.18.2 */
        for (i=0; i < n; i++) {
          lua_pushnumber(L, b->data.us[start + i]);
        }
        break;
      case 4:  /* 2.22.1 */
        for (i=0; i < n; i++) {
          lua_pushnumber(L, b->data.ui[start + i]);
        }
        break;
      case 5:  /* 2.35.0 */
        for (i=0; i < n; i++) {
          createdlong(L, b->data.ld[start + i]);
        }
        break;
      default: lua_assert(0);
    }
    return n;
  } else {  /* a[start to stop] indexing */
    NumArray *a;
    char type;  /* 2.31.3 fix */
    if (stop == 0 || start < 0 || start + 1 > stop || start + n - 1 >= b->size) {
      luaL_error(L, "Error in " LUA_QS ": indices %d to %d out of range.", "numarray.subarray", (int)start + 1, (int)stop);
    }
    type = (char)b->datatype;
    createarray(L, a, type, n);
    switch (type) {
      case 0:
        for (i=0; i < n; i++) {
          a->data.c[i] = b->data.c[start + i];
        }
        break;
      case 1:
        for (i=0; i < n; i++) {
          a->data.n[i] = b->data.n[start + i];
        }
        break;
      case 2:
        for (i=0; i < n; i++) {
          a->data.i[i] = b->data.i[start + i];
        }
        break;
      case 3:  /* 2.18.2 */
        for (i=0; i < n; i++) {
          a->data.us[i] = b->data.us[start + i];
        }
        break;
      case 4:  /* 2.22.1 */
        for (i=0; i < n; i++) {
          a->data.ui[i] = b->data.ui[start + i];
        }
        break;
      case 5:  /* 2.35.0 */
        for (i=0; i < n; i++) {
          a->data.ld[i] = b->data.ld[start + i];
        }
        break;
      default: lua_assert(0);
    }
    return 1;
  }
}


static INLINE void aux_copy (NumArray *to, NumArray *from) {
  int64_t i, n;
  n = from->size;
  switch (from->datatype) {
    case 0:
      for (i=0; i < n; i++) {
        to->data.c[i] = from->data.c[i];
      }
      break;
    case 1:
      for (i=0; i < n; i++) {
        to->data.n[i] = from->data.n[i];
      }
      break;
    case 2:
      for (i=0; i < n; i++) {
        to->data.i[i] = from->data.i[i];
      }
      break;
    case 3:
      for (i=0; i < n; i++) {
        to->data.us[i] = from->data.us[i];
      }
      break;
    case 4:
      for (i=0; i < n; i++) {
        to->data.ui[i] = from->data.ui[i];
      }
      break;
    case 5:
      for (i=0; i < n; i++) {
        to->data.ld[i] = from->data.ld[i];
      }
      break;
    default: lua_assert(0);
  }
}

static int numarray_replicate (lua_State *L) {  /* 2.31.3, based on numarray.subarray */
  NumArray *a = NULL;
  NumArray *b = checknumarray(L, 1);
  int64_t n = b->size;
  char type = (char)b->datatype;
  createarray(L, a, type, n);
  aux_copy(a, b);
  return 1;
}


static INLINE void aux_introsort (lua_State *L, NumArray *a) {
  switch (a->datatype) {
    case 0: tools_dintrosort_uchar(a->data.c, 0, a->size - 1, 0, 2*sun_log2(a->size)); break;
    case 1: tools_dintrosort(a->data.n, 0, a->size - 1, 0, 2*sun_log2(a->size)); break;
    case 2: tools_dintrosort_int32(a->data.i, 0, a->size - 1, 0, 2*sun_log2(a->size)); break;
    case 3: tools_dintrosort_ushort(a->data.us, 0, a->size - 1, 0, 2*sun_log2(a->size)); break;
    case 4: tools_dintrosort_uint32(a->data.ui, 0, a->size - 1, 0, 2*sun_log2(a->size)); break;
    case 5: tools_dintrosortl(a->data.ld, 0, a->size - 1, 0, 2*sun_log2(a->size)); break;
    default:
      luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.sort(ed)");
  }
}

/* Sorts a numarray of doubles in ascending order, in-place. The function returns nothing. 2.17.9 */
static int numarray_sort (lua_State *L) {
  NumArray *a = checknumarray(L, 1);
  aux_introsort(L, a);
  return 0;
}

static int numarray_sorted (lua_State *L) {  /* 2.31.8, based on numarray.replicate and numarray.sort */
  NumArray *a = NULL;
  NumArray *b = checknumarray(L, 1);
  int64_t n = b->size;
  char type = (char)b->datatype;
  createarray(L, a, type, n);
  aux_copy(a, b);
  aux_introsort(L, a);
  return 1;
}


#define NBITS (int64_t)( sizeof(unsigned char) * 8)  /* ! we need the cast to int64_t ! */

/* Source: https://stackoverflow.com/questions/2525310/how-to-define-and-work-with-an-array-of-bits-in-c */

/* numarray.setbit (a, idx, x)
   Sets bit n at index position idx of uchar array a. n must be either 0 or 1. idx starts from 1, the rightmost bit,
   not zero. The function returns nothing. */
static int numarray_setbit (lua_State *L) {  /* set a bit (0 or 1) to the given slot of a uchar array; 2.12.0 RC 4 */
  int64_t idx, x;
  NumArray *a = checknumarray(L, 1);
  if (a->size == 0)
    luaL_error(L, "Error in " LUA_QS ": array is of size zero.", "numarray.setbit");
  if (a->datatype != 0)
    luaL_error(L, "Error in " LUA_QS ": can set bits only with uchar arrays.", "numarray.setbit");
  idx = tools_posrelat(agn_checkinteger(L, 2), a->size) - 1;  /* realign Lua/Agena index to C index, starting from 0, extended 2.15.1 */
  x = agn_checkinteger(L, 3);  /* realign Lua/Agena index to C index, starting from 0 */
  if (idx < 0 || idx >= a->size * NBITS)
    luaL_error(L, "Error in " LUA_QS ": index %d out of range.", "numarray.setbit", idx + 1);
  if (x < 0 || x > 1)
    luaL_error(L, "Error in " LUA_QS ": bit must be 0 or 1, got %d.", "numarray.setbit", x);
  if (x)
    /* void SetBit( int A[],  int k ) {  A[k/32] |= 1 << (k%32); }  set the bit at the k-th position in A[i] */
    a->data.c[idx / NBITS] |= 1 << (idx % NBITS);
  else
    /* #define ClearBit(A,k)   ( A[(k/32)] &= ~(1 << (k%32)) ) */
    a->data.c[idx / NBITS] &= ~(1 << (idx % NBITS));
  return 0;
}

/*
import numarray as n
a := n.uchar(1)
for i to 8 do print(n.getbit(a, i)) od
n.get(a, 1):
*/
/* numarray.getbit (a, i)
   Returns the bit at index position i of uchar array a. i starts from 1, the rightmost bit, not zero.
   The return is either 0 or 1. */
static int numarray_getbit (lua_State *L) {  /* get a bit from the given Lua/Agena index. */
  int64_t idx;
  NumArray *a = checknumarray(L, 1);
  if (a->size == 0)
    luaL_error(L, "Error in " LUA_QS ": array is of size zero.", "numarray.getbit");
  if (a->datatype != 0)
    luaL_error(L, "Error in " LUA_QS ": can get bits only with uchar arrays.", "numarray.getbit");
  idx = tools_posrelat(agn_checkinteger(L, 2), a->size) - 1;  /* realign Lua/Agena index to C index, starting from 0, extended 2.15.1 */
  if (idx < 0 || idx >= a->size * NBITS)
    luaL_error(L, "Error in " LUA_QS ": index %d out of range.", "numarray.getbit", idx + 1);
  /* #define TestBit(A,k)    ( A[(k/32)] & (1 << (k%32)) ) */
  lua_pushnumber(L, ((a->data.c[idx / NBITS]) & (1 << (idx % NBITS))) != 0);
  return 1;
}


#define aux_purge(L,datastruct,datatype) { \
  x = datastruct[idx]; \
  for (i=idx + 1; i < a->size - 1; i++) \
    datastruct[i] = datastruct[i + 1]; \
  if (resize) { \
    if ((datastruct = realloc(datastruct, (--a->size + zerosize) * sizeof(datatype))) == NULL) \
      luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "numarray.purge"); \
  } else \
    datastruct[a->size - 1] = 0; \
}

static int numarray_purge (lua_State *L) {  /* 2.9.1, delete a number at a given slot. */
  int64_t idx;
  size_t i;
  int resize, zerosize;  /* realloc array after deletion ? */
  lua_Number x = 0;
  NumArray *a = checknumarray(L, 1);
  idx = tools_posrelat(agn_checkinteger(L, 2), a->size) - 1;  /* realign Lua/Agena index to C index, starting from 0, extended 2.15.1 */
  resize = agnL_optboolean(L, 3, 1);
  if (idx < 0 || idx >= a->size || a->size < 1)
    luaL_error(L, "Error in " LUA_QS ": index %d out-of-range or array empty.", "numarray.purge", idx + 1);
  zerosize = a->size == 1;  /* shrinking to zero blocks causes segmentation faults, so prevent this */
  switch (a->datatype) {
    case 0: aux_purge(L, a->data.c, unsigned char); break;
    case 1: aux_purge(L, a->data.n, lua_Number); break;
    case 2: aux_purge(L, a->data.i, int32_t); break;
    case 3: aux_purge(L, a->data.us, uint16_t); break;  /* 2.18.2 */
    case 4: aux_purge(L, a->data.ui, uint32_t); break;  /* 2.22.1 */
    case 5: aux_purge(L, a->data.ld, long double); break;  /* 2.23.12 */
    default: luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.purge");
  }
  lua_pushnumber(L, x);
  return 1;
}


/* numarray.resize (a, n): The function resizes a numarray userdata structure a to the given number of entries n. Thus you
   can extend or shrink a numarray. [DELETED :: When shrinking a numarray, the function first overwrites the entries to be `deleted`
   with zeros.] When extending, it fills the new array slots with zeros.

   The function returns the new size, an integer.

   The function issues an error if the new size is less than 1 (an array cannot be shrunk to zero bytes). */

#define aux_resize(L,datastruct,datatype) { \
  if ((datastruct = realloc(datastruct, (n + zerosize) * sizeof(datatype))) == NULL) \
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "numarray.resize"); \
  if (n > a->size) { for (i=a->size; i < n; i++) datastruct[i] = 0; }; \
}  /* last line: fill added space with zeros */

static int numarray_resize (lua_State *L) {
  int64_t n, i;
  int zerosize;
  NumArray *a = checknumarray(L, 1);
  zerosize = 0;  /* added 2.9.1 */
  n = agnL_checkint(L, 2);  /* absolute new size */
  if (n > MAX_INT)
    luaL_error(L, "Error in " LUA_QS ": array too large, max. of 2147483645 elements accepted.", "numarray.resize");
  if (n  < 0)
    luaL_error(L, "Error in " LUA_QS ": invalid new size %d.", "numarray.resize", n);
  else if (a->size == 0 && n == 0) {  /* do nothing */
    lua_pushinteger(L, 0);
    return 1;
  } else if (n == 0)
    zerosize = 1;  /* trying to realloc to 0 bytes will cause segmentation faults later, so leave at least one block assigned */
  if (n != a->size) {  /* extend or shrink (or do nothing if a->size == n) */
    switch (a->datatype) {
      case 0: aux_resize(L, a->data.c,  unsigned char); break;
      case 1: aux_resize(L, a->data.n,  lua_Number); break;
      case 2: aux_resize(L, a->data.i,  int32_t); break;
      case 3: aux_resize(L, a->data.us, uint16_t); break;
      case 4: aux_resize(L, a->data.ui, uint32_t); break;
      case 5: aux_resize(L, a->data.ld, long double); break;
      default:
        luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.resize");
    }
    a->size = n + zerosize;  /* 2.15.1 fix */
  }
  lua_pushinteger(L, a->size);
  return 1;
}


/* numarray.iterate (a [, i [, p]])
Returns an iterator function that when called returns the next value in the numarray userdata structure
a, or null if there are no further entries in the structure.

If an index i is passed, the first call to the iterator function returns the i-th element in
the numarray list and with subsequent calls, the respective elements after index i.

You may also pass a positive integer step p to the iterator function: If given, then in subsequent
calls the p-th element after the respective current one is returned, equivalent to giving an optional
step size in numeric for loops.

Example:

> import numarray

> a := numarray.double(3)

> for i to 3 do a[i] := i * Pi od

> f := numarray.iterate(a, 2):  # return all values starting with index 2
procedure(01CDC200)

> f():
6.2831853071796

> f():
9.4247779607694

> f():  # no more values in a
null

import numarray as n
a := n.uchar(1)
for i to 8 do n.setbit(a, i, 1) od
n.get(a, 1):

f := numarray.iterate(a, 1, 1, true)
f():
*/

static int iteratebits (lua_State *L) {  /* 2.12.0 RC 4 */
  NumArray *a;
  int64_t pos, step;
  a = lua_touserdata(L, lua_upvalueindex(1));
  pos = lua_tointeger(L, lua_upvalueindex(2));
  step = lua_tointeger(L, lua_upvalueindex(3));
  if (pos > a->size * NBITS)
    lua_pushnil(L);
  else {
    lua_pushnumber(L, (a->data.c[(pos - 1) / NBITS] & ((1 << ((pos - 1) % NBITS)))) != 0);
    lua_pushinteger(L, pos + step);
    lua_replace(L, lua_upvalueindex(2));
  }
  return 1;
}

static int iterate (lua_State *L) {
  NumArray *a;
  size_t pos, step;
  a = lua_touserdata(L, lua_upvalueindex(1));
  pos = lua_tointeger(L, lua_upvalueindex(2));
  step = lua_tointeger(L, lua_upvalueindex(3));
  if (pos > a->size)
    lua_pushnil(L);
  else {
    switch (a->datatype) {
      case 0: lua_pushnumber(L, a->data.c[pos - 1]);  break;
      case 1: lua_pushnumber(L, a->data.n[pos - 1]);  break;
      case 2: lua_pushnumber(L, a->data.i[pos - 1]);  break;
      case 3: lua_pushnumber(L, a->data.us[pos - 1]); break;  /* 2.18.2 */
      case 4: lua_pushnumber(L, a->data.ui[pos - 1]); break;  /* 2.22.1 */
      case 5: { createdlong(L, a->data.ui[pos - 1]); break; } /* 2.35.0 */
      default:
        luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.iterate");
    }
    lua_pushinteger(L, pos + step);
    lua_replace(L, lua_upvalueindex(2));
  }
  return 1;
}

static int numarray_iterate (lua_State *L) {
  int64_t pos, step;
  int bits;
  NumArray *a = checknumarray(L, 1);
  pos = tools_posrelat(agnL_optinteger(L, 2, 1), a->size);  /* extended 2.15.1 */
  step = agnL_optinteger(L, 3, 1);
  bits = agnL_optboolean(L, 4, 0);  /* 2.12.0 RC 4 */
  if (bits == 0) {
    if (pos < 1 || (pos > a->size && a->size != 0))  /* when iterating, return null if the array is empty */
      luaL_error(L, "Error in " LUA_QS ": index %d out of range.", "numarray.iterate", pos);
  } else {
    if (pos < 1 || ((pos * NBITS) > (a->size * NBITS) && a->size != 0))  /* when iterating, return null if the array is empty */
      luaL_error(L, "Error in " LUA_QS ": index %d out of range.", "numarray.iterate", pos);
  }
  if (step < 1)
    luaL_error(L, "Error in " LUA_QS ": step size must be a positive.", "numarray.iterate");
  lua_pushvalue(L, 1);
  lua_pushinteger(L, pos);  /* we iterate from index position 1, not 0 */
  lua_pushinteger(L, step);
  lua_pushcclosure(L, (bits) ? &iteratebits : &iterate, 3);
  return 1;
}


static int cyclebits (lua_State *L) {
  NumArray *a;
  int64_t pos, step;
  a = lua_touserdata(L, lua_upvalueindex(1));
  pos = lua_tointeger(L, lua_upvalueindex(2));
  step = lua_tointeger(L, lua_upvalueindex(3));
  if (pos > a->size * NBITS)
    pos = lua_tointeger(L, lua_upvalueindex(4));
  lua_pushnumber(L, (a->data.c[(pos - 1) / NBITS] & ((1 << ((pos - 1) % NBITS)))) != 0);
  lua_pushinteger(L, pos + step);
  lua_replace(L, lua_upvalueindex(2));
  return 1;
}

static int cycle (lua_State *L) {
  NumArray *a;
  size_t pos, step;
  a = lua_touserdata(L, lua_upvalueindex(1));
  pos = lua_tointeger(L, lua_upvalueindex(2));
  step = lua_tointeger(L, lua_upvalueindex(3));
  if (pos > a->size)
    pos = lua_tointeger(L, lua_upvalueindex(4));
  switch (a->datatype) {
    case 0: lua_pushnumber(L, a->data.c[pos - 1]);  break;
    case 1: lua_pushnumber(L, a->data.n[pos - 1]);  break;
    case 2: lua_pushnumber(L, a->data.i[pos - 1]);  break;
    case 3: lua_pushnumber(L, a->data.us[pos - 1]); break;  /* 2.18.2 */
    case 4: lua_pushnumber(L, a->data.ui[pos - 1]); break;  /* 2.22.1 */
    case 5: { createdlong(L, a->data.ui[pos - 1]); break; } /* 2.35.0 */
    default:
      luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.iterate");
  }
  lua_pushinteger(L, pos + step);
  lua_replace(L, lua_upvalueindex(2));
  return 1;
}

static int numarray_cycle (lua_State *L) {  /* 2.15.5, based on numarray_iterate */
  int64_t pos, step;
  int bits;
  NumArray *a = checknumarray(L, 1);
  pos = tools_posrelat(agnL_optinteger(L, 2, 1), a->size);
  step = agnL_optinteger(L, 3, 1);
  bits = agnL_optboolean(L, 4, 0);
  if (bits == 0) {
    if (pos < 1 || (pos > a->size && a->size != 0))
      luaL_error(L, "Error in " LUA_QS ": index %d out of range.", "numarray.cycle", pos);
  } else {
    if (pos < 1 || ((pos * NBITS) > (a->size * NBITS) && a->size != 0))
      luaL_error(L, "Error in " LUA_QS ": index %d out of range.", "numarray.cycle", pos);
  }
  if (step < 1)
    luaL_error(L, "Error in " LUA_QS ": step size must be a positive.", "numarray.cycle");
  lua_pushvalue(L, 1);
  lua_pushinteger(L, pos);  /* we cycle from index position 1, not 0 */
  lua_pushinteger(L, step);
  lua_pushinteger(L, pos);  /* original start value */
  lua_pushcclosure(L, (bits) ? &cyclebits : &cycle, 4);
  return 1;
}


static int mt_getsize (lua_State *L) {  /* returns the number of allocated slots in the userdata numeric array */
  NumArray *a = checknumarray(L, 1);
  lua_pushinteger(L, a->size);
  return 1;
}


static int mt_empty (lua_State *L) {  /* 2.15.4 */
  NumArray *a = checknumarray(L, 1);
  lua_pushboolean(L, a->size == 0);
  return 1;
}


static int mt_filled (lua_State *L) {  /* 2.15.4 */
  NumArray *a = checknumarray(L, 1);
  lua_pushboolean(L, a->size != 0);
  return 1;
}


static int mt_array2string (lua_State *L) {  /* at the console, the array is formatted as follows: */
  NumArray *a = checknumarray(L, 1);
  if (agn_getutype(L, 1)) {
    lua_pushfstring(L, "(%u)", a->size);
    lua_concat(L, 2);
  } else
    luaL_error(L, "Error in " LUA_QS ": invalid array.", "numarray.__tostring");
  return 1;
}


static int mt_arraygc (lua_State *L) {  /* please do not forget to garbage collect deleted userdata */
  NumArray *a = checknumarray(L, 1);
  switch (a->datatype) {
    case 0: xfree(a->data.c);  break;
    case 1: xfree(a->data.n);  break;
    case 2: xfree(a->data.i);  break;
    case 3: xfree(a->data.us); break;  /* 2.18.2 */
    case 4: xfree(a->data.ui); break;  /* 2.22.1 */
    case 5: xfree(a->data.ld); break;  /* 2.35.0 */
    default:
      luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray garbage collection");
  }
  lua_setmetatabletoobject(L, 1, NULL, 1);
  luaL_unref(L, LUA_REGISTRYINDEX, a->registry);  /* 2.15.1, delete registry table */
  return 0;
}


/* Returns the index for a given value what in the numarray a. By default, the search starts at the beginning of the array,
   but you may pass any valid position pos (a positive integer) to determine where to start the search. The return is
   the index position, a positive number, or `null` if `what` could not be found in a.

   By default, the function checks for exact equality to detect the existence of a value. By passing the fourth argument
   eps, a non-negative number, the function also compares the values approximately with the given maximum deviation eps.
   See approx for more details. */

#define aux_whereis(L,datastruct,what,cond) { \
  for (i=pos; i < a->size; i++) { \
    if (datastruct[i] == (what) || (cond)) { \
      lua_pushinteger(L, i + 1); \
      return 1; \
    } \
  } \
}
  /* 2.15.4 change, isapprox deleted for this does not make sense with non-doubles */

static int numarray_whereis (lua_State *L) {
  int64_t i, pos, isapprox;
  lua_Number eps;
  union ldshape d;
  NumArray *a;
  a = checknumarray(L, 1);  /* 2.15.5 fix */
  getdlongnum(L, d, 2, "numarray.whereis");
  /* what = agn_checknumber(L, 2); */
  pos = tools_posrelat(agnL_optinteger(L, 3, 1), a->size) - 1;  /* extended 2.15.1 */
  eps = luaL_optnumber(L, 4, 0);
  if (eps < 0)
    luaL_error(L, "Error in " LUA_QS ": fourth argument must be non-negative.", "numarray.whereis");
  isapprox = lua_gettop(L) == 4;
  if (pos < 0 || pos >= a->size)
    luaL_error(L, "Error in " LUA_QS ": index %d out-of-range.", "numarray.whereis", pos + 1);
  switch (a->datatype) {
    case 0: aux_whereis(L, a->data.c,  d.d, 0);  break;
    case 1: aux_whereis(L, a->data.n,  d.d, (isapprox && tools_approx(a->data.n[i], d.d, eps))); break;
    case 2: aux_whereis(L, a->data.i,  d.d, 0); break;
    case 3: aux_whereis(L, a->data.us, d.d, 0); break;
    case 4: aux_whereis(L, a->data.ui, d.d, 0); break;
    case 5: {
      long double v = tolong(L, 2, d);
      aux_whereis(L, a->data.ld, v, (isapprox && tools_approxl(a->data.ld[i], v, eps)));
      break;
    }
    default: luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.wheris");
  }
  lua_pushnil(L);
  return 1;
}


#define aux_arrayeq(L, fn, ds1, ds2, eps) { \
  for (i=0; i < a->size; i++) { \
    if (!fn(ds1[i], ds2[i], (eps))) { \
      flag = 0; \
      goto endofarrayeq; \
    } \
  } \
}
/* goto statement: exit in case of inequality */

static void arrayeq (lua_State *L, int (*fn)(lua_Number, lua_Number, lua_Number)) {  /* based on the linalg package (function meq) */
  size_t i;
  int flag;
  lua_Number eps;
  NumArray *a, *b;
  a = checknumarray(L, 1);
  b = checknumarray(L, 2);
  flag = 1;
  if (a->size != b->size || a->datatype != b->datatype)
    lua_pushfalse(L);
  else {  /* arrays have same dimension */
    eps = agn_getepsilon(L);
    switch (a->datatype) {
      case 0: aux_arrayeq(L, fn, a->data.c, b->data.c, eps);   break;
      case 1: aux_arrayeq(L, fn, a->data.n, b->data.n, eps);   break;
      case 2: aux_arrayeq(L, fn, a->data.i, b->data.i, eps);   break;
      case 3: aux_arrayeq(L, fn, a->data.us, b->data.us, eps); break;  /* 2.18.2 */
      case 4: aux_arrayeq(L, fn, a->data.ui, b->data.ui, eps); break;  /* 2.22.1 */
      case 5: aux_arrayeq(L, fn, a->data.ld, b->data.ui, eps); break;  /* 2.35.0 */
      default: luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray equality operation");
    }
  }
endofarrayeq:
  lua_pushboolean(L, flag);
}

static int equal (lua_Number x, lua_Number y, lua_Number eps) {  /* strict equality, eps does not matter */
  return x == y;
}

static int mt_aeq (lua_State *L) {
  arrayeq(L, tools_approx);
  return 1;
}


static int mt_eeq (lua_State *L) {
  arrayeq(L, equal);
  return 1;
}


/* numarray.read (fh [, bufsize])

Reads data from the file denoted by its filehandle fh and returns a numarray userdata structure
of unsigned C chars.

The file should before be opened with `binio.open` and should finally be closed with `binio.close`.

In general, the function reads in a limited amount of bytes per call. If only fh is passed, the
number of bytes read is determined by the environ.kernel('buffersize') setting, usually 512 bytes.

You can pass the second argument bufsize, a positive integer, to read less or more bytes. Passing
the bufsize argument may also be necessary if your platform requires that an internal input buffer
is aligned to a certain block size.

The function increments the file position thereafter so that the next bytes in the file can be read
with a new call to numarray.read.

If the end of the file has been reached, or there is nothing to read at all, `null` is returned.
In case of an error, it quits with the respective error.

See also: numarray.readdoubles, numarray.readintegers, numarray.readuchars. */

/* based on binio.readbytes. */

/* 2.15.5 Hotfixes, get binio userdata as file handles, not integers */

static int numarray_read (lua_State *L) {
  int64_t s, i;
  int hnd, n, en;
  NumArray *a;
  unsigned char *buffer;
  hnd = agn_tofileno(L, 1);
  s = luaL_optinteger(L, 2, agn_getbuffersize(L));  /* reading a lot of bytes at one stroke will cause block size errors, so keep the buffer small */
  if (s == -1)
    luaL_error(L, "Error in " LUA_QS " with file #%d: file I/O error.", "numarray.read", hnd);
  else if (s <= 0)
    s = agn_getbuffersize(L);  /* 2.34.9 adaption */
  buffer = malloc(s * sizeof(unsigned char));
  if (buffer == NULL)
    luaL_error(L, "Error in " LUA_QS " with file #%d: memory allocation failed.", "numarray.read", hnd);
  n = read(hnd, buffer, s * sizeof(unsigned char));
  en = errno;
  if (n > 0) {
    createnumarrayfileio(unsigned char, "uchar", 0, n);
    for (i=0; i < n; i++)
      a->data.c[i] = buffer[i];
  }
  else if (n == 0)
    lua_pushnil(L);
  else {
    xfree(buffer);
    luaL_error(L, "Error in " LUA_QS " with file #%d: %s.", "numarray.read", hnd, my_ioerror(en));
  }
  xfree(buffer);
  return 1;
}


/* numarray.write (fh, a [, pos [, nvalues]])

based on `binio.writebytes`; writes unsigned chars, doubles, or integers stored in a numarray userdata a to the file denoted by its
numeric file handle fh. The file should be opened with binio.open and closed with binio.close.

The start position `pos` is 1 by default but can be changed to any other valid position in the numarray.

The number of values (not bytes 1) `nvalues` to be written can be changed by passing an optional fourth argument, a positive
number, and by default equals the total number of entries in `a`. Depending on the data type stored to the numarray, the
function automatically computes the number of bytes to be written. Passing the `nvalues` argument may also be necessary if your
platform requires that buffers must be aligned to a particular block size.

The function returns the index of the next start position (an integer) for a further call to write a entirely, or `null`,
if the end of the numarray has been reached. When the function returns `null`, then it also automatically flushes all unwritten
content to the file so that you do not have to call `binio.sync` manually if you want to read the file subsequently.

No further information is stored to the file created, so you always must know the type of data you want to read in later.

Example on how to write an entire array of 4,096 integers piece-by-piece:

> a := numarray.integer(4 * 1024);
> fd := binio.open('uchar.bin');
> pos := 1;
> do
>    pos := numarray.write(fd, a, pos, 1024)  # write 1024 values per each call
> until pos = null;
> binio.close(fd);

Use `binio.sync` if you want to make sure that any unwritten content is written to the file when calling `numarray.write`
multiple times on one numarray. */

static size_t aux_fillbuffer (unsigned char *buffer, NumArray *a, int64_t start, int64_t nvals, size_t buffersize) {
  int64_t i;
  int j;
  size_t c;
  unsigned char *dst;
  memset(buffer, 0, buffersize);
  c = 0;
  switch (a->datatype) {
    case 0:
      for (i=start; i < start + nvals; i++) {
        buffer[i - start] = a->data.c[i]; c++;
      }
      break;
    case 1: {
      uint64_t u;
      for (i=start; i < start + nvals; i++) {
#if BYTE_ORDER == BIG_ENDIAN
        u = tools_doubletouint64andswap(a->data.n[i]);  /* 2.16.6 fix, convert to Little Endian `integer` */
#else
        u = tools_doubletouint64(a->data.n[i]);  /* convert to Little Endian `integer` */
#endif
        dst = (unsigned char *)&u;
        for (j=0; j < sizeof(lua_Number); j++)
          buffer[c++] = dst[j];
      }
      break;
    }
    case 2: {
      int32_t u;
      for (i=start; i < start + nvals; i++) {
        u = a->data.i[i];
#if BYTE_ORDER == BIG_ENDIAN  /* 2.16.1 fix */
        tools_swapint32_t(&u);  /* convert to Little Endian 32-bit `integer` */
#endif
        dst = (unsigned char *)&u;
        for (j=0; j < sizeof(int32_t); j++)
          buffer[c++] = dst[j];
      }
      break;
    }
    case 3: {  /* 2.18.2 */
      uint16_t u;
      for (i=start; i < start + nvals; i++) {
        u = a->data.us[i];
#if BYTE_ORDER == BIG_ENDIAN  /* 2.16.1 fix */
        tools_swapuint16_t(&u);  /* convert to Little Endian 32-bit `integer` */
#endif
        dst = (unsigned char *)&u;
        for (j=0; j < sizeof(uint16_t); j++)
          buffer[c++] = dst[j];
      }
      break;
    }
    case 4: {  /* 2.22.1 */
      uint32_t u;
      for (i=start; i < start + nvals; i++) {
        u = a->data.ui[i];
#if BYTE_ORDER == BIG_ENDIAN  /* 2.16.1 fix */
        tools_swapuint32_t(&u);  /* convert to Little Endian 32-bit `integer` */
#endif
        dst = (unsigned char *)&u;
        for (j=0; j < sizeof(uint32_t); j++)
          buffer[c++] = dst[j];
      }
      break;
    }
    case 5: {  /* 2.35.0 */
      union ldshape d;
      for (i=start; i < start + nvals; i++) {
        /* memset(d.c, 0, SIZEOFLDBL);
        d.i.m = d.i.se = d.i.junk = 0; */
        d.f = a->data.ld[i];
        for (j=0; j < SIZEOFLDBL; j++) {
#if BYTE_ORDER == BIG_ENDIAN
          buffer[c++] = d.c[SIZEOFLDBL - j - 1];
#else
          buffer[c++] = d.c[j];
#endif
        }
      }
      break;
    }
    default:
      lua_assert(0);
  }
  return c;
}

static int numarray_write (lua_State *L) {
  int hnd, en, bufdatasize;
  int64_t nvals, start;
  ssize_t nbyteswritten;
  size_t valuestowrite, nbytes, bytestowrite;
  unsigned char *buffer;
  NumArray *a;
  hnd = agn_tofileno(L, 1);
  a = checknumarray(L, 2);
  switch (a->datatype) {
    case 0: nbytes = sizeof(unsigned char); break;
    case 1: nbytes = sizeof(lua_Number); break;
    case 2: nbytes = sizeof(int32_t); break;
    case 3: nbytes = sizeof(uint16_t); break;  /* 2.18.2 */
    case 4: nbytes = sizeof(uint32_t); break;  /* 2.22.1 */
    case 5: nbytes = SIZEOFLDBL; break;  /* 2.35.0 */
    default: {
      nbytes = 0;
      luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.write");  /* avoid compiler warnings */
    }
  }
  bufdatasize = agn_getbuffersize(L)/nbytes;
  if (bufdatasize < 1)
    luaL_error(L, "Error in " LUA_QS ": buffer too small, increase it with environ.kernel/buffersize.", "numarray.write", hnd);
  start = agnL_optinteger(L, 3, 1) - 1;
  if (start < 0 || start >= a->size)
    luaL_error(L, "Error in " LUA_QS " with file #%d: start position %d out-of-range.", "numarray.write", hnd, start + 1);
  nvals = agnL_optinteger(L, 4, a->size);  /* number of values (not bytes) in array to be written */
  if (start + nvals > a->size) nvals = a->size - start;  /* adjust to actual size */
  if (nvals < 1)
    luaL_error(L, "Error in " LUA_QS " with file #%d: invalid number of bytes to be written.", "numarray.write", hnd);
  /* from here on rewritten 2.18.0 RC 2 */
  valuestowrite = (nvals > bufdatasize) ? bufdatasize : nvals;
  bytestowrite = valuestowrite*nbytes*sizeof(unsigned char);
  buffer = (unsigned char *)malloc(bytestowrite);
  if (buffer == NULL)
    luaL_error(L, "Error in " LUA_QS " with file #%d: memory allocation failed.", "numarray.write", hnd);
  do {
//printf("valuestowrite=%d bytestowrite=%d\n", (int)valuestowrite, (int)bytestowrite);
    if (aux_fillbuffer(buffer, a, start, valuestowrite, bytestowrite) != bytestowrite) 
      luaL_error(L, "Error in " LUA_QS " with file #%d: memory error.", "numarray.write", hnd);
    if ( (nbyteswritten = write(hnd, buffer, bytestowrite)) == -1) {
      en = errno;
      xfree(buffer);
      luaL_error(L, "Error in " LUA_QS " with file #%d: %s.", "numarray.write", hnd, my_ioerror(en));
    }
    if (nbyteswritten != bytestowrite)
      luaL_error(L, "Error in " LUA_QS " with file #%d: I/O error.", "numarray.write", hnd);
    nvals -= valuestowrite;
    start += valuestowrite;
    if (nvals < valuestowrite) {
      valuestowrite = nvals;
      bytestowrite = valuestowrite*nbytes;
    }
  } while (nvals > 0);
  start++;
  if (start > a->size)  /* no more bytes to be written ? */
    lua_pushnil(L);
  else  /* return the next start position for a further call to write further parts of the numarray; 2.17.8 change */
    lua_pushnumber(L, start);
  tools_fsync(hnd);  /* 2.17.8, 2.35.0 */
  xfree(buffer);
  return 1;
}


/* numarray.toseq (a)

Receives a numarray userdata structure a and converts it into a sequence of numbers, the return. */

static int numarray_toseq (lua_State *L) {
  size_t i;
  NumArray *a = checknumarray(L, 1);
  agn_createseq(L, a->size);
  switch (a->datatype) {
    case 0:
      for (i=0; i < a->size; i++)
        agn_seqsetinumber(L, -1, i + 1, a->data.c[i]);
      break;
    case 1:
      for (i=0; i < a->size; i++)
        agn_seqsetinumber(L, -1, i + 1, a->data.n[i]);
      break;
    case 2:
      for (i=0; i < a->size; i++)
        agn_seqsetinumber(L, -1, i + 1, a->data.i[i]);
      break;
    case 3:  /* 2.18.2 */
      for (i=0; i < a->size; i++)
        agn_seqsetinumber(L, -1, i + 1, a->data.us[i]);
      break;
    case 4:  /* 2.22.1 */
      for (i=0; i < a->size; i++)
        agn_seqsetinumber(L, -1, i + 1, a->data.ui[i]);
      break;
    case 5:  /* 2.35.0 */
      for (i=0; i < a->size; i++) {
        createdlong(L, a->data.ld[i]);
        lua_seqseti(L, -2, i + 1);
      }
      break;
    default:
      luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.toseq");
  }
  return 1;
}


/* numarray.toreg (a)

Receives a numarray userdata structure a and converts it into a register of numbers, the return. */

static int numarray_toreg (lua_State *L) {
  size_t i;
  NumArray *a = checknumarray(L, 1);
  agn_createreg(L, a->size);
  switch (a->datatype) {
    case 0:
      for (i=0; i < a->size; i++) {
        agn_regsetinumber(L, -1, i + 1, a->data.c[i]);
      }
      break;
    case 1:
      for (i=0; i < a->size; i++) {
        agn_regsetinumber(L, -1, i + 1, a->data.n[i]);
      }
      break;
    case 2:
      for (i=0; i < a->size; i++) {
        agn_regsetinumber(L, -1, i + 1, a->data.i[i]);
      }
      break;
    case 3:  /* 2.18.2 */
      for (i=0; i < a->size; i++) {
        agn_regsetinumber(L, -1, i + 1, a->data.us[i]);
      }
      break;
    case 4:  /* 2.22.1 */
      for (i=0; i < a->size; i++) {
        agn_regsetinumber(L, -1, i + 1, a->data.ui[i]);
      }
      break;
    case 5:  /* 2.35.0 */
      for (i=0; i < a->size; i++) {
        createdlong(L, a->data.ld[i]);
        agn_regseti(L, -2, i + 1);
      }
      break;
    default:
      luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.toreg");
  }
  return 1;
}

/* numarray.toarray(o [, option])

   Converts a sequence or register o of numbers into a numarray. By default, a double array is returned;
   if the second argument `option` is the string 'uchar', an unsigned char array is created; if it is the
   string 'integer', an integer array is returned. By default, option is the string 'double'.

   If a value in o is not a number, zero is written to the array. */

static int numarray_toarray (lua_State *L) {
  int64_t i, n;
  int what;
  NumArray *a;
  static const char *const datatypes[] = {"uchar", "double", "int32", "ushort", "uint32", "longdouble", NULL};
  what = agnL_checkoption(L, 2, "double", datatypes, 0);
  switch (lua_type(L, 1)) {
    case LUA_TTABLE: {  /* 2.31.3 */
      n = agn_asize(L, 1);
      createarray(L, a, what, n);
      switch (a->datatype) {
        case 0:
          for (i=0; i < n; i++)
            a->data.c[i] = (unsigned char)agn_getinumber(L, 1, i + 1);
          break;
        case 1:
          for (i=0; i < n; i++)
            a->data.n[i] = agn_getinumber(L, 1, i + 1);
          break;
        case 2:
          for (i=0; i < n; i++)
            a->data.i[i] = (int32_t)agn_getinumber(L, 1, i + 1);
          break;
        case 3:
          for (i=0; i < n; i++)
            a->data.us[i] = (uint16_t)agn_getinumber(L, 1, i + 1);
          break;
        case 4:
          for (i=0; i < n; i++)
            a->data.ui[i] = (uint32_t)agn_getinumber(L, 1, i + 1);
          break;
        case 5:
          for (i=0; i < n; i++) {
            lua_rawgeti(L, 1, i + 1);
            a->data.ld[i] = agnL_checkdlongnum(L, -1, "numarray.toarray");
            agn_poptop(L);
          }
          break;
        default:
          luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.toarray");
      }
      break;
    }
    case LUA_TSEQ: {
      n = agn_seqsize(L, 1);
      createarray(L, a, what, n);
      switch (a->datatype) {
        case 0:
          for (i=0; i < n; i++)
            a->data.c[i] = (unsigned char)lua_seqrawgetinumber(L, 1, i + 1);
          break;
        case 1:
          for (i=0; i < n; i++)
            a->data.n[i] = lua_seqrawgetinumber(L, 1, i + 1);
          break;
        case 2:
          for (i=0; i < n; i++)
            a->data.i[i] = (int32_t)lua_seqrawgetinumber(L, 1, i + 1);
          break;
        case 3:  /* 2.18.2 */
          for (i=0; i < n; i++)
            a->data.us[i] = (uint16_t)lua_seqrawgetinumber(L, 1, i + 1);
          break;
        case 4:  /* 2.22.1 */
          for (i=0; i < n; i++)
            a->data.ui[i] = (uint32_t)lua_seqrawgetinumber(L, 1, i + 1);
          break;
        case 5:
          for (i=0; i < n; i++) {
            lua_seqrawgeti(L, 1, i + 1);
            a->data.ld[i] = agnL_checkdlongnum(L, -1, "numarray.toarray");
            agn_poptop(L);
          }
          break;
        default:
          luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.toarray");
      }
      break;
    }
    case LUA_TREG: {
      n = agn_reggettop(L, 1);
      createarray(L, a, what, n);
      switch (a->datatype) {
        case 0:
          for (i=0; i < n; i++)
            a->data.c[i] = (unsigned char)agn_reggetinumber(L, 1, i + 1);
          break;
        case 1:
          for (i=0; i < n; i++)
            a->data.n[i] = agn_reggetinumber(L, 1, i + 1);
          break;
        case 2:
          for (i=0; i < n; i++)
            a->data.i[i] = (int32_t)agn_reggetinumber(L, 1, i + 1);
          break;
        case 3:  /* 2.18.2 */
          for (i=0; i < n; i++)
            a->data.us[i] = (uint16_t)agn_reggetinumber(L, 1, i + 1);
          break;
        case 4:  /* 2.22.1 */
          for (i=0; i < n; i++)
            a->data.ui[i] = (uint32_t)agn_reggetinumber(L, 1, i + 1);
          break;
        case 5:
          for (i=0; i < n; i++) {
            agn_reggeti(L, 1, i + 1);
            a->data.ld[i] = agnL_checkdlongnum(L, -1, "numarray.toarray");
            agn_poptop(L);
          }
          break;
        default:
          luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.toarray");
      }
      break;
    }
    default:
      luaL_error(L, "Error in " LUA_QS ": table, sequence or register expected, got %s.",
        "numarray.toarray", luaL_typename(L, 1));
  }
  return 1;
}

/* numarray.include(a, pos, b)

   Copies all values in the numarray b into the numarray a, starting at index pos (a number) of a. The function
   returns nothing. Both numarrays must by of the same type: either be uchar, integer, or double arrays. */

#define aux_include_ud(L,ds1,ds2) { \
  for (i=pos; i < pos + b->size; i++) \
    ds1[i] = ds2[i - pos]; \
}

#define aux_include_ln(L,datastruct,x,datatype) { \
  if ((datastruct = realloc(datastruct, (a->size + 1) * sizeof(datatype))) == NULL) \
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "numarray.include"); \
  a->size++; \
  for (i=a->size - 1; i > pos; i--) \
    datastruct[i] = datastruct[i - 1]; \
  datastruct[pos] = x; \
}

static int numarray_include (lua_State *L) {
  int64_t i, pos;
  NumArray *a, *b;
  a = checknumarray(L, 1);
  pos = agn_checkinteger(L, 2);  /* extended 2.15.1, 2.15.2 fix, 2.15.5 fix */
  if (pos == 0)
    luaL_error(L, "Error in " LUA_QS ": index 0 is out-of-range.", "numarray.include");
  pos = tools_posrelat(pos, a->size) - 1;  /* 2.15.5 fix */
  if (lua_type(L, 3) == LUA_TUSERDATA && !agnL_isdlong(L, 3)) {
    b = checknumarray(L, 3);
    if (b->size + pos > a->size)
      luaL_error(L, "Error in " LUA_QS ": first numarray not large enough.", "numarray.include");
    if (b->size == 0)
      luaL_error(L, "Error in " LUA_QS ": second numarray is empty.", "numarray.include");
    if (a->datatype != b->datatype)
      luaL_error(L, "Error in " LUA_QS ": both numarrays must by of the same type.", "numarray.include");
    switch (a->datatype) {  /* if size of b is 0, nothing happens */
      case 0: aux_include_ud(L, a->data.c,  b->data.c);  break;
      case 1: aux_include_ud(L, a->data.n,  b->data.n);  break;
      case 2: aux_include_ud(L, a->data.i,  b->data.i);  break;
      case 3: aux_include_ud(L, a->data.us, b->data.us); break;
      case 4: aux_include_ud(L, a->data.ui, b->data.ui); break;
      case 5: aux_include_ud(L, a->data.ld, b->data.ld); break;
      default: luaL_error(L, "Error in " LUA_QS ": this should not happen/1.", "numarray.include");
    }
  } else if (lua_isnumber(L, 3) || agnL_isdlong(L, 3)) {  /* added 2.9.1/2.35.0 */
    union ldshape d;
    if (pos > a->size)  /* allow index pos = a->size + 1 */
      luaL_error(L, "Error in " LUA_QS ": index %d is out-of-range.", "numarray.include", pos + 1);
    /* enlarge array by one slot and shift values to open space */
    getdlongnum(L, d, 3, "numarray.include");
    switch (a->datatype) {
      case 0: aux_include_ln(L, a->data.c,  d.d, unsigned char); break;
      case 1: aux_include_ln(L, a->data.n,  d.d, lua_Number);    break;
      case 2: aux_include_ln(L, a->data.i,  d.d, int32_t);       break;
      case 3: aux_include_ln(L, a->data.us, d.d, uint16_t);      break;
      case 4: aux_include_ln(L, a->data.ui, d.d, uint32_t);      break;
      case 5: aux_include_ln(L, a->data.ld, tolong(L, 3, d), long double);   break;
      default: luaL_error(L, "Error in " LUA_QS ": this should not happen/2.", "numarray.include");
    }
  } else
    luaL_error(L, "Error in " LUA_QS ": argument #3 must be a numarray, number or longdouble, got %d.",
      "numarray.include", luaL_typename(L, 3));
  return 0;
}


/* Returns the number of bytes consumed by the given array. */

static int numarray_used (lua_State *L) {
  int64_t r;
  size_t size;
  NumArray *a = checknumarray(L, 1);
  GCObject *o = index2gco(L, 1);
  r = sizeof(NumArray) + sizeudata(gco2u(o));
  size = a->size;
  switch (a->datatype) {
    case 0:
      if (size == 0 && a->data.c != NULL) size = 1;
      r += size * sizeof(unsigned char);
      break;
    case 1:
      if (size == 0 && a->data.n != NULL) size = 1;
      r += size * sizeof(lua_Number);
      break;
    case 2:
      if (size == 0 && a->data.i != NULL) size = 1;
      r += size * sizeof(int32_t);
      break;
    case 3:  /* 2.18.2 */
      if (size == 0 && a->data.us != NULL) size = 1;
      r += size * sizeof(uint16_t);
      break;
    case 4:  /* 2.22.1 */
      if (size == 0 && a->data.ui != NULL) size = 1;
      r += size * sizeof(uint32_t);
      break;
    case 5:  /* 2.35.0 */
      if (size == 0 && a->data.ld != NULL) size = 1;
      r += size * SIZEOFLDBL;
      break;
    default:
      luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.used");
  }
  lua_pushnumber(L, r);
  return 1;
}


static int mt_in (lua_State *L) {  /* `__in` mt, 2.18.0 */
  NumArray *a;
  union ldshape d;
  size_t i;
  getdlongnum(L, d, 1, "numarray.__in");
  /* x = agn_checknumber(L, 1); */
  a = checknumarray(L, 2);
  switch (a->datatype) {
    case 0:
      for (i=0; i < a->size && a->data.c[i] != d.d; i++) { }
      break;
    case 1:
      for (i=0; i < a->size && a->data.n[i] != d.d; i++) { }
      break;
    case 2:
      for (i=0; i < a->size && a->data.i[i] != d.d; i++) { }
      break;
    case 3:  /* 2.18.2 */
      for (i=0; i < a->size && a->data.us[i] != d.d; i++) { }
      break;
    case 4:  /* 2.22.1 */
      for (i=0; i < a->size && a->data.ui[i] != d.d; i++) { }
      break;
    case 5:  /* 2.35.0 */
      for (i=0; i < a->size && a->data.ld[i] != tolong(L, 1, d); i++) { }
      break;
    default:
      i = 0;
      luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray in operation");;
  }
  lua_pushboolean(L, i != a->size);
  return 1;
}


static int mt_notin (lua_State *L) {  /* `__notin` mt, 2.18.0 */
  NumArray *a;
  union ldshape d;
  size_t i;
  getdlongnum(L, d, 1, "numarray.__notin");
  a = checknumarray(L, 2);
  switch (a->datatype) {
    case 0:
      for (i=0; i < a->size && a->data.c[i] != d.d; i++) { }
      break;
    case 1:
      for (i=0; i < a->size && a->data.n[i] != d.d; i++) { }
      break;
    case 2:
      for (i=0; i < a->size && a->data.i[i] != d.d; i++) { }
      break;
    case 3:  /* 2.18.2 */
      for (i=0; i < a->size && a->data.us[i] != d.d; i++) { }
      break;
    case 4:  /* 2.22.1 */
      for (i=0; i < a->size && a->data.ui[i] != d.d; i++) { }
      break;
    case 5:  /* 2.35.0 */
      for (i=0; i < a->size && a->data.ld[i] != tolong(L, 1, d); i++) { }
      break;
    default:
      i = 0;
      luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray notin operation");
  }
  lua_pushboolean(L, i == a->size);
  return 1;
}


/* Checks whether all elements in the numarray are zeros and returns `true` or `false`. 2.17.9 */
static int mt_zero (lua_State *L) {
  size_t i;
  NumArray *a = checknumarray(L, 1);
  switch (a->datatype) {
    case 0:
      for (i=0; i < a->size && a->data.c[i] == 0; i++) { }
      break;
    case 1:
      for (i=0; i < a->size && a->data.n[i] == 0; i++) { }
      break;
    case 2:
      for (i=0; i < a->size && a->data.i[i] == 0; i++) { }
      break;
    case 3:  /* 2.18.2 */
      for (i=0; i < a->size && a->data.us[i] == 0; i++) { }
      break;
    case 4:  /* 2.22.1 */
      for (i=0; i < a->size && a->data.ui[i] == 0; i++) { }
      break;
    case 5:  /* 2.35.0 */
      for (i=0; i < a->size && a->data.ld[i] == 0.0L; i++) { }
      break;
    default:
      i = 0;
      luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray zero operation");
  }
  lua_pushboolean(L, i == a->size);
  return 1;
}


/* Checks whether all elements in the numarray are non-zeros and returns `true` or `false`. 2.17.9 */
static int mt_nonzero (lua_State *L) {  /* 2.17.9 */
  size_t i;
  NumArray *a = checknumarray(L, 1);
  switch (a->datatype) {
    case 0:
      for (i=0; i < a->size && a->data.c[i] != 0; i++) { }
      break;
    case 1:
      for (i=0; i < a->size && a->data.n[i] != 0; i++) { }
      break;
    case 2:
      for (i=0; i < a->size && a->data.i[i] != 0; i++) { }
      break;
    case 3:  /* 2.18.2 */
      for (i=0; i < a->size && a->data.us[i] != 0; i++) { }
      break;
    case 4:  /* 2.22.1 */
      for (i=0; i < a->size && a->data.ui[i] != 0; i++) { }
      break;
    case 5:  /* 2.35.0 */
      for (i=0; i < a->size && a->data.ld[i] != 0.0L; i++) { }
      break;
    default:
      i = 0;
      luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray nonzero operation");
  }
  lua_pushboolean(L, i == a->size);
  return 1;
}


/* numarray.satisfy (f, a [, ])
   With any numarray `a`, checks each element by calling function `f` which should return `true` or `false`. If at least
   one element in `a` does not satisfy the condition checked by `f`, the result is `false`, and `true` otherwise. 2.17.9 */

#define fn_checkresult(L,fn,procname,type,vals) { \
  if (!fn(L, -1)) { \
    lua_pop(L,vals); \
    luaL_error(L, "Error in " LUA_QS ": procedure must evaluate to a %s, returned %s.", \
      procname, type, luaL_typename(L, -1)); \
  } \
}

#define aux_satisfy(L,fn,datastruct) { \
  for (i=0; i < a->size; i++) { \
    lua_pushvalue(L, 1); \
    fn(L, datastruct[i]); \
    for (j=3; j <= nargs; j++) lua_pushvalue(L, j); \
    lua_call(L, nargs - 1, 1); \
    fn_checkresult(L, lua_isboolean, "numarray.satisfy", "boolean", 1); \
    if (agn_isfalse(L, -1)) return 1; \
  } \
}

static int numarray_satisfy (lua_State *L) {
  NumArray *a;
  int i, j;
  size_t nargs = lua_gettop(L);
  if (nargs < 2)
    luaL_error(L, "Error in " LUA_QS ": requires at least two arguments.", "numarray.satisfy");
  luaL_checkstack(L, nargs, "too many arguments");
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = checknumarray(L, 2);
  switch (a->datatype) {
    case 0: aux_satisfy(L, lua_pushnumber, a->data.c);  break;
    case 1: aux_satisfy(L, lua_pushnumber, a->data.n);  break;
    case 2: aux_satisfy(L, lua_pushnumber, a->data.i);  break;
    case 3: aux_satisfy(L, lua_pushnumber, a->data.us); break;  /* 2.18.2 */
    case 4: aux_satisfy(L, lua_pushnumber, a->data.ui); break;  /* 2.22.1 */
    case 5: aux_satisfy(L, createdlong,    a->data.ld); break;  /* 2.35.0 */
    default: luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.satisfy");
  }
  lua_pushtrue(L);
  return 1;
}


/* numarray.map (f, a [, ])
   Maps a function f on each element in the numarray `a` and returns a new numarray with the mapped results, i.e. the new
   array includes the values f(a[1]), f(a[2]), f(a[3]), etc. `f` must always return a number. See also: `numarray.convert`.
   2.17.9 */

#define aux_map(L,luapush,check,convert,ds1,ds2) { \
  for (i=0; i < n; i++) { \
    lua_pushvalue(L, 1); \
    luapush(L, ds2[i]); \
    for (j=3; j <= nargs; j++) lua_pushvalue(L, j); \
    lua_call(L, nargs - 1, 1); \
    fn_checkresult(L, check, "numarray.map", "number or longdouble", 2); \
    ds1[i] = convert(L, -1); \
    agn_poptop(L); \
  } \
}

static int numarray_map (lua_State *L) {
  NumArray *a, *b;
  int i, j, newstruct, equality, nargs;
  char type;  /* 2.31.3 fix */
  size_t n;
  nargs = lua_gettop(L);
  if (nargs < 2)
    luaL_error(L, "Error in " LUA_QS ": requires at least two arguments.", "numarray.map");
  luaL_checkstack(L, nargs, "too many arguments");
  luaL_checktype(L, 1, LUA_TFUNCTION);
  numarray_checkoptions(L, &nargs, &newstruct, &equality, "numarray.map");
  b = checknumarray(L, 2);
  type = b->datatype;
  n = b->size;
  if (newstruct) {
    createarray(L, a, type, n);
    switch (b->datatype) {
      case 0: aux_map(L, lua_pushinteger, lua_isnumber, agn_tonumber, a->data.c,  b->data.c);  break;
      case 1: aux_map(L, lua_pushnumber,  lua_isnumber, agn_tonumber, a->data.n,  b->data.n);  break;
      case 2: aux_map(L, lua_pushnumber,  lua_isnumber, agn_tonumber, a->data.i,  b->data.i);  break;
      case 3: aux_map(L, lua_pushnumber,  lua_isnumber, agn_tonumber, a->data.us, b->data.us); break;
      case 4: aux_map(L, lua_pushnumber,  lua_isnumber, agn_tonumber, a->data.ui, b->data.ui); break;
      case 5: aux_map(L, createdlong,    agnL_isdlong, getdlongvalue, a->data.ld, b->data.ld); break;
      default:
        luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.map");
    }  /* return new numarray */
  } else {  /* in-place operation, 2.31.3 */
    lua_pushvalue(L, 2);  /* push numarray for later return */
    switch (b->datatype) {
      case 0: aux_map(L, lua_pushinteger, lua_isnumber, agn_tonumber, b->data.c,  b->data.c);  break;
      case 1: aux_map(L, lua_pushnumber,  lua_isnumber, agn_tonumber, b->data.n,  b->data.n);  break;
      case 2: aux_map(L, lua_pushnumber,  lua_isnumber, agn_tonumber, b->data.i,  b->data.i);  break;
      case 3: aux_map(L, lua_pushnumber,  lua_isnumber, agn_tonumber, b->data.us, b->data.us); break;
      case 4: aux_map(L, lua_pushnumber,  lua_isnumber, agn_tonumber, b->data.ui, b->data.ui); break;
      case 5: aux_map(L, createdlong,     agnL_isdlong, getdlongvalue, b->data.ld, b->data.ld); break;
      default:
        luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.map");
    }  /* return existing numarray */
  }
  return 1;
}


#define aux_select(L,luapush,ds1,ds2) { \
  for (i=0; i < n; i++) { \
    lua_pushvalue(L, 1); \
    luapush(L, ds2[i]); \
    for (j=3; j <= nargs; j++) lua_pushvalue(L, j); \
    lua_call(L, nargs - 1, 1); \
    fn_checkresult(L, lua_isboolean, "numarray.select", "boolean", 2); \
    if (agn_istrue(L, -1)) ds1[c++] = ds2[i]; \
    agn_poptop(L); \
  } \
}

#define aux_resizeselect(L,datastruct,datatype,c,procname) { \
  if ((c) != 0) { \
    datastruct = realloc(datastruct, (c)*sizeof(datatype)); \
    if (datastruct == NULL) \
      luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", procname); \
  } \
}

static int numarray_select (lua_State *L) {  /* 2.31.3 */
  NumArray *a, *b;
  int i, j, c, nargs, newstruct, equality;
  char type;
  size_t n;
  nargs = lua_gettop(L);
  if (nargs < 2)
    luaL_error(L, "Error in " LUA_QS ": requires at least two arguments.", "numarray.select");
  luaL_checkstack(L, nargs, "too many arguments");
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = NULL;
  b = checknumarray(L, 2);
  type = b->datatype;
  n = b->size;
  numarray_checkoptions(L, &nargs, &newstruct, &equality, "numarray.select");
  c = 0;
  if (newstruct) {
    createarray(L, a, type, n);
    switch (b->datatype) {
      case 0:
        aux_select(L, lua_pushinteger, a->data.c, b->data.c);
        aux_resizeselect(L, a->data.c, unsigned char, c, "numarray.select");
        break;
      case 1:
        aux_select(L, lua_pushnumber, a->data.n, b->data.n);
        aux_resizeselect(L, a->data.n, lua_Number, c, "numarray.select");
        break;
      case 2:
        aux_select(L, lua_pushnumber, a->data.i, b->data.i);
        aux_resizeselect(L, a->data.i, int32_t, c, "numarray.select");
        break;
      case 3:
        aux_select(L, lua_pushnumber, a->data.us, b->data.us);
        aux_resizeselect(L, a->data.us, uint16_t, c, "numarray.select");
        break;
      case 4:
        aux_select(L, lua_pushnumber, a->data.ui, b->data.ui);
        aux_resizeselect(L, a->data.ui, uint32_t, c, "numarray.select");  /* 2.35.0 patch */
        break;
      case 5:
        aux_select(L, createdlong, a->data.ld, b->data.ld);
        aux_resizeselect(L, a->data.ld, long double, c, "numarray.select");
        break;
      default:
        luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.select");
    }  /* return new numarray or null */
  } else {
    lua_pushvalue(L, 2);  /* push numarray for later return */
    switch (b->datatype) {
      case 0:
        aux_select(L, lua_pushinteger, b->data.c, b->data.c);
        aux_resizeselect(L, b->data.c, unsigned char, c, "numarray.select");
        break;
      case 1:
        aux_select(L, lua_pushnumber, b->data.n, b->data.n);
        aux_resizeselect(L, b->data.n, lua_Number, c, "numarray.select");
        break;
      case 2:
        aux_select(L, lua_pushnumber, b->data.i, b->data.i);
        aux_resizeselect(L, b->data.i, int32_t, c, "numarray.select");
        break;
      case 3:
        aux_select(L, lua_pushnumber, b->data.us, b->data.us);
        aux_resizeselect(L, b->data.us, uint16_t, c, "numarray.select");
        break;
      case 4:
        aux_select(L, lua_pushnumber, b->data.ui, b->data.ui);
        aux_resizeselect(L, b->data.ui, uint32_t, c, "numarray.select");  /* 2.35.0 patch */
        break;
      case 5:
        aux_select(L, createdlong, b->data.ld, b->data.ld);
        aux_resizeselect(L, b->data.ld, long double, c, "numarray.select");
        break;
      default:
        luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.select");
    }  /* return numarray or null */
  }
  if (c == 0) {
    agn_poptop(L);
    lua_pushnil(L);
  } else if (newstruct)
    a->size = c;
  else
    b->size = c;
  return 1;
}


#define aux_remove(L,luapush,ds1,ds2) { \
  for (i=0; i < n; i++) { \
    lua_pushvalue(L, 1); \
    luapush(L, ds2[i]); \
    for (j=3; j <= nargs; j++) lua_pushvalue(L, j); \
    lua_call(L, nargs - 1, 1); \
    fn_checkresult(L, lua_isboolean, "numarray.remove", "boolean", 2); \
    if (agn_isfalse(L, -1)) ds1[c++] = ds2[i]; \
    agn_poptop(L); \
  } \
}

static int numarray_remove (lua_State *L) {  /* 2.31.3 */
  NumArray *a, *b;
  int i, j, c, newstruct, equality, nargs;
  char type;
  size_t n;
  nargs = lua_gettop(L);
  if (nargs < 2)
    luaL_error(L, "Error in " LUA_QS ": requires at least two arguments.", "numarray.remove");
  luaL_checkstack(L, nargs, "too many arguments");
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = NULL;
  b = checknumarray(L, 2);
  type = b->datatype;
  n = b->size;
  c = 0;
  numarray_checkoptions(L, &nargs, &newstruct, &equality, "numarray.remove");
  if (newstruct) {
    createarray(L, a, type, n);
    switch (b->datatype) {
      case 0:
        aux_remove(L, lua_pushinteger, a->data.c, b->data.c);
        aux_resizeselect(L, a->data.c, unsigned char, c, "numarray.remove");
        break;
      case 1:
        aux_remove(L, lua_pushnumber, a->data.n, b->data.n);
        aux_resizeselect(L, a->data.n, lua_Number, c, "numarray.remove");
        break;
      case 2:
        aux_remove(L, lua_pushnumber, a->data.i, b->data.i);
        aux_resizeselect(L, a->data.i, int32_t, c, "numarray.remove");
        break;
      case 3:
        aux_remove(L, lua_pushnumber, a->data.us, b->data.us);
        aux_resizeselect(L, a->data.us, uint16_t, c, "numarray.remove");  /* 2.35.0 patch */
        break;
      case 4:
        aux_remove(L, lua_pushnumber, a->data.ui, b->data.ui);
        aux_resizeselect(L, a->data.ui, uint32_t, c, "numarray.remove");
        break;
      case 5:
        aux_remove(L, createdlong, a->data.ld, b->data.ld);
        aux_resizeselect(L, a->data.ld, long double, c, "numarray.remove");
        break;
      default:
        luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.remove");
    }  /* return new numarray or null */
  } else {
    lua_pushvalue(L, 2);  /* push numarray for later return */
    switch (b->datatype) {
      case 0:
        aux_remove(L, lua_pushinteger, b->data.c, b->data.c);
        aux_resizeselect(L, b->data.c, unsigned char, c, "numarray.remove");
        break;
      case 1:
        aux_remove(L, lua_pushnumber, b->data.n, b->data.n);
        aux_resizeselect(L, b->data.n, lua_Number, c, "numarray.remove");
        break;
      case 2:
        aux_remove(L, lua_pushnumber, b->data.i, b->data.i);
        aux_resizeselect(L, b->data.i, int32_t, c, "numarray.remove");
        break;
      case 3:
        aux_remove(L, lua_pushnumber, b->data.us, b->data.us);
        aux_resizeselect(L, b->data.us, uint16_t, c, "numarray.remove");
        break;
      case 4:
        aux_remove(L, lua_pushnumber, b->data.ui, b->data.ui);
        aux_resizeselect(L, b->data.ui, uint32_t, c, "numarray.remove");  /* 2.35.0 patch */
        break;
      case 5:
        aux_remove(L, createdlong, b->data.ld, b->data.ld);
        aux_resizeselect(L, b->data.ld, long double, c, "numarray.remove");
        break;
      default:
        luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.remove");
    }  /* return new numarray or null */
  }
  if (c == 0) {
    agn_poptop(L);
    lua_pushnil(L);
  } else if (newstruct)
    a->size = c;
  else
    b->size = c;
  return 1;
}


static int numarray_subs (lua_State *L) {
  NumArray *a, *b, *y;
  int i, j, eq, nargs, newstruct, equality;
  char type;
  size_t n;
  double p, q;
  long double pl, ql, xl;
  pl = ql = 0.0L;
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  numarray_checkoptions(L, &nargs, &newstruct, &equality, "numarray.subs");
  for (i=1; i < nargs; i++)
    luaL_checktype(L, i, LUA_TPAIR);
  a = checknumarray(L, nargs);
  n = a->size;
  type = a->datatype;
  if (newstruct) {
    createarray(L, b, type, n); y = b;
    aux_copy(b, a);
  } else {  /* push structure explicitly, we cannot play with lua_settop(L, 2) unfortunately INPLACE */
    b = NULL; lua_pushvalue(L, nargs); y = a;
  }
  /* start traversal */
  for (i=0; i < n; i++) {  /* array */
    switch (type) {
      case 0: p  = a->data.c[i];  break;
      case 1: p  = a->data.n[i];  break;
      case 2: p  = a->data.i[i];  break;
      case 3: p  = a->data.us[i]; break;
      case 4: p  = a->data.ui[i]; break;
      case 5: pl = a->data.ld[i]; break;
      default:
        p = 0;
        luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.subs");
    }
    if (type < 5) {  /* no long doubles ? */
      for (j=1; j < nargs; j++) {  /* subst list */
        q = agnL_pairgetinumber(L, "numarray.subs", j, 1);
        eq = (equality) ? p == q : tools_approx(p, q, L->Eps);
        if (eq) {
          lua_Number x = agnL_pairgetinumber(L, "numarray.subs", j, 2);
          switch (type) {
            case 0: y->data.c[i]  = x; break;
            case 1: y->data.n[i]  = x; break;
            case 2: y->data.i[i]  = x; break;
            case 3: y->data.us[i] = x; break;
            case 4: y->data.ui[i] = x; break;
            default: {
              luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.subs");
            }
          }
          break;
        }  /* of equality check  */
      }
    } else {  /* long doubles */
      for (j=1; j < nargs; j++) {  /* subst list */
        agnL_pairgetilongnumbers(L, "numarray.subs", j, &ql, &xl);
        eq = (equality) ? pl == ql : tools_approxl(pl, ql, L->hEps);
        if (eq) {
          y->data.ld[i] = xl;
          break;
        }  /* of equality check  */
      }
    }
  }
  return 1;  /* return numarray */
}


/* numarray.convert (f, a [, ])
   Same as `numarray.map` but processes in-place: Maps a function f on each element in the numarray `a` and changes the entries accordingly,
   i.e. the array elements will be transformed from a[1] to f(a[1]), etc. `f` must always return a number. */

#define aux_convert(L,luapush,check,convert,ds) { \
  for (i=0; i < n; i++) { \
    lua_pushvalue(L, 1); \
    luapush(L, ds[i]); \
    for (j=3; j <= nargs; j++) lua_pushvalue(L, j); \
    lua_call(L, nargs - 1, 1); \
    fn_checkresult(L, (check), "numarray.convert", "number or longdouble", 2); \
    ds[i] = convert(L, -1); \
    agn_poptop(L); \
  } \
}

static int numarray_convert (lua_State *L) {  /* 2.17.9 */
  NumArray *a;
  int i, j;
  size_t n, nargs;
  nargs = lua_gettop(L);
  if (nargs < 2)
    luaL_error(L, "Error in " LUA_QS ": requires at least two arguments.", "numarray.convert");
  luaL_checkstack(L, nargs, "too many arguments");
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = checknumarray(L, 2);
  n = a->size;
  switch (a->datatype) {
    case 0: aux_convert(L, lua_pushinteger, lua_isnumber, agn_tonumber, a->data.c); break;
    case 1: aux_convert(L, lua_pushnumber,  lua_isnumber, agn_tonumber, a->data.n);  break;
    case 2: aux_convert(L, lua_pushnumber,  lua_isnumber, agn_tonumber, a->data.i);  break;
    case 3: aux_convert(L, lua_pushnumber,  lua_isnumber, agn_tonumber, a->data.us); break;
    case 4: aux_convert(L, lua_pushnumber,  lua_isnumber, agn_tonumber, a->data.ui); break;
    case 5: aux_convert(L, createdlong,     agnL_isdlong, getdlongvalue, a->data.ld); break;
    default:
      luaL_error(L, "Error in " LUA_QS ": this should not happen.", "numarray.convert");
  }
  return 0;  /* return nothing */
}


static const struct luaL_Reg numarray_arraylib [] = {  /* metamethods for numeric userdata `n` */
  {"__aeq",      mt_aeq},              /* approximate equality mt */
  {"__eq",       mt_eeq},              /* equality mt */
  {"__eeq",      mt_eeq},              /* strict equality mt */
  {"__index",    numarray_subarray},   /* n[p], with p the index, counting from 1, NO support of n[p to q] indexing since the VM can only return one value */
  {"__writeindex", numarray_setitem},  /* n[p] := value, with p the index, counting from 1 */
  {"__tostring", mt_array2string},     /* for output at the console, e.g. print(n) */
  {"__size",     mt_getsize},          /* metamethod for `size` operator */
  {"__empty",    mt_empty},            /* metamethod for `empty` operator */
  {"__filled",   mt_filled},           /* metamethod for `filled` operator */
  {"__zero",     mt_zero},             /* metamethod for `zero` operator */
  {"__nonzero",  mt_nonzero},          /* metamethod for `nonzero` operator */
  {"__gc",       mt_arraygc},          /* please do not forget garbage collection */
  {"__in",       mt_in},               /* metamethod for `in` operator */
  {"__notin",    mt_notin},            /* metamethod for `notin` operator */
  {NULL, NULL}
};


static const luaL_Reg numarraylib[] = {
  {"convert",    numarray_convert},
  {"cycle",      numarray_cycle},
  {"double",     numarray_double},
  {"replicate",  numarray_replicate},
  {"getitem",    numarray_getitem},
  {"getbit",     numarray_getbit},
  {"getsize",    mt_getsize},
  {"include",    numarray_include},
  {"int32",      numarray_int32},
  {"iterate",    numarray_iterate},
  {"longdouble", numarray_longdouble},
  {"map",        numarray_map},
  {"purge",      numarray_purge},
  {"read",       numarray_read},
  {"remove",     numarray_remove},
  {"resize",     numarray_resize},
  {"satisfy",    numarray_satisfy},
  {"select",     numarray_select},
  {"setbit",     numarray_setbit},
  {"setitem",    numarray_setitem},
  {"sort",       numarray_sort},
  {"sorted",     numarray_sorted},
  {"subs",       numarray_subs},
  {"subarray",   numarray_subarray},
  {"toarray",    numarray_toarray},
  {"toreg",      numarray_toreg},
  {"toseq",      numarray_toseq},
  {"uchar",      numarray_uchar},
  {"uint32",     numarray_uint32},
  {"ushort",     numarray_ushort},
  {"used",       numarray_used},
  {"whereis",    numarray_whereis},
  {"write",      numarray_write},
  {NULL, NULL}
};


/*
** Open numarray library
*/
LUALIB_API int luaopen_numarray (lua_State *L) {
  luaL_newmetatable(L, "numarray");  /* metatable for numarray userdata, adds it to the registry with key 'numarray' */
  luaL_register(L, NULL, numarray_arraylib);  /* assign C metamethods to this metatable */
  luaL_register(L, AGENA_NUMARRAYLIBNAME, numarraylib);
  return 1;
}

