/*
** Rings: Multiple Lua States
** $Id: rings.c,v 1.24 2008/06/30 17:52:31 carregal Exp $

** Licence

Rings is free software: it can be used for both academic and commercial purposes at absolutely no cost. There are no
royalties or GNU-like "copyleft" restrictions. Rings qualifies as {Open Source} software. Its licenses are compatible
with {GPL}. Rings is not in the public domain and the {Kepler Project} keeps its copyright. The legal details are below.

The spirit of the license is that you are free to use Rings for any purpose at no cost without having to ask us. The
only requirement is that if you do use Rings, then you should give us credit by including the appropriate copyright
notice somewhere in your product or its documentation.

The Rings library is designed and implemented by Roberto Ierusalimschy and Toms Guisasola. The implementation is not
derived from licensed software.

------------------------------------
Copyright  2006-2017 Kepler Project.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of
the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Source taken from: http://keplerproject.github.io/rings

Some modifications have been made to access the arguments and return the result in old-fashioned `varargs' Lua 5.0
architecture. */

/* from: http://keplerproject.github.io/rings/manual.html

Rings is a library which provides a way to create new Agena states from within Agena. It also offers a simple way to
communicate between the creator (master) and the created (slave) states.

_Master functions_

Rings offers a single function which creates a new Agena state and returns an object representing it. The state which
creates other states is called the master and the created ones are called slaves. The master can execute code in any
of its slaves but each slave only has direct access to its master (or its own slaves).

All standard Agena libraries are opened automatically in a new state; other libraries have to be loaded explicitly.

The object representing a slave state has a method (dostring) which can execute Agena code in the corresponding state.
This method can receive arguments (only numbers, strings, booleans and userdata, which are converted to lightuserdata)
and always returns a boolean indicating whether the code executed correctly or not, followed by eventual return values
or an error message.

rings.open (env)
    Returns a newly created Agena state. Takes an optional environment to be used by remotedostring. If the environment is
    null, it defaults to the master_M or _G tables.

state:close () or rings.close (state)
    Closes the state.

state:dostring (string, ...) or  rings.dostring (state, string, ...)
    Executes a string in the slave state. In the slave, all the arguments are in a table named 'varargs'. Valid types of
    arguments and return values are: number, string, boolean, null and userdata (which are converted to lightuserdata).
    All other data types cannot be passed between states and will all be set to `null` if you try to exchange them.
    Returns a boolean indicating the status of the operation, followed by the returned values or an error message in
    case of error.

_Slave function_

The following function is registered in the newly created slave state:

remotedostring (string, ...)
    Executes a string in the master state. Behaves exactly as the method dostring except that it acts in the master state.


# test case: reverse arguments
code := "local args_reversed := [];

for _, value in ipairs(varargs) do
   put(args_reversed, 1, value)
end

return unpack(args_reversed)"

import rings;

a := rings.open()

rings.dostring(a, code, 1, 2, 3):

rings.close(a);
*/

#include "agena.h"
#include "agenalib.h"
#include "agnxlib.h"

#define RINGS_STATE     "rings state"
#define RINGS_ENV       "rings environment"
#define STATE_NAME      "rings Lua state"
#define RINGS_CACHE     "rings cache"

/* default initialization for fresh states */
#define RINGS_INITIALIZE luaL_openlibs

#if !(defined(LUA_DOS) || defined(__OS2__) || defined(LUA_ANSI))
#define AGENA_RINGSLIBNAME "rings"
LUALIB_API int (luaopen_rings) (lua_State *L);
#endif


typedef struct {
  lua_State *L;
} state_data;


/*
** Get a State object from the first call stack position.
*/
static state_data *getstate (lua_State *L) {
  state_data *s = (state_data *)luaL_checkudata(L, 1, STATE_NAME);
  luaL_argcheck(L, s != NULL, 1, "not a Lua State");
  luaL_argcheck(L, s->L, 1, "already closed state");
  return s;
}


static int state_tostring (lua_State *L) {
  state_data *s = (state_data *)luaL_checkudata(L, 1, STATE_NAME);
  lua_pushfstring(L, "Lua State (%p)", s);
  return 1;
}


/*
** Copies values from State src to State dst.
*/
static int copy_values (lua_State *dst, lua_State *src, int i, int top) {
  int nargs = top - i + 1;
  lua_checkstack(dst, nargs);
  for (; i <= top; i++) {
    switch (lua_type(src, i)) {
      case LUA_TNUMBER:
        if (lua_isinteger(src, i)) {
          lua_pushinteger(dst, lua_tointeger(src, i));
        } else {
          lua_pushnumber(dst, lua_tonumber(src, i));
        }
        break;
      case LUA_TBOOLEAN:
        lua_pushboolean(dst, lua_toboolean(src, i));
        break;
      case LUA_TSTRING: {
        size_t length;
        const char *string = lua_tolstring(src, i, &length);
        lua_pushlstring(dst, string, length);
        break;
      }
      case LUA_TLIGHTUSERDATA: {
        lua_pushlightuserdata(dst, lua_touserdata(src, i));
        break;
      }
      case LUA_TFUNCTION: {
        lua_CFunction f = lua_tocfunction(src, i);
        if (f)
          lua_pushcfunction(dst, f);
        else
          lua_pushnil(dst);
      }
      case LUA_TNIL:
      default:
        lua_pushnil(dst);
        break;
    }
  }
  return nargs;
}


/*
** Obtains a function which is the compiled string in the given state.
** It also caches the resulting function to optimize future uses.
** Leaves the compiled function on top of the stack or the error message
** produced by luaL_loadbuffer.
*/
static int compile_string (lua_State *L, void *cache, const char *str, size_t str_len) {
  if (cache == NULL) {
    lua_pushliteral(L, RINGS_CACHE);
  } else {
    lua_pushlightuserdata(L, cache);
  }
  lua_gettable(L, LUA_REGISTRYINDEX); /* push cache table */
  lua_getfield(L, -1, str);  /* cache[str] */
  if (!lua_isfunction(L, -1)) {
    int status;
    lua_pop(L, 1);  /* remove cache[str] (= nil) from top of the stack */
    status = luaL_loadbuffer(L, str, str_len, str); /* Compile */
    if (status != 0) { /* error? */
      lua_remove(L, -2);  /* removes cache table; leaves the error message */
      return status;
    }
    /* Sets environment */
    lua_getfield(L, LUA_REGISTRYINDEX, RINGS_ENV);
    if (cache == NULL) {
      lua_pushliteral(L, RINGS_CACHE);
    } else {
      lua_pushlightuserdata(L, cache);
    }
    lua_gettable(L, -2);
    if (!lua_isnil(L, -1)) {
      lua_setfenv(L, -3);
      lua_pop(L, 1);
    } else lua_pop(L, 2);
    /* Stores the produced function at cache[str] */
    lua_pushlstring(L, str, str_len);
    lua_pushvalue(L, -2);
    lua_settable(L, -4);  /* cache[str] = func */
  }
  lua_remove(L, -2);  /* removes cache table; leaves the function */
  return 0;
}


/*
** Executes a string of code from State src into State dst.
** idx is the index of the string of code.
*/
static int dostring (lua_State *dst, lua_State *src, void *cache, int idx) {
  int base;
  size_t str_len;
  const char *str = luaL_checklstring(src, idx, &str_len);
  lua_getfield(dst, LUA_REGISTRYINDEX, "rings_traceback");
  base = lua_gettop(dst);
  idx++;  /* ignore first argument (string of code) */
  if (compile_string(dst, cache, str, str_len) == 0) {  /* Compile OK? => push function */
    int i, arg_top, nargs, ret_base;
    arg_top = lua_gettop(src);
    nargs = copy_values(dst, src, idx, arg_top);  /* Push arguments to dst stack */
    lua_checkstack(dst, 2);
    /* we now create a `varargs' table of all the arguments pushed before */
    agn_createtable(dst, nargs, 0);
    lua_insert(dst, -nargs - 1);
    for (i=nargs; i > 0; i--) {
      lua_pushvalue(dst, -i);
      lua_rawseti(dst, -nargs - 2, nargs - i + 1);
    }
    lua_pop(dst, nargs);  /* drop all the arguments */
    lua_setglobal(dst, "varargs");  /* set table to global `varargs' */
    /* if (lua_pcall(dst, arg_top - idx + 1, LUA_MULTRET, base) == 0) { */ /* run OK? */
    ret_base = lua_gettop(dst);
    if (lua_pcall(dst, 0, LUA_MULTRET, base) == 0) {  /* run OK? */
      int nrets, ret_top;
      ret_top = lua_gettop(dst);
      lua_pushboolean(src, 1);  /* Push status = OK */
      /* nrets = copy_values(src, dst, base + 1, ret_top); */ /* Return values to src */
      nrets = copy_values(src, dst, ret_base, ret_top); /* return result values to master */
      lua_pop(dst, ret_top - base + 1);  /* clean up slave state */
      return 1 + nrets;  /* Return true (success) plus return values */
    }
  }
  lua_pushboolean(src, 0); /* Push status = ERR */
  lua_pushstring(src, lua_tostring(dst, -1));
  lua_pop(dst, 2); /* pops debug.traceback and result from dst state */
  return 2;
}


/*
** Executes a string of Lua code in the master state.
*/
static int master_dostring (lua_State *S) {
  lua_State *M;
  void *C;
  lua_getfield(S, LUA_REGISTRYINDEX, RINGS_STATE);
  M = (lua_State *)lua_touserdata(S, -1);
  lua_pop(S, 1);
  C = lua_touserdata(S, lua_upvalueindex(1));
  return dostring(M, S, C, 1);
}


/*
** Executes a string of Lua code in a given slave state.
*/
static int slave_dostring (lua_State *M) {
  state_data *s = getstate(M); /* S == s->L */
  lua_pushlightuserdata(s->L, M);
  lua_setfield(s->L, LUA_REGISTRYINDEX, RINGS_STATE);
  return dostring(s->L, M, NULL, 2);
}


/*
** Creates a weak table in the registry.
*/
static void create_cache (lua_State *L) {
  lua_newtable(L);
  lua_newtable(L); /* cache metatable */
  lua_pushliteral(L, "v");
  lua_setfield(L, -2, "__weak"); /* metatable.__mode = "v" */
  lua_setmetatable(L, -2);
  lua_settable(L, LUA_REGISTRYINDEX);
}


/*
** Creates a new Lua State and returns an userdata that represents it.
*/
static int state_new (lua_State *L) {
  state_data *s;
  //luaL_error(L, "Error in " LUA_QS ": this package is still in development.", "rings.open");
  if (lua_gettop(L) == 0) {
    lua_getglobal(L, "_M");
    if (lua_isnil(L, 1)) {
      lua_settop(L, 0);
      lua_getglobal(L, "_G");
      if (lua_isnil(L, 1)) {
        lua_settop(L, 0);
        lua_newtable(L);
      }
    }
  }
  s = (state_data *)lua_newuserdata(L, sizeof(state_data));
  if (s == NULL) {
    lua_pushliteral(L, "rings: could not create state data");
    lua_error(L);
  }
  s->L = NULL;
  luaL_getmetatable(L, STATE_NAME);
  lua_setmetatable(L, -2);
  s->L = luaL_newstate();
  if (s->L == NULL) {
    lua_pushliteral(L, "rings: could not create new state");
    lua_error(L);
  }
  /* Initialize environment */
  lua_getfield(L, LUA_REGISTRYINDEX, RINGS_ENV);
  lua_pushlightuserdata(L, s->L);
  lua_pushvalue(L, 1);
  lua_settable(L, -3);
  lua_pop(L, 1);
   /* load base libraries */
  RINGS_INITIALIZE(s->L);
  /* define dostring function (which runs strings on the master state) */
  lua_pushlightuserdata(s->L, s->L);
  lua_pushcclosure(s->L, master_dostring, 1);
  lua_setglobal(s->L, "remotedostring");
  /* fetches debug.traceback to registry */
  lua_getglobal(s->L, "debug");
  lua_getfield(s->L, -1, "traceback");
  lua_setfield(s->L, LUA_REGISTRYINDEX, "rings_traceback");
  /* Create caches */
  lua_pushlightuserdata(L, s->L);
  create_cache(L);
  lua_pushliteral(s->L, RINGS_CACHE);
  create_cache(s->L);
  lua_pushliteral(s->L, RINGS_ENV);
  create_cache(s->L);
  /* load library.agn and agena.ini files */
  lua_gc(s->L, LUA_GCSTOP, 0);  /* stop collector again during initialisation phase 2, Agena 1.0.4 */
  agnL_setLibname(s->L, 1, 0);  /* determine value of libname */
  agnL_initialise(s->L, 0, 0, 0);  /* load and run library.agn and optionally initialisation files */
  lua_gc(s->L, LUA_GCRESTART, 0);  /* restart collector */
  return 1;
}


/*
** Closes a Lua State.
** Returns `true' in case of success; `nil' when the state was already closed.
*/
static int slave_close (lua_State *L) {
  lua_State *K;
  state_data *s;
  s = (state_data *)luaL_checkudata(L, 1, STATE_NAME);
  K = s->L;
  luaL_argcheck(K, s != NULL, 1, "not a Lua State");
  if (K != NULL) {
    lua_getfield(L, LUA_REGISTRYINDEX, RINGS_ENV);
    lua_pushlightuserdata(L, K);
    lua_pushnil(L);
    lua_settable(L, -3);
    lua_pop(L, 1);
    lua_pushnil(K);
    lua_setglobal(K, "remotedostring");
//    lua_pushnil(K);
//    lua_setglobal(K, "debug");
#if !defined(LUA_DOS) && !defined(__OS2__) // && !defined(_WIN32) && !defined(__SOLARIS)
    lua_closestate(K);  /* the function must be called to prevent lost memory in Linux for whatever reasons */
#endif
    K = NULL;
  }
  return 0;
}


/*
** Creates the metatable for the state on top of the stack.
*/
static int state_createmetatable (lua_State *L) {
  /* State methods */
  struct luaL_Reg methods[] = {
    {"close", slave_close},
    {"dostring", slave_dostring},
    {NULL, NULL},
  };
  /* State metatable */
  if (!luaL_newmetatable(L, STATE_NAME)) return 0;
  /* define methods */
  luaL_setfuncs(L, methods, 0);
  /* define metamethods */
  lua_pushcfunction(L, slave_close);
  lua_setfield(L, -2, "__gc");
  lua_pushvalue(L, -1);
  lua_setfield(L, -2, "__index");
  lua_pushcfunction(L, state_tostring);
  lua_setfield(L, -2, "__tostring");
  lua_pushliteral(L, "You're not allowed to get the metatable of a Lua State");
  lua_setfield(L, -2, "__metatable");
  return 1;
}

static void set_info (lua_State *L) {
  lua_pushliteral(L, "Copyright (C) 2006 Kepler Project");
  lua_setfield(L, -2, "_COPYRIGHT");
  lua_pushliteral(L, "Rings: Multiple Lua States");
  lua_setfield(L, -2, "_DESCRIPTION");
  lua_pushliteral(L, "Rings 1.3.0");
  lua_setfield(L, -2, "_VERSION");
}

/*
** Opens library.
*/
int luaopen_rings (lua_State *L) {
  /* Library functions */
  struct luaL_Reg rings[] = {
    {"open", state_new},
    {"close", slave_close},
    {"dostring", slave_dostring},
    {NULL, NULL},
  };
  if (!state_createmetatable(L)) return 0;
  lua_pop(L, 1);
  /* define library functions */
  lua_newtable(L);
  luaL_newlib(L, rings);
  lua_pushvalue(L, -1);
  lua_setglobal(L, AGENA_RINGSLIBNAME);  /* = "rings" */
  lua_newtable(L);
  lua_setfield(L, LUA_REGISTRYINDEX, RINGS_ENV);
  set_info(L);
  /* fetches debug.traceback to registry */
  lua_getglobal(L, "debug");
  if (!lua_isnil(L, -1)) {
    lua_getfield(L, -1, "traceback");
    lua_setfield(L, LUA_REGISTRYINDEX, "rings_traceback");
  }
  lua_pop(L, 1);
  return 1;
}

