/* Tuple Library, based on code written by Roberto Ierusalimschy, and published on Lua-list on March 25, 2004. Initiated 16/11/2022.
   See: http://lua-users.org/lists/lua-l/2004-03/msg00353.html

   "To add more fun to the discussion, follows a small C module that adds tuples to Lua. In some quick tests, they outperform tables
   both in space and creation time, but are slower to access.

   Manual:

   a := tuple.ew(v1, v2, ..., vn) --> creates a new tuple
   a(i)                           --> returns vi
   a()                            --> returns v1, v2, ... vn

   -- Roberto" */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define tuple_c
#define LUA_LIB

#include "agena.h"
#include "agnxlib.h"
#include "agenalib.h"
#include "agncmpt.h"
#include "agnhlps.h"
#include "lapi.h"
#include "lstate.h"

#if !(defined(LUA_DOS) || defined(__OS2__) || defined(LUA_ANSI))
#define AGENA_TUPLELIBNAME "tuple"
LUALIB_API int (luaopen_tuple) (lua_State *L);
#endif


#define istuple(L, idx) (lua_isfunction(L, idx) && agn_isutype(L, idx, "tuple"))

#define checktuple(L,idx,pn) { \
  if (!(istuple(L, idx))) { \
    luaL_error(L, "Error in " LUA_QS ": expected a tuple, got %s.", \
      pn, luaL_typename(L, idx)); \
  } \
}


static int readtuple (lua_State *L) {
  int a = luaL_optint(L, 1, 0);
  if (a == 0) {  /* return all elements? */
    int i;
    for (i=1; lua_type(L, lua_upvalueindex(i)) != LUA_TNONE; i++)
      lua_pushvalue(L, lua_upvalueindex(i));
    return i - 1;
  } else {
    if (a < 0) {  /* return |a|-th field from the right */
      int i;
      for (i=1; lua_type(L, lua_upvalueindex(i)) != LUA_TNONE; i++);
      a = tools_posrelat(a, i - 1);
    }
    luaL_argcheck(L, 1 <= a && lua_type(L, lua_upvalueindex(a)) != LUA_TNONE,
                     1, "invalid index");
    lua_pushvalue(L, lua_upvalueindex(a));
    return 1;
  }
}


/* Creates a new tuple. It receives none, one or more values and returns a factory that when called returns the given values. e.g.

   > t := tuple.new(10, 20, 30);
   t():  # return all values
   10      20      30

   > t(1), t(2), t(3):  # return the values at indices 1, 2 and 3
   10      20      30

   The latter is equivalent to:

   > t[1], t[2], t[3]:
   10      20      30

   See also: tuple.getitem, tuple.setitem. */

static int tuple_new (lua_State *L) {
  int n = lua_gettop(L);
  luaL_argcheck(L, n < 256 || lua_checkstack(L, n), n, "too many values");
  lua_pushcclosure(L, readtuple, n);
  lua_setmetatabletoobject(L, -1, "tuple", 0);
  agn_setutypestring(L, -1, "tuple");
  return 1;
}


static int tuple_getsize (lua_State *L) {
  int nops;
  checktuple(L, 1, "tuple.getsize");
  nops = lua_nupvalues(L, 1);
  lua_pushinteger(L, nops);
  return 1;
}


static int tuple_tostring (lua_State *L) {  /* at the console, the tuple is formatted as follows: */
  int nops;
  checktuple(L, 1, "tuple.tostring");
  nops = lua_nupvalues(L, 1);
  if (agn_getutype(L, 1)) {
    lua_pushfstring(L, "(%u)", nops);
    lua_concat(L, 2);
  } else
    luaL_error(L, "Error in " LUA_QS ": invalid tuple.", "tuple.tostring");
  return 1;
}


/* tuple.getitem (t, i)
With any tuple, returns the value at a[i], where i, the index, is an integer counting from 1. The function is provided to avoid the index metamethod
overhead.

See also: tuple.setitem. */

static int tuple_getitem (lua_State *L) {
  int a, nops;
  checktuple(L, 1, "tuple.getitem");
  nops = lua_nupvalues(L, 1);
  a = agn_checkinteger(L, 2);
  if (a == 0) {  /* return all elements? */
    int i;
    luaL_checkstack(L, nops, "tuple.getitem");
    for (i=1; i <= nops; i++) lua_getupvalue(L, 1, i);
    return i - 1;
  } else {
    a = tools_posrelat(a, nops);
    luaL_argcheck(L, a > 0 && a <= nops, 1, "invalid index");
    lua_getupvalue(L, 1, a);
    return 1;
  }
}


/* tuple.setitem (t, i, v)
With any tuple, sets value v to a[i], where i, the index, is an integer counting from 1. The function is provided to avoid the index metamethod
overhead.

See also: tuple.getitem. */

static int tuple_setitem (lua_State *L) {
  int a, nops;
  if (lua_gettop(L) != 3)
    luaL_error(L, "Error in " LUA_QS ": expected three arguments.", "tuple.setitem");
  checktuple(L, 1, "tuple.setitem");
  nops = lua_nupvalues(L, 1);
  a = tools_posrelat(agn_checkinteger(L, 2), nops);
  luaL_argcheck(L, a > 0 && a <= nops, 1, "invalid index");
  lua_getupvalue(L, 1, a);
  lua_pushvalue(L, 3);
  lua_setupvalue(L, 1, a);
  return 1;
}

/* Taken from lbaselib.c; check options for map, select, remove, subs; externalised 2.31.5 */
static void aux_fcheckoptions (lua_State *L, int *nargs, int *newstruct, int *multipass, int *equality, int *newarray, const char *procname) {
  int checkoptions;
  *newstruct = 1;  /* returns a new structure, do not work in-place */
  *multipass = 1;  /* 1 = always substitute at each pass, 0 = substitute a position only once */
  *equality = 1;   /* 1 = strict equality, 0 = approximate equality */
  *newarray = 0;   /* 0 = leave holes in return as they are, 1 = remove holes and return an array */
  /* check for options, here `map in-place` */
  if (*nargs > 2 && lua_istrue(L, *nargs)) {
    (*nargs)--; *newstruct = 0;
  }
  checkoptions = 4;  /* check n options; CHANGE THIS if you add/delete options, 2.29.0 */
  while (checkoptions-- && *nargs > 2 && lua_ispair(L, *nargs)) {
    agn_pairgeti(L, *nargs, 1);  /* get left value, set to stack index -2 */
    agn_pairgeti(L, *nargs, 2);  /* get right value, set to stack index  -1 */
    if (agn_isstring(L, -2)) {
      const char *option = agn_tostring(L, -2);
      if (tools_streq("inplace", option)) {
        *newstruct = !agn_checkboolean(L, -1);
      } else if (tools_streq("multipass", option)) {
        *multipass = agn_checkboolean(L, -1);
      } else if (tools_streq("strict", option)) {
        *equality = agn_checkboolean(L, -1);
      } else if (tools_streq("newtuple", option)) {
        *newarray = agn_checkboolean(L, -1);
      } else {
        agn_poptoptwo(L);
        luaL_error(L, "Error in " LUA_QS ": unknown option " LUA_QS ".", procname, option);
      }
    }
    /* do not call lua_settop as it would corrupt the argument stack since we have already pushed values */
    (*nargs)--;
    agn_poptoptwo(L);
  }
}

static int tuple_map (lua_State *L) {
  int i, j, nargs, nops, newstruct, multipass, equality, newarray;
  nargs = lua_gettop(L);
  if (nargs < 2)
    luaL_error(L, "Error in " LUA_QS ": expected at least two arguments.", "tuple.map");
  luaL_checktype(L, 1, LUA_TFUNCTION);
  checktuple(L, 2, "tuple.map");
  aux_fcheckoptions(L, &nargs, &newstruct, &multipass, &equality, &newarray, "tuple.map");
  nops = lua_nupvalues(L, 2);
  if (newstruct) {
    luaL_checkstack(L, nops > nargs ? nops : nargs, "not enough stack space");
    for (i=1; i <= nops; i++) lua_getupvalue(L, 2, i);
    lua_pushcclosure(L, readtuple, nops);  /* create new tuple */
    lua_setmetatabletoobject(L, -1, "tuple", 0);
    agn_setutypestring(L, -1, "tuple");
  } else {
    luaL_checkstack(L, nargs, "too many arguments");
    lua_pushvalue(L, 2);  /* push existing tuple */
  }
  for (i=1; i <= nops; i++) {
    lua_pushvalue(L, 1);        /* push function */
    lua_getupvalue(L, -2, i);   /* push i-th upvalue */
    for (j=3; j <= nargs; j++)  /* push optional function arguments */
      lua_pushvalue(L, j);
    lua_call(L, nargs - 1, 1);
    lua_setupvalue(L, -2, i);   /* pops result of the function call */
  }
  return 1;
}


/* tuple.subs (x:v [, ], t [, true])
Substitutes all occurrences of the value x in tuple t with value v. Multiple substitution pairs can be given. The substitutions are performed sequentially and by default simultaneously starting with the first pair.

> t := tuple.subs(1:3, 2:4, tuple.new(1, 2, -1)):
returns a tuple with values 3, 4, -1.

If the last argument is the option inplace=true, or the Boolean true, then the operation will be done in-place, modifying the original tuple, but saving memory. After completion, the function returns the modified tuple.

By default, tuple.subs is replacing the elements in a structure with _all_ the replacements given, including intermediate substitutions. So if we have an expression like

> t := tuple.subs(1:2, 2:3, 3:4, tuple.new(1, 2, 3))

we will get a tuple with values 4, 4, 4.

By passing the multipass=false option, the rest of the replacement list will be skipped as soon as a substitution has been done:

> t := tuple.subs(1:2, 2:3, 3:4, tuple.new(1, 2, 3), multipass=false):

returns a new tuple with values 2, 3, 4.

You can check numbers for approximate instead of strict equality by passing the new strict=false option.

See also: tuple.map, tuple.remove, tuple.select. */

static int tuple_subs (lua_State *L) {
  int i, k, flag, nargs, nops, newstruct, multipass, equality, newarray;
  nargs = lua_gettop(L);
  aux_fcheckoptions(L, &nargs, &newstruct, &multipass, &equality, &newarray, "tuple.subs");
  for (i=1; i < nargs; i++)
    luaL_checktype(L, i, LUA_TPAIR);
  checktuple(L, nargs, "tuple.subs");
  nops = lua_nupvalues(L, nargs);
  luaL_checkstack(L, (nargs > nops) ? nargs : nops, "too many arguments");
  if (newstruct) {
     luaL_checkstack(L, nops > nargs ? nops : nargs, "not enough stack space");
     for (i=1; i <= nops; i++) lua_getupvalue(L, nargs, i);
     lua_pushcclosure(L, readtuple, nops);  /* create new tuple */
     lua_setmetatabletoobject(L, -1, "tuple", 0);
     agn_setutypestring(L, -1, "tuple");
  } else
    lua_pushvalue(L, nargs);
  /* start traversal */
  for (k=1; k <= nops; k++) {
    lua_getupvalue(L, nargs, k);
    flag = 0;
    for (i=1; i < nargs; i++) {
      agn_pairgeti(L, i, 1);
      if (equality ? lua_equal(L, -1, -2) : lua_rawaequal(L, -1, -2)) {
        agn_poptoptwo(L);  /* delete value and lhs */
        agn_pairgeti(L, i, 2);  /* push rhs */
        flag = 1;
        if (!multipass) break;
      } else {
        agn_poptop(L);  /* delete lhs */
      }
    }
    if (!flag && !newstruct)  /* replace only if needed */
      agn_poptop(L);
    else
      lua_setupvalue(L, -2, k);  /* store result in tuple */
  }
  return 1;
}


/* tuple.select (f, t [,  ])
Returns all values in tuple t that satisfy a condition determined by function f in a new tuple.

If f has only one argument, then only the function and the tuple are passed to tuple.select.

> t := tuple.select(<< x -> x > 1 >>, tuple.new(1, 2, 3)):
returns a new tuple including the value 2 and 3.

If the function has more than one argument, then all arguments except the first are passed right after the name of t.

> t := tuple.select(<< x, y -> x > y >>, tuple.new(1, 2, 3), 1):   # 1 for parameter y
returns a new tuple including the value 2 and 3.

See also: tuple.map, tuple.remove, tuple.subs. */

static int tuple_select (lua_State *L) {
  size_t i, j, c;
  int nargs, nops;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  nargs = lua_gettop(L);
  checktuple(L, 2, "tuple.select");
  nops = lua_nupvalues(L, 2);
  luaL_checkstack(L, (5 + nargs - 3 + 1) > nops ? (5 + nargs - 3 + 1) : nops, "not enough stack space");
  /* very rough guess whether table is an array or dictionary */
  c = 0;
  /* start traversal */
  for (i=1; i <= nops; i++) {
    lua_pushvalue(L, 1);
    lua_getupvalue(L, 2, i);
    for (j=3; j <= nargs; j++)
      lua_pushvalue(L, j);
    lua_call(L, nargs - 1, 1);  /* call function with nargs-1 argument(s) and one result */
    if (agn_istrue(L, -1)) {
      agn_poptop(L);
      lua_getupvalue(L, 2, i);
      c++;
    } else {  /* in-place operation: delete non-satisfying value from table */
      agn_poptop(L);  /* delete result */
    }
  }
  lua_pushcclosure(L, readtuple, c);  /* create new tuple */
  lua_setmetatabletoobject(L, -1, "tuple", 0);
  agn_setutypestring(L, -1, "tuple");
  return 1;
}


/* tuple.remove (f, t [,  ])
Returns all values in tuple t that do not satisfy a condition determined by function f, as a new tuple.

If the function has only one argument, then only the function and the tuple are passed to tuple.remove.

> tuple.remove(<< x -> x > 1 >>, tuple.new(1, 2, 3));
returns a new tuple including the value 1 only.

If the function has more than one argument, then all arguments except the first are
passed right after the name of t.

> tuple.remove(<< x, y -> x > y >>, tuple.new(1, 2, 3), 1): # 1 for parameter y
returns a new tuple including the value 1 only.

See also: tuple.map, tuple.select, tuple.subs. */

static int tuple_remove (lua_State *L) {
  size_t i, j, c;
  int nargs, nops;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  nargs = lua_gettop(L);
  checktuple(L, 2, "tuple.remove");
  nops = lua_nupvalues(L, 2);
  luaL_checkstack(L, (5 + nargs - 3 + 1) > nops ? (5 + nargs - 3 + 1) : nops, "not enough stack space");
  /* very rough guess whether table is an array or dictionary */
  c = 0;
  /* start traversal */
  for (i=1; i <= nops; i++) {
    lua_pushvalue(L, 1);
    lua_getupvalue(L, 2, i);
    for (j=3; j <= nargs; j++)
      lua_pushvalue(L, j);
    lua_call(L, nargs - 1, 1);  /* call function with nargs-1 argument(s) and one result */
    if (agn_istrue(L, -1)) {
      agn_poptop(L);  /* delete result */
    } else {  /* in-place operation: delete non-satisfying value from table */
      agn_poptop(L);
      lua_getupvalue(L, 2, i);
      c++;
    }
  }
  lua_pushcclosure(L, readtuple, c);  /* create new tuple */
  lua_setmetatabletoobject(L, -1, "tuple", 0);
  agn_setutypestring(L, -1, "tuple");
  return 1;
}


/* Receives tuple t and puts its contents into a table, the return. */
static int tuple_totable (lua_State *L) {
  int i, nops;
  checktuple(L, 1, "tuple.totable");
  nops = lua_nupvalues(L, 1);
  agn_createtable(L, nops, 0);
  for (i=1; i <= nops; i++) {
    lua_getupvalue(L, 1, i);
    lua_rawseti(L, -2, i);
  }
  return 1;
}


/* Receives tuple t and puts its contents into a sequence, the return. */
static int tuple_toseq (lua_State *L) {
  int i, nops;
  checktuple(L, 1, "tuple.toseq");
  nops = lua_nupvalues(L, 1);
  agn_createseq(L, nops);
  for (i=1; i <= nops; i++) {
    lua_getupvalue(L, 1, i);
    lua_seqseti(L, -2, i);
  }
  return 1;
}


/* Receives tuple t and puts its contents into a register, the return. */
static int tuple_toreg (lua_State *L) {
  int i, nops;
  checktuple(L, 1, "tuple.toreg");
  nops = lua_nupvalues(L, 1);
  agn_createreg(L, nops);
  for (i=1; i <= nops; i++) {
    lua_getupvalue(L, 1, i);
    agn_regseti(L, -2, i);
  }
  return 1;
}


static void tupleeq (lua_State *L, int (*fn)(lua_Number, lua_Number, lua_Number), const char *procname) {  /* based on the linalg package (function meq) */
  int i, flag, xn, yn;
  lua_Number eps;
  checktuple(L, 1, procname);
  checktuple(L, 2, procname);
  eps = agn_getepsilon(L);
  xn = lua_nupvalues(L, 1);
  yn = lua_nupvalues(L, 2);
  flag = xn == yn;
  luaL_checkstack(L, 2, "not enough stack space");
  for (i=1; i <= xn && flag; i++) {
    lua_getupvalue(L, 1, i);
    lua_getupvalue(L, 2, i);
    if (agn_isnumber(L, -1) && agn_isnumber(L, -2))
      flag = fn(agn_tonumber(L, -1), agn_tonumber(L, -2), eps);
    else
      flag = lua_equal(L, -1, -2);
    agn_poptoptwo(L);
  }
  lua_pushboolean(L, flag);
}

static int equal (lua_Number x, lua_Number y, lua_Number eps) {  /* strict equality, eps does not matter */
  return x == y;
}

static int mt_aeq (lua_State *L) {
  tupleeq(L, tools_approx, "tuple.__aeq");
  return 1;
}


static int mt_eeq (lua_State *L) {
  tupleeq(L, equal, "tuple.__(e)eq");
  return 1;
}


static int mt_empty (lua_State *L) {
  int nops;
  checktuple(L, 1, "empty");
  nops = lua_nupvalues(L, 1);
  lua_pushboolean(L, nops == 0);
  return 1;
}


static int mt_filled (lua_State *L) {
  int nops;
  checktuple(L, 1, "filled");
  nops = lua_nupvalues(L, 1);
  lua_pushboolean(L, nops != 0);
  return 1;
}


/* Checks whether all elements in the tuple are zeros and returns `true` or `false`. */
static int mt_zero (lua_State *L) {
  int i, rc, nops;
  checktuple(L, 1, "zero");
  nops = lua_nupvalues(L, 1);
  rc = 1;
  for (i=1; (i <= nops) && rc; i++) {
    lua_getupvalue(L, 1, i);
    rc = agn_isnumber(L, -1) && agn_tonumber(L, -1) == 0;
    agn_poptop(L);
  }
  lua_pushboolean(L, nops && rc);
  return 1;
}


/* Checks whether all elements in the tuple are non-zeros and returns `true` or `false`. */
static int mt_nonzero (lua_State *L) {
  int i, rc, nops;
  checktuple(L, 1, "nonzero");
  nops = lua_nupvalues(L, 1);
  rc = 0;
  for (i=1; (i <= nops) && !rc; i++) {
    lua_getupvalue(L, 1, i);
    rc = agn_isnumber(L, -1) && agn_tonumber(L, -1) == 0;
    agn_poptop(L);
  }
  lua_pushboolean(L, !rc);
  return 1;
}


static int mt_in (lua_State *L) {
  int i, rc, nops;
  checktuple(L, 2, "in");
  nops = lua_nupvalues(L, 2);
  rc = 0;
  for (i=1; (i <= nops) && !rc; i++) {
    lua_getupvalue(L, 2, i);
    rc = lua_equal(L, -1, 1);
    agn_poptop(L);
  }
  lua_pushboolean(L, rc);
  return 1;
}


static int mt_notin (lua_State *L) {
  int i, rc, nops;
  checktuple(L, 2, "notin");
  nops = lua_nupvalues(L, 2);
  rc = 1;
  for (i=1; (i <= nops) && rc; i++) {
    lua_getupvalue(L, 2, i);
    rc = !lua_equal(L, -1, 1);
    agn_poptop(L);
  }
  lua_pushboolean(L, rc);
  return 1;
}


static int mt_union (lua_State *L) {
  int i, xn, yn, nops;
  checktuple(L, 1, "union");
  checktuple(L, 2, "union");
  xn = lua_nupvalues(L, 1);
  yn = lua_nupvalues(L, 2);
  nops = xn + yn;
  if (nops > 255)
    luaL_error(L, "Error in " LUA_QS ": too many elements in result: %d instead of max. 255.", "union", nops);
  luaL_checkstack(L, nops + 2, "not enough stack space");
  for (i=1; i <= xn; i++) {
    lua_getupvalue(L, 1, i);
  }
  for (i=1; i <= yn; i++) {
    lua_getupvalue(L, 2, i);
  }
  lua_pushcclosure(L, readtuple, nops);
  lua_setmetatabletoobject(L, -1, "tuple", 0);
  agn_setutypestring(L, -1, "tuple");
  return 1;
}


static int mt_intersect (lua_State *L) {
  int i, j, c, xn, yn, rc;
  checktuple(L, 1, "intersect");
  checktuple(L, 2, "intersect");
  xn = lua_nupvalues(L, 1);
  yn = lua_nupvalues(L, 2);
  if (xn > 255 || yn > 255)
    luaL_error(L, "Error in " LUA_QS ": too many elements in result, maxmum of 255 possible.", "intersect");
  luaL_checkstack(L, 2 + ((xn > yn) ? xn : yn), "not enough stack space");
  c = 0;
  for (i=1; i <= xn; i++) {
    rc = 0;
    lua_getupvalue(L, 1, i);
    for (j=1; j <= yn; j++) {
      lua_getupvalue(L, 2, j);
      rc = lua_equal(L, -2, -1);
      agn_poptop(L);
      if (rc) {  /* leave value on stack for later inclusion */
        c++;
        break;
      }
    }
    if (!rc) agn_poptop(L);
  }
  lua_pushcclosure(L, readtuple, c);
  lua_setmetatabletoobject(L, -1, "tuple", 0);
  agn_setutypestring(L, -1, "tuple");
  return 1;
}


static int mt_minus (lua_State *L) {
  int i, j, c, xn, yn, rc;
  checktuple(L, 1, "minus");
  checktuple(L, 2, "minus");
  xn = lua_nupvalues(L, 1);
  yn = lua_nupvalues(L, 2);
  if (xn > 255)
    luaL_error(L, "Error in " LUA_QS ": too many elements in result, maxmum of 255 possible.", "minus");
  luaL_checkstack(L, xn + 2, "not enough stack space");
  c = 0;
  for (i=1; i <= xn; i++) {
    rc = 0;
    lua_getupvalue(L, 1, i);
    for (j=1; j <= yn; j++) {
      lua_getupvalue(L, 2, j);
      rc = lua_equal(L, -2, -1);
      agn_poptop(L);
      if (rc) break;
    }
    if (!rc) c++;
    else agn_poptop(L);
  }
  lua_pushcclosure(L, readtuple, c);
  lua_setmetatabletoobject(L, -1, "tuple", 0);
  agn_setutypestring(L, -1, "tuple");
  return 1;
}


static int mt_gc (lua_State *L) {  /* please do not forget to collect garbage */
  checktuple(L, 1, "(gc)");
  if (lua_nupvalues(L, 1) != -1) {
    lua_setmetatabletoobject(L, 1, NULL, 1);
  }
  return 0;
}


static const struct luaL_Reg mt_tuplelib [] = {  /* metamethods for numeric userdata `n` */
  {"__index",      tuple_getitem},   /* n[p], with p the index, counting from 1, NO support of n[p to q] indexing since the VM can only return one value */
  {"__writeindex", tuple_setitem},   /* n[p] := value, with p the index, counting from 1 */
  {"__tostring",   tuple_tostring},  /* for output at the console, e.g. print(n) */
  {"__size",       tuple_getsize},   /* metamethod for `size` operator */
  {"__empty",      mt_empty},        /* metamethod for `empty` operator */
  {"__filled",     mt_filled},       /* metamethod for `filled` operator */
  {"__zero",       mt_zero},         /* metamethod for `zero` operator */
  {"__nonzero",    mt_nonzero},      /* metamethod for `nonzero` operator */
  {"__in",         mt_in},           /* metamethod for `in` operator */
  {"__notin",      mt_notin},        /* metamethod for `notin` operator */
  {"__intersect",  mt_intersect},    /* metamethod for `intersect` operator */
  {"__minus",      mt_minus},        /* metamethod for `minus` operator */
  {"__union",      mt_union},        /* metamethod for `union` operator */
  {"__eq",         mt_eeq},          /* equality mt */
  {"__aeq",        mt_aeq},          /* approximate equality mt */
  {"__gc",         mt_gc},           /* garbage collection */
  {NULL, NULL}
};


static const luaL_Reg tuplelib[] = {
  {"getitem",    tuple_getitem},
  {"getsize",    tuple_getsize},
  {"map",        tuple_map},
  {"new",        tuple_new},
  {"remove",     tuple_remove},
  {"select",     tuple_select},
  {"setitem",    tuple_setitem},
  {"subs",       tuple_subs},
  {"toreg",      tuple_toreg},
  {"toseq",      tuple_toseq},
  {"tostring",   tuple_tostring},
  {"totable",    tuple_totable},
  {NULL, NULL}
};


/*
** Open tuple library
*/
LUALIB_API int luaopen_tuple (lua_State *L) {
  /* metamethods */
  luaL_newmetatable(L, "tuple");
  luaL_register(L, NULL, mt_tuplelib);
  /* register library */
  luaL_register(L, AGENA_TUPLELIBNAME, tuplelib);
  return 1;
}

